{%mainunit ../BASE_Engine.pas}

{***************************************************
 Package pl_TitanScript.pkg
 This file is part of CodeTyphon Studio
****************************************************}

constructor TTitanBaseScripter.Create(Owner: TObject);
begin
  CurrScripter := Self;

  _ObjCount := 0;

  IgnoreBreakpoints := False;
  try
    _BeginWrite;
    ScripterList.Add(Self);
  finally
    _EndWrite;
  end;
  NameList := TTitanNameList.Create;
  LocalDefinitions := TTitanDefinitionList.Create(False);

  Visited := TList.Create;
  ExtraModuleList := TStringList.Create;
  DefList := TStringList.Create;
  RunList := TTitanStack.Create;
  fSearchPathes := TStringList.Create;
  DefaultParameterList := TDefaultParameterList.Create;
  UnknownTypes := TTitanIDRecList.Create;
  CallRecList := TTitanCallRecList.Create;
  TypeAliasList := TTitanAssociativeList.Create(False);

  CancelMessage := '';

  AllowedEvents := True;

  ArrayClassRec := nil;

  LastResultID := 0;

  Self.Owner := Owner;

  Modules := TTitanModules.Create(Self);

  VariantStack := TTitanVariantStack.Create;
  NegVarList := TTitanVarList.Create;

  ParamList := TTitanParamList.Create;

  PrototypeNameIndex := NameList.Add('prototype');
  ConstructorNameIndex := NameList.Add('constructor');

  fStackSize := DefaultStackSize;

  SymbolTable := TTitanSymbolTable.Create(Self);

  ClassList := TTitanClassList.Create(Self);
  Code := TTitanCode.Create(Self);
  ClassList.AddClass(SymbolTable.RootNamespaceID, RootNamespaceName, '', '', [modSTATIC], ckClass, True);

  SignDefs := False;
  fLongStrLiterals := True;

  MethodBody := TTitanMethodBody.Create(Self);
  EventHandlerList := TTitanEventHandlerList.Create;
  ScriptObjectList := TTitanScriptObjectList.Create(Self);
  ActiveXObjectList := TTitanObjectList.Create;

  CompileTimeHeap := TTitanCompileTimeHeap.Create;

  ErrorInstance := TTitanError.Create;

  LocalDefinitions.AddObject('ErrorObject', ErrorInstance, nil, Self);

  RegisteredFieldList := TTitanFieldList.Create;
  ParserList := TTitanParserList.Create;

  DiscardError;

  ExtraCodeList := TTitanCodeList.Create;
  DoNotDestroyList := TTitanIds.Create(False);
  TempObjectList := TTitanAssocList.Create;
  ForbiddenPublishedProperties := TList.Create;
  ForbiddenPublishedPropertiesEx := TStringList.Create;

  EvalCount := 0;

  fTotalLineCount := 0;

  State := _ssInit;

end;

destructor TTitanBaseScripter.Destroy;
var
  I: integer;
begin
  _BeginWrite;
  try
    I := ScripterList.IndexOf(Self);
    if I <> -1 then
      ScripterList.Delete(I);
  finally
    _EndWrite;
  end;

  ScriptObjectList.Free;
  CompileTimeHeap.Free;
  ActiveXObjectList.Free;

  SymbolTable.Free;
  Code.Free;
  VariantStack.Free;
  NegVarList.Free;
  ClassList.Free;
  MethodBody.Free;
  EventHandlerList.Free;

  RegisteredFieldList.Free;
  ParserList.Free;

  ExtraCodeList.Free;
  DoNotDestroyList.Free;
  TempObjectList.Free;
  ForbiddenPublishedProperties.Free;
  ForbiddenPublishedPropertiesEx.Free;

  Modules.Free;

  ParamList.Free;
  Visited.Free;
  ExtraModuleList.Free;
  DefList.Free;
  RunList.Free;
  fSearchPathes.Free;
  DefaultParameterList.Free;
  UnknownTypes.Free;
  CallRecList.Free;
  TypeAliasList.Free;

  NameList.Free;
  LocalDefinitions.Free;

  ErrorInstance.Free;

  inherited;
end;

procedure TTitanBaseScripter.SetState(Value: TTitanScripterState);
begin
  if Assigned(OnChangeState) then
    if fState <> Value then
      OnChangeState(Owner, fState, Value);

  fState := Value;
  case fState of
    _ssReadyToCompile:
    begin
      SymbolTable.Card := FirstSymbolCard;
      SymbolTable.MemBoundVar := SymbolTable.CreateMemBoundVar;
    end;
  end;
end;

procedure TTitanBaseScripter.ResetCompileStage;
begin
  CancelMessage := '';

  EventHandlerList.ClearHandlers;

  ScriptObjectList.ResetCompileStage;
  CompileTimeHeap.ResetCompileStage;
  ActiveXObjectList.Clear;

  SymbolTable.ResetCompileStage;
  ClassList.ResetCompileStage;
  VariantStack.Clear;
  Code.ResetCompileStage;
  State := _ssReadyToCompile;

  ExtraCodeList.Clear;
  DoNotDestroyList.Clear;
  TempObjectList.Clear;
  ExtraModuleList.Clear;

  LastResultID := 0;

  AllowedEvents := True;

  DefList.Clear;
  DefaultParameterList.Clear;
  UnknownTypes.Clear;
  CallRecList.Clear;
  TypeAliasList.Clear;

  ScriptObjectList.ResetRunStage;

  Dump();
end;


procedure TTitanBaseScripter.ResetScripterEx;
begin
  CancelMessage := '';

  EventHandlerList.ClearHandlers;

  ScriptObjectList.ResetCompileStage(True);
  CompileTimeHeap.ResetCompileStage;
  ActiveXObjectList.Clear;

  SymbolTable.ResetCompileStage;
  ClassList.ResetCompileStage;
  VariantStack.Clear;
  Code.ResetCompileStage;
  State := _ssReadyToCompile;

  ExtraCodeList.Clear;
  DoNotDestroyList.Clear;
  TempObjectList.Clear;
  ExtraModuleList.Clear;

  LastResultID := 0;

  AllowedEvents := True;

  DefList.Clear;
  DefaultParameterList.Clear;
  UnknownTypes.Clear;
  CallRecList.Clear;
  TypeAliasList.Clear;

  ScriptObjectList.ResetRunStage(True);

  Dump();
end;


procedure TTitanBaseScripter.InitRunStage;
begin
  ClassList.InitRunStage;
  SymbolTable.InitRunStage;
  Code.InitRunStage;

  CreateRunList;
end;

procedure TTitanBaseScripter.ResetRunStage;
begin
  CancelMessage := '';

  ClassList.ResetRunStage;
  SymbolTable.ResetRunStage;
  Code.ResetRunStage;
  ScriptObjectList.ResetRunStage;
  ActiveXObjectList.Clear;
  TempObjectList.Clear;

  VariantStack.Clear;

  AllowedEvents := True;

  RunList.Clear;
end;

procedure TTitanBaseScripter.AddLocalDefs;
var
  I: integer;
  SO: TTitanScriptObject;
  ID: integer;
  ClassRec: TTitanClassRec;
begin
  for I := 0 to LocalDefinitions.Count - 1 do
    if LocalDefinitions.Records[I] is TTitanObjectDefinition then
      with LocalDefinitions.Records[I] as TTitanObjectDefinition do
      begin
        SO := DelphiInstanceToScriptObject(Instance, Self);
        Value := ScriptObjectToVariant(SO);

        if Owner <> nil then
          ClassRec := ClassList.FindClassByName(Owner.Name)
        else
          ClassRec := ClassList[0];

        ID := ClassRec.MemberList.GetMemberID(Name);
        if ID > 0 then
          ClassRec.DeleteMember(ID);

        AddToScripter(Self);
      end
    else if LocalDefinitions.Records[I] is TTitanVirtualObjectDefinition then
      with LocalDefinitions.Records[I] as TTitanVirtualObjectDefinition do
      begin
        if Owner <> nil then
          ClassRec := ClassList.FindClassByName(Owner.Name)
        else
          ClassRec := ClassList[0];

        ID := ClassRec.MemberList.GetMemberID(Name);
        if ID > 0 then
          ClassRec.DeleteMember(ID);

        AddToScripter(Self);
      end
    else if LocalDefinitions.Records[I] is TTitanInterfaceVarDefinition then
      with LocalDefinitions.Records[I] as TTitanInterfaceVarDefinition do
      begin
        if Owner <> nil then
          ClassRec := ClassList.FindClassByName(Owner.Name)
        else
          ClassRec := ClassList[0];

        ID := ClassRec.MemberList.GetMemberID(Name);
        if ID > 0 then
          ClassRec.DeleteMember(ID);

        AddToScripter(Self);
      end
    else if LocalDefinitions.Records[I] is TTitanConstantDefinition then
      with LocalDefinitions.Records[I] as TTitanConstantDefinition do
      begin
        if Owner <> nil then
          ClassRec := ClassList.FindClassByName(Owner.Name)
        else
          ClassRec := ClassList[0];

        ID := ClassRec.MemberList.GetMemberID(Name);
        if ID > 0 then
          ClassRec.DeleteMember(ID);
        AddToScripter(Self);
      end
    else if LocalDefinitions.Records[I] is TTitanVariableDefinition then
      with LocalDefinitions.Records[I] as TTitanVariableDefinition do
      begin
        if Owner <> nil then
          ClassRec := ClassList.FindClassByName(Owner.Name)
        else
          ClassRec := ClassList[0];

        ID := ClassRec.MemberList.GetMemberID(Name);
        if ID > 0 then
          ClassRec.DeleteMember(ID);
        AddToScripter(Self);
      end;
end;

procedure TTitanBaseScripter.AddDefs;
var
  I, N: integer;
  SO: TTitanScriptObject;
  ID: integer;
  DOB: TTitanObjectDefinition;
  D: TTitanDefinition;
begin
  if SignLoadOnDemand then
    N := GlobalDefListInitialCount
  else
    N := GlobalDefinitionList.Count;

  try
    _BeginRead;
    if not SignDefs then
    begin
      ClassList.AddDefinitionList(GlobalDefinitionList, N);
      SignDefs := True;
    end
    else
    begin
      for I := 0 to N - 1 do
      begin
        if I >= LastDefinitionListCount then
        begin
          D := GlobalDefinitionList.Records[I];
          D.AddToScripter(Self);
          continue;
        end;

        if GlobalDefinitionList.Records[I] is TTitanObjectDefinition then
        begin
          DOB := TTitanObjectDefinition(GlobalDefinitionList.Records[I]);
          with DOB do
            if IsObject(Value) then
            begin
              SO := VariantToScriptObject(Value);
              if SO.Instance <> Instance then
              begin
                SO := DelphiInstanceToScriptObject(Instance, Self);
                Value := ScriptObjectToVariant(SO);
              end;
            end
            else
            begin
              SO := DelphiInstanceToScriptObject(Instance, Self);
              Value := ScriptObjectToVariant(SO);

              ID := ClassList[0].MemberList.GetMemberID(Name);
              if ID = 0 then
                AddToScripter(Self);
            end;
        end
        else if GlobalDefinitionList.Records[I] is TTitanConstantDefinition then
          with GlobalDefinitionList.Records[I] as TTitanConstantDefinition do
          begin
            ID := ClassList[0].MemberList.GetMemberID(Name);
            if ID = 0 then
              AddToScripter(Self);
          end
        else if GlobalDefinitionList.Records[I] is TTitanVariableDefinition then
          with GlobalDefinitionList.Records[I] as TTitanVariableDefinition do
          begin
            ID := ClassList[0].MemberList.GetMemberID(Name);
            if ID = 0 then
              AddToScripter(Self);
          end;
      end;
    end;
  finally
    LastDefinitionListCount := GlobalDefinitionList.Count;
    _EndRead;
  end;
end;

function TTitanBaseScripter.AddModule(const ModuleName, LanguageName: string): integer;
begin
  Result := Modules.IndexOf(ModuleName);
  if Result = -1 then
    Result := Modules.Add(ModuleName, LanguageName);
end;

procedure TTitanBaseScripter.AddCode(const ModuleName, Code: string);
var
  I: integer;
  Module: TTitanModule;
begin
  I := Modules.IndexOf(ModuleName);
  if I = -1 then
    raise TTitanScriptFailure.Create(Format(errModuleNotFound, [ModuleName]));
  Module := Modules.Items[I];
  Module.Text := Module.Text + Code;
end;

procedure TTitanBaseScripter.AddCodeFromFile(const ModuleName, FileName: string);
var
  L: TStringList;
  I: integer;
  Module: TTitanModule;
begin
  if not FileExists(FileName) then
    Exit;
  L := TStringList.Create;
  try
    L.LoadFromFile(FileName);
    AddCode(ModuleName, L.Text);
    I := Modules.IndexOf(ModuleName);
    Module := Modules.Items[I];
    Module.FileName := FileName;
  finally
    L.Free;
  end;
end;

procedure TTitanBaseScripter.CopyLevelStack(Parser: TTitanParser);
var
  I, ID, L: integer;
begin
  Parser.LevelStack.Clear;
  Parser.LevelStack.Push(0);

  for I := 1 to Code.LevelStack.Card do
  begin
    ID := Code.LevelStack[I];
    if ID > 0 then
    begin
      L := SymbolTable.Level[ID];
      if SymbolTable.Kind[L] = KindTYPE then
        Parser.LevelStack.Push(L);
      Parser.LevelStack.Push(ID);
    end;
  end;

  if Parser.LevelStack.Card = 1 then
    Parser.LevelStack.Push(SymbolTable.RootNamespaceID);
end;

function TTitanBaseScripter.Eval(const SourceCode: string; Parser: TTitanParser): variant;
var
  StartPos, ID, TempCount: integer;
  Success: boolean;
  CodeStackCard: integer;
begin
  Parser.SetScripter(Self);

  Inc(EvalCount);
  StartPos := Code.Card;

  Code.SaveState;
  SymbolTable.SaveState;

  Success := True;
  Parser.Scanner.SourceCode := SourceCode;

  CopyLevelStack(Parser);

  TempCount := Parser.TempCount;
  Parser.UsingList.CopyFrom(Code.UsingList);
  Parser.WithStack.CopyFrom(Code.WithStack);
  ID := Parser.NewVar;
  try
    Parser.Call_SCANNER;
    Parser.Gen(OP_ASSIGN, ID, Parser.Parse_EvalExpression, ID);
  except
    Success := False;
  end;
  Parser.TempCount := TempCount;

  CodeStackCard := Code.Stack.Card;
  if Success then
  begin
    Parser.Gen(OP_HALT, 0, 0, 0);

    Code.DeclareOn := False;

    Code.N := StartPos;
    Code.SignFOP := False;
    Code.Terminated := False;
    Code.Run;

    Result := SymbolTable.VariantValue[ID];
  end;

  SymbolTable.RestoreState;
  Code.RestoreState;
  Code.Stack.Card := CodeStackCard;

  Dec(EvalCount);
end;

procedure TTitanBaseScripter.EvalStatementList(const SourceCode: string; Parser: TTitanParser);
var
  StartPos, TempCount: integer;
  Success: boolean;
  CodeStackCard, TempCodeCard, TempSymbolCard, TempClassCount: integer;
begin
  Parser.SetScripter(Self);

  Inc(EvalCount);
  StartPos := Code.Card;

  TempCodeCard := Code.Card;
  TempSymbolCard := SymbolTable.Card;
  TempClassCount := ClassList.Count;

  Code.SaveState;
  SymbolTable.SaveState;

  Success := True;
  Parser.Scanner.SourceCode := SourceCode;

  CopyLevelStack(Parser);

  TempCount := Parser.TempCount;
  Parser.UsingList.CopyFrom(Code.UsingList);
  Parser.WithStack.CopyFrom(Code.WithStack);
  try
    Parser.Call_SCANNER;
    Parser.Parse_StmtList;
  except
    Success := False;
  end;
  Parser.TempCount := TempCount;

  Dump();

  CodeStackCard := Code.Stack.Card;
  if Success then
  begin
    Parser.Gen(OP_HALT, 0, 0, 0);

    Code.DeclareOn := False;

    Code.LinkGoTo(TempCodeCard + 1, Code.Card);
    ClassList.CreateClassObjects(TempClassCount);
    ClassList.InitStaticFields(TempClassCount);
    SymbolTable.SetupSubs(TempSymbolCard + 1);

    Code.N := StartPos;
    Code.SignFOP := False;
    Code.Terminated := False;
    Code.Run;
  end;

  SymbolTable.RestoreState;
  Code.RestoreState;
  Code.Stack.Card := CodeStackCard;

  Dec(EvalCount);
end;

function TTitanBaseScripter.CompileModule(const Name: string; Parser: TTitanParser;
  SyntaxCheckOnly: boolean = False): boolean;
var
  I, ID: integer;
  Module: TTitanModule;
begin
  Result := True;
  Parser.SetScripter(Self);

  AddDefs;

  I := Modules.IndexOf(Name);
  if I = -1 then
    Exit;

  Module := Modules.Items[I];

  Module.S1 := SymbolTable.Card + 1;
  Module.P1 := Code.Card + 1;
  Module.C1 := ClassList.Count + 1;

  Parser.ModuleID := I;
  Parser.Scanner.SourceCode := Module.Text;
  Parser.SyntaxCheckOnly := SyntaxCheckOnly;

  with Parser do
  begin
    Gen(OP_BEGIN_WITH, WithStack.Push(SymbolTable.RootNamespaceID), 0, 0);

    UsingList.Push(SymbolTable.RootNamespaceID);

    if SignLoadOnDemand then
      ID := ClassList.FindClassByName(LanguageName + 'Namespace').ClassId
    else
      ID := LookUpID(LanguageName + 'Namespace');

    if ID <> 0 then
    begin
      Gen(OP_USE_NAMESPACE, UsingList.Push(ID), 0, 0);
      Gen(OP_USE_LANGUAGE_NAMESPACE, ID, 0, 0);
    end;
    Gen(OP_SEPARATOR, I, 0, 0);
  end;

  try
    Parser.Call_SCANNER;
    Parser.Parse_Program;

    if Parser.Scanner.DefStack.Count > 0 then
      raise TTitanScriptFailure.Create(errMissingENDIFdirective);

    Dump();
  except
    on E: Exception do
    begin
      Result := False;
      Code.N := Code.Card;
      CreateErrorObject(E.Message, Parser.Scanner.PosNumber);

      Dump;
      if not SyntaxCheckOnly then
      begin
        if Assigned(fOnShowError) then
          fOnShowError(Owner)
        else
          ShowError;
      end;
    end;
  end;
{
  ClassRec := ClassList[0];
  for I:=0 to ClassRec.MemberList.Count - 1 do
  begin
    MemberRec := ClassRec.MemberList[I];
    if MemberRec.Kind = KindVAR then
      if (MemberRec.ID >= Module.S1) and (MemberRec.ID <= SymbolTable.Card) then
         Parser.Gen(OP_DESTROY_LOCAL_VAR, MemberRec.ID, 0, 0);
  end;
}
  //  for I:=1 to Parser.UsingList.Card do
  //    Parser.Gen(OP_END_OF_NAMESPACE, Parser.UsingList[I], 0, 0);
  Parser.Gen(OP_HALT, 0, 0, 0);

  Module.S2 := SymbolTable.Card;
  Module.P2 := Code.Card;
  Module.C2 := ClassList.Count;

  if not IsError then
  begin
    Code.LinkGoTo(Module.P1, Module.P2);
    Code.Optimization(Module.P1, Module.P2);
  end;
end;

procedure TTitanBaseScripter.CheckForUndeclared;
var
  I, ID, N, P: integer;
  S: string;
begin
  for I := 0 to UnknownTypes.Count - 1 do
  begin
    ID := UnknownTypes[I].ID;
    ID := TypeAliasList.Convert(ID);

    N := UnknownTypes[I].N;
    P := UnknownTypes[I].Pos;
    S := SymbolTable.Name[ID];

    if (ID >= 0) and (ID < PaxTypes.Count) then
      Continue;

    if ClassList.FindClassByName(S) = nil then
    begin
      S := Format(errUndeclaredIdentifier, [S]);
      Code.N := N;
      CreateErrorObject(S, P);
      Exit;
    end;
  end;

  CheckCalls;
end;

function TTitanBaseScripter.MatchAssignment(T1, T2: integer): boolean;
begin
  Result := True;

  if T1 = T2 then
    Exit;

  if T1 = typeVARIANT then
    Exit;

  if T2 = typeVARIANT then
    Exit;

  if T1 > PAXTypes.Count then
    Exit;

  if T2 > PAXTypes.Count then
    Exit;

  if T1 <= 0 then
    Exit;

  if T2 <= 0 then
    Exit;

  if AssResultTypes[T1, T2] then
    Exit;

  Result := False;
end;

function TTitanBaseScripter.OpResultType(T1, T2: integer): integer;
begin
  Result := 0;
  if T1 > PAXTypes.Count then
    Exit;
  if T2 > PAXTypes.Count then
    Exit;
  Result := OpResultTypes[T1, T2];
  if Result = 0 then
    Result := OpResultTypes[T2, T1];
end;

function TTitanBaseScripter.MatchTypes(T1, T2: integer): integer;
begin
  if T1 = T2 then
    Result := T1
  else if T1 = typeVARIANT then
    Result := T1
  else if T2 = typeVARIANT then
    Result := T2
  else
    Result := OpResultType(T1, T2);
end;

function TTitanBaseScripter.MatchAssignmentStrict(T1, T2: integer): boolean;
begin
  if T1 = T2 then
    Result := True
  else if T1 > PAXTypes.Count then
    Result := False
  else if T2 > PAXTypes.Count then
    Result := False
  else
    Result := AssResultTypes[T1, T2];
end;

function TTitanBaseScripter.strIncompatibleTypes(T1, T2: integer): string;
begin
  Result := Format(errIncompatibleTypesExt, [SymbolTable.Name[T1], SymbolTable.Name[T2]]);
end;

procedure TTitanBaseScripter.CheckCalls;
var
  Assoc1, Assoc2, Assoc3, Assoc4: TTitanIds;
  Members: TList;
  UsingList: TTitanUsingList;
  WithStack: TTitanWithStack;

  JSOpers: boolean;

  function A1A2(ID: integer): integer;
  var
    I: integer;
  begin
    ID := TypeAliasList.Convert(ID);

    I := Assoc1.IndexOf(ID);
    if I <> -1 then
      Result := Assoc2[I]
    else
      Result := ID;
  end;

  function A2A1(ID: integer): integer;
  var
    I: integer;
  begin
    ID := TypeAliasList.Convert(ID);

    I := Assoc2.IndexOf(ID);
    if I <> -1 then
      Result := Assoc1[I]
    else
      Result := ID;
  end;

  function A3A4(ID: integer): integer;
  var
    I: integer;
  begin
    ID := TypeAliasList.Convert(ID);

    I := Assoc3.IndexOf(ID);
    if I <> -1 then
      Result := Assoc4[I]
    else
      Result := ID;
  end;

  function GetMemberRec(SubID: integer; L: TList): TTitanMemberRec;
  var
    MemberRec: TTitanMemberRec;
    I: integer;
  begin
    Result := nil;
    for I := 0 to L.Count - 1 do
    begin
      MemberRec := TTitanMemberRec(L[I]);
      if MemberRec.ID = SubID then
      begin
        Result := MemberRec;
        Exit;
      end;
    end;
  end;

  function HasComponent(Instance: TObject; const S: string; var ClsType: string): boolean;
  var
    C: TComponent;
    I: integer;
  begin
    Result := False;
    if Instance = nil then
      Exit;
    if not Instance.InheritsFrom(TComponent) then
      Exit;

    C := TComponent(Instance);

    for I := 0 to C.ComponentCount - 1 do
      if StrEql(C.Components[I].Name, S) then
      begin
        clsType := C.Components[I].ClassName;
        Result := True;
        Exit;
      end;
  end;

  function EvalID(ID: integer; UpCaseOn: boolean; var IsPublished: boolean; var IsVariant: boolean): TTitanMemberRec;
  var
    I, J, K, ObjectID, ClassID: integer;
    ClassRec: TTitanClassRec;
    D, DDest: TTitanDefinition;
    Instance: TObject;
    clsType: string;

    DefList: TTitanDEfinitionList;
    DOB: TTitanObjectDefinition;
    DI: TTitanInterfaceVarDefinition;
    CR: TTitanClassRec;

    S: string;
    L: TList;
  begin
    Result := nil;
    IsPublished := False;
    IsVariant := False;

    for I := WithStack.Card downto 1 do
    begin
      ObjectID := WithStack[I];
      K := SymbolTable.Kind[ObjectID];
      if K = KindTYPE then
      begin
        ClassRec := ClassList.FindClass(ObjectID);
        if ClassRec <> nil then
        begin
          Result := ClassRec.MemberList.GetMemberRec(SymbolTable.Name[ID], UpCaseON);
          if Result <> nil then
            Exit;
          if ClassRec.HasPublishedProperty(SymbolTable.Name[ID]) then
          begin
            IsPublished := True;
            Exit;
          end;
          if (ObjectId = SymbolTable.RootNamespaceID) and SignLoadOnDemand then
          begin
            CR := ClassList.FindClassByName(SymbolTable.Name[ID]);
            if CR <> nil then
            begin
              Result := ClassRec.MemberList.GetMemberRecByID(CR.ClassID);
              if Result <> nil then
                Exit;
            end;
          end;
        end;
      end
      else if K = KindHOSTOBJECT then
      begin
        J := SymbolTable.GetVariant(ObjectID);

        if J > 0 then
        begin
          DefList := GlobalDefinitionList;
          DOB := TTitanObjectDefinition(DefList[J]);
        end
        else
        begin
          DefList := LocalDefinitions;
          DOB := TTitanObjectDefinition(DefList.GetRecordByIndex(-J));
        end;

        Instance := DOB.Instance;

        if HasComponent(Instance, SymbolTable.Name[ID], clsType) then
        begin
          IsPublished := True;
          Exit;
        end;
        ClassRec := ClassList.FindClassByName(Instance.ClassName);
        if ClassRec <> nil then
        begin
          Result := ClassRec.MemberList.GetMemberRec(SymbolTable.Name[ID], UpCaseON);
          if Result <> nil then
            Exit;
          if ClassRec.HasPublishedProperty(SymbolTable.Name[ID]) then
          begin
            IsPublished := True;
            Exit;
          end;
        end;

      end
      else if K = KindVIRTUALOBJECT then
      begin
        SymbolTable.GetVariant(ObjectID);
      end
      else if K = KindHOSTINTERFACEVAR then
      begin
        J := SymbolTable.GetVariant(ObjectID);

        if J > 0 then
        begin
          DefList := GlobalDefinitionList;
          DI := TTitanInterfaceVarDefinition(DefList[J]);
        end
        else
        begin
          DefList := LocalDefinitions;
          DI := TTitanInterfaceVarDefinition(DefList.GetRecordByIndex(-J));
        end;

        ClassRec := ClassList.FindClassByGuid(DI.Guid);
        if ClassRec <> nil then
        begin
          Result := ClassRec.MemberList.GetMemberRec(SymbolTable.Name[ID], UpCaseON);
          if Result <> nil then
            Exit;
        end;
      end
      else if K = KindVAR then
      begin
        if ObjectID < 0 then
        begin
          _BeginRead;
          try
            D := GlobalDefinitionList.GetRecordByIndex(-ObjectID);
          finally
            _EndRead;
          end;
          if D.DefKind = dkObject then
          begin
            Instance := TTitanObjectDefinition(D).Instance;
            if HasComponent(Instance, SymbolTable.Name[ID], clsType) then
            begin
              IsPublished := True;
              Exit;
            end;
            ClassRec := ClassList.FindClassByName(Instance.ClassName);
            if ClassRec <> nil then
            begin
              Result := ClassRec.MemberList.GetMemberRec(SymbolTable.Name[ID], UpCaseON);
              if Result <> nil then
                Exit;
            end;
          end;
        end;

        ClassID := SymbolTable.PType[ObjectID];
        ClassID := A1A2(ClassID);
        ClassRec := ClassList.FindClass(ClassID);

        if ClassID = typeVARIANT then
          IsVariant := True;

        if ClassRec <> nil then
        begin
          if ClassRec.HasRunTimeProperties then
            IsVariant := True;

          Result := ClassRec.MemberList.GetMemberRec(SymbolTable.Name[ID], UpCaseON);
          if Result <> nil then
            Exit;
          if ClassRec.HasPublishedProperty(SymbolTable.Name[ID]) then
          begin
            IsPublished := True;
            Exit;
          end;
        end;
      end;
    end;

    for I := UsingList.Card downto 1 do
    begin
      ObjectID := UsingList[I];
      ClassRec := ClassList.FindClass(ObjectID);
      if ClassRec <> nil then
      begin
        Result := ClassRec.MemberList.GetMemberRec(SymbolTable.Name[ID], UpCaseON);
        if Result <> nil then
          Exit;
        if ClassRec.HasPublishedProperty(SymbolTable.Name[ID]) then
        begin
          IsPublished := True;
          Exit;
        end;
      end;
    end;

    if Result = nil then
      if SignLoadOnDemand then
      begin
        S := SymbolTable.Name[ID];
        for I := UsingList.Card downto 1 do
        begin
          ObjectID := UsingList[I];
          if ObjectID < 0 then
          begin
            D := GlobalDefinitionList.GetRecordByIndex(-ObjectID);
            L := GlobalDefinitionList.LookupAll(S, dkAny, D);
            for J := 0 to L.Count - 1 do
            begin
              DDest := TTitanDefinition(L[J]);
              ClassRec := ClassList.FindClass(ObjectID);

              if DDest.DefKind = dkConstant then
                Result := ClassRec.AddHostConstant(DDest as TTitanConstantDefinition)
              else if DDest.DefKind = dkVariable then
                Result := ClassRec.AddHostVariable(DDest as TTitanVariableDefinition)
              else if DDest.DefKind = dkObject then
                Result := ClassRec.AddHostObject(DDest as TTitanObjectDefinition)
              else if DDest.DefKind = dkVirtualObject then
                Result := ClassRec.AddVirtualObject(DDest as TTitanVirtualObjectDefinition)
              else if DDest.DefKind = dkMethod then
                Result := ClassRec.AddHostMethod(DDest as TTitanMethodDefinition);

              break;
            end;
            L.Free;
          end;
        end;

        if Result = nil then
        begin
          L := GlobalDefinitionList.LookupAll(S, dkAny, nil);
          for I := 0 to L.Count - 1 do
          begin
            DDest := TTitanDefinition(L[I]);

            ClassRec := ClassList[0];

            if DDest.DefKind = dkConstant then
              Result := ClassRec.AddHostConstant(DDest as TTitanConstantDefinition)
            else if DDest.DefKind = dkVariable then
              Result := ClassRec.AddHostVariable(DDest as TTitanVariableDefinition)
            else if DDest.DefKind = dkObject then
              Result := ClassRec.AddHostObject(DDest as TTitanObjectDefinition)
            else if DDest.DefKind = dkVirtualObject then
              Result := ClassRec.AddVirtualObject(DDest as TTitanVirtualObjectDefinition)
            else if DDest.DefKind = dkMethod then
              Result := ClassRec.AddHostMethod(DDest as TTitanMethodDefinition);
          end;
          L.Free;
        end;
      end;
  end;

  procedure AssignBinaryOperator(const OperName: string; I: integer);
  var
    T1, T2, ResType, SubID, ResID: integer;
    ClassRec: TTitanClassRec;
  begin
    T1 := A1A2(SymbolTable.PType[Code.Prog[I].Arg1]);
    T2 := A1A2(SymbolTable.PType[Code.Prog[I].Arg2]);
    if T1 > PaxTypes.Count then
    begin
      ClassRec := ClassList.FindClass(T1);
      if ClassRec <> nil then
        Code.Prog[I].AltArg1 := ClassRec.FindBinaryOperatorID(OperName, T1, T2);
    end
    else
    if T2 > PaxTypes.Count then
    begin
      ClassRec := ClassList.FindClass(T2);
      if ClassRec <> nil then
        Code.Prog[I].AltArg1 := ClassRec.FindBinaryOperatorID(OperName, T1, T2);
    end;

    if Code.Prog[I].AltArg1 = 0 then
    begin
      if JSOpers then
        Exit;

      ResType := MatchTypes(T1, T2);
      if ResType > 0 then
      begin
        if RelationalOperators.IndexOf(OperName) <> -1 then
          SymbolTable.PType[Code.Prog[I].Res] := typeBOOLEAN
        else
          SymbolTable.PType[Code.Prog[I].Res] := ResType;
      end
      else
      begin
        Dump();
        Code.N := I;
        CreateErrorObject(strIncompatibleTypes(T1, T2), Code.Prog[I].LinePos);
      end;
    end
    else
    begin
      SubID := Code.Prog[I].AltArg1;
      ResID := SymbolTable.GetResultID(SubID);
      ResType := SymbolTable.PType[ResID];
      SymbolTable.PType[Code.Prog[I].Res] := ResType;
    end;
  end;

  procedure AssignUnaryOperator(const OperName: string; I: integer);
  var
    T1, ResType, SubID, ResID: integer;
    ClassRec: TTitanClassRec;
  begin
    T1 := A1A2(SymbolTable.PType[Code.Prog[I].Arg1]);
    if T1 > PaxTypes.Count then
    begin
      ClassRec := ClassList.FindClass(T1);
      if ClassRec <> nil then
        Code.Prog[I].AltArg1 := ClassRec.FindUnaryOperatorID(OperName, T1);
    end;
    if Code.Prog[I].AltArg1 = 0 then
    begin
      if JSOpers then
        Exit;
      SymbolTable.PType[Code.Prog[I].Res] := SymbolTable.PType[Code.Prog[I].Arg1];
    end
    else
    begin
      SubID := Code.Prog[I].AltArg1;
      ResID := SymbolTable.GetResultID(SubID);
      ResType := SymbolTable.PType[ResID];
      SymbolTable.PType[Code.Prog[I].Res] := ResType;
    end;
  end;

var
  Index, I, J, K, ID, NP, NP1, SubID, ClassID, ExprID, ParamID, T1, T2: integer;
  ClassRec: TTitanClassRec;
  Ids: TTitanIds;
  S: string;
  D: TTitanDefinition;
  MemberRec: TTitanMemberRec;
  SubCount, BoundID: integer;
  DeclareON: boolean;
  UpCaseON: boolean;
  ma: TTitanMemberAccess;
  IsPublished: boolean;
  IsVariant: boolean;
  Instance: TObject;
  OP: integer;
  DM: TTitanMethodDefinition;
  I1: integer;
  DP: TTitanPropertyDefinition;
  S1, S2, clsType: string;
  C1, C2: TTitanClassRec;
  DC1, DC2, DelphiClassEx: TClass;
  K1: integer;
  DOB: TTitanObjectDefinition;
  IsOverloaded: boolean;
  Module1, Module2: integer;
  declared: boolean;
  ResId: integer;
label
  Del, patch_label;
begin
  Dump();

  JSOpers := False;
  UsingList := TTitanUsingList.Create;
  WithStack := TTitanWithStack.Create;
  Ids := TTitanIds.Create(False);
  Assoc1 := TTitanIds.Create(True);
  Assoc2 := TTitanIds.Create(True);
  Assoc3 := TTitanIds.Create(True);
  Assoc4 := TTitanIds.Create(True);
  Members := TList.Create;

  DeclareON := True;
  UpCaseON := True;
  IsPublished := False;

  try

    for I := 1 to Code.Card do
    begin
      Index := -1;

      OP := Code.Prog[I].Op;

      if (OP = OP_CALL) or (OP = OP_CALL_CONSTRUCTOR) then
        Index := CallRecList.IndexOf(I)
      else if OP = OP_SET_TYPE then
      begin
        S := _GetName(Code.Prog[I].Arg2, Self);
        ClassRec := ClassList.FindClassByName(S);
        if ClassRec = nil then
        begin
          if DeclareON then
          begin
            Dump();
            S := Format(errUndeclaredIdentifier, [S]);
            Code.N := I;
            CreateErrorObject(S, Code.Prog[I].LinePos);
            Exit;
          end;
        end
        else
        begin
          ID := Code.Prog[I].Arg1;
          SymbolTable.PType[ID] := ClassRec.ClassID;
        end;
      end
      else if (OP = OP_PLUS) or (OP = OP_PLUS_EX) then
      begin
        AssignBinaryOperator('+', I);
        if IsError then
          Exit;
      end
      else if (OP = OP_UNARY_PLUS) then
      begin
        AssignUnaryOperator('+', I);
        if IsError then
          Exit;
      end
      else if (OP = OP_MINUS) or (OP = OP_MINUS_EX) then
      begin
        AssignBinaryOperator('-', I);
        if IsError then
          Exit;
      end
      else if (OP = OP_UNARY_MINUS) or (OP = OP_UNARY_MINUS_EX) then
      begin
        AssignUnaryOperator('-', I);
        if IsError then
          Exit;
      end
      else if (OP = OP_MULT) or (OP = OP_MULT_EX) then
      begin
        AssignBinaryOperator('*', I);
        if IsError then
          Exit;
      end
      else if (OP = OP_DIV) or (OP = OP_DIV_EX) then
      begin
        AssignBinaryOperator('/', I);
        if IsError then
          Exit;
      end
      else if (OP = OP_MOD) or (OP = OP_MOD_EX) then
      begin
        AssignBinaryOperator('%', I);
        if IsError then
          Exit;
      end
      else if (OP = OP_LEFT_SHIFT) or (OP = OP_LEFT_SHIFT_EX) then
      begin
        AssignBinaryOperator('<<', I);
        if IsError then
          Exit;
      end
      else if (OP = OP_RIGHT_SHIFT) or (OP = OP_RIGHT_SHIFT_EX) then
      begin
        AssignBinaryOperator('>>', I);
        if IsError then
          Exit;
      end
      else if (OP = OP_EQ) or (OP = OP_EQ_EX) then
      begin
        AssignBinaryOperator('==', I);
        if IsError then
          Exit;
      end
      else if (OP = OP_NE) or (OP = OP_NE_EX) then
      begin
        AssignBinaryOperator('!=', I);
        if IsError then
          Exit;
      end
      else if (OP = OP_LT) or (OP = OP_LT_EX) then
      begin
        AssignBinaryOperator('<', I);
        if IsError then
          Exit;
      end
      else if (OP = OP_GT) or (OP = OP_GT_EX) then
      begin
        AssignBinaryOperator('>', I);
        if IsError then
          Exit;
      end
      else if (OP = OP_LE) or (OP = OP_LE_EX) then
      begin
        AssignBinaryOperator('<=', I);
        if IsError then
          Exit;
      end
      else if (OP = OP_GE) or (OP = OP_GE_EX) then
      begin
        AssignBinaryOperator('>=', I);
        if IsError then
          Exit;
      end
      else if OP = OP_JS_OPERS_ON then
        JSOpers := True
      else if OP = OP_JS_OPERS_OFF then
        JSOpers := False
      else if OP = OP_DECLARE_ON then
        DeclareON := True
      else if OP = OP_DECLARE_OFF then
        DeclareON := False
      else if OP = OP_UPCASE_ON then
        UpcaseON := True
      else if OP = OP_UPCASE_OFF then
        UpcaseON := False
      else if OP = OP_BEGIN_WITH then
      begin
        ID := Code.Prog[I].Arg1;
        ID := A1A2(ID);
        ID := A3A4(ID);
        WithStack.Push(ID);
      end
      else if OP = OP_END_WITH then
        WithStack.Pop
      else if OP = OP_USE_NAMESPACE then
      begin
        ID := Code.Prog[I].Arg1;
        ID := A1A2(ID);
        UsingList.PushUnique(ID);
      end
      else if OP = OP_END_OF_NAMESPACE then
      begin
        ID := Code.Prog[I].Arg1;
        ID := A2A1(ID);
        UsingList.Delete(ID);
      end
      else if OP = OP_CHECK_CLASS then
      begin
        if Code.Prog[I + 1].Op = OP_CREATE_OBJECT then
        begin
          ID := Code.Prog[I + 1].Arg1;
          ClassID := A1A2(ID);
          Code.Prog[I].Op := OP_NOP;
          ClassRec := ClassList.FindClass(ClassID);
          if ClassRec <> nil then
            if ClassRec.ck = ckClass then
            begin
              Code.Prog[I + 1].Op := OP_NOP;
              Code.Prog[I + 1].Arg1 := 0;
              Code.Prog[I + 1].Res := 0;
            end;
        end;
      end
      else if OP = OP_CREATE_RESULT then
      begin
        ID := Code.Prog[I].Arg1;
        ID := A1A2(ID);
        if (ID >= 0) and (ID < PaxTypes.Count - 1) then
        begin
          Code.Prog[I].Op := OP_NOP;
          Code.Prog[I].Arg1 := 0;
          Code.Prog[I].Res := 0;
        end
        else
          Code.Prog[I].Arg1 := ID;
      end
      else if OP = OP_CREATE_OBJECT then
      begin
        ID := Code.Prog[I].Arg1;
        ID := A1A2(ID);
        if (ID >= 0) and (ID < PaxTypes.Count - 1) then
        begin
          Code.Prog[I].Op := OP_NOP;
          Code.Prog[I].Arg1 := 0;
          Code.Prog[I].Res := 0;
        end
        else
          Code.Prog[I].Arg1 := ID;
      end
      else if OP = OP_ASSIGN_SIMPLE then
      begin
        //        Code.Prog[I].Op := OP_ASSIGN;
      end
      else if OP = OP_ASSIGN then
        with Code.Prog[I], SymbolTable do
        begin
          if (Kind[Arg1] = KindHostConst) then
          begin
            Self.Dump();
            Code.N := I;
            S := errLeftSideCannotBeAssignedTo;
            CreateErrorObject(S, Code.Prog[I].LinePos);
            Exit;
          end;

          ClassID := PType[Arg2];
          ClassID := A1A2(ClassID);

          if not MatchAssignment(A1A2(PType[Arg1]), ClassID) then
          begin
            Self.Dump();
            Code.N := I;
            S := errIncompatibleTypes;
            CreateErrorObject(S, Code.Prog[I].LinePos);
            Exit;
          end;

          if ClassID <> typeVARIANT then
          begin
            if (PType[Arg1] = typeVARIANT) or (PType[Arg1] = typeVOID) then
              if not SymbolTable.IsFormalParamID(Arg1) then
              begin
                declared := False;
                for J := 1 to I do
                  if (Code.Prog[J].Op = OP_DECLARE) and (Code.Prog[J].Arg1 = Arg1) then
                  begin
                    declared := True;
                    break;
                  end;

                if not declared then
                  PType[Arg1] := ClassID;
              end;
          end;
        end
      else if OP = OP_CREATE_REF then
      begin
        ID := Code.Prog[I].Arg1;
        ID := A1A2(ID);
        ID := A3A4(ID);

        Instance := nil;

        if ID < 0 then
        begin
          D := GlobalDefinitionList.GetRecordByIndex(-ID);
          if D.DefKind = dkObject then
          begin
            Instance := TTitanObjectDefinition(D).Instance;

            S := Instance.ClassName;
            ClassRec := ClassList.FindClassByName(S);
          end
          else if D.DefKind = dkInterface then
          begin
            ClassRec := ClassList.FindClassByGuid(TTitanInterfaceVarDefinition(D).guid);
          end
          else if D.DefKind = dkClass then
          begin
            ClassRec := ClassList.FindClassByName(D.Name);
          end
          else if D.DefKind = dkMethod then
          begin
            ClassRec := ClassList.FindClassByName(TTitanMethodDefinition(D).StrTypes[0]);
          end
          else
            ClassRec := nil;
        end
        else
        begin
          if SymbolTable.Kind[ID] = KindTYPE then
            ClassID := ID
          else
          begin
            ClassID := SymbolTable.PType[ID];
            ClassID := A1A2(ClassID);
          end;
          ClassRec := ClassList.FindClass(ClassID);

          if SymbolTable.Kind[ID] = KindHOSTOBJECT then
            with SymbolTable do
            begin
              J := GetVariant(ID);
              if J > 0 then
                DOB := TTitanObjectDefinition(GlobalDefinitionList[J])
              else
                DOB := TTitanObjectDefinition(LocalDefinitions.GetRecordByIndex(-J));
              Instance := DOB.Instance;
            end
          else if SymbolTable.Kind[ID] = KindVIRTUALOBJECT then
            with SymbolTable do
            begin
              A[Code.Prog[I].Res].Owner := ID;
            end;
        end;

        ma := TTitanMemberAccess(Code.Prog[I].Arg2);
        if ma = maMyBase then
          if ClassRec <> nil then
            ClassRec := ClassList.FindClassByName(ClassRec.AncestorName);

        if ClassRec <> nil then
        begin
          ID := Code.Prog[I].Res;
          MemberRec := ClassRec.MemberList.GetMemberRec(SymbolTable.Name[ID], UpCaseON);

          if MemberRec <> nil then
          begin
            if MemberRec.ID > 0 then
            begin
              Module1 := Code.GetModuleID(I);
              Module2 := SymbolTable.Module[MemberRec.ID];
              if Module1 <> Module2 then
                if MemberRec.IsImplementationSection then
                begin
                  S := SymbolTable.Name[MemberRec.ID];
                  S := Format(errMemberIsNotAccessible, [S]);
                  Code.N := I;
                  CreateErrorObject(S, Code.Prog[I].LinePos);
                  break;
                end;
            end;

            if (MemberRec.Kind = KindPROP) and (MemberRec.Definition <> nil) then
            begin
              DP := TTitanPropertyDefinition(MemberRec.Definition);
              ClassRec := ClassList.FindClassByName(DP.PropType);
              if ClassRec <> nil then
              begin
                Assoc3.Add(ID);
                Assoc4.Add(ClassRec.ClassID);
              end;
            end
            else
            begin
              S := SymbolTable.Name[ID];

              DelphiClassEx := ClassRec.DelphiClassEx;
              if DelphiClassEx <> nil then
              begin
                if HasPublishedProperty(DelphiClassEx, S, Self, False) then
                  if HasForbiddenPublishedProperty(DelphiClassEx, S) then
                  begin
                    S := Format(errMemberIsNotAccessible, [S]);
                    Code.N := I;
                    CreateErrorObject(S, Code.Prog[I].LinePos);
                    break;
                  end;
              end;

              if ClassRec.HasPublishedPropertyEx(S, clsType) then
              begin
                ClassRec := ClassList.FindClassByName(clsType);
                if ClassRec <> nil then
                begin
                  Assoc3.Add(ID);
                  Assoc4.Add(ClassRec.ClassID);
                  SymbolTable.PType[ID] := ClassRec.ClassID;
                end;
              end
              else
              begin
                Assoc1.Add(ID);
                Assoc2.Add(MemberRec.ID);
                Members.Add(MemberRec);
              end;

              S := SymbolTable.Name[ID];
              if StrEql(S, 'Create') then
                SymbolTable.PType[ID] := ClassRec.ClassID;

            end;
          end
          else
          begin
            S := SymbolTable.Name[ID];
            if StrEql(S, 'NEW') or StrEql(S, ClassRec.Name) or StrEql(ClassRec.Name, 'ACTIVEXOBJECT') or
              StrEql(ClassRec.Name, 'OBJECT') or StrEql(ClassRec.Name, 'STRING') or
              StrEql(ClassRec.Name, 'ARRAY') or ClassRec.HasRunTimeProperties then
            begin
              SymbolTable.PType[ID] := ClassRec.ClassID;
            end
            else if Pos('$$', S) > 0 then
            begin
            end
            else if ClassRec.HasPublishedPropertyEx(S, clsType) then
            begin
              ClassRec := ClassList.FindClassByName(clsType);
              if ClassRec <> nil then
              begin
                Assoc3.Add(Code.Prog[I].Res);
                Assoc4.Add(ClassRec.ClassID);
              end;
            end
            else if HasComponent(Instance, S, clsType) then
            begin
              ClassRec := ClassList.FindClassByName(clsType);
              if ClassRec <> nil then
              begin
                Assoc3.Add(Code.Prog[I].Res);
                Assoc4.Add(ClassRec.ClassID);
              end;
            end
            else if DeclareON and (not ClassRec.HasRunTimeProperties) then
            begin
              Dump();
              S := Format(errPropertyIsNotFound, [S]);
              Code.N := I;
              CreateErrorObject(S, Code.Prog[I].LinePos);
              Exit;
            end;
          end;
        end;
      end
      else if OP = OP_EVAL_WITH then
      begin
        if SymbolTable.Name[Code.Prog[I].Res] = '' then
        begin
          Code.Prog[I].Op := OP_NOP;
          Code.Prog[I].Res := 0;
          Continue;
        end;

        ClassID := TypeAliasList.Convert(Code.Prog[I].Res);

        if (ClassID >= 0) and (ClassID < PaxTypes.Count - 1) then
        begin
          Code.Prog[I].Op := OP_NOP;
          Code.Prog[I].Res := 0;
          Continue;
        end;

        ID := ClassID;
        Code.Prog[I].Res := ID;

        MemberRec := EvalID(ID, UpCaseOn, IsPublished, IsVariant);

        if MemberRec = nil then
          if not IsPublished then
            if DeclareON then
            begin
              if IsVariant then
                continue;

              S := SymbolTable.Name[ID];

              if MemberRec = nil then
              begin
                Dump();
                S := Format(errPropertyIsNotFound, [S]);
                Code.N := I;
                CreateErrorObject(S, Code.Prog[I].LinePos);
                break;
              end;
            end;

        if MemberRec <> nil then
          if ID <> MemberRec.ID then
          begin
            if MemberRec.ID > 0 then
              if SymbolTable.Kind[MemberRec.ID] = KindHOSTOBJECT then
              begin
                Code.Prog[I].Op := OP_NOP;
                Code.ReplaceID(Code.Prog[I].Res, MemberRec.ID);
                Dump();

                continue;
              end;

            if MemberRec.ID > 0 then
            begin
              Module1 := Code.GetModuleID(I);
              Module2 := SymbolTable.Module[MemberRec.ID];
              if Module1 <> Module2 then
                if MemberRec.IsImplementationSection then
                begin
                  S := SymbolTable.Name[MemberRec.ID];
                  S := Format(errMemberIsNotAccessible, [S]);
                  Code.N := I;
                  CreateErrorObject(S, Code.Prog[I].LinePos);
                  break;
                end;
            end;

            Assoc1.Add(ID);
            Assoc2.Add(MemberRec.ID);
            Members.Add(MemberRec);
          end;
      end;

      if Index <> -1 then
        with CallRecList[Index] do
        begin
          if (Code.Prog[CallN].Op <> OP_CALL) and (Code.Prog[CallN].Op <> OP_CALL_CONSTRUCTOR) then
            Continue;

          SubID := Code.Prog[CallN].Arg1;

          if SymbolTable.Kind[SubID] <> KindSUB then
          begin
            SubID := A1A2(SubID);
          end;

          patch_label:

            if SymbolTable.Kind[SubID] <> KindSUB then
            begin
              if SymbolTable.Kind[SubID] = KindPROP then
              begin
                ClassRec := ClassList.FindClass(SymbolTable.Level[SubID]);
                if ClassRec <> nil then
                begin
                  MemberRec := ClassRec.MemberList.GetMemberRec(SymbolTable.Name[SubID]);
                  if MemberRec <> nil then
                    if MemberRec.ReadID <> 0 then
                    begin
                      SubId := MemberRec.ReadID;
                      ResId := Code.Prog[I].Res;
                      SymbolTable.Ptype[ResId] := SymbolTable.PType[SubId];
                      goto patch_label;
                    end;
                end;
              end;

              S := SymbolTable.Name[SubID];
              if StrEql(S, 'New') then
              begin
                with SymbolTable, Code.Prog[CallN] do
                  if PType[Res] = typeVARIANT then
                    PType[Res] := PType[Arg1];
              end
              else
              begin
                ClassRec := ClassList.FindClassByName(S);
                if ClassRec <> nil then
                  with SymbolTable, Code.Prog[CallN] do
                  begin
                    if PType[Res] = typeVARIANT then
                      PType[Res] := PType[Arg1];
                    if ClassRec.IsStatic then
                      Code.Prog[CallN].Op := OP_CALL
                    else if Code.Prog[CallN].Op = OP_CALL_CONSTRUCTOR then
                      Code.Prog[CallN].Op := OP_CALL
                    else
                      Code.Prog[CallN].Op := OP_TYPE_CAST;
                  end
                else if (SubId > 0) and (SubId <= PaxTypes.Count) then
                begin
                  with SymbolTable, Code.Prog[CallN] do
                  begin
                    if PType[Res] = typeVARIANT then
                      PType[Res] := PType[Arg1];
                    Code.Prog[CallN].Op := OP_TYPE_CAST;
                  end;
                end;
              end;

              Continue;
              //          S := Format(errPropertyIsNotFound, [SymbolTable.Name[SubID]]);
              //          Code.N := CallN;
              //          CreateErrorObject(S, CallP);
              //          Exit;
            end;

          MemberRec := GetMemberRec(SubID, Members);

          NP := Code.Prog[CallN].Arg2;

          if SubID > 0 then
          begin
            ClassID := SymbolTable.Level[SubID];
            ClassRec := ClassList.FindClass(ClassID);
          end
          else
          begin
            D := GlobalDefinitionList.GetRecordByIndex(-SubID);
            if D.Owner = nil then
              ClassRec := ClassList[0]
            else
              ClassRec := ClassList.FindClassByName(D.Owner.Name);
          end;

          if ClassRec = nil then
          begin
            ClassID := SymbolTable.Level[SubID];
            if SymbolTable.Kind[ClassID] = KindSUB then
              continue; // nested subroutine

            Dump();
            Code.N := CallN;
            S := Format(errPropertyIsNotFound, [SymbolTable.Name[SubID]]);
            CreateErrorObject(S, CallP);
            Exit;
          end;

          Ids.Clear;
          ClassRec.FindOverloadedSubList(SymbolTable.NameIndex[SubID], maAny, Ids);

          SubCount := Ids.Count;

          IsOverloaded := Ids.Count > 1;


          for J := Ids.Count - 1 downto 0 do
          begin
            ID := Ids[J];

            if SymbolTable.Count[ID] = -1 then
              Continue;

            if SymbolTable.Count[ID] > NP then
            begin
              if ID > 0 then
              begin
                NP1 := NP;

                BoundID := SymbolTable.GetParamID(ID, NP1);
                K := DefaultParameterList.FindFirst(ID);
                while K <> -1 do
                begin
                  if DefaultParameterList[K].ID > BoundID then
                    Inc(NP1);
                  K := DefaultParameterList.FindNext(K, ID);
                end;
                if NP1 <> SymbolTable.Count[ID] then
                  goto Del;
              end
              else
              begin
                DM := TTitanMethodDefinition(GlobalDefinitionList.GetRecordByIndex(-ID));
                if SymbolTable.Count[ID] - DM.DefParamList.Count > NP then
                  goto Del;
              end;
            end
            else if SymbolTable.Count[ID] < NP then
              goto Del;

            for K := 0 to ArgsN.Count - 1 do
            begin
              ExprID := Code.Prog[ArgsN[K]].Arg1;
              T2 := SymbolTable.PType[ExprID];

              S1 := '';
              S2 := SymbolTable.Name[T2];

              if T2 < 0 then
              begin
                if Copy(S2, 1, 1) = 'I' then
                  T2 := typeINTERFACE
                else
                  T2 := typeCLASS;
              end;

              D := nil;

              if ID > 0 then
              begin
                ParamID := SymbolTable.GetParamID(ID, K + 1);
                T1 := SymbolTable.PType[ParamID];
              end
              else
              begin
                D := GlobalDefinitionList.GetRecordByIndex(-ID);
                if TTitanMethodDefinition(D).NP < 0 then
                  T1 := typeVARIANT
                else
                begin
                  if K < Length(TTitanMethodDefinition(D).Types) then
                  begin
                    T1 := TTitanMethodDefinition(D).Types[K];
                    S1 := TTitanMethodDefinition(D).StrTypes[K];
                  end
                  else
                    T1 := typeVARIANT;
                end;
              end;

              if StrEql(S1, S2) then
                continue;

              if not MatchAssignment(T1, T2) then
                goto Del;

              if (T1 = typeVARIANT) and (T2 = typeCLASS) and (ID < 0) and (Ids.Count > 1) then
                goto Del;


              if SymbolTable.Kind[T2] <> KindTYPE then
              begin
                ClassRec := ClassList.FindClassByName(SymbolTable.Name[T2]);
                if ClassRec <> nil then
                begin
                  if (T1 = TypeINTERFACE) and (ClassRec.ck = ckInterface) then
                    continue;

                  if (T1 > 0) and (T1 < PaxTypes.Count) then
                    if not (T1 in [typeCLASS, typeVARIANT]) then
                      goto Del;

                  if SymbolTable.Kind[T1] <> KindTYPE then
                  begin
                    ClassRec := ClassList.FindClassByName(SymbolTable.Name[T1]);
                    if ClassRec = nil then
                      goto Del;
                  end;
                end;
              end;


              if Copy(S2, 1, 2) = '__' then // script-defined array
              begin
                if D <> nil then // host proc
                begin
                  if K < Length(TTitanMethodDefinition(D).Types) then
                  begin
                    if TTitanMethodDefinition(D).ExtraTypes[K] = typeDYNAMICARRAY then
                      continue
                    else
                      goto Del;
                  end;
                end;
              end;

              if (T1 = typeCLASS) and (T2 = typeCLASS) and (S1 <> '') and (S2 <> '') and (S1 <> S2) then
              begin
                C1 := ClassList.FindClassByName(S1);
                C2 := ClassList.FindClassByName(S2);
                if (C1 <> nil) and (C2 <> nil) then
                begin
                  DC1 := C1.DelphiClass;
                  DC2 := C2.DelphiClass;
                  if (DC1 <> nil) and (DC2 <> nil) then
                    if not DC2.InheritsFrom(DC1) then
                      goto Del;
                end;
              end;

              // choose the best method
            end;

            Continue;
            Del:
              Ids.Delete(J);
            Continue;
          end;

          if IDs.Count = 0 then
          begin
            Code.N := CallN;

            if SubCount = 0 then
              S := Format(errPropertyIsNotFound, [SymbolTable.Name[SubID]])
            else if SubCount = 1 then
            begin
              K := SymbolTable.Count[SubID];

              if NP < K then
              begin
                Dump();
                S := errNotEnoughParameters;
              end
              else if NP > K then
                S := errTooManyParameters
              else
                S := errIncompatibleTypes;
            end
            else
              S := Format(errOverloadedMethodNotFound, [SymbolTable.Name[SubID]]);

            CreateErrorObject(S, CallP);
            Exit;
          end
          else if IDs.Count >= 1 then
          begin
            SubID := Ids[0];

            if Ids.Count > 1 then
            begin
              for K := 1 to Ids.Count - 1 do
                if SymbolTable.Count[SubId] = SymbolTable.Count[Ids[K]] then
                  IsOverloaded := True;
            end;

{          if MemberRec = nil then
            Code.Prog[CallN].Arg1 := SubID
          else if SymbolTable.Kind[Code.Prog[CallN].Arg1] = KindREF then
            Code.Prog[CallN].AltArg1 := SubID
          else if MemberRec.IsStatic then
            Code.Prog[CallN].Arg1 := SubID
          else }
            if SymbolTable.IsVirtual(SubId) and (not IsOverloaded) then
              Code.Prog[CallN].AltArg1 := 0
            else
            begin
              Code.Prog[CallN].AltArg1 := SubID;
            end;

            if MemberRec <> nil then
              with SymbolTable, Code.Prog[CallN] do
                if PType[Res] = typeVARIANT then
                begin
                  PType[Res] := GetTypeID(SubId);
                  //                 PType[Res] := PType[Arg1];

                  if StrEql(SymbolTable.Name[SubID], 'Create') then
                    if Code.Prog[CallN + 1].Op = OP_ASSIGN then
                      if Code.Prog[CallN + 1].Arg2 = Res then
                      begin
                        declared := False;
                        for J := 1 to I do
                          if (Code.Prog[J].Op = OP_DECLARE) and (Code.Prog[J].Arg1 = Code.Prog[CallN + 1].Arg1) then
                          begin
                            declared := True;
                            break;
                          end;

                        //                       if not declared then
                        PType[Code.Prog[CallN + 1].Arg1] := PType[Arg1];
                      end;
                end
                else
                begin
                  S := SymbolTable.Name[SubID];
                  if StrEql(S, 'New') then
                    with SymbolTable, Code.Prog[CallN] do
                      if PType[Res] = typeVARIANT then
                      begin
                        PType[Res] := PType[Arg1];
                      end;
                end;
          end
          else
          begin
            Code.N := CallN;
            S := Format(errIdentifierIsRedeclared, [SymbolTable.Name[SubID]]);
            CreateErrorObject(S, CallP);
            Exit;
          end;
        end;
    end;

  finally
    UsingList.Free;
    WithStack.Free;
    Ids.Free;
    Assoc1.Free;
    Assoc2.Free;
    Assoc3.Free;
    Assoc4.Free;
    Members.Free;
    Dump();
  end;
end;

procedure TTitanBaseScripter.CompileExtraCode;
var
  Parser: TTitanParser;
  S: string;
  ModuleName, LanguageName: string;
  I: integer;
begin
  for I := 0 to ParserList.Count - 1 do
  begin
    LanguageName := ParserList[I].LanguageName;
    S := ExtraCodeList.GetCode(LanguageName);
    if S <> '' then
    begin
      Parser := ParserList.FindParser(LanguageName);
      Parser.SetScripter(Self);

      ModuleName := '###' + LanguageName;
      AddModule(ModuleName, LanguageName);
      AddCode(ModuleName, S);

      CompileModule(ModuleName, Parser);
    end;
  end;
end;

procedure TTitanBaseScripter.Link(reallocate: boolean);
begin
  if ExtraCodeList.Count > 0 then
    CompileExtraCode;

  SymbolTable.ReallocateMem(SymbolTable.MemBoundVar + fStackSize + SymbolTable.GetSubCount *
    SymbolTable.MaxLocalVars * _SizeVariant);

  Code.SaveOP;

  ClassList.CreateClassObjects(0);
  SymbolTable.AllocateSubroutines;
  SymbolTable.SetupSubs(1);

  State := _ssReadyToRun;
end;

procedure TTitanBaseScripter.Run(RunMode: integer = _rmRun; DestLine: integer = 0);
label
  Again;
begin

  Again:

    if IsError then
      Exit;

  State := _ssRunning;

  if (Code.N = 0) and (RunList.Top = -1) then
  begin
    RunList.Pop;
    ClassList.InitStaticFields(0);
    Code.UsingList.Push(SymbolTable.RootNamespaceID);
    Code.N := RunList.Pop;

    Code.BreakPointList.InitCurrPassCounts;
  end;

  if not IsError then
  begin
    if (RunMode = _rmRun) and ((Code.BreakpointList.Count = 0) or IgnoreBreakpoints) then
      Code.DebugState := False;
    Code.Terminated := False;
    Code.Run(RunMode, DestLine);
    Code.DebugState := True;
  end;

  if IsError then
  begin
    if Assigned(fOnShowError) then
      fOnShowError(Owner)
    else
      ShowError;
  end;

  if Code.fTerminated then
  begin
    if not IsError then
      if RunList.Card > 0 then
        if RunList.Top <> -1 then
        begin
          Code.N := RunList.Pop;
          goto Again;
        end;

    State := _ssTerminated;
    Code.N := 0;
  end
  else
    State := _ssPaused;
end;

procedure TTitanBaseScripter.ShowError;
begin
  with ErrorInstance do
    ErrMessageBox(Format(errScript, [Description, ModuleName, LineNumber, Line]));
end;

function TTitanBaseScripter.GetProperty(const ScriptObject: variant; PropertyName: string): variant;
var
  SO: TTitanScriptObject;
begin
  SO := VariantToScriptObject(ScriptObject);
  Result := SO.GetProperty(CreateNameIndex(PropertyName, Self), 0);
end;

procedure TTitanBaseScripter.Dump;
begin
  if _IsDump = False then
    Exit;        //Main Dump Function ====================================

  SymbolTable.Dump('SymbolTable.dmp');
  Code.Dump('Code.dmp');
  ClassList.Dump('ClassList.dmp');
  Modules.Dump('modules.dmp');
  GlobalDefinitionList.Dump('DefinitionList.dmp');
  UnresolvedTypes.SaveToFile('unresolved.dmp');
end;

procedure TTitanBaseScripter.DiscardError;
begin
  VarClear(Error);
end;

function TTitanBaseScripter.IsError: boolean;
begin
  Result := not (VarType(Error) = varEmpty);
end;

function TTitanBaseScripter.AddBreakpoint(const ModuleName: string; SourceLineNumber: integer;
  const Condition: string; PassCount: integer): boolean;
var
  PCodeLineNumber: integer;
begin
  Result := False;

  PCodeLineNumber := SourceLineToPCodeLine(ModuleName, SourceLineNumber);
  if PCodeLineNumber = -1 then
    Exit;

  if not Code.IsExecutableLine(PCodeLineNumber) then
    Exit;

  Result := True;

  Code.BreakpointList.AddBreakpoint(PCodeLineNumber, Condition, PassCount);
end;

function TTitanBaseScripter.RemoveBreakpoint(const ModuleName: string; SourceLineNumber: integer;
  const Condition: string; PassCount: integer): boolean;
var
  PCodeLineNumber: integer;
begin
  Result := False;

  PCodeLineNumber := SourceLineToPCodeLine(ModuleName, SourceLineNumber);
  if PCodeLineNumber = -1 then
    Exit;

  if not Code.IsExecutableLine(PCodeLineNumber) then
    Exit;

  Result := Code.BreakpointList.RemoveBreakpoint(PCodeLineNumber);
end;

procedure TTitanBaseScripter.RemoveAllBreakpoints;
begin
  Code.BreakpointList.Clear;
end;

function TTitanBaseScripter.SourceLineToPCodeLine(const ModuleName: string; SourceLineNumber: integer): integer;
var
  I, ModuleID: integer;
begin
  Result := -1;

  ModuleID := Modules.IndexOf(ModuleName);
  if (ModuleID = -1) or (SourceLineNumber = 0) then
    Exit;

  for I := 1 to Code.Card do
    with Code.Prog[I] do
      if Op = OP_SEPARATOR then
        if ModuleID = Arg1 then
          if SourceLineNumber = Arg2 then
          begin
            Result := I;
            Exit;
          end;
end;

function TTitanBaseScripter.GetName(NameIndex: integer): string;
begin
  Result := NameList[NameIndex];
end;

procedure TTitanBaseScripter.CreateErrorObject(const Message: string; PosNumber: integer);
var
  ModuleID, LineID: integer;
  Module: TTitanModule;
begin
  case fState of
    _ssCompiling:
      ErrorInstance.ScriptTime := 'Compile-time';
    _ssRunning:
      ErrorInstance.ScriptTime := 'Run-time';
  end;

  ModuleID := Code.ModuleID;
  LineID := Code.LineID;
  Module := Modules.Items[ModuleID];

  ErrorInstance.Description := Message;
  ErrorInstance.MethodId := Code.CurrMethodID;

  if Module <> nil then
    if EvalCount = 0 then
    begin
      ErrorInstance.ModuleName := Module.Name;
      if Module.FileName <> '' then
        ErrorInstance.FileName := Module.FileName;
      if LineID >= 0 then
      begin
        if LineID < Module.Count then
          ErrorInstance.Line := Module.Strings[LineID];
        ErrorInstance.LineNumber := LineID + 1;
        ErrorInstance.Position := PosNumber;
        ErrorInstance.TextPosition := Module.GetTextPos(LineID + 1, PosNumber);
      end;
    end;

  with SymbolTable do
    Error := LocalDefinitions.Records[0].DefValue;
end;

procedure TTitanBaseScripter.RunEx(ExtendedRun: boolean);
var
  M: TMethod;
begin
  State := _ssRunning;
  M := OwnerEventHandlerMethod;
  if Assigned(M.Code) and ExtendedRun then
  begin
    State := _ssPaused;
    AllowedEvents := False;

  {$IFNDEF SOLARIS}

    {$IFDEF CPUARM}
        // ct9999 MUST CODE THIS
    asm
    end;
    {$ELSE}

    asm
             MOV     EAX,DWORD PTR M.Data;
             CALL    M.Code;
    end;

    {$ENDIF}
  {$ELSE} // NOT SOLARIS
         // ct9999 MUST CODE THIS
    asm
    end;
  {$ENDIF}

  end
  else
    Code.Run;

  if IsError then
  begin
    if Assigned(fOnShowError) then
      fOnShowError(Owner)
    else
      ShowError;
  end;
end;

{$define THREADS}//By Sternas Stefanos

function TTitanBaseScripter.CallMethod(SubID: integer; const This: variant; const P: array of const; IsEventHandler: boolean = False): variant;
var
  StartPos, I, ResultID, ParamID, TypeID, NP, Temp, K1, K2: integer;
  PV: PVariant;
  S: string;
  ClassRec: TTitanClassRec;
  SO: TTitanScriptObject;
  V: variant;
  _Upcase: boolean;
begin

  if SubID > 0 then
    if SymbolTable.Kind[SubID] = KindTYPE then
      while SymbolTable.Kind[SubID] <> KindSUB do
        Inc(SubID);

  StartPos := Code.Card;
  NP := Length(P);

  Code.SaveState;
  SymbolTable.SaveState;

  ResultID := SymbolTable.GetResultID(SubID);
  if SubId < 0 then
    ResultId := SymbolTable.IDundefined;

  LastResultID := ResultID;

  for I := 0 to NP - 1 do
  begin
    PV := VariantStack.Push(Undefined);
    if P[I].VType = vtVariant then
    begin
      Code.Stack.Push(P[I].VVariant);
    end
    else if P[I].VType = vtInterface then
    begin
      Code.Stack.Push(InterfaceToScriptObject(IUnknown(P[I].VInterface), Self));
    end
    else if P[I].VType = vtPointer then
    begin
      ParamID := SymbolTable.GetParamID(SubID, I + 1);
      TypeID := SymbolTable.PType[ParamID];
      S := SymbolTable.Name[TypeID];
      ClassRec := ClassList.FindClassByName(S);
      if ClassRec <> nil then
      begin
        if ClassRec.ck = ckDynamicArray then
        begin
          SO := ClassRec.CreateScriptObject;
          SO.Instance := SO;
          SO.ExtraPtr := Pointer(P[I].VPointer);
          SO.ExternalExtraPtr := True;
          PV^ := ScriptObjectToVariant(SO);
          Code.Stack.Push(PV);
          SymbolTable.ByRef[ResultID] := 1;
        end
        else if ClassRec.ck = ckStructure then
        begin
          SO := ClassRec.CreateScriptObject;
          SO.FreeExtraPtr;
          SO.Instance := SO;
          SO.ExtraPtr := P[I].VPointer;
          SO.ExternalExtraPtr := True;
          PV^ := ScriptObjectToVariant(SO);
          Code.Stack.Push(PV);
          SymbolTable.ByRef[ResultID] := 1;
        end;
      end
      else
      begin
        Code.Stack.Push(P[I].VPointer);
        SymbolTable.ByRef[ResultID] := 1;
      end;
    end
    else
    begin
      PV^ := TVarRecToVariant(P[I], Self);
      Code.Stack.Push(PV);
    end;
  end;

  Code.N := StartPos;

  Temp := Code.Prog[StartPos].Vars;
  Code.Prog[StartPos].Vars := $FFFFFFF;

  Code.SignFOP := False;
  Code.Terminated := False;
  Code.CallSub(SubID, NP, @This, ResultID);
  Code.Add(OP_HALT, 0, 0, 0);
  Dec(Code.N);

  K1 := ScriptObjectList.Count;

  _Upcase := Code.Upcaseon;
  if not _Upcase then
    Code.UpcaseOn := Code.GetUpcase(Code.Card);

  Code.UsingList.Push(Code.GetLanguageNamespaceID);
  RunEx((Code.BreakpointList.Count > 0) and (IgnoreBreakpoints = False));
  Code.UsingList.Pop;

  Code.Upcaseon := _Upcase;

  K2 := ScriptObjectList.Count;

  Result := SymbolTable.VariantValue[ResultID];

{$IFDEF WINDOWS}
  {$IFDEF THREADS}
  if K1 <> K2 then
    ScriptObjectList.RemoveTail(K1, GetCurrentThreadID());
  {$ELSE}
  if K1 <> K2 then
    ScriptObjectList.RemoveTail(K1);
  {$ENDIF}
{$ENDIF}

  fLastResult := Result;
  Code.Prog[StartPos].Vars := Temp;

  if IsEventHandler then
  begin
    State := _ssRunning;
    if Code.N <> Code.Card then
      State := _ssPaused;
    Code.fTerminated := False;

    if Code.IsAborted then
      Code.fTerminated := True;
  end
  else if (Code.Prog[Code.N].Op = OP_HALT) and (Code.SubRunCount = 0) then
    State := _ssTerminated
  else if (Code.Prog[Code.N].Op = OP_HALT) and (Code.SubRunCount > 0) then
  begin
    SymbolTable.ClearVariant(ResultID);
    State := _ssRunning;
    for I := NP - 1 downto 0 do
      VariantStack.Pop;
    SymbolTable.RestoreState;
    Code.RestoreState;

    Exit;
  end
  else
    State := _ssPaused;

  if (Code.N = Code.Card) then
    if (State = _ssTerminated) or IsEventHandler then
    begin
      for I := NP - 1 downto 0 do
      begin
        if SubId > 0 then
        begin
          ParamID := SymbolTable.GetParamID(SubID, I + 1);
          if SymbolTable.ByRef[ParamID] > 0 then
          begin
            V := VariantStack.Top;

            if P[I].VType = vtVariant then
              V := P[I].VVariant^;

            SymbolTable.PutVariant(ParamID, V);
          end;
        end;
        VariantStack.Pop;
      end;
      SymbolTable.RestoreState;
      Code.RestoreState;
    end;

  if SubId < 0 then
    SymbolTable.ClearVariant(SymbolTable.IDundefined);
end;

function TTitanBaseScripter.CallMethodEx(SubID: integer; const This: variant;
  const P: array of const; const StrTypes: array of string;
  IsEventHandler: boolean = False): variant;
var
  StartPos, I, ResultID, ParamID, TypeID, NP, Temp, K1, K2: integer;
  PV: PVariant;
  S: string;
  ClassRec: TTitanClassRec;
  SO: TTitanScriptObject;
begin
  if SubID > 0 then
    if SymbolTable.Kind[SubID] = KindTYPE then
      while SymbolTable.Kind[SubID] <> KindSUB do
        Inc(SubID);

  StartPos := Code.Card;
  NP := Length(P);

  Code.SaveState;
  SymbolTable.SaveState;

  ResultID := SymbolTable.GetResultID(SubID);
  LastResultID := ResultID;

  for I := 0 to NP - 1 do
  begin
    PV := VariantStack.Push(Undefined);
    if P[I].VType = vtVariant then
    begin
      Code.Stack.Push(P[I].VVariant);
    end
    else if P[I].VType = vtInterface then
    begin
      SO := InterfaceToScriptObject(IUnknown(P[I].VInterface), Self, StrTypes[I]);
      PV^ := ScriptObjectToVariant(SO);
      Code.Stack.Push(PV);
    end
    else if P[I].VType = vtPointer then
    begin
      ParamID := SymbolTable.GetParamID(SubID, I + 1);
      TypeID := SymbolTable.PType[ParamID];
      S := SymbolTable.Name[TypeID];
      ClassRec := ClassList.FindClassByName(S);
      if ClassRec <> nil then
      begin
        if ClassRec.ck = ckDynamicArray then
        begin
          SO := ClassRec.CreateScriptObject;
          SO.Instance := SO;
          SO.ExtraPtr := Pointer(P[I].VPointer);
          SO.ExternalExtraPtr := True;
          PV^ := ScriptObjectToVariant(SO);
          Code.Stack.Push(PV);
          SymbolTable.ByRef[ResultID] := 1;
        end
        else if ClassRec.ck = ckStructure then
        begin
          SO := ClassRec.CreateScriptObject;
          SO.FreeExtraPtr;
          SO.Instance := SO;
          SO.ExtraPtr := P[I].VPointer;
          SO.ExternalExtraPtr := True;
          PV^ := ScriptObjectToVariant(SO);
          Code.Stack.Push(PV);
          SymbolTable.ByRef[ResultID] := 1;
        end;
      end
      else
      begin
        Code.Stack.Push(P[I].VPointer);
        SymbolTable.ByRef[ResultID] := 1;
      end;
    end
    else
    begin
      PV^ := TVarRecToVariant(P[I], Self);
      Code.Stack.Push(PV);
    end;
  end;

  Code.N := StartPos;

  Temp := Code.Prog[StartPos].Vars;
  Code.Prog[StartPos].Vars := $FFFFFFF;

  Code.SignFOP := False;
  Code.Terminated := False;
  Code.CallSub(SubID, NP, @This, ResultID);
  Code.Add(OP_HALT, 0, 0, 0);
  Dec(Code.N);

  K1 := ScriptObjectList.Count;

  Code.UsingList.Push(Code.GetLanguageNamespaceID);
  RunEx(Code.BreakpointList.Count > 0);
  Code.UsingList.Pop;

  K2 := ScriptObjectList.Count;

  Result := SymbolTable.VariantValue[ResultID];

  {$IFDEF WINDOWS}
  if K1 <> K2 then
    ScriptObjectList.RemoveTail(K1, GetCurrentThreadID());
  {$ELSE}
  if K1 <> K2 then
    ScriptObjectList.RemoveTail(K1);
  {$ENDIF}

  fLastResult := Result;
  Code.Prog[StartPos].Vars := Temp;

  if IsEventHandler then
  begin
    State := _ssRunning;
    if Code.N <> Code.Card then
      State := _ssPaused;
    Code.fTerminated := False;
  end
  else if (Code.Prog[Code.N].Op = OP_HALT) and (Code.SubRunCount = 0) then
    State := _ssTerminated
  else if (Code.Prog[Code.N].Op = OP_HALT) and (Code.SubRunCount > 0) then
  begin
    SymbolTable.ClearVariant(ResultID);
    State := _ssRunning;
    for I := NP - 1 downto 0 do
      VariantStack.Pop;
    SymbolTable.RestoreState;
    Code.RestoreState;

    Exit;
  end
  else
    State := _ssPaused;

  if (Code.N = Code.Card) then
    if (State = _ssTerminated) or IsEventHandler then
    begin
      for I := NP - 1 downto 0 do
      begin
        ParamID := SymbolTable.GetParamID(SubID, I + 1);
        if SymbolTable.ByRef[ParamID] > 0 then
          SymbolTable.PutVariant(ParamID, VariantStack.Top);
        VariantStack.Pop;
      end;
      SymbolTable.RestoreState;
      Code.RestoreState;
    end;
end;

function TTitanBaseScripter.AssignedObject(SO: TTitanScriptObject): boolean;
begin
  Result := ScriptObjectList.HasObject(SO);
end;

procedure TTitanBaseScripter.SaveCompiledModuleToStream(M: TTitanModule; S: TStream);
var
  I, N, Temp, PosExtra, ToWrite: integer;
  InitList: TTitanIds;

  ModuleSizePos, ModuleSize: integer;
begin
  ModuleSizePos := S.Position;
  SaveInteger(0, S); // module size
  SaveInteger(_CompiledModuleVersion, S);
  Temp := S.Position;
  SaveInteger(0, S);  // position of extra data
  M._SaveToStream(S);

  PosExtra := S.Position;

  S.Position := Temp;
  if Assigned(OnWriteExtraData) then
    ToWrite := PosExtra
  else
    ToWrite := 0;
  SaveInteger(ToWrite, S); // position of extra data
  S.Position := PosExtra;

  if Assigned(OnWriteExtraData) then
    OnWriteExtraData(Owner, S, M.Name);

  Code.SaveToStream(S, M.P1, M.P2);
  SymbolTable.SaveToStream(S, M.S1, M.S2);

  InitList := TTitanIds.Create(False);
  for I := 0 to ClassList[0].UsingInitList.Count - 1 do
  begin
    N := ClassList[0].UsingInitList[I];
    if (N >= M.P1) and (N <= M.P2) then
      InitList.Add(N);
  end;
  InitList.SaveToStream(S);
  DefaultParameterList.SaveToStream(S);

  Temp := S.Position;
  ModuleSize := S.Position - ModuleSizePos;
  S.Position := ModuleSizePos;
  SaveInteger(ModuleSize, S);
  S.Position := Temp;

  InitList.Free;
end;

procedure TTitanBaseScripter.LoadCompiledModuleFromStream(M: TTitanModule; S: TStream);
var
  DS, DP, I, N, TempPos, Version: integer;
  InitList: TTitanIds;
  ModuleSize, ModuleSizePos: integer;
begin
  if M.BuffStream <> S then
  begin
    ModuleSizePos := S.Position;

    M.BuffStream.Position := 0;
    TempPos := S.Position;
    M.BuffStream.CopyFrom(S, S.Size - S.Position);
    S.Position := TempPos;
    ModuleSize := LoadInteger(S);
    LoadInteger(S); // compiled module version
    LoadInteger(S); // position of extra data
    M._LoadFromStream(S);

    S.Position := ModuleSizePos + ModuleSize;
    Exit;
  end;

  LoadInteger(S); // module size
  Version := LoadInteger(S); // version
  if Version <> _CompiledModuleVersion then
    raise TTitanScriptFailure.Create(Format(errIncorrectCompiledModuleVersion, [Version, _CompiledModuleVersion]));

  LoadInteger(S); // position of extra data
  M.Clear();
  M._LoadFromStream(S);

  if Assigned(OnReadExtraData) then
  begin
    OnReadExtraData(Owner, S, M.Name);
  end;

  DS := SymbolTable.Card - M.S1 + 1;
  DP := Code.Card - M.P1 + 1;

  Code.LoadFromStream(S, DS, DP);
  SymbolTable.LoadFromStream(S, DS, DP);

  if DS <> 0 then
    with M do
      for I := 0 to Namespaces.Count - 1 do
        Namespaces[I] := Namespaces[I] + DS;

  InitList := TTitanIds.Create(False);
  InitList.LoadFromStream(S);
  for I := 0 to InitList.Count - 1 do
  begin
    N := InitList[I] + DP;
    ClassList[0].UsingInitList.Add(N);
  end;

  InitList.Free;

  DefaultParameterList.LoadFromStream(S);
  for I := 0 to DefaultParameterList.Count - 1 do
    Inc(DefaultParameterList[I].ID, DS);

  Inc(M.S1, DS);
  Inc(M.S2, DS);
  Inc(M.P1, DP);
  Inc(M.P2, DP);
end;

procedure TTitanBaseScripter.SaveToStream(S: TStream);
var
  I: integer;
  M: TTitanModule;
begin
  SaveInteger(Modules.Count, S);
  for I := 0 to Modules.Count - 1 do
  begin
    M := Modules.Items[I];
    SaveCompiledModuleToStream(M, S);
  end;
  //  ClassList[0].UsingInitList.SaveToStream(S);
end;

procedure TTitanBaseScripter.LoadFromStream(S: TStream);
var
  I, K: integer;
  M: TTitanModule;
begin
  K := LoadInteger(S);
  for I := 0 to K - 1 do
  begin
    M := TTitanModule.Create(Self);
    LoadCompiledModuleFromStream(M, S);
    Modules.AddObject(M.Name, M);
  end;
end;

procedure TTitanBaseScripter.DisconnectObjects;
begin
  EventHandlerList.ClearHandlers;
  RegisteredFieldList.Clear;
end;

procedure TTitanBaseScripter.AddExtraModule(const ModuleName, SourceCode: string; SyntaxCheckOnly: boolean;
  const PaxLanguage: string);
var
  I, P: integer;
  Found: boolean;
  M: TTitanModule;
  S, FileName: string;
  ClassRec: TTitanClassRec;

  DelphiUnitName, DelphiDfmName, Script: string;
begin
  if SourceCode <> '' then
  begin
    for I := 0 to ExtraModuleList.Count - 1 do
    begin
      S := ExtraModuleList[I];
      P := Pos('=', S);
      if P > 0 then
        S := Copy(S, 1, P - 1);
      if StrEql(ModuleName, S) then
        Exit;
    end;

    for I := 0 to Modules.Count - 1 do
    begin
      M := Modules.Items[I];
      if StrEql(M.Name, ModuleName) then
      begin
        M.Text := SourceCode;
        Exit;
      end;
    end;

    ExtraModuleList.Add(ModuleName + '=' + SourceCode);

    Exit;
  end;

  FileName := FindFullName(ModuleName);

  if FileName <> '' then
  begin
    Found := False;

    for I := 0 to ExtraModuleList.Count - 1 do
      if StrEql(ModuleName, ExtraModuleList[I]) then
      begin
        Found := True;
        Break;
      end;

    for I := 0 to Modules.Count - 1 do
    begin
      M := Modules.Items[I];
      if StrEql(M.Name, ModuleName) then
      begin
        Found := True;
        Break;
      end;
    end;

    if not Found then
      ExtraModuleList.Add(ModuleName);
  end
  else
  begin
    S := ModuleName;
    if Pos('.', S) > 0 then
      S := Copy(S, 1, Pos('.', S) - 1);

    DelphiUnitName := FindFullName(S + '.pas');
    DelphiDfmName := FindFullName(S + '.dfm');

    if DelphiDfmName <> '' then
    begin
      if DelphiUnitName <> '' then
      begin
        Script := ConvertDelphiForm1(DelphiDfmName, DelphiUnitName, PaxLanguage);
        ExtraModuleList.Add(S + '=' + Script);
        Exit;
      end;
    end;

    ClassRec := ClassList.FindClassByName(S);
    if ClassRec = nil then
      if not SyntaxCheckOnly then
        raise TTitanScriptFailure.Create(Format(errFileNotFound, [ModuleName]));
  end;
end;

function TTitanBaseScripter.FindFullName(const FileName: string): string;

  function FindFullNameBase(const FileName: string; UsePathes: boolean): string;

    function RemoveLastDelimeter(const S: string): string;
    var
      I: integer;
    begin
      Result := S;
      for I := Length(S) downto 1 do
        if S[I] = PathDelim then
        begin
          Result := Copy(S, 1, I - 1);
          Exit;
        end;
    end;

  var
    I: integer;
    S, FName: string;
  begin
    Result := '';

    if Pos(PathDelim, FileName) = 1 then
    begin
      S := GetCurrentDir + FileName;
      if FileExists(S) then
      begin
        Result := S;
        Exit;
      end;
      Exit;
    end;

    if Pos('..' + PathDelim, FileName) = 1 then
    begin
      S := GetCurrentDir;
      FName := FileName;

      while Pos('..' + PathDelim, FName) = 1 do
      begin
        S := RemoveLastDelimeter(S);
        FName := Copy(FName, 3, Length(FName) - 2);
      end;

      S := S + FName;

      if FileExists(S) then
      begin
        Result := S;
        Exit;
      end;
      Exit;
    end;

    if FileExists(FileName) then
    begin
      Result := FileName;
      Exit;
    end;

    if not UsePathes then
      Exit;

    for I := 0 to fSearchPathes.Count - 1 do
    begin
      S := fSearchPathes[I] + FileName;
      Result := FindFullNameBase(S, False);
      if Result <> '' then
        Exit;
    end;
  end;

begin
  Result := FindFullNameBase(FileName, True);
end;

function TTitanBaseScripter.IsCompiledScript(const FileName: string): boolean;
var
  S: TFileStream;
  I: integer;
  FullName: string;
begin
  FullName := FindFullName(FileName);

  if FullName = '' then
  begin
    Result := False;
    Exit;
  end;

  S := TFileStream.Create(FullName, fmOpenRead);
  try
    I := LoadInteger(S);
    Result := I < 32;
  finally
    S.Free;
  end;
end;

procedure TTitanBaseScripter.CreateRunList;
var
  I: integer;
begin
  RunList.Clear;
  for I := 0 to Code.FinalizationList.Count - 1 do
    RunList.Push(Code.FinalizationList[I]);
  RunList.Push(0);
  for I := 0 to Code.InitializationList.Count - 1 do
    RunList.Push(Code.InitializationList[I]);
  RunList.Push(-1);
end;

function TTitanBaseScripter.ConvertDelphiForm1(const DfmFileName, UnitFileName, PaxLanguage: string): string;
var
  L, Src, UsedUnits: TStringList;
  I, J: integer;
  S: string;
  P: TPascalParser;
begin
  L := TStringList.Create;
  Src := TStringList.Create;
  UsedUnits := TStringList.Create;

  try
    if UnitFileName <> '' then
      Src.LoadFromFile(UnitFileName);

    for I := 0 to Src.Count - 1 do
      if Pos('USES', UpperCase(Src[I])) = 1 then
      begin
        S := '';
        for J := I to Src.Count - 1 do
          S := S + Src[J];
        P := TPascalParser.Create;
        P.Scanner.SourceCode := S;
        P.Call_SCANNER;
        P.ParseUsesClause(UsedUnits);

        break;
      end;
    for J := UsedUnits.Count - 1 downto 0 do
    begin
      S := UsedUnits[J];
      if GlobalDefinitionList.FindClassDefByName(S) = nil then
        UsedUnits.Delete(J); // not registered
    end;

    if UnitFileName <> '' then
      ConvertDfmFile(DfmFileName, UsedUnits, L, True, Src, PaxLanguage)
    else
      ConvertDfmFile(DfmFileName, UsedUnits, L, True, nil, PaxLanguage);
    Result := L.Text;
  finally
    L.Free;
    Src.Free;
    UsedUnits.Free;
  end;
end;

function TTitanBaseScripter.ConvertDelphiForm2(const DfmFileName: string; UsedUnits: TStringList;
  const PaxLanguage: string): string;
var
  L: TStringList;
begin
  L := TStringList.Create;
  try
    ConvertDfmFile(DfmFileName, UsedUnits, L, True, nil, PaxLanguage);
    Result := L.Text;
  finally
    L.Free;
  end;
end;

procedure TTitanBaseScripter.AddDelphiForm(const DfmFileName, UnitFileName, PaxLanguage: string);
var
  ModuleName: string;
  I: integer;
begin
  ModuleName := ExtractFileName(DfmFileName);
  I := Pos('.', ModuleName);
  if I > 0 then
    ModuleName := Copy(ModuleName, 1, I - 1);

  AddModule(ModuleName, PaxLanguage);
  AddCode(ModuleName, ConvertDelphiForm1(DfmFileName, UnitFileName, PaxLanguage));

  State := _ssReadyToCompile;
end;

function TTitanBaseScripter.GetTypeID(ID: integer): integer;
var
  D: TTitanDefinition;
  S: string;
  C: TTitanClassRec;
begin
  if (ID > 0) and (ID <= SymbolTable.Card) then
  begin
    if ID <= PaxTypes.Count then
      Result := ID
    else
      Result := SymbolTable.PType[ID];
  end
  else
  begin
    D := GlobalDefinitionList[-ID];
    case D.DefKind of
      dkVariable: Result := TTitanVariableDefinition(D).TypeID;
      dkConstant: Result := TTitanConstantDefinition(D).TypeID;
      dkMethod:
      begin
        S := TTitanMethodDefinition(D).ResultType;
        Result := PaxTypes.GetTypeID(S);
        if Result = -1 then
        begin
          C := ClassList.FindClassByName(S);
          if C <> nil then
            Result := C.ClassID
          else
            Result := typeVARIANT;
        end;
      end;
      else
        Result := 0;
    end;
  end;
end;

function TTitanBaseScripter.HasForbiddenPublishedProperty(C: TClass; const PropName: string): boolean;
var
  I: integer;
  AClass: TClass;
begin
  Result := ForbiddenPublishedProperties.IndexOf(C) >= 0;

  if not Result then
  begin
    for I := 0 to ForbiddenPublishedPropertiesEx.Count - 1 do
    begin
      AClass := TClass(ForbiddenPublishedPropertiesEx.Objects[I]);
      if AClass = C then
        if StrEql(PropName, ForbiddenPublishedPropertiesEx[I]) then
        begin
          Result := True;
          Exit;
        end;
    end;
  end;
end;

//=====================================================

function _Eval(const SourceCode: string; Scripter: TTitanBaseScripter; Parser: TTitanParser): variant;

  procedure _CopyLevelStack;
  var
    I, ID, L: integer;
  begin
    Parser.LevelStack.Clear;
    Parser.LevelStack.Push(0);

    with Scripter do
      for I := 1 to Code.LevelStack.Card do
      begin
        ID := Code.LevelStack[I];
        if ID > 0 then
        begin
          L := SymbolTable.Level[ID];
          if SymbolTable.Kind[L] = KindTYPE then
            Parser.LevelStack.Push(L);
          Parser.LevelStack.Push(ID);
        end;
      end;

    if Parser.LevelStack.Card = 1 then
      Parser.LevelStack.Push(Scripter.SymbolTable.RootNamespaceID);
  end;

var
  StartPos, TempCodeCard, TempSymbolCard, TempClassCount: integer;
  Success: boolean;
begin
  with Scripter do
  begin
    TempCodeCard := Code.Card;
    TempSymbolCard := SymbolTable.Card;
    TempClassCount := ClassList.Count;

    Code.SaveState;

    Inc(EvalCount);
    StartPos := Code.Card;

    Success := True;
    Parser.Scanner.SourceCode := SourceCode + ';';

    _CopyLevelStack;

    Parser.UsingList.CopyFrom(Code.UsingList);
    Parser.WithStack.CopyFrom(Code.WithStack);
    try
      with TTitanJavaScriptParser(Parser) do
      begin
        Call_SCANNER;
        Gen(OP_DECLARE_OFF, 0, 0, 0);
        Parse_SourceElements;
        Gen(OP_DECLARE_ON, 0, 0, 0);
        Gen(OP_HALT, 0, 0, 0);
      end;
    except
      Success := False;
    end;

    if Success then
    begin
      Code.LinkGoTo(TempCodeCard + 1, Code.Card);
      ClassList.CreateClassObjects(TempClassCount);
      ClassList.InitStaticFields(TempClassCount);
      SymbolTable.SetupSubs(TempSymbolCard + 1);

      Code.N := StartPos;
      Code.Terminated := False;
      Code.Run;

      Result := Code.ResultValue;
    end;

    Dec(EvalCount);

    Code.RestoreState;
  end;
end;

