{%mainunit ../BASE_Engine.pas}

{***************************************************
 Package pl_TitanScript.pkg
 This file is part of CodeTyphon Studio
****************************************************}

function DelphiInstanceToScriptObject(Instance: TObject; Scripter: Pointer;
                                      RaiseException: boolean = true): TTitanScriptObject;
var
  S: String;
  ClassRec: TTitanClassRec;
  W: TTitanScriptObjectList;
begin
  W := TTitanBaseScripter(Scripter).ScriptObjectList;
  result := W.FindScriptObject(Instance);
  if result <> nil then
    Exit;

  ClassRec := TTitanBaseScripter(Scripter).ClassList.FindClassByInstance(Instance);

  if ClassRec = nil then
  if not RaiseException then
    Exit;

  if ClassRec = nil then
  begin
    S := Instance.ClassName;
    raise TTitanScriptFailure.Create(Format(errClassIsNotImported, [S]));
  end;
  result := ClassRec.CreateScriptObject;
  result.Instance := Instance;

//  if Instance.ClassType = TTitanArray then
    result.RefCount := 1;
end;

function DelphiClassToScriptObject(AClass: TClass; Scripter: Pointer): TTitanScriptObject;
var
  S: String;
  ClassRec: TTitanClassRec;
begin
  S := AClass.ClassName;
  ClassRec := TTitanBaseScripter(Scripter).ClassList.FindClassByName(S);
  if ClassRec = nil then
    raise TTitanScriptFailure.Create(Format(errClassIsNotImported, [S]));
  result := ClassRec.CreateScriptObject;
  result.PClass := AClass;
end;

function InterfaceToScriptObject(const I: IUnknown; Scripter: Pointer;
                                 const InterfaceClassName: String = ''): TTitanScriptObject;
var
  Instance: TObject;
  S: String;
  ClassRec: TTitanClassRec;
begin
  result := nil;
  Instance := GetImplementorOfInterface(I);
  if Instance = nil then
    Exit
  else
  begin
    if InterfaceClassName <> '' then
      S := InterfaceClassName
    else
    begin
      S := Instance.ClassName;
      S[1] := 'I';
    end;

    ClassRec := TTitanBaseScripter(Scripter).ClassList.FindClassByName(S);
    if ClassRec <> nil then
    begin
      result := ClassRec.CreateScriptObject;
      Move(I, result.Intf, 4);
      result.RefCount := 1;
    end
    else
    begin
      ClassRec := TTitanBaseScripter(Scripter).ClassList.FindClassByName('IUnknown');
      if ClassRec <> nil then
      begin
        result := ClassRec.CreateScriptObject;
        Move(I, result.Intf, 4);
        result.RefCount := 1;
      end;
    end;
  end;
end;

function IsHostObject(const V: Variant): Boolean;
var
  SO: TTitanScriptObject;
begin
  if IsObject(V) then
  begin
    SO := VariantToScriptObject(V);
    result := SO.Instance <> nil;
  end
  else
    result := false;
end;

function IsPaxArray(const V: Variant): Boolean;
var
  SO: TTitanScriptObject;
begin
  if IsObject(V) then
  begin
    SO := VariantToScriptObject(V);
    result := SO.ClassRec.ck = ckArray;
  end
  else
    result := false;
end;

function IsDynamicArray(const V: Variant): Boolean;
var
  SO: TTitanScriptObject;
begin
  if IsObject(V) then
  begin
    SO := VariantToScriptObject(V);
    result := SO.ClassRec.ck = ckDynamicArray;
  end
  else
    result := false;
end;

function IsDateObject(const V: Variant): Boolean;
var
  SO: TTitanScriptObject;
  ClassList: TTitanClassList;
begin
  if IsObject(V) then
  begin
    SO := VariantToScriptObject(V);
    ClassList := TTitanBaseScripter(SO.Scripter).ClassList;
    result := SO.ClassRec = ClassList.DateClassRec;
  end
  else
    result := false;
end;

function IsFunctionObject(const V: Variant): Boolean;
var
  SO: TTitanScriptObject;
  ClassList: TTitanClassList;
begin
  if IsObject(V) then
  begin
    SO := VariantToScriptObject(V);
    ClassList := TTitanBaseScripter(SO.Scripter).ClassList;
    result := SO.ClassRec = ClassList.FunctionClassRec;
  end
  else
    result := false;
end;

function IsStringObject(const V: Variant): Boolean;
var
  SO: TTitanScriptObject;
  ClassList: TTitanClassList;
begin
  if IsObject(V) then
  begin
    SO := VariantToScriptObject(V);
    ClassList := TTitanBaseScripter(SO.Scripter).ClassList;
    result := SO.ClassRec = ClassList.StringClassRec;
  end
  else
    result := false;
end;

function IsBooleanObject(const V: Variant): Boolean;
var
  SO: TTitanScriptObject;
  ClassList: TTitanClassList;
begin
  if IsObject(V) then
  begin
    SO := VariantToScriptObject(V);
    ClassList := TTitanBaseScripter(SO.Scripter).ClassList;
    result := SO.ClassRec = ClassList.BooleanClassRec;
  end
  else
    result := false;
end;

function IsNumberObject(const V: Variant): Boolean;
var
  SO: TTitanScriptObject;
  ClassList: TTitanClassList;
begin
  if IsObject(V) then
  begin
    SO := VariantToScriptObject(V);
    ClassList := TTitanBaseScripter(SO.Scripter).ClassList;
    result := SO.ClassRec = ClassList.NumberClassRec;
  end
  else
    result := false;
end;

constructor TTitanError.Create;
begin
  inherited;

  fScriptTime := '';
  fDescription := '';
  fModuleName := '';
  fFileName := '';
  fLine := '';
  fLineNumber := 0;
  fPosition := 0;
  fTextPosition := 0;
  fErrClassType := nil;
end;

function CreateNameIndex(const Name: String; Scripter: Pointer): Integer;
begin
  result := TTitanBaseScripter(Scripter).NameList.Add(Name);
end;

function NameIndexToUpperCaseIndex(NameIndex: Integer; Scripter: Pointer): Integer;
var
  S: String;
  I: PtrInt;
begin
  I := PtrInt(TTitanBaseScripter(Scripter).NameList.Objects[NameIndex]);
  if I = 0 then
  begin
    S := UpperCase(TTitanBaseScripter(Scripter).NameList[NameIndex]);
    result := TTitanBaseScripter(Scripter).NameList.Add(S);
    TTitanBaseScripter(Scripter).NameList.Objects[NameIndex] := TObject(result);
  end
  else
    result := I;
end;

constructor TTitanMemberList.Create(Owner: TTitanClassRec);
begin
  Self.Owner := Owner;
  inherited Create;
end;

function TTitanMemberList.Scripter: Pointer;
begin
  result :=Owner.Scripter;
end;

procedure TTitanMemberList.SaveToStream(S: TStream);
var
  I: Integer;
begin
  SaveInteger(Count, S);
  for I:=0 to Count - 1 do
    Records[I].SaveToStream(S);
end;

procedure TTitanMemberList.LoadFromStream(S: TStream;
                                        DS: Integer = 0; DP: Integer = 0);
var
  I, Index, K: Integer;
  MemberRec: TTitanMemberRec;
begin
  K := LoadInteger(S);
  for I:=0 to K - 1 do
  begin
    MemberRec := TTitanMemberRec.Create(0, Owner);
    MemberRec.LoadFromStream(S, DS, DP);

    Index := CreateNameIndex(MemberRec.GetName, Scripter);
    AddObject(Index, MemberRec);
  end;
end;

function TTitanMemberList.GetRecord(Index: Integer): TTitanMemberRec;
begin
  result := TTitanMemberRec(Objects[Index]);
end;

function TTitanMemberList.UpperCaseIndexOf(UpCaseIndex: Integer): Integer;
var
  I: Integer;
  R: TTitanMemberRec;
begin
  result := -1;
  if UpCaseIndex <= 0 then
    Exit;
  for I:=0 to Count - 1 do
  begin
    R := GetRecord(I);
    if R.UpCaseIndex = UpCaseIndex then
    begin
      result := I;
      Exit;
    end;
  end;
end;

function TTitanMemberList.GetMemberID(const Name: String; UpCase: Boolean = true): Integer;
var
  NameIndex, UpCaseIndex, Index: Integer;
  P: TTitanClassRec;
begin
  result := 0;
  NameIndex := CreateNameIndex(Name, Scripter);

  if UpCase then
     UpCaseIndex := NameIndexToUpperCaseIndex(NameIndex, Scripter)
  else
     UpCaseIndex := -1;

  P := Owner;

  while P <> nil do
  begin
    Index := P.MemberList.IndexOf(NameIndex);
    if Index >= 0 then
    begin
      result := P.MemberList[Index].ID;
      Exit;
    end
    else
    begin
      Index := P.MemberList.UpperCaseIndexOf(UpCaseIndex);
      if Index >= 0 then
      begin
        result := P.MemberList[Index].ID;
        Exit;
      end;
    end;
    P := P.AncestorClassRec;
  end;
end;

function TTitanMemberList.IndexOfMember(const Name: String; UpCase: Boolean = true): Integer;
var
  NameIndex, UpCaseIndex, Index: Integer;
  P: TTitanClassRec;
begin
  result := -1;
  NameIndex := CreateNameIndex(Name, Scripter);

  if UpCase then
     UpCaseIndex := NameIndexToUpperCaseIndex(NameIndex, Scripter)
  else
     UpCaseIndex := -1;

  P := Owner;

  while P <> nil do
  begin
    Index := P.MemberList.IndexOf(NameIndex);
    if Index >= 0 then
    begin
      result := Index;
      Exit;
    end
    else
    begin
      Index := P.MemberList.UpperCaseIndexOf(UpCaseIndex);
      if Index >= 0 then
      begin
        result := Index;
        Exit;
      end;
    end;
    P := P.AncestorClassRec;
  end;
end;

function TTitanMemberList.GetMemberRec(const Name: String; UpCase: Boolean = true): TTitanMemberRec;
var
  NameIndex, UpCaseIndex, Index: Integer;
  P: TTitanClassRec;
begin
  result := nil;
  NameIndex := CreateNameIndex(Name, Scripter);

  if UpCase then
     UpCaseIndex := NameIndexToUpperCaseIndex(NameIndex, Scripter)
  else
     UpCaseIndex := -1;

  P := Owner;

  while P <> nil do
  begin
    Index := P.MemberList.IndexOf(NameIndex);
    if Index >= 0 then
    begin
      result := P.MemberList[Index];
      Exit;
    end
    else
    begin
      Index := P.MemberList.UpperCaseIndexOf(UpCaseIndex);
      if Index >= 0 then
      begin
        result := P.MemberList[Index];
        Exit;
      end;
    end;
    P := TTitanBaseScripter(P.Scripter).ClassList.FindClassByName(P.AncestorName);
  end;
end;

procedure TTitanMemberList.DeleteMember(const Name: String; UpCase: Boolean = true);
var
  NameIndex, UpCaseIndex, Index: Integer;
  P: TTitanClassRec;
begin
  NameIndex := CreateNameIndex(Name, Scripter);

  if UpCase then
     UpCaseIndex := NameIndexToUpperCaseIndex(NameIndex, Scripter)
  else
     UpCaseIndex := -1;

  P := Owner;

  while P <> nil do
  begin
    Index := P.MemberList.IndexOf(NameIndex);
    if Index >= 0 then
    begin
      P.MemberList.DeleteObject(Index);
      Exit;
    end
    else
    begin
      Index := P.MemberList.UpperCaseIndexOf(UpCaseIndex);
      if Index >= 0 then
      begin
        P.MemberList.Delete(Index);
        Exit;
      end;
    end;
    P := TTitanBaseScripter(P.Scripter).ClassList.FindClassByName(P.AncestorName);
  end;
end;

function TTitanMemberList.GetMemberRecByID(MemberID: Integer): TTitanMemberRec;
var
  I: Integer;
  P: TTitanClassRec;
begin
  P := Owner;
  while P <> nil do
  begin
    for I:=0 to P.MemberList.Count - 1 do
    begin
      result := P.MemberList[I];
      if result.ID = MemberID then
        Exit;
    end;
    P := P.AncestorClassRec;
  end;
  result := nil;
end;

constructor TTitanHeapItem.Create(H: TTitanObjectList);
begin
  inherited Create;
  H.Add(Self);
end;

constructor TTitanProperty.Create(SO: TTitanScriptObject; PValue: PVariant;
                                MemberRec: TTitanMemberRec);
begin
  inherited Create;

  Self.SO := SO;
  Self.MemberRec := MemberRec;
  Self.Scripter := SO.Scripter;

  if MemberRec = nil then
  begin
    Val := PValue^;
    fValue := @ Val;
    fUpCaseIndex := 0;
  end
  else
  begin
    if MemberRec.IsStatic then
    begin
      fValue := PValue
    end
    else if IsUndefined(PValue^) then
      fValue := @ Val
    else
    begin
      Val := PValue^;
      fValue := @ Val;
    end;

    fUpCaseIndex := MemberRec.UpCaseIndex;
  end;

  ExtraDef := nil;
end;

function TTitanProperty.GetKind: Integer;
begin
  if MemberRec <> nil then
    result := MemberRec.Kind
  else
    result := KindPROP;
end;

function TTitanProperty.PTerminalValue: PVariant;
begin
  result := GetTerminal(fValue);
end;

function TTitanProperty.GetValue(NParams: Integer): Variant;
var
  P: TTitanProperty;
  SubID, ResID: Integer;
  V: Variant;
  MethodDef: TTitanMethodDefinition;
  S: String;
  I: Integer;
  Indexes: array of Integer;
  SO1: TTitanScriptObject;
  Prop1: TTitanProperty;
  temp: boolean;
begin
  if ReadID = 0 then
  begin
    if IsImported then
    begin
      case Definition.DefKind of
        dkConstant:
          result := TTitanConstantDefinition(Definition).Value;
        dkField:
          result := TTitanFieldDefinition(Definition).GetFieldValue(SO);
        dkRecordField:
          if SO.ExtraPtr = nil then
            result := PTerminalValue^
          else
            result := TTitanRecordFieldDefinition(Definition).GetFieldValue(SO.Scripter, SO.ExtraPtr);
        dkVariable:
          result := TTitanVariableDefinition(Definition).GetValue(Scripter);
        dkObject:
          result := fValue^;
        dkInterface:
          result := fValue^;
        dkProperty:
        begin
          if TTitanPropertyDefinition(Definition).ReadDef.DefKind = dkField then
          begin
            result := TTitanFieldDefinition(TTitanPropertyDefinition(Definition).ReadDef).GetFieldValue(SO);
            Exit;
          end;

          MethodDef := TTitanMethodDefinition(TTitanPropertyDefinition(Definition).ReadDef);
          if MethodDef <> nil then
          begin
            V := ScriptObjectToVariant(SO);
            with TTitanBaseScripter(SO.ClassRec.Scripter) do
            begin
              Code.SaveState;
              Code.CallHostSub(MethodDef, @V, @result, MethodDef.NP);
              Code.RestoreState;
            end;
          end;
        end;
      dkMethod:
      begin
        SO1 := TTitanScriptObject.Create(TTitanBaseScripter(Scripter).ClassList.DelegateClassRec);
        SO1.Instance := TTitanDelegate.Create(Scripter, - Definition.Index,
                        TTitanBaseScripter(Scripter).Code.N, TTitanMethodDefinition(Definition));
        SO1.PClass := TTitanDelegate;
        SO1.PutProperty(CreateNameIndex('Name', Scripter), Definition.Name, 0);
        result := ScriptObjectToVariant(SO1);
      end;
      else
        raise TTitanScriptFailure.Create(Format(ErrPropertyIsNotFound,
          [Definition.Name]));
      end; // case
      Exit;
    end;

    if NParams = 0 then
      result := PTerminalValue^
    else
    begin
      if IsString(fValue^) then
      begin
        if ParamCount > 1 then
          raise TTitanScriptFailure.Create(errTooManyActualParameters);

        I := TTitanBaseScripter(SO.ClassRec.Scripter).Code.PopVariant;
        if TTitanBaseScripter(SO.ClassRec.Scripter).Code.SignZERO_BASED_STRINGS then
           Inc(I);

        S := fValue^;
        result := S[I];
      end
      else if IsPaxArray(PTerminalValue^) then
      begin
        SetLength(Indexes, NParams);
        for I:=1 to NParams do
          Indexes[NParams - I] := TTitanBaseScripter(SO.ClassRec.Scripter).Code.PopVariant;
        result := TTitanArray(VariantToScriptObject(PTerminalValue^).Instance).Get(Indexes);
      end
      else if IsVBArray(fValue^) then
      begin
        SetLength(Indexes, NParams);
        for I:=1 to NParams do
          Indexes[NParams - I] := TTitanBaseScripter(SO.ClassRec.Scripter).Code.PopVariant;

       if VarArrayDimCount(fValue^) = 1 then
       begin
         I := Indexes[0];
         result := Variant(fValue^)[I];
       end
       else
         result := Variant(ArrayGet(fValue, Indexes)^);
      end
      else if IsObject(fValue^) then
      begin
        SO1 := VariantToScriptObject(fValue^);
        Prop1 := SO1.PropertyList.GetDefaultProperty;
        if Prop1 = nil then
        begin
          S := ToStr(SO1.ClassRec.Scripter, TTitanBaseScripter(SO.ClassRec.Scripter).Code.PopVariant);
          result := SO1.GetProperty(CreateNameIndex(S, Scripter), NParams - 1);
          Exit;
        end;
      end;
    end;

    Exit;
  end;

  with TTitanBaseScripter(SO.ClassRec.Scripter) do
  begin
    if SymbolTable.Kind[ReadID] = KindSUB then
    begin
      temp := Code.Terminated;
      Code.Terminated := false;

      V := ScriptObjectToVariant(SO);
      SymbolTable.SaveState;
      SubID := SymbolTable.AppReference(V, SymbolTable.NameIndex[ReadID], maAny);
      ResID := Code.EvalGetFunction(SubID, NParams);
      result := SymbolTable.GetValue(ResID);
      SymbolTable.RestoreState;

      Code.Terminated := temp;
      Exit;
    end;

    P := SO.PropertyList.FindProperty(SymbolTable.NameIndex[ReadID]);
    if P = nil then
      raise TTitanScriptFailure.Create(Format(errPropertyIsNotFound, [SymbolTable.Name[ReadID]]));

    result := P.fValue^;
  end;
end;

function TTitanProperty.GetAddress(NParams: Integer): PVariant;
var
  P: TTitanProperty;
  SubID, ResID: Integer;
  V: Variant;
  I: Integer;
  Indexes: array of Integer;
begin
  result := nil;
  if ReadID = 0 then
  begin
    if IsImported then
    begin
      case Definition.DefKind of
        dkConstant:
//        raise TTitanScriptFailure.Create(errOperatorNotApplicable);
          result := @ TTitanConstantDefinition(Definition).DefValue;
        dkField:
          result := TTitanFieldDefinition(Definition).GetFieldAddress(SO);
        dkRecordField:
          if SO.ExtraPtr = nil then
            result := PTerminalValue
          else
            result := TTitanRecordFieldDefinition(Definition).GetFieldAddress(SO.Scripter, SO.ExtraPtr);
        dkVariable:
          result := TTitanVariableDefinition(Definition).GetAddress;
        dkObject:
          result := fValue;
        dkInterface:
          result := fValue;
        dkProperty:
        begin
          if TTitanPropertyDefinition(Definition).ReadDef.DefKind = dkField then
          begin
            result := TTitanFieldDefinition(TTitanPropertyDefinition(Definition).ReadDef).GetFieldAddress(SO);
            Exit;
          end;

          result := nil;
{
          MethodDef := TTitanMethodDefinition(TTitanPropertyDefinition(Definition).ReadDef);
          if MethodDef <> nil then
          begin
            V := ScriptObjectToVariant(SO);
            with TTitanBaseScripter(SO.ClassRec.Scripter) do
            begin
              Code.SaveState;
              Code.CallHostSub(MethodDef, @V, @result, MethodDef.NP);
              Code.RestoreState;
            end;
          end;
}
        end;
      else
        raise TTitanScriptFailure.Create(Format(ErrPropertyIsNotFound,
          [Definition.Name]));
      end; // case
      Exit;
    end;

    if NParams = 0 then
      result := PTerminalValue
    else
    begin
      if IsString(fValue^) then
      begin
        raise TTitanScriptFailure.Create(errOperatorNotApplicable);
      end
      else if IsPaxArray(PTerminalValue^) then
      begin
        SetLength(Indexes, NParams);
        for I:=1 to NParams do
          Indexes[NParams - I] := TTitanBaseScripter(SO.ClassRec.Scripter).Code.PopVariant;
        result := TTitanArray(VariantToScriptObject(PTerminalValue^).Instance).GetPtr(Indexes);
      end
      else if IsVBArray(fValue^) then
      begin
        SetLength(Indexes, NParams);
        for I:=1 to NParams do
          Indexes[NParams - I] := TTitanBaseScripter(SO.ClassRec.Scripter).Code.PopVariant;

        result := ArrayGet(fValue, Indexes);
     end;
    end;

    Exit;
  end;

  with TTitanBaseScripter(SO.ClassRec.Scripter) do
  begin
    if SymbolTable.Kind[ReadID] = KindSUB then
    begin
      V := ScriptObjectToVariant(SO);
      SymbolTable.SaveState;
      SubID := SymbolTable.AppReference(V, SymbolTable.NameIndex[ReadID], maAny);
      ResID := Code.EvalGetFunction(SubID, NParams);
      result := SymbolTable.Address[ResID];
      SymbolTable.RestoreState;
      Exit;
    end;

    P := SO.PropertyList.FindProperty(SymbolTable.NameIndex[ReadID]);
    if P = nil then
      raise TTitanScriptFailure.Create(Format(errPropertyIsNotFound, [SymbolTable.Name[ReadID]]));

    result := P.fValue;
  end;
end;

procedure TTitanProperty.PutValue(NParams: Integer; const AValue: Variant);
var
  P: TTitanProperty;
  SubID: Integer;
  V: Variant;
  MethodDef: TTitanMethodDefinition;
  S: String;
  ch: Char;
  I: Integer;
  Indexes: array of Integer;

  SO1: TTitanScriptObject;
  Prop1: TTitanProperty;
begin
  if WriteID = 0 then
  begin
    if IsImported then
    begin
      case Definition.DefKind of
        dkConstant:
          raise TTitanScriptFailure.Create(Format(errCannotWriteReadOnlyProperty, [Definition.Name]));
        dkProperty:
        begin
          if TTitanPropertyDefinition(Definition).WriteDef.DefKind = dkField then
          begin
            TTitanFieldDefinition(TTitanPropertyDefinition(Definition).WriteDef).SetFieldValue(SO, AValue);
            Exit;
          end;

          MethodDef := TTitanMethodDefinition(TTitanPropertyDefinition(Definition).WriteDef);
          if MethodDef <> nil then
          begin
            V := ScriptObjectToVariant(SO);
            with TTitanBaseScripter(SO.ClassRec.Scripter) do
            begin
              Code.Stack.Push(@AValue);
              Code.SaveState;
              Code.CallHostSub(MethodDef, @V, nil, NParams + 1);
              Code.RestoreState;
            end;
          end;
        end;
      dkField:
        TTitanFieldDefinition(Definition).SetFieldValue(SO, AValue);
      dkRecordField:
        if SO.ExtraPtr = nil then
          PTerminalValue^ := AValue
        else
          TTitanRecordFieldDefinition(Definition).SetFieldValue(SO.Scripter, SO.ExtraPtr, AValue);
      dkVariable:
        TTitanVariableDefinition(Definition).SetValue(Scripter, AValue);
      dkObject:
        fValue^ := AValue;
      dkInterface:
        fValue^ := AValue;
      else
        raise TTitanScriptFailure.Create(Format(ErrPropertyIsNotFound,
          [Definition.Name]));
      end; // case
      Exit;
    end;

    if NParams = 0 then
    begin
      if MemberRec <> nil then
      if MemberRec.ID > 0 then
        if VarType(AValue) = varString then if TTitanBaseScripter(SO.Scripter).SymbolTable.Count[MemberRec.ID] > 0 then
        begin
          PTerminalValue^ := Copy(AValue, 1, TTitanBaseScripter(SO.Scripter).SymbolTable.Count[MemberRec.ID]);
          Exit;
        end;

      PTerminalValue^ := AValue;
    end
    else
    begin
      if IsString(fValue^) then
      begin
        if NParams > 1 then
          raise TTitanScriptFailure.Create(errTooManyActualParameters);

        S := AValue;
        Ch := S[1];

        I := TTitanBaseScripter(SO.ClassRec.Scripter).Code.PopVariant;
        if TTitanBaseScripter(SO.ClassRec.Scripter).Code.SignZERO_BASED_STRINGS then
           Inc(I);

        S := fValue^;
        S[I] := Ch;
        fValue^ := S;
      end
      else if IsPaxArray(PTerminalValue^) then
      begin
        SetLength(Indexes, NParams);
        for I:=1 to NParams do
          Indexes[NParams - I] := TTitanBaseScripter(SO.ClassRec.Scripter).Code.PopVariant;
        TTitanArray(VariantToScriptObject(PTerminalValue^).Instance).Put(Indexes, AValue);
      end
      else if IsVBArray(fValue^) then
      begin
        SetLength(Indexes, NParams);
        for I:=1 to NParams do
          Indexes[NParams - I] := TTitanBaseScripter(SO.ClassRec.Scripter).Code.PopVariant;

       if VarArrayDimCount(fValue^) = 1 then
       begin
         I := Indexes[0];
         Variant(fValue^)[I] := AValue;
       end
       else
         ArrayPut(fValue, Indexes, AValue);
      end
      else if IsObject(fValue^) then
      begin
        SO1 := VariantToScriptObject(fValue^);
        Prop1 := SO1.PropertyList.GetDefaultProperty;
        if Prop1 = nil then
        begin
          S := ToStr(SO1.ClassRec.Scripter, TTitanBaseScripter(SO1.ClassRec.Scripter).Code.PopVariant);
          SO1.PutProperty(CreateNameIndex(S, Scripter), AValue, NParams - 1);
          Exit;
        end;
      end;
    end;

    Exit;
  end;

  with TTitanBaseScripter(SO.ClassRec.Scripter) do
  begin
    if SymbolTable.Kind[WriteID] = KindSUB then
    begin
      V := ScriptObjectToVariant(SO);
      SymbolTable.SaveState;
      SubID := SymbolTable.AppReference(V, SymbolTable.NameIndex[WriteID], maAny);
      Code.EvalSetProcedure(SubID, NParams, AValue);
      SymbolTable.RestoreState;
      Exit;
    end;

    P := SO.PropertyList.FindProperty(SymbolTable.NameIndex[WriteID]);
    if P = nil then
      raise TTitanScriptFailure.Create(Format(errPropertyIsNotFound, [SymbolTable.Name[WriteID]]));

    P.fValue^ := AValue;
  end;
end;

function TTitanProperty.IsImported: Boolean;
begin
  result := Definition <> nil;
end;

function TTitanProperty.IsStatic: Boolean;
begin
  result := false;
  if MemberRec <> nil then
    result := modStatic in MemberRec.ml;
end;

function TTitanProperty.Definition: TTitanDefinition;
begin
  if MemberRec = nil then
  begin
    result := nil;
    Exit;
  end;

  result := MemberRec.Definition;

  if ExtraDef <> nil then
    result := ExtraDef;
end;

function TTitanProperty.GetReadID: Integer;
begin
  if MemberRec = nil then
  begin
    result := 0;
    Exit;
  end;

  result := MemberRec.ReadID;
end;

function TTitanProperty.GetWriteID: Integer;
begin
  if MemberRec = nil then
  begin
    result := 0;
    Exit;
  end;

  result := MemberRec.WriteID;
end;

constructor TTitanPropertyList.Create(Scripter: Pointer);
begin
  Self.Scripter := Scripter;
  inherited Create;
end;

function TTitanPropertyList.GetProperty(I: Integer): TTitanProperty;
begin
  result := TTitanProperty(Objects[I]);
end;

function TTitanPropertyList.GetName(I: Integer): String;
var
  NameIndex: Integer;
begin
  NameIndex := NameID[I];
  result := _GetName(NameIndex, Scripter);
end;

function TTitanPropertyList.FindProperty(PropertyNameIndex: Integer): TTitanProperty;
var
  I, UpCaseIndex: Integer;
  P: TTitanProperty;
begin
  result := TTitanProperty(GetObject(PropertyNameIndex));

  if result = nil then
  begin
    UpCaseIndex := NameIndexToUpperCaseIndex(PropertyNameIndex, Scripter);
    for I:=0 to Count - 1 do
    begin
      P := Properties[I];
      if P.UpCaseIndex = UpCaseIndex then
      begin
        result := P;
        Exit;
      end;
    end;
  end;
end;

function TTitanPropertyList.GetDefaultProperty: TTitanProperty;
var
  I: Integer;
begin
  result := nil;
  for I:=0 to Count - 1 do
    with TTitanProperty(Objects[I]) do
      if MemberRec <> nil then
        if modDEFAULT in MemberRec.ml then
        begin
          result := Properties[I];
          Exit;
        end;
end;

constructor TTitanScriptObject.Create(ClassRec: TTitanClassRec);
begin
  inherited Create;

  Inc(TTitanBaseScripter(ClassRec.Scripter)._ObjCount);
  _ObjCount := TTitanBaseScripter(ClassRec.Scripter)._ObjCount;

  Self.ClassRec := ClassRec;

  PropertyList := TTitanPropertyList.Create(ClassRec.Scripter);

  Instance := nil;
  PClass := nil;
  RefCount := 0;
  IsClass := false;
  ExtraPtr := nil;
  ExtraPtrSize := 0;
  ExternalExtraPtr := false;
  Intf := nil;
  PIntf := nil;

{$IFDEF WINDOWS}
  ThreadID := GetCurrentThreadID();
{$ELSE}
  ThreadID := 0;
{$ENDIF}

  TTitanBaseScripter(ClassRec.Scripter).ScriptObjectList.Add(Self);
  HasFordiddenProperties := (TTitanBaseScripter(Scripter).ForbiddenPublishedProperties.Count > 0) or
                            (TTitanBaseScripter(Scripter).ForbiddenPublishedPropertiesEx.Count > 0);
end;

procedure TTitanScriptObject.CallAutoDestructor;
begin
  Pointer(Intf) := nil;

  if ClassRec = nil then
    Exit;
  if ClassRec.AutoDestructorID = 0 then
    Exit;

//  TTitanBaseScripter(Scripter).CallMethod(ClassRec.AutoDestructorID,
//    ScriptObjectToVariant(Self), []);
end;

function TTitanScriptObject.ExtraInstance: TObject;
begin
  result := Instance;
end;

function TTitanScriptObject.GetPropertyName(Index: Integer): String;
var
  NameIndex: Integer;
begin
  with PropertyList do
    if (Index >= 0) and (Index < Count) then
    begin
      NameIndex := NameID[Index];
      result := _GetName(NameIndex, Scripter)
    end
    else
      result := '';
end;

procedure TTitanScriptObject.SetDefaultValue(const V: Variant);
begin
end;

function TTitanScriptObject.DefaultValue: Variant;
begin
  result := Undefined;
end;

procedure TTitanScriptObject.FreeExtraPtr;
var
  D: TTitanClassDefinition;
  P: Pointer;
begin
  if ExtraPtr = nil then
    Exit;
  if ExternalExtraPtr then
    Exit;

  if ExtraPtrSize > 0 then
  begin
    if ClassRec.ck = ckDynamicArray then
      ExtraPtr := ShiftPointer(ExtraPtr, - SizeOf(Pointer) - SizeOf(Pointer));
    FreeMem(ExtraPtr, ExtraPtrSize);
    Exit;
  end
  else if ExtraPtrSize < 0 then
  begin
    P := ShiftPointer(ExtraPtr, ExtraPtrSize);
    FreeMem(P, - ExtraPtrSize);
  end;

  D := GetClassDef;
  if D = nil then
    Exit;

  if D.ClassKind <> ckDynamicArray then
     Exit;
end;

destructor TTitanScriptObject.Destroy;
begin
  PropertyList.Free;

  if (Assigned(Instance) and (ClassRec.ck = ckArray)) then
  begin
    Instance.Free;
    Instance := nil;
  end;

  FreeExtraPtr;

  inherited;
end;

function TTitanScriptObject.Scripter: Pointer;
begin
  result := ClassRec.Scripter;
end;

function TTitanScriptObject.IsImported: Boolean;
begin
  if ClassRec = nil then
    result := false
  else
    result := ClassRec.IsImported;
end;

function TTitanScriptObject.GetClassDef: TTitanClassDefinition;
begin
  if not IsImported then
  begin
    result := nil;
    Exit;
  end;
  result := ClassRec.GetClassDef;
end;


procedure TTitanScriptObject.PutPublishedProperty(const PropertyName: String; const Value: Variant);
var
  PropInfo: PPropInfo;
  I: Integer;
  M: TMethod;
  pti, pti1: PTypeInfo;
  ptd: PTypeData;
  EnumName, S: String;
  SO: TTitanScriptObject;
  EventHandler: TTitanEventHandler;
  SubID: Integer;
  This: Variant;
  ByteSet: TByteSet;
begin
  if HasFordiddenProperties then
    if TTitanBaseScripter(Scripter).HasForbiddenPublishedProperty(Instance.ClassType, PropertyName) then
      raise TTitanScriptFailure.Create(Format(errPropertyNotFoundInClass, [PropertyName, Instance.ClassName]));

  pti := PTypeInfo(Instance.ClassType.ClassInfo);

  if pti = nil then
    PropInfo := nil  else
    PropInfo := GetPropInfo(pti, PropertyName);

  if PropInfo <> nil then
  begin
   {$ifdef fpc}
    pti1 := PropInfo.PropType;
   {$else}
    pti1 := PropInfo.PropType^;
   {$endif}
    case pti1^.Kind of
      tkInteger, tkChar, tkWChar:
        SetOrdProp(Instance, PropInfo, TVarData(Value).VInteger);
      tkClass:
        SetOrdProp(Instance, PropInfo, PtrInt(VariantToScriptObject(Value).Instance));
      tkEnumeration:
      begin
        case VarType(Value) of
          varString:
          begin
         {$ifdef fpc}
            pti := PropInfo.PropType;
         {$else}
            pti := PropInfo.PropType^;
         {$endif}
            ptd := GetTypeData(pti);
            S := ToStr(Scripter, Value);
            with ptd^ do
              for I:= MinValue to MaxValue do
              begin
                EnumName := GetEnumName(pti, I);
                if StrEql(S, EnumName) then
                begin
                  SetOrdProp(Instance, PropInfo, I);
                  Break;
                end;
              end;
          end;
          varBoolean:
          begin
            if TVarData(Value).VInteger <> 0 then
              SetOrdProp(Instance, PropInfo, Ord(True)) else
              SetOrdProp(Instance, PropInfo, 0);
          end;
        else
          SetOrdProp(Instance, PropInfo, TVarData(Value).VInteger);
        end;
      end;
      tkFloat:
        SetFloatProp(Instance, PropInfo, Value);
      tkString, tkLString, tkWString:
        case TVarData(Value).VType of
          varInteger: SetStrProp(Instance, PropInfo, Chr(TVarData(Value).VInteger));
          varString: SetStrProp(Instance, PropInfo, String(TVarData(Value).VString));
        else
          SetStrProp(Instance, PropInfo, ToStr(Scripter, Value));
        end;
      tkVariant:
        SetVariantProp(Instance, PropInfo, GetTerminal(@Value)^);
      tkSet:
      begin
        ByteSet := PaxArrayToByteSet(Value);
        I := 0;
        Move(ByteSet, I, SizeOf(ByteSet));
        SetOrdProp(Instance, PropInfo, I);
      end;
      tkMethod:
      begin
        M := GetMethodProp(Instance, PropInfo);

        if IsUndefined(Value) then
        begin
          M.Code := nil;
          M.Data := nil;
          SetMethodProp(Instance, PropInfo, M);
          Exit;
        end;

        SO := VariantToScriptObject(Value);

        if SO.ClassRec.AncestorName = 'Function' then
          SubID := SO.ClassRec.GetConstructorID else
          SubID := TTitanDelegate(SO.Instance).SubID;

        This := ScriptObjectToVariant(Self);

       {$ifdef fpc}
        EventHandler := TTitanEventHandler.Create(Scripter, @PropInfo.PropType, SubID, This);
       {$else}
        EventHandler := TTitanEventHandler.Create(Scripter, PropInfo.PropType^, SubID, This);
       {$endif}
        EventHandler.DelphiInstance := Instance;
        EventHandler.PropInfo := PropInfo;
        EventHandler.This := TTitanBaseScripter(Scripter).Code._This;
        EventHandler.HostHandler :=M;
        EventHandler.OverrideHandlerMode := TTitanBaseScripter(Scripter).OverrideHandlerMode;

        M.Code := @TTitanEventHandler.HandleEvent;
        M.Data := Pointer(EventHandler);

        TTitanBaseScripter(Scripter).EventHandlerList.Add(EventHandler);

        SetMethodProp(Instance, PropInfo, M);

        if IsUndefined(EventHandler.This) then
        begin
          SubID := TTitanBaseScripter(Scripter).Code.LevelStack.Top;
          with TTitanBaseScripter(Scripter).SymbolTable do
            EventHandler.This := GetVariant(GetThisID(SubID));
        end;


      end;
    end
  end
  else
  begin
    raise TTitanScriptFailure.Create(Format(errPropertyIsNotFoundInClass, [PropertyName, Instance.ClassName]));
  end;
end;

function TTitanScriptObject.GetPublishedProperty(const PropertyName: String; NParams: Integer): Variant;
var
  PropInfo: PPropInfo;
  S: String;
  I: Integer;
  TypeData: PTypeData;
  pti, pti1: PTypeInfo;
  C: TComponent;
  X: ActiveXObject;
  Indexes: array of Integer;
  V: Variant;
  P: Pointer;
  This: Variant;
  Def: TTitanMethodDefinition;
  ByteSet: TByteSet;
{$IFNDEF VARIANTS}
  D: Double;
{$ENDIF}
begin
  if HasFordiddenProperties then
    if TTitanBaseScripter(Scripter).HasForbiddenPublishedProperty(Instance.ClassType, PropertyName) then
      raise TTitanScriptFailure.Create(Format(errPropertyIsNotFoundInClass, [PropertyName, Instance.ClassName]));

  pti := PTypeInfo(Instance.ClassType.ClassInfo);
  if pti = nil then
    PropInfo := nil else
    PropInfo := GetPropInfo(pti, PropertyName);

  if PropInfo <> nil then
  begin
{$ifdef fpc}

    TypeData := GetTypeData(PropInfo^.PropType);
    pti1 := PropInfo.PropType;
{$else}
    TypeData := GetTypeData(PropInfo^.PropType^);
    pti1 := PropInfo.PropType^;
{$endif}

    // return the right type

    case pti1^.Kind of
      tkInteger: result := GetOrdProp(Instance, PropInfo);
      tkInt64:
      begin
        {$IFDEF VARIANTS}
        result := GetInt64Prop(Instance, PropInfo);
        {$ELSE}
        D := GetInt64Prop(Instance, PropInfo);
        result := D;
        {$ENDIF}
      end;
      tkFloat: result := GetFloatProp(Instance, PropInfo);
      tkEnumeration:
{$ifdef fpc}
        if TypeData^.BaseType = TypeInfo(Boolean) then
{$else}
        if TypeData^.BaseType^ = TypeInfo(Boolean) then
{$Endif}
          result := Boolean(GetOrdProp(Instance, PropInfo))
        else
        begin
          result := GetOrdProp(Instance, PropInfo);
{
          I := GetOrdProp(Instance, PropInfo);
          result := GetEnumName(PropInfo^.PropType^, I); }
        end;
      tkSet:
      begin
        I := GetOrdProp(Instance, PropInfo);
        ByteSet := [];
        Move(I, ByteSet, SizeOf(ByteSet));
        result := ByteSetToPaxArray(ByteSet, Scripter);
      end;
      tkClass:
      begin
        I := GetOrdProp(Instance, PropInfo);
        if I <> 0 then
        begin
          result := ScriptObjectToVariant(DelphiInstanceToScriptObject(TObject(I), Scripter));
        end;
      end;
      tkChar:
      begin
        S := Chr(GetOrdProp(Instance, PropInfo));
        result := S;
      end;
      tkString, tkLString, tkWString:
      begin
        S := GetStrProp(Instance, PropInfo);
        result := S;
      end;
      tkVariant:
      begin
        result := GetVariantProp(Instance, PropInfo);

        if VarType(result) = varDispatch then
        begin
          X := ActiveXObject.Create(Scripter);
          X.D := result;
          result := ScriptObjectToVariant(DelphiInstanceToScriptObject(X, Scripter));
        end
        else
        if (VarType(result) >= varArray) and (Nparams > 0) then
        begin
          SetLength(Indexes, NParams);
          for I:=1 to NParams do
            Indexes[NParams - I] := TTitanBaseScripter(Scripter).Code.PopVariant;
          V := Variant(ArrayGet(@result, Indexes)^);
          result := V;
        end;
      end;
    end;
  end
  else
  begin
    if Instance.InheritsFrom(TComponent) then
    begin
      C := TComponent(Instance);
      for I:=0 to C.ComponentCount - 1 do
        if StrEql(C.Components[I].Name, PropertyName) then
        begin
          result := ScriptObjectToVariant(DelphiInstanceToScriptObject(C.Components[I], Scripter));
          Exit;
        end;
    end;

    P := Instance.MethodAddress(PropertyName);
    if P <> nil then
    begin
      Def := TTitanMethodDefinition.Create(GlobalDefinitionList, '', nil, -1, nil, true);

      try
        with Def do
        begin
          DirectProc := P;
          SetLength(Types, NParams + 1);
          SetLength(ExtraTypes, NParams + 1);
          SetLength(ByRefs, NParams + 1);

          for I:=0 to NParams do
          begin
            Types[I] := typeVOID;
            ByRefs[I] := false;
          end;
        end;

        This := ScriptObjectToVariant(DelphiInstanceToScriptObject(Instance, Scripter));
        TTitanBaseScripter(Scripter).Code.CallHostSub(Def, @This, @Result, NParams);
      finally
        Def.Free;
      end;
      Exit;
    end;

    raise TTitanScriptFailure.Create(Format(errPropertyIsNotFoundInClass, [PropertyName, Instance.ClassName]));
  end;
end;

function TTitanScriptObject.HasProperty(const PropertyName: String): Boolean;
var
  P: TTitanProperty;
begin
  P := PropertyList.FindProperty(CreateNameIndex(PropertyName, Scripter));
  if P = nil then
    result := HasPublishedProperty(PropertyName)
  else
    result := true;
end;

function TTitanScriptObject.HasPublishedProperty(const PropertyName: String): Boolean;
var
  PropInfo: PPropInfo;
  pti: PTypeInfo;
  C: TComponent;
  I: Integer;
begin
  result := false;

  if Instance = nil then Exit;

  if HasFordiddenProperties then
    if TTitanBaseScripter(Scripter).HasForbiddenPublishedProperty(Instance.ClassType, PropertyName) then
       Exit;

  if Instance.MethodAddress(PropertyName) <> nil then
  begin
    result := true;
    Exit;
  end;

  pti := PTypeInfo(Instance.ClassType.ClassInfo);
  if pti = nil then
    Exit;
  PropInfo := GetPropInfo(pti, PropertyName);
  if PropInfo <> nil then
  begin
    result := true;
    Exit;
  end;
  if Instance.InheritsFrom(TComponent) then
  begin
    C := TComponent(Instance);
    for I:=0 to C.ComponentCount - 1 do
      if StrEql(C.Components[I].Name, PropertyName) then
      begin
        result := true;
        Exit;
      end;
  end;
end;

function TTitanScriptObject.Get(PropertyNameIndex: Integer): TTitanProperty;
begin
  result := PropertyList.FindProperty(PropertyNameIndex);
  if result = nil then
  begin
    TTitanBaseScripter(Scripter).Dump;
    raise TTitanScriptFailure.Create(Format(errPropertyIsNotFound,
        [_GetName(PropertyNameIndex, Scripter)]));
  end;
end;

function TTitanScriptObject.SafeGet(PropertyNameIndex: Integer): TTitanProperty;
begin
  result := PropertyList.FindProperty(PropertyNameIndex);
end;

function TTitanScriptObject.Put(PropertyNameIndex: Integer;
                           const Value: Variant; NParams: Integer): TTitanProperty;
var
  S: String;
  V: Variant;
  D: TTitanMethodDefinition;
  MemberRec: TTitanMemberRec;
begin
  result := PropertyList.FindProperty(PropertyNameIndex);
  if result <> nil then
  begin
    if result.MemberRec <> nil then
      if result.MemberRec.IsPublished then
      begin
        S := _GetName(PropertyNameIndex, Scripter);
        PutPublishedProperty(S, Value);
        Exit;
      end;

    result.PutValue(NParams, Value);
    Exit;
  end;
  if IsImported then
  begin
    S := _GetName(PropertyNameIndex, Scripter);
    if HasPublishedProperty(S) then
      PutPublishedProperty(S, Value)
    else if ClassDef.PutPropDef <> nil then
    begin
      D := ClassDef.PutPropDef;
      D.Name := S;
      with TTitanBaseScripter(ClassRec.Scripter) do
      begin
//        if ClassRec.AncestorName = 'Function' then
//          Instance := Self;

        V := ScriptObjectToVariant(Self);
        Code.SaveState;
        Code.Stack.Push(PtrInt(@Value));
        Code.CallHostSub(D, @V, nil, NParams + 1);
        Code.RestoreState;
      end;
      D.Name := '';
    end
    else
      Get(PropertyNameIndex); // exception
  end
  else
  begin
    MemberRec := ClassRec.FindMember(PropertyNameIndex, maAny, true);
    if MemberRec = nil then
      Get(PropertyNameIndex) // exception
    else
    begin
      result := CreateProperty(PropertyNameIndex,
                               TTitanBaseScripter(ClassRec.Scripter).SymbolTable.Address[MemberRec.Id],
                               MemberRec);
      while (NParams > 0) do
      begin
        TTitanBaseScripter(ClassRec.Scripter).Code.PopVariant;
        Dec(NParams);
      end;
    end;
  end;
end;

function TTitanScriptObject.GetProperty(PropertyNameIndex: Integer;
                                      NParams: Integer): Variant;
var
  P: TTitanProperty;
  S: String;
  D: TTitanMethodDefinition;
  V: Variant;
  MemberRec: TTitanMemberRec;
begin
  P := PropertyList.FindProperty(PropertyNameIndex);
  if P <> nil then
  begin
    if P.MemberRec <> nil then
      if P.MemberRec.IsPublished then
      begin
        S := _GetName(PropertyNameIndex, Scripter);
        result := GetPublishedProperty(S, NParams);
        Exit;
      end;

    result := P.GetValue(NParams);
    Exit;
  end;
  if IsImported then
  begin
    S := _GetName(PropertyNameIndex, Scripter);
    if HasPublishedProperty(S) then
      result := GetPublishedProperty(S, NParams)
    else if ClassDef.GetPropDef <> nil then
    begin
//      if ClassRec.AncestorName = 'Function' then
//        Instance := Self;

      D := ClassDef.GetPropDef;
      D.Name := S;
      with TTitanBaseScripter(ClassRec.Scripter) do
      begin
        V := ScriptObjectToVariant(Self);
        Code.SaveState;
        Code.CallHostSub(D, @V, @result, NParams);
        Code.RestoreState;
      end;
      D.Name := '';
    end
    else
      Get(PropertyNameIndex); // exception
  end
  else
  begin
    if SignLoadOnDemand then
    begin
      MemberRec := ClassRec.GetMember(PropertyNameIndex);
      if MemberRec <> nil then
      begin
        result := TTitanBaseScripter(ClassRec.Scripter).SymbolTable.GetVariant(MemberRec.Id);
        Exit;
      end;
    end;

    Get(PropertyNameIndex); // exception
  end;
end;

procedure TTitanScriptObject.ClearProperty(PropertyNameIndex: Integer);
var
  P: TTitanProperty;
begin
  P := PropertyList.FindProperty(PropertyNameIndex);
  if P <> nil then
    VarClear(P.fValue^);
end;

function TTitanScriptObject.GetAddress(PropertyNameIndex: Integer;
                                     NParams: Integer): PVariant;
var
  P: TTitanProperty;
  S: String;
  D: TTitanMethodDefinition;
  V: Variant;
  MemberRec: TTitanMemberRec;
begin
  result := nil;
  P := PropertyList.FindProperty(PropertyNameIndex);
  if P <> nil then
  begin
    if P.MemberRec <> nil then
      if P.MemberRec.IsPublished then
      begin
        result := nil;
//        raise TTitanScriptFailure.Create(errOperatorNotApplicable);
        Exit;
      end;

    result := P.GetAddress(NParams);
    Exit;
  end;
  if IsImported then
  begin
    S := _GetName(PropertyNameIndex, Scripter);
    if HasPublishedProperty(S) then
      result := nil
    else if ClassDef.GetPropDef <> nil then
    begin
      D := ClassDef.GetPropDef;
      D.Name := S;
      with TTitanBaseScripter(ClassRec.Scripter) do
      begin
        V := ScriptObjectToVariant(Self);
        Code.SaveState;
        Code.CallHostSub(D, @V, @TempVariant, NParams);
        Code.RestoreState;

        result := @TempVariant;
      end;
      D.Name := '';
    end
    else
      Get(PropertyNameIndex); // exception
  end
  else
  begin
    if SignLoadOnDemand then
    begin
      MemberRec := ClassRec.GetMember(PropertyNameIndex);
      if MemberRec <> nil then
      begin
        result := TTitanBaseScripter(ClassRec.Scripter).SymbolTable.Address[MemberRec.Id];
        Exit;
      end;
    end;

    Get(PropertyNameIndex); // exception
  end;
end;

procedure TTitanScriptObject.PutProperty(PropertyNameIndex: Integer;
                                    const Value: Variant; NParams: Integer);
begin
  Put(PropertyNameIndex, Value, NParams);
end;

function TTitanScriptObject.GetDefaultProperty(NParams: Integer): Variant;
var
  P: TTitanProperty;
begin
  P := PropertyList.GetDefaultProperty;
  if P = nil then
  begin
    raise TTitanScriptFailure.Create(errDefaultPropertyIsNotFound);
   end;
  result := P.GetValue(NParams);
end;

procedure TTitanScriptObject.PutDefaultProperty(const Value: Variant; NParams: Integer);
var
  P: TTitanProperty;
begin
  P := PropertyList.GetDefaultProperty;
  if P = nil then
    raise TTitanScriptFailure.Create(errDefaultPropertyIsNotFound);
  P.PutValue(NParams, Value);
end;

function TTitanScriptObject.CreateProperty(PropertyNameIndex: Integer;
                                         PValue: PVariant;
                                         MemberRec: TTitanMemberRec): TTitanProperty;
var
  I: Integer;
begin
  I := PropertyList.IndexOf(PropertyNameIndex);
  if I <> -1 then
  begin
    result := PropertyList.Properties[I];
    Exit;
  end;

  result := TTitanProperty.Create(Self, PValue, MemberRec);
  PropertyList.AddObject(PropertyNameIndex, result);
end;

function TTitanScriptObject.Duplicate: TTitanScriptObject;
var
  I, PropertyNameIndex: Integer;
  SourceProperty, DestProperty: TTitanProperty;
  P: PVariant;
  SO: TTitanScriptObject;
  V: Variant;
begin
  result := TTitanScriptObject.Create(ClassRec);
  result.RefCount := 1;

  if Instance = nil then
    result.Instance := Instance;

  result.PClass := PClass;

  if ExtraPtrSize > 0 then
  begin
    result.ExtraPtrSize := ExtraPtrSize;
    result.ExtraPtr := AllocMem(ExtraPtrSize);
    Move(ExtraPtr^, result.ExtraPtr^, ExtraPtrSize);
  end;

  for I:=0 to PropertyList.Count - 1 do
  begin
    PropertyNameIndex := PtrInt(PropertyList[I]);

    SourceProperty := TTitanProperty(PropertyList.Objects[I]);

    if not SourceProperty.IsStatic then
    begin
      P := SourceProperty.fValue;

      if P <> nil then
        if IsPaxArray(P^) then
        begin
          SO := VariantToScriptObject(P^);
          V := (SO.Instance as TTitanArray).Duplicate;
          P := @V;
        end
        else if IsObject(P^) then
        begin
          SO := VariantToScriptObject(P^);
          SO := SO.Duplicate;
          V := ScriptObjectToVariant(SO);
          P := @V;
        end;
      DestProperty := TTitanProperty.Create(result, P, SourceProperty.MemberRec);
      result.PropertyList.AddObject(PropertyNameIndex, DestProperty);
    end;
  end;
end;

function TTitanScriptObject.ToString: String;
var
  I, K, T, L: PtrInt;
  P: TTitanProperty;
  LPtr, CurrPtr: Pointer;
  V: Variant;
  D: TTitanClassDefinition;
begin
  if ClassRec <> nil then
    if ClassRec.ck = ckArray then
    begin
       if Instance <> nil then
       begin
         result := TTitanArray(Instance).ToString;
         Exit;
       end;
    end
    else if ClassRec.ck = ckDynamicArray then
    begin
      LPtr := ExtraPtr;
      if LPtr = nil then
      begin
        result := '()';
        Exit;
      end
      else
      begin
        LPtr := ShiftPointer(LPtr, - SizeOf(Integer));
        L := PtrInt(LPtr^);
        if L = 0 then
        begin
          result := '()';
          Exit;
        end;
        D := ClassRec.GetClassDef;
        if D = nil then
        begin
          result := '()';
          Exit;
        end;
        if D.ElSize = _SizeVariant then
        begin
          result := '(';
          CurrPtr := ExtraPtr;
          for I := 0 to L - 1 do
          begin
            result := result + ToStr(Scripter, Variant(CurrPtr^));
            if I <> L - 1 then
              result := result + ',';
            Inc(PtrInt(CurrPtr), _SizeVariant);
          end;
          result := result + ')';
        end
        else
        begin
          result := '(';
          for I := 0 to L - 1 do
          begin
            CurrPtr := ShiftPointer(ExtraPtr, I * D.ElSize);
            result := result + ToStr(Scripter, GetVariantValue(Scripter,
                CurrPtr, D.ElType, D.ElTypeName));
            if I <> L - 1 then
              result := result + ',';
          end;
          result := result + ')';
        end;
      end;

      Exit;
    end;

  if (Instance = nil) and (PClass <> nil) then
    result := 'Class of ' + ClassRec.Name + ' = {'
  else
    result := 'Object of ' + ClassRec.Name + ' = {';

  if Instance <> nil then
  begin
    if Instance.InheritsFrom(TTitanScriptObject) then
    begin
      V := TTitanScriptObject(Instance).DefaultValue;
      if VarType(V) <> varUndefined then
      begin
        result := ToStr(Scripter, V);
        Exit;
      end;
    end;

    if Instance.InheritsFrom(RegExp) then
    begin
      result := RegExp(Instance).toString;
      Exit;
    end;
  end;

  K := 0;
  for I:=0 to PropertyList.Count - 1 do
  begin
    P := TTitanProperty(PropertyList.Objects[I]);
    if P.MemberRec <> nil then
      if P.MemberRec.Kind in [KindVAR] then
      begin
        Inc(K);

        if P.MemberRec.ID > 0 then
        with TTitanBaseScripter(Scripter).SymbolTable do
        begin
          T := PType[P.MemberRec.ID];
          if StrEql(Name[T], ClassRec.Name) then
          begin
            result := result + 'Object of ' + ClassRec.Name;
            continue;
          end;
        end;

        result := result + ToStr(Scripter, P.Value[0]) + ',';
      end;
  end;

  if K = 0 then
    result := result + '}'
  else
    result := Copy(result, 1, Length(result) - 1) + ')';
end;

constructor TTitanMemberRec.Create(ID: Integer; ClassRec: TTitanClassRec);
begin
  Self.ID := ID;
  Self.fClassrec := ClassRec;
  ReadID := 0;
  WriteID := 0;
  InitN := 0;
  ml := [];
  Kind := 0;
  Definition := nil;
  UpCaseIndex := 0;
  IsPublished := false;
  IsSource := ID >= FirstSymbolCard;
  NParams := 0;
  IsImplementationSection := false;
end;

destructor TTitanMemberRec.Destroy;
begin
  inherited;
end;

function TTitanMemberRec.Scripter: Pointer;
begin
  result := fClassRec.Scripter;
end;

procedure TTitanMemberRec.SaveToStream(S: TStream);
begin
  SaveInteger(ID, S);
  SaveInteger(ReadID, S);
  SaveInteger(WriteID, S);
  SaveInteger(InitN, S);
  S.WriteBuffer(ml, SizeOf(ml));
  SaveInteger(Kind, S);
  SaveInteger(NParams, S);
end;

procedure TTitanMemberRec.LoadFromStream(S: TStream;
                                       DS: Integer = 0; DP: Integer = 0);
begin
  ID := LoadInteger(S);
  if ID > 0 then
    Inc(ID, DS);
  ReadID := LoadInteger(S);
  if ReadID > 0 then
    Inc(ReadID, DS);
  WriteID := LoadInteger(S);
  if WriteID > 0 then
    Inc(WriteID, DS);
  InitN := LoadInteger(S);
  if InitN > 0 then
    Inc(InitN, DP);
  S.ReadBuffer(ml, SizeOf(ml));
  Kind := LoadInteger(S);
  NParams := LoadInteger(S);

  UpCaseIndex := NameIndexToUpperCaseIndex(GetNameIndex, Scripter);
end;

procedure TTitanMemberRec.CheckAccess;
var
  _ClassID, _SubID, L, I, T: Integer;
  Found: Boolean;
  TempClassRec: TTitanClassRec;
begin
  if ID = 0 then
    Exit;
  if fClassRec = nil then
    Exit;
  if not (modPRIVATE in ml) then
    Exit;
  Found := false;
  with TTitanBaseScripter(fClassRec.Scripter) do
  begin
    if fClassRec.ClassID = SymbolTable.RootNamespaceID then
      Exit;

    if State = _ssRunning then
    begin
      _ClassID := fClassRec.ClassID;
      with Code do
        for I:=1 to LevelStack.Card do
        if LevelStack[I] <> 0 then
        begin
          _SubID := LevelStack.fItems[I];
          L := SymbolTable.Level[_SubID];
          if L = 0 then
            Continue;
          if L = _ClassID then
          begin
            Found := true;
            Break;
          end;
          TempClassRec := TTitanBaseScripter(fClassRec.Scripter).ClassList.FindClass(L);
          if TempClassRec.InheritsFromClass(_ClassID) then
          begin
            Found := true;
            Break;
          end;
        end;
    end;
  end;
  if not Found then
  begin
    with TTitanBaseScripter(fClassRec.Scripter).SymbolTable do
    begin
      T := PType[ID];
      if T > PAXTypes.Count then
        if Kind[T] <> KindType then
        begin
          with TTitanBaseScripter(fClassRec.Scripter).Code do
          begin
            N := FindCreateObjectStmt(T);
          end;
          raise TTitanScriptFailure.Create(Format(errUndeclaredIdentifier, [Name[T]]));
        end;
    end;
//  raise TTitanScriptFailure.Create(Format(errNotAccessToPrivateMember, [GetName, fClassRec.Name]));
  end;
end;

function TTitanMemberRec.GetName: String;
begin
  if fClassRec = nil then
    result := ''
  else
    result := TTitanBaseScripter(fClassRec.Scripter).SymbolTable.Name[ID];
end;

function TTitanMemberRec.GetNameIndex: Integer;
begin
  if fClassRec = nil then
    result := 0
  else
    result := TTitanBaseScripter(fClassRec.Scripter).SymbolTable.NameIndex[ID];
end;

function TTitanMemberRec.IsStatic: Boolean;
begin
  result := modSTATIC in ml;
end;

function TTitanMemberRec.IsDefault: Boolean;
begin
  result := modDEFAULT in ml;
end;

function TTitanMemberRec.IsImported: Boolean;
begin
  result := (Definition <> nil);
end;

function TTitanMemberRec.IsImportedObject: Boolean;
begin
  result := IsImported;
  if result then
    result := Definition.DefKind = dkObject;
end;

constructor TTitanClassRec.Create(AScripter: Pointer; kc: TTitanClassKind);
begin
  inherited Create;
  Scripter := AScripter;
  Self.ck := ck;
  MemberList := TTitanMemberList.Create(Self);
  fClassDef := nil;
  OwnerClassRec := nil;
  AncestorClassRec := nil;
  NameIdx := 0;
  UpCaseIndex := 0;
  UsingInitList := TTitanIDs.Create(false);
  AutoDestructorID := 0;
  IsSet := false;
  PtiSet := nil;
  fHasRunTimeProperties := false;
end;

function TTitanClassRec.HasRunTimeProperties: Boolean;
var
  C: TTitanClassRec;
begin
  result := fHasRunTimeProperties;
  if not result then
    if AncestorName <> '' then
    begin
      C := TTitanBaseScripter(Scripter).ClassList.FindClassByName(AncestorName);
      if C <> nil then
        result := C.HasRunTimeProperties;
    end;
end;

destructor TTitanClassRec.Destroy;
begin
  MemberList.Free;
  UsingInitList.Free;

  inherited;
end;

function TTitanClassRec.FindBinaryOperatorID(const OperName: String; T1, T2: Integer): Integer;
var
  I, SubID, ParamID1, ParamID2, TypeID1, TypeID2: Integer;
  SymbolTable: TTitanSymbolTable;
  NameIndex: Integer;
  TempID: Integer;
begin
  NameIndex := CreateNameIndex(OperName, Scripter);
  TempID := 0;
  SymbolTable := TTitanBaseScripter(Scripter).SymbolTable;

  for I:=MemberList.Count - 1 downto 0 do
  if MemberList[I].Kind = KindSUB then
  if MemberList.NameID[I] = NameIndex then
  begin
    SubID := MemberList[I].ID;
    if SubID > 0 then
      if SymbolTable.Count[SubID] = 2 then
      begin
        ParamID1 := SymbolTable.GetParamID(SubID, 1);
        ParamID2 := SymbolTable.GetParamID(SubID, 2);
        TypeID1 := SymbolTable.PType[ParamID1];
        TypeID2 := SymbolTable.PType[ParamID2];

        with TTitanBaseScripter(Scripter) do
        if MatchAssignmentStrict(TypeID1, T1) and MatchAssignmentStrict(TypeID2, T2) then
        begin
          TempID := SubID;
          if (TypeID1 = T1) and (TypeID2 = T2) then
            break;
        end;
      end;
  end;
  result := TempID;
end;

function TTitanClassRec.FindUnaryOperatorID(const OperName: String; T: Integer): Integer;
var
  I, SubID, ID: Integer;
  SymbolTable: TTitanSymbolTable;
  NameIndex: Integer;
begin
  NameIndex := CreateNameIndex(OperName, Scripter);
  result := 0;
  SymbolTable := TTitanBaseScripter(Scripter).SymbolTable;

  for I:=MemberList.Count - 1 downto 0 do
  if MemberList[I].Kind = KindSUB then
  if MemberList.NameID[I] = NameIndex then
  begin
    SubID := MemberList[I].ID;
    if SubID > 0 then
      if SymbolTable.Count[SubID] = 1 then
      begin
        ID := SymbolTable.GetParamID(SubID, 1);
        if (SymbolTable.PType[ID] = T) then
        begin
          result := SubID;
          Exit;
        end;
      end;
  end;
end;

function TTitanClassRec.IsStatic: Boolean;
begin
  result := modSTATIC in ml;
end;

function TTitanClassRec.ModuleName: String;
begin
  with TTitanBaseScripter(Scripter) do
    if (ModuleID >= 0) and (ModuleID < Modules.Count - 1) then
      result := Modules[ModuleID];
end;

function TTitanClassRec.GetClassList: TTitanClassList;
begin
  result := TTitanBaseScripter(Scripter).ClassList;
end;

procedure TTitanClassRec.ResetCompileStage;
var
  I: Integer;
begin
  for I:=MemberList.Count - 1 downto 0 do
    if MemberList[I].IsSource or (MemberList[I].ID > FirstSymbolCard) then
      MemberList.DeleteObject(I);
  UsingInitList.Clear;
end;

procedure TTitanClassRec.SaveToStream(S: TStream);
begin
  SaveString(Name, S);
  SaveString(OwnerName, S);
  SaveString(AncestorName, S);

  SaveInteger(ClassID, S);
  SaveInteger(ModuleID, S);
  SaveInteger(AutoDestructorID, S);

  S.WriteBuffer(ck, SizeOf(ck));

  MemberList.SaveToStream(S);
  UsingInitList.SaveToStream(S);
end;

procedure TTitanClassRec.LoadFromStream(S: TStream;
                                      DS: Integer = 0; DP: Integer = 0);
var
  I: Integer;
begin
  Name := LoadString(S);
  OwnerName := LoadString(S);
  AncestorName := LoadString(S);

  ClassID := LoadInteger(S);
  if ClassID > 0 then
    Inc(ClassID, DS);

  ModuleID := LoadInteger(S);

  AutoDestructorID := LoadInteger(S);
  if AutoDestructorID > 0 then
    Inc(AutoDestructorID, DS);

  S.ReadBuffer(ck, SizeOf(ck));

  MemberList.LoadFromStream(S, DS, DP);
  UsingInitList.LoadFromStream(S);
  if DP <> 0 then
    for I:=0 to UsingInitList.Count - 1 do
      if UsingInitList[I] > 0 then
        UsingInitList[I] := UsingInitList[I] + DP;

  NameIdx := CreateNameIndex(Name, Scripter);
  UpCaseIndex :=NameIndexToUpperCaseIndex(NameIdx, Scripter);
end;

function TTitanClassRec.InheritsFromClass(AClassID: Integer): Boolean;
var
  P: TTitanClassRec;
begin
  if ClassID = AClassID then
  begin
    result := true;
    Exit;
  end;
  result := false;
  P := AncestorClassRec;
  while P <> nil do
  begin
    if P.ClassID = AClassID then
    begin
      result := true;
      Exit;
    end;
    P := P.AncestorClassRec;
  end;
end;

function TTitanClassRec.IsImported: Boolean;
begin
  result := GetClassDef <> nil;
end;

function TTitanClassRec.DelphiClass: TClass;
var
  D: TTitanClassDefinition;
begin
  result := nil;
  D := GetClassDef;
  if D = nil then
    Exit;
  result := D.PClass;
end;

function TTitanClassRec.DelphiClassEx: TClass;
var
  ClassRec: TTitanClassRec;
begin
  result := DelphiClass;
  if result = nil then if AncestorName <> '' then
  begin
    ClassRec := TTitanBaseScripter(scripter).ClassList.FindClassByName(AncestorName);
    if ClassRec = nil then
      Exit;
    result := ClassRec.DelphiClassEx;
  end;
end;

function TTitanClassRec.HasPublishedProperty(const PropertyName: String): Boolean;
var
  C: TClass;
  CR: TTitanClassRec;
begin
  result := false;
  C := DelphiClass;
  if C = nil then
  begin
    if AncestorName <> '' then
    begin
      CR := TTitanBaseScripter(Scripter).ClassList.FindClassByName(AncestorName);
      if CR <> nil then
        result := CR.HasPublishedProperty(PropertyName);
    end;
    Exit;
  end;
  result := BASE_Engine.HasPublishedProperty(C, PropertyName, Scripter);
end;

function TTitanClassRec.HasPublishedPropertyEx(const PropertyName: String; var AClassName: String): Boolean;
var
  C: TClass;
  CR: TTitanClassRec;
begin
  result := false;
  C := DelphiClass;
  if C = nil then
  begin
    if AncestorName <> '' then
    begin
      CR := TTitanBaseScripter(Scripter).ClassList.FindClassByName(AncestorName);
      if CR <> nil then
        result := CR.HasPublishedPropertyEx(PropertyName, AClassName);
    end;
    Exit;
  end;
  result := BASE_Engine.HasPublishedPropertyEx(C, PropertyName, AClassName, Scripter);
end;

function TTitanClassRec.GetClassDef: TTitanClassDefinition;
begin
  result := fClassDef;
  if result = nil then
    if AncestorClassRec <> nil then
      result := AncestorClassRec.GetClassDef;
end;

function TTitanClassRec.GetMember(MemberNameIndex: Integer; UpCaseON: Boolean = true): TTitanMemberRec;
var
  U, I: Integer;
begin
  result := TTitanMemberRec(MemberList.GetObject(MemberNameIndex));

  if result = nil then
    if UpCaseON then
    begin
      U := NameIndexToUpperCaseIndex(MemberNameIndex, Scripter);
      I := MemberList.UpperCaseIndexOf(U);
      if I >= 0 then
        result := MemberList[I];
    end;
end;

function TTitanClassRec.FindMember(MemberNameIndex: Integer;
                                 ma: TTitanMemberAccess;
                                 UpCaseON: Boolean = true): TTitanMemberRec;
begin
  result := nil;
  case ma of
    maAny:
    begin
      result := GetMember(MemberNameIndex, UpCaseON);
      if result = nil then
        if AncestorClassRec <> nil then
          result := AncestorClassRec.FindMember(MemberNameIndex, maAny, UpCaseON);
    end;
    maMyClass:
      result := GetMember(MemberNameIndex, UpCaseON);
    maMyBase:
    begin
      if AncestorClassRec <> nil then
        result := AncestorClassRec.FindMember(MemberNameIndex, maAny, UpCaseON);
    end;
  end;
end;

function TTitanClassRec.FindMemberEx(MemberNameIndex: Integer;
                                   ma: TTitanMemberAccess;
                                   var FoundInBaseClass: Boolean;
                                   UpCaseON: Boolean = true): TTitanMemberRec;
begin
  FoundInBaseClass := false;

  result := nil;
  case ma of
    maAny:
    begin
      result := GetMember(MemberNameIndex,UpCaseON);
      if result = nil then
        if AncestorClassRec <> nil then
        begin
          result := AncestorClassRec.FindMember(MemberNameIndex, maAny, UpCaseON);
          FoundInBaseClass := true;
        end;
    end;
    maMyClass:
      result := GetMember(MemberNameIndex, UpCaseON);
    maMyBase:
    begin
      if AncestorClassRec <> nil then
        result := AncestorClassRec.FindMember(MemberNameIndex, maAny, UpCaseON);
    end;
  end;
end;

procedure TTitanClassRec.DeleteMember(ID: Integer);
var
  I: Integer;
begin
  for I:=0 to MemberList.Count - 1 do
   if MemberList.Records[I].ID = ID then
   begin
     MemberList.DeleteObject(I);
     Exit;
   end;
end;

function TTitanClassRec.AddField(ID: Integer; ml: TTitanModifierList): TTitanMemberRec;
var
  NameIndex: Integer;
begin
  NameIndex := TTitanBaseScripter(Scripter).SymbolTable.NameIndex[ID];

  result := TTitanMemberRec.Create(ID, Self);
  result.ml := ml;
  result.Kind := KindVAR;

  if UpCaseIndex > 0 then
    result.UpCaseIndex := NameIndexToUpperCaseIndex(NameIndex, Scripter);

  MemberList.AddObject(NameIndex, result);
end;

function TTitanClassRec.AddMethod(ID: Integer; ml: TTitanModifierList): TTitanMemberRec;
var
  NameIndex: Integer;
begin
  NameIndex := TTitanBaseScripter(Scripter).SymbolTable.NameIndex[ID];

  result := TTitanMemberRec.Create(ID, Self);
  result.ml := ml;
  result.Kind := KindSUB;

  if UpCaseIndex > 0 then
    result.UpCaseIndex := NameIndexToUpperCaseIndex(NameIndex, Scripter);

  MemberList.AddObject(NameIndex, result);
end;

function TTitanClassRec.AddHostMethod(D: TTitanMethodDefinition): TTitanMemberRec;
var
  SubID, NameIndex: Integer;
begin
  NameIndex := CreateNameIndex(D.Name, Scripter);

  SubID := - D.Index;

  result := TTitanMemberRec.Create(SubID, Self);
  result.ml := D.ml;
  result.Kind := KindSUB;
  result.Definition := D;
  result.UpCaseIndex := NameIndexToUpperCaseIndex(NameIndex, Scripter);
  MemberList.AddObject(NameIndex, result);
end;

function TTitanClassRec.AddHostConstant(D: TTitanConstantDefinition): TTitanMemberRec;
var
  ID, I, NameIndex: Integer;
begin
  NameIndex := CreateNameIndex(D.Name, Scripter);

  I := MemberList.IndexOf(NameIndex);
  if I <> -1 then
    result := MemberList[I]
  else
  begin
// NWR

//  ID := - D.Index;

{    if D.DefList = DefinitionList then
      ID := TTitanBaseScripter(Scripter).SymbolTable.AppVariant(D.Index)
    else
      ID := TTitanBaseScripter(Scripter).SymbolTable.AppVariant(-D.Index);

    TTitanBaseScripter(Scripter).SymbolTable.NameIndex[ID] := NameIndex;
    TTitanBaseScripter(Scripter).SymbolTable.Kind[ID] := KindHostConst; }

    if D.DefList = GlobalDefinitionList then
      ID := - D.Index
    else
    begin
      ID := TTitanBaseScripter(Scripter).SymbolTable.LookUpId(D.Name, ClassID);
      if ID = 0 then
        ID := TTitanBaseScripter(Scripter).SymbolTable.AppVariant(-D.Index)
      else
        TTitanBaseScripter(Scripter).SymbolTable.PutVariant(ID, -D.Index);

      TTitanBaseScripter(Scripter).SymbolTable.NameIndex[ID] := NameIndex;
      TTitanBaseScripter(Scripter).SymbolTable.Kind[ID] := KindHostConst;
      TTitanBaseScripter(Scripter).SymbolTable.Imported[ID] := true;
      TTitanBaseScripter(Scripter).SymbolTable.Level[ID] := ClassID;
      TTitanBaseScripter(Scripter).SymbolTable.PType[ID] := D.TypeID;
    end;

// NWR

    result := TTitanMemberRec.Create(ID, Self);
    MemberList.AddObject(NameIndex, result);
    result.ml := D.ml;
    result.Kind := KindCONST;
    result.Definition := D;
    result.IsSource := false;
    result.UpCaseIndex := NameIndexToUpperCaseIndex(NameIndex, Scripter);
  end;
end;

function TTitanClassRec.AddHostVariable(D: TTitanVariableDefinition): TTitanMemberRec;
var
  ID, I, TypeId, NameIndex: Integer;
  C: TTitanClassRec;
  SO: TTitanScriptObject;
  V: Variant;
begin
  NameIndex := CreateNameIndex(D.Name, Scripter);

  I := MemberList.IndexOf(NameIndex);
  if I <> -1 then
    result := MemberList[I]
  else
  begin
// NWR
//  ID := - D.Index;

{    if D.DefList = DefinitionList then
      ID := TTitanBaseScripter(Scripter).SymbolTable.AppVariant(D.Index)
    else
      ID := TTitanBaseScripter(Scripter).SymbolTable.AppVariant(-D.Index);

    TTitanBaseScripter(Scripter).SymbolTable.NameIndex[ID] := NameIndex;
    TTitanBaseScripter(Scripter).SymbolTable.Kind[ID] := KindHostVar; }

    if D.DefList = GlobalDefinitionList then
      ID := - D.Index
    else
    begin
      ID := TTitanBaseScripter(Scripter).SymbolTable.LookUpId(D.Name, ClassID);
      if ID = 0 then
        ID := TTitanBaseScripter(Scripter).SymbolTable.AppVariant(-D.Index)
      else
        TTitanBaseScripter(Scripter).SymbolTable.PutVariant(ID, -D.Index);

      TypeID := D.TypeID;
      if TypeID = -1 then
      begin
        C := TTitanBaseScripter(Scripter).ClassList.FindClassByName(D.TypeName);
        if C <> nil then
        begin
          TypeID := C.ClassID;
          D.TypeID := typeCLASS;
        end;
      end;

      TTitanBaseScripter(Scripter).SymbolTable.NameIndex[ID] := NameIndex;
      TTitanBaseScripter(Scripter).SymbolTable.Kind[ID] := KindHostVar;
      TTitanBaseScripter(Scripter).SymbolTable.PType[ID] := TypeID;
      TTitanBaseScripter(Scripter).SymbolTable.Imported[ID] := true;
      TTitanBaseScripter(Scripter).SymbolTable.Level[ID] := ClassID;
    end;

// NWR

    result := TTitanMemberRec.Create(ID, Self);
    MemberList.AddObject(NameIndex, result);
    result.UpCaseIndex := NameIndexToUpperCaseIndex(NameIndex, Scripter);

    if D.TypeID = -1 then
    begin
      C := TTitanBaseScripter(Scripter).ClassList.FindClassByName(D.TypeName);
      if C <> nil then
      begin
        if C.ck = ckDynamicArray then
        begin
          SO := C.CreateScriptObject;
          SO.Instance := SO;

          if C.IsStaticArray then
            SO.ExtraPtr := Pointer(D.Address)
          else
            SO.ExtraPtr := Pointer(D.Address^);
          SO.ExternalExtraPtr := true;
          V := ScriptObjectToVariant(SO);
          result.ID := TTitanBaseScripter(Scripter).SymbolTable.AppVariant(V);
          TTitanBaseScripter(Scripter).SymbolTable.Imported[result.ID] := true;

          if D.DefList <> GlobalDefinitionList then
            TTitanBaseScripter(Scripter).SymbolTable.Name[ID] := '';

          Exit;
        end
        else if C.ck = ckStructure then
        begin
          SO := C.CreateScriptObject;
          SO.FreeExtraPtr;

          SO.Instance := SO;
          SO.ExtraPtr := D.Address;
          SO.ExternalExtraPtr := true;
          V := ScriptObjectToVariant(SO);
          result.ID := TTitanBaseScripter(Scripter).SymbolTable.AppVariant(V);
          TTitanBaseScripter(Scripter).SymbolTable.Imported[result.ID] := true;
          Exit;
        end;
      end;
    end;

    result.ml := D.ml;
    result.Kind := KindVAR;
    result.Definition := D;
    result.IsSource := false;
  end;
end;

function TTitanClassRec.AddHostObject(D: TTitanObjectDefinition): TTitanMemberRec;
var
  ID, NameIndex, I: Integer;
  SO: TTitanScriptObject;
  ClassRec: TTitanClassRec;
begin
  NameIndex := CreateNameIndex(D.Name, Scripter);

  I := MemberList.IndexOf(NameIndex);
  if I <> -1 then
    result := MemberList[I]
  else
  begin
// NWR
//  ID := - D.Index;

{    if D.DefList = DefinitionList then
      ID := TTitanBaseScripter(Scripter).SymbolTable.AppVariant(D.Index)
    else
      ID := TTitanBaseScripter(Scripter).SymbolTable.AppVariant(-D.Index);

    TTitanBaseScripter(Scripter).SymbolTable.NameIndex[ID] := NameIndex;
    TTitanBaseScripter(Scripter).SymbolTable.Kind[ID] := KindHostObject;  }

    if D.DefList = GlobalDefinitionList then
      ID := - D.Index
    else
    begin
      ID := TTitanBaseScripter(Scripter).SymbolTable.LookUpId(D.Name, ClassID);
      if ID = 0 then
        ID := TTitanBaseScripter(Scripter).SymbolTable.AppVariant(-D.Index)
      else
        TTitanBaseScripter(Scripter).SymbolTable.PutVariant(ID, -D.Index);

      TTitanBaseScripter(Scripter).SymbolTable.NameIndex[ID] := NameIndex;
      TTitanBaseScripter(Scripter).SymbolTable.Kind[ID] := KindHostObject;
      TTitanBaseScripter(Scripter).SymbolTable.PType[ID] := typeCLASS;
      TTitanBaseScripter(Scripter).SymbolTable.Imported[ID] := true;
      TTitanBaseScripter(Scripter).SymbolTable.Level[ID] := ClassID;

      ClassRec := TTitanBaseScripter(Scripter).ClassList.FindClassByName(D.Instance.ClassName);
      if ClassRec <> nil then
        TTitanBaseScripter(Scripter).SymbolTable.PType[ID] := ClassRec.ClassID;
    end;

// NWR

    result := TTitanMemberRec.Create(ID, Self);
    MemberList.AddObject(NameIndex, result);
    result.ml := D.ml;
    result.Kind := KindVAR;
    result.Definition := D;
    result.IsSource := false;
    result.UpCaseIndex := NameIndexToUpperCaseIndex(NameIndex, Scripter);
  end;

  SO := DelphiInstanceToScriptObject(D.Instance, Scripter);
  SO.RefCount := 0;
  D.Value := ScriptObjectToVariant(SO);
end;

function TTitanClassRec.AddVirtualObject(D: TTitanVirtualObjectDefinition): TTitanMemberRec;
var
  ID, NameIndex, I: Integer;
begin
  NameIndex := CreateNameIndex(D.Name, Scripter);

  I := MemberList.IndexOf(NameIndex);
  if I <> -1 then
    result := MemberList[I]
  else
  begin
    ID := TTitanBaseScripter(Scripter).SymbolTable.LookUpId(D.Name, ClassID);
    if ID = 0 then
      ID := TTitanBaseScripter(Scripter).SymbolTable.AppVariant(-D.Index)
    else
      TTitanBaseScripter(Scripter).SymbolTable.PutVariant(ID, -D.Index);

    TTitanBaseScripter(Scripter).SymbolTable.NameIndex[ID] := NameIndex;
    TTitanBaseScripter(Scripter).SymbolTable.Kind[ID] := KindVirtualObject;
    TTitanBaseScripter(Scripter).SymbolTable.PType[ID] := typeVARIANT;
    TTitanBaseScripter(Scripter).SymbolTable.Imported[ID] := true;
    TTitanBaseScripter(Scripter).SymbolTable.Level[ID] := ClassID;

    result := TTitanMemberRec.Create(ID, Self);
    MemberList.AddObject(NameIndex, result);
    result.ml := D.ml;
    result.Kind := KindVAR;
    result.Definition := D;
    result.IsSource := false;
    result.UpCaseIndex := NameIndexToUpperCaseIndex(NameIndex, Scripter);
  end;
end;

function TTitanClassRec.AddHostInterfaceVar(D: TTitanInterfaceVarDefinition): TTitanMemberRec;
var
  ID, NameIndex, I: Integer;
  SO: TTitanScriptObject;
  ClassRec: TTitanClassRec;
  ClassDef: TTitanClassDefinition;
begin
  NameIndex := CreateNameIndex(D.Name, Scripter);

  ClassRec := nil;

  I := MemberList.IndexOf(NameIndex);
  if I <> -1 then
    result := MemberList[I]
  else
  begin
    if D.DefList = GlobalDefinitionList then
      ID := - D.Index
    else
    begin
      ID := TTitanBaseScripter(Scripter).SymbolTable.LookUpId(D.Name, ClassID);
      if ID = 0 then
        ID := TTitanBaseScripter(Scripter).SymbolTable.AppVariant(-D.Index)
      else
        TTitanBaseScripter(Scripter).SymbolTable.PutVariant(ID, -D.Index);

      TTitanBaseScripter(Scripter).SymbolTable.NameIndex[ID] := NameIndex;
      TTitanBaseScripter(Scripter).SymbolTable.Kind[ID] := KindHostInterfaceVar;
      TTitanBaseScripter(Scripter).SymbolTable.PType[ID] := typeCLASS;
      TTitanBaseScripter(Scripter).SymbolTable.Imported[ID] := true;
      TTitanBaseScripter(Scripter).SymbolTable.Level[ID] := ClassID;

      ClassDef := GlobalDefinitionList.FindInterfaceTypeDef(D.guid);

      ClassRec := TTitanBaseScripter(Scripter).ClassList.FindClassByName(ClassDef.Name);
      if ClassRec <> nil then
        TTitanBaseScripter(Scripter).SymbolTable.PType[ID] := ClassRec.ClassID;
    end;

// NWR

    result := TTitanMemberRec.Create(ID, Self);
    MemberList.AddObject(NameIndex, result);
    result.ml := D.ml;
    result.Kind := KindVAR;
    result.Definition := D;
    result.IsSource := false;
    result.UpCaseIndex := NameIndexToUpperCaseIndex(NameIndex, Scripter);

    if ClassRec <> nil then
    begin
      SO := TTitanBaseScripter(Scripter).ScriptObjectList.FindScriptObjectByIntf(D.PIntf);
      if SO = nil then
      begin
        SO := ClassRec.CreateScriptObject();
        SO.PIntf := D.PIntf;
        SO.RefCount := 0;
      end;
      D.Value := ScriptObjectToVariant(SO);
    end;
  end;
end;

function TTitanClassRec.AddHostField(D: TTitanFieldDefinition): TTitanMemberRec;
var
  ID, NameIndex, I: Integer;
begin
  NameIndex := CreateNameIndex(D.Name, Scripter);

  I := MemberList.IndexOf(NameIndex);
  if I <> -1 then
    result := MemberList[I]
  else
  begin
    ID := - D.Index;
    result := TTitanMemberRec.Create(ID, Self);
    MemberList.AddObject(NameIndex, result);
    result.ml := D.ml;
    result.Kind := KindVAR;
    result.Definition := D;
    result.IsSource := false;
    result.UpCaseIndex := NameIndexToUpperCaseIndex(NameIndex, Scripter);
  end;
end;

function TTitanClassRec.AddHostRecordField(D: TTitanRecordFieldDefinition): TTitanMemberRec;
var
  ID, NameIndex, I: Integer;
begin
  NameIndex := CreateNameIndex(D.Name, Scripter);

  I := MemberList.IndexOf(NameIndex);
  if I <> -1 then
    result := MemberList[I]
  else
  begin
    ID := - D.Index;
    result := TTitanMemberRec.Create(ID, Self);
    MemberList.AddObject(NameIndex, result);
    result.ml := D.ml;
    result.Kind := KindVAR;
    result.Definition := D;
    result.IsSource := false;
    result.UpCaseIndex := NameIndexToUpperCaseIndex(NameIndex, Scripter);
  end;
end;

function TTitanClassRec.AddHostProperty(D: TTitanPropertyDefinition): TTitanMemberRec;
var
  NameIndex: Integer;
begin
  NameIndex := CreateNameIndex(D.Name, Scripter);
{
  I := MemberList.IndexOf(NameIndex);
  if I <> -1 then
  begin
    result := MemberList[I];
    Exit;
  end;
}
  result := TTitanMemberRec.Create(0, Self);
  result.ml := D.ml;
  result.Kind := KindPROP;
  result.Definition := D;
  result.NParams := D.NP;

  result.UpCaseIndex := NameIndexToUpperCaseIndex(NameIndex, Scripter);

  MemberList.AddObject(NameIndex, result);
end;

function TTitanClassRec.AddProperty(ID: Integer; ml: TTitanModifierList): TTitanMemberRec;
var
  I, NameIndex: Integer;
begin
  NameIndex := TTitanBaseScripter(Scripter).SymbolTable.NameIndex[ID];

  I := MemberList.IndexOf(NameIndex);

  if I = -1 then
  begin
    result := TTitanMemberRec.Create(ID, Self);
    result.ml := ml;
    result.Kind := KindPROP;

    if UpCaseIndex > 0 then
      result.UpCaseIndex := NameIndexToUpperCaseIndex(NameIndex, Scripter);

    MemberList.AddObject(NameIndex, result);
  end
  else
    result := MemberList.Objects[I];
end;

function TTitanClassRec.AddNestedClass(ID: Integer; ml: TTitanModifierList): TTitanMemberRec;
var
  NameIndex, I: Integer;
begin
  NameIndex := TTitanBaseScripter(Scripter).SymbolTable.NameIndex[ID];

  I := MemberList.IndexOf(NameIndex);
  if (I = -1) or ((I >= 0) and (MemberList[I].Kind <> KindTYPE)) then
  begin
    result := TTitanMemberRec.Create(ID, Self);
    result.ml := ml;
    result.Kind := KindTYPE;

    if UpCaseIndex > 0 then
      result.UpCaseIndex := NameIndexToUpperCaseIndex(NameIndex, Scripter);

    MemberList.AddObject(NameIndex, result);
  end
  else
    result := MemberList[I];
end;

procedure TTitanClassRec.FindOverloadedSubList(NameIndex: Integer;
                                             ma: TTitanMemberAccess;
                                             Ids: TTitanIds);
var
  I: Integer;
begin
  case ma of
    maAny:
    begin
      FindOverloadedSubList(NameIndex, maMyClass, Ids);
      if Ids.Count > 0 then
        Exit;
      FindOverloadedSubList(NameIndex, maMyBase, Ids);
    end;
    maMyClass:
    begin
      for I:=0 to MemberList.Count - 1 do
        if MemberList[I].Kind = KindSUB then
        begin
          if MemberList.NameID[I] = NameIndex then
            Ids.Add(MemberList[I].ID);
        end;
    end;
    maMyBase:
    begin
      if AncestorName <> '' then
         AncestorClassRec := TTitanBaseScripter(Scripter).ClassList.FindClassByName(AncestorName);
      if AncestorClassRec <> nil then
        AncestorClassRec.FindOverloadedSubList(NameIndex, maAny, Ids);
    end;
  end;
end;

function TTitanClassRec.FindOverloadedSubID(NameIndex: Integer; ma: TTitanMemberAccess; NP: Integer): Integer;
var
  I, ID: Integer;
begin
  result := 0;
  case ma of
    maAny:
    begin
      result := FindOverloadedSubID(NameIndex, maMyClass, NP);
      if result = 0 then
        result := FindOverloadedSubID(NameIndex, maMyBase, NP);
    end;
    maMyClass:
    begin
      for I:=0 to MemberList.Count - 1 do
        if MemberList[I].Kind = KindSUB then
          if MemberList.NameID[I] = NameIndex then
          begin
            ID := MemberList[I].ID;
            if TTitanBaseScripter(Scripter).SymbolTable.Count[ID] = NP then
            begin
              result := ID;
              Exit;
            end;
          end;
    end;
    maMyBase:
      if AncestorClassRec <> nil then
        result := AncestorClassRec.FindOverloadedSubID(NameIndex, maAny, NP);
  end;
end;

function TTitanClassRec.CreateScriptObject(const ObjectName: String = ''): TTitanScriptObject;
var
  MemberRec: TTitanMemberRec;
  I, RefID: Integer;
  CR: TTitanClassRec;
  R: TTitanFieldRec;
  P: TTitanProperty;
//........................
  AR:TTitanArray;
  Bounds:array of Integer;
begin
  result := TTitanScriptObject.Create(Self);

  //=============================== By Sternas ====================================================================
 if fClassDef <> nil then
  if (fClassDef.ClassKind=ckArray) and (fClassDef.IsStaticArray) then
    begin
      SetLength(Bounds,1);
      Bounds[0] :=fClassDef.ElMax-fClassDef.ElMin;
      Ar:=TTitanArray.Create(Bounds,fClassDef.ElType);
      AR.SetMinMax(fClassDef.ElMin,fClassDef.ElMax);
      result.Instance:=AR;
    end;
  //=============================== By Sternas ====================================================================

  if fClassDef <> nil then
    if fClassDef.RecordSize > 0 then
    begin
      result.ExtraPtrSize := fClassDef.RecordSize;
      result.ExtraPtr := AllocMem(result.ExtraPtrSize);
    end;

  CR := Self;

  while CR <> nil do
  with CR do
  begin
    for I:=0 to MemberList.Count - 1 do
    if MemberList[I].Kind = KindVAR then
    begin
      MemberRec := MemberList[I];
      if MemberRec.IsStatic then
        with TTitanBaseScripter(Scripter).SymbolTable do
          result.CreateProperty(MemberList.NameID[I], Address[MemberRec.ID], MemberRec)
      else
      begin
        R := nil;
        with TTitanBaseScripter(Scripter) do
          if RegisteredFieldList.Count > 0 then
            R := RegisteredFieldList.FindRecord(ObjectName, Self.Name,
                     NameList[MemberList.NameID[I]]);

        P := result.CreateProperty(MemberList.NameID[I], @Undefined, MemberRec);

        if R <> nil then
          P.ExtraDef := TTitanVariableDefinition.Create(nil, '', R.FieldType, R.Address, nil);
      end;
    end;

    for I:=0 to MemberList.Count - 1 do
    if MemberList[I].Kind = KindPROP then
    begin
      MemberRec := MemberList[I];
      if not MemberRec.IsPublished then
      begin
        P := result.PropertyList.FindProperty(MemberList.NameID[I]);
        if P = nil then
          result.CreateProperty(MemberList.NameID[I], @Undefined, MemberRec);
      end
      else
        result.CreateProperty(MemberList.NameID[I], @Undefined, MemberRec);
    end;

    for I:=0 to MemberList.Count - 1 do
    if MemberList[I].Kind = KindSUB then
    begin
      MemberRec := MemberList[I];
      if not MemberRec.IsStatic then
      with TTitanBaseScripter(Scripter).SymbolTable do
        result.CreateProperty(MemberList.NameID[I], @Undefined {Address[MemberRec.ID]}, MemberRec);
    end;

    for I:=0 to MemberList.Count - 1 do
    if MemberList[I].Kind = KindVAR then
    begin
      MemberRec := MemberList[I];
      if not MemberRec.IsStatic then
        if MemberRec.InitN > 0 then
          with TTitanBaseScripter(Scripter).Code do
          begin
            SaveState;

            N := MemberRec.InitN;

            SignFOP := false;
            Run;

            if TTitanBaseScripter(Scripter).IsError then
            begin
              N := ErrorN;
              raise  TTitanScriptFailure.Create(TTitanBaseScripter(Scripter).ErrorInstance.Description);
            end;

            RestoreState;
            result.PutProperty(MemberList.NameID[I], SymbolTable.GetVariant(MemberRec.ID), 0);
          end;
    end;

    if IsImported and (AncestorClassRec = nil) then
      CR := TTitanBaseScripter(Scripter).ClassList.FindClassByName(AncestorName)
    else
      CR := AncestorClassRec;
  end;

  CR := Self;

  while CR <> nil do
  with CR do
  begin
    for I:=0 to MemberList.Count - 1 do
    if MemberList[I].Kind = KindSUB then
    begin
      MemberRec := MemberList[I];
      if not MemberRec.IsStatic then
        if MemberRec.InitN > 0 then
        with TTitanBaseScripter(Scripter).Code do
        begin
          SymbolTable.SaveState;
          SaveState;

          RefID := SymbolTable.AppReference(ScriptObjectToVariant(result),
                                            MemberList.NameID[I],
                                            maAny);

          N := Card;
          Add(OP_CALL, RefID, 0, 0);
          Add(OP_HALT, 0, 0, 0);

          SignFOP := false;
          fTerminated := false;
          Run;

          if TTitanBaseScripter(Scripter).IsError then
          begin
            N := ErrorN;
            raise TTitanScriptFailure.Create(TTitanBaseScripter(Scripter).ErrorInstance.Description);
          end;

          RestoreState;
          SymbolTable.RestoreState;

          Exit;
        end;
    end;
    CR := AncestorClassRec;
  end;
end;

procedure TTitanClassRec.InitStaticFields;
var
  MemberRec: TTitanMemberRec;
  I: Integer;
begin
  for I:=0 to UsingInitList.Count - 1 do
    with TTitanBaseScripter(Scripter) do
    begin
      Code.SaveState;
      Code.N := UsingInitList[I];
      Code.WithStack.Push(SymbolTable.RootNamespaceID);
      Code.SignFOP := false;
      Code.fTerminated := false;
      Code.SignInitStage := true;

      with Code do
      while (Prog[N].Op = OP_JS_OPERS_ON) or
            (Prog[N].Op = OP_JS_OPERS_OFF) or
            (Prog[N].Op = OP_UPCASE_ON) or
            (Prog[N].Op = OP_UPCASE_OFF) or
            (Prog[N].Op = OP_VBARRAYS_ON) or
            (Prog[N].Op = OP_VBARRAYS_OFF) or
            (Prog[N].Op = OP_DECLARE_ON) or
            (Prog[N].Op = OP_DECLARE_OFF) or
            (Prog[N].Op = OP_TRY_ON) or
            (Prog[N].Op = OP_ZERO_BASED_STRINGS_ON) or
            (Prog[N].Op = OP_ZERO_BASED_STRINGS_OFF) do
            Dec(N);

      Code.Run;

      Code.WithStack.Pop;
      Code.RestoreState;

      if IsError then
        Exit;
    end;

  for I:=0 to MemberList.Count - 1 do
  if MemberList[I].Kind = KindVAR then
  begin
    MemberRec := MemberList[I];
    if MemberRec.IsStatic and (MemberRec.InitN > 0) then
    with TTitanBaseScripter(Scripter).Code do
    begin
      SaveState;
      N := MemberRec.InitN;

      WithStack.Push(GetLanguageNamespaceID);
      WithStack.Push(SymbolTable.RootNamespaceID);

      SignFOP := false;
      fTerminated := false;

      Run;

      WithStack.Pop;
      WithStack.Pop;

      RestoreState;

      if TTitanBaseScripter(Scripter).IsError then
         Exit;

      if ClassObject <> nil then
        ClassObject.PutProperty(MemberList.NameID[I], SymbolTable.GetVariant(MemberRec.ID), 0);
    end;
  end;
end;

procedure TTitanClassRec.CreateClassObject;
var
  MemberRec: TTitanMemberRec;
  I: Integer;
  CR: TTitanClassRec;

  V: Variant;
  NameIndex: Integer;
  PropList: TTitanPropertyList;
begin
  with TTitanBaseScripter(Scripter).SymbolTable do
    V := GetVariant(ClassID);

  if IsObject(V) then
    ClassObject := VariantToScriptObject(V)
  else
  begin
    ClassObject := TTitanScriptObject.Create(Self);
    ClassObject.IsClass := true;
  end;

  PropList := ClassObject.PropertyList;

  CR := Self;

  while CR <> nil do
  begin
    for I:=0 to CR.MemberList.Count - 1 do
    if CR.MemberList[I].Kind in [KindVAR, KindCONST] then
    begin

     NameIndex := CR.MemberList.NameID[I];
     if PropList.FindProperty(NameIndex) <> nil then
        Continue;
      MemberRec := CR.MemberList[I];

      if MemberRec.IsStatic then
        with TTitanBaseScripter(Scripter) do
        begin
          ClassObject.CreateProperty(NameIndex, SymbolTable.Address[MemberRec.ID], MemberRec);
          DoNotDestroyList.Add(MemberRec.ID);
        end;
    end;

    for I:=0 to CR.MemberList.Count - 1 do
    if CR.MemberList[I].Kind = KindPROP then
    begin
      NameIndex := CR.MemberList.NameID[I];
      if PropList.FindProperty(NameIndex) <> nil then
        Continue;

      MemberRec := CR.MemberList[I];
      if MemberRec.IsStatic then
        ClassObject.CreateProperty(NameIndex, @Undefined, MemberRec);
    end;

    for I:=0 to CR.MemberList.Count - 1 do
    if CR.MemberList[I].Kind = KindSUB then
    begin
      NameIndex := CR.MemberList.NameID[I];
      if PropList.FindProperty(NameIndex) <> nil then
        Continue;

      MemberRec := CR.MemberList[I];
      if MemberRec.IsStatic then
      begin
        if MemberRec.ID = 0 then
          ClassObject.CreateProperty(NameIndex, @Undefined, MemberRec)
        else
          ClassObject.CreateProperty(NameIndex,
           TTitanBaseScripter(Scripter).SymbolTable.Address[MemberRec.ID], MemberRec);
      end;
    end;

    for I:=0 to CR.MemberList.Count - 1 do
    if CR.MemberList[I].Kind = KindTYPE then
    begin
      NameIndex := CR.MemberList.NameID[I];
      if PropList.FindProperty(NameIndex) <> nil then
        Continue;

      MemberRec := CR.MemberList[I];
      MemberRec.ml := MemberRec.ml + [modSTATIC];
      ClassObject.CreateProperty(NameIndex,
           TTitanBaseScripter(Scripter).SymbolTable.Address[MemberRec.ID], MemberRec);
    end;

    CR := CR.AncestorClassRec;
  end;

  with TTitanBaseScripter(Scripter).SymbolTable do
  begin
    Kind[ClassID] := kindTYPE;
    PutVariant(ClassID, ScriptObjectToVariant(ClassObject));
  end;
end;

function TTitanClassRec.GetConstructorID: Integer;
begin
  result := ClassID;
  with TTitanBaseScripter(Scripter).SymbolTable do
    while (Kind[result] <> KindSUB) do
      Inc(result);
end;

function TTitanClassRec.GetConstructorIDEx: Integer;
var
  C: TTitanClassRec;
begin
  result := ClassID;
  with TTitanBaseScripter(Scripter).SymbolTable do
  begin
    repeat
       if Kind[result] = KindSUB then
         if TypeSub[result] = tsConstructor then
           Exit;
       Inc(result);
     until result = Card;
  end;
  if AncestorName = '' then
    result := 0
  else
  begin
    C := TTitanBaseScripter(Scripter).ClassList.FindClassByName(AncestorName);
    if C = nil then
      result := 0
    else
      result := C.GetConstructorIDEx;
  end;
end;

function TTitanClassRec.GetConstructorIDEx(const Params: array of const): Integer;
var
  C: TTitanClassRec;
  I, T, ParamId, L: Integer;
  ok: Boolean;
  D: TTitanMethodDefinition;

label
  Ancestor;

begin
  result := ClassID;
  L := Length(Params);
  with TTitanBaseScripter(Scripter).SymbolTable do
  begin
    repeat
       if Kind[result] = KindSUB then
         if (TypeSub[result] = tsConstructor) and (Count[result] = L) then
         begin
            if result < 0 then
            begin
              D := GlobalDefinitionList[-result];
              if (D.PClass <> nil) and (DelphiClass <> nil) then
                if D.PClass <> DelphiClass then
                   goto Ancestor;
            end;

           ok := true;

           for I:=0 to L - 1 do
           begin
             ParamID := GetParamId(result, I + 1);

             T := PType[ParamId];

             case Params[I].VType of
               vtBoolean:
               begin
                  if not (T in [typeBOOLEAN, typeVARIANT]) then
                    ok := false;
               end;
               vtInteger:
               begin
                  if not (T in [typeINTEGER, typeBYTE, typeWORD, typeCARDINAL, typeINT64, typeSHORTINT, typeSMALLINT, typeVARIANT]) then
                    ok := false;
               end;
               vtAnsiString, vtString, vtChar:
               begin
                  if not (T in [typeSTRING, typeVARIANT]) then
                    ok := false;
               end;
               vtExtended, vtCurrency:
               begin
                  if not (T in [typeDOUBLE, typeVARIANT]) then
                    ok := false;
               end;
             end;
           end;

           if ok then
           begin
             Exit;
           end;
         end;
       if result > 0 then
         Inc(result)
       else
         Dec(result);

       if result < 0 then
         if (-result) >= GlobalDefinitionList.Count then
           goto Ancestor;

     until result = Card;
  end;

Ancestor:

  if AncestorName = '' then
    result := 0
  else
  begin
    C := TTitanBaseScripter(Scripter).ClassList.FindClassByName(AncestorName);
    if C = nil then
      result := 0
    else
      result := C.GetConstructorIDEx(Params);
  end;
end;

function TTitanClassRec.IsNestedClass(ClassID: Integer): Boolean;
var
  I: Integer;
  MemberRec: TTitanMemberRec;
begin
  result := false;
  for I:=0 to MemberList.Count - 1 do
  if MemberList[I].Kind = KindTYPE then
  begin
    MemberRec := MemberList[I];
    if MemberRec.ID = ClassID then
    begin
      result := true;
      Exit;
    end;
  end;
end;

function TTitanClassRec.FindNestedClassID(ClassNameIndex: Integer): Integer;
var
  I, U: Integer;
  MemberRec: TTitanMemberRec;
begin
  result := 0;
  for I:=0 to MemberList.Count - 1 do
  begin
    MemberRec := MemberList[I];
    if MemberRec.Kind = KindTYPE then
      if MemberList.NameID[I] = ClassNameIndex then
      begin
        result := MemberRec.ID;
        Exit;
      end;
  end;

  U := NameIndexToUpperCaseIndex(ClassNameIndex, Scripter);
  for I:=0 to MemberList.Count - 1 do
  begin
    MemberRec := MemberList[I];
    if MemberRec.Kind = KindTYPE then
      if MemberRec.UpCaseIndex = U then
      begin
        result := MemberRec.ID;
        Exit;
      end;
  end;
end;

constructor TTitanClassList.Create(AScripter: Pointer);
begin
  inherited Create;
  Scripter := AScripter;

  ObjectClassRec := nil;
  BooleanClassRec := nil;
  StringClassRec := nil;
  NumberClassRec := nil;
  DateClassRec := nil;
  FunctionClassRec := nil;
  FunctionClassRec := nil;
  ArrayClassRec := nil;
  ActiveXClassRec := nil;

  DelegateClassRec := nil;
end;

destructor TTitanClassList.Destroy;
begin
  inherited;
end;

function TTitanClassList.GetSourceClassList: TList;

procedure AddClasses(R: TTitanClassRec);
var
  MemberRec: TTitanMemberRec;
  I: Integer;
  P: Pointer;
begin
  for I:=0 to R.MemberList.Count - 1 do
  begin
    MemberRec := R.MemberList[I];
    if MemberRec.IsSource then
      if MemberRec.Kind = KindTYPE then
      begin
        P := Pointer(MemberRec.ID);
        if result.IndexOf(P) = -1 then
          result.Add(P);
        AddClasses(FindClass(MemberRec.ID));
      end;
  end;
end;

begin
  result := TList.Create;
  AddClasses(Records[0]);
end;

procedure TTitanClassList.ResetCompileStage;
var
  L: TList;
  I, ClassID, Index: PtrInt;
begin
  L := GetSourceClassList;
  Records[0].ResetCompileStage;
  for I:=0 to L.Count - 1 do
  begin
    ClassID := PtrInt(L[I]);
    Index := IndexOf(ClassID);
    DeleteObject(Index);
  end;
  L.Free;
end;

procedure TTitanClassList.SaveToStream(S: TStream; I1, I2: Integer);
var
  I: Integer;
begin
  SaveInteger(I1, S);
  SaveInteger(I2, S);
  for I:=I1 to I2 do
    Records[I].SaveToStream(S);
end;

procedure TTitanClassList.LoadFromStream(S: TStream);
var
  I, I1, I2: Integer;
  ClassRec: TTitanClassRec;
begin
  I1 := LoadInteger(S);
  I2 := LoadInteger(S);
  for I:=I1 to I2 do
  begin
    ClassRec := TTitanClassRec.Create(Scripter, ckClass);
    ClassRec.LoadFromStream(S);
    AddObject(ClassRec.ClassID, ClassRec);
  end;
end;

function TTitanClassList.GetRecord(Index: Integer): TTitanClassRec;
begin
  result := TTitanClassRec(Objects[Index]);
end;

procedure TTitanClassList.InitRunStage;
begin
  SaveCount := Count;
end;

procedure TTitanClassList.ResetRunStage;
var
  I: Integer;
begin
  for I:= Count - 1 downto SaveCount do
  begin
    TTitanBaseScripter(Scripter).SymbolTable.NameIndex[Records[I].ClassID];
    DeleteObject(I);
  end;
end;

procedure TTitanClassList.CreateClassObjects(StartRecNo: Integer);
var
  I: Integer;
  ClassRec: TTitanClassRec;
  V: Variant;
begin
  for I:=StartRecNo to Count - 1 do
  begin
    ClassRec := Records[I];

    V := TTitanBaseScripter(Scripter).SymbolTable.GetVariant(ClassRec.ClassID);

    if not IsObject(V) then
    begin
      ClassRec.CreateClassObject;
      ClassRec.AncestorClassRec := FindClassByName(ClassRec.AncestorName);

{     if (ClassRec.AncestorClassRec = nil) and (ClassRec.AncestorName <> '') then
      begin
        ErrMessageBox('Class ' + ClassRec.AncestorName + ' is not found!');
        ClassRec.AncestorName := '';
      end;
}
      ClassRec.OwnerClassRec := FindClassByName(ClassRec.OwnerName);

      if StartRecNo > 0 then
        if ClassRec.OwnerClassRec <> nil then
          ClassRec.OwnerClassRec.CreateClassObject;
    end;

    if ClassRec.Name = 'Object' then
      ObjectClassRec := ClassRec
    else if ClassRec.Name = 'Number' then
      NumberClassRec := ClassRec
    else if ClassRec.Name = 'Date' then
      DateClassRec := ClassRec
    else if ClassRec.Name = 'Function' then
      FunctionClassRec := ClassRec
    else if ClassRec.Name = 'String' then
      StringClassRec := ClassRec
    else if ClassRec.Name = 'Boolean' then
      BooleanClassRec := ClassRec
    else if ClassRec.Name = 'Function' then
      FunctionClassRec := ClassRec
    else if ClassRec.Name = 'TTitanDelegate' then
      DelegateClassRec := ClassRec
    else if ClassRec.Name = 'Array' then
      ArrayClassRec := ClassRec
    else if ClassRec.Name = 'ActiveXObject' then
      ActiveXClassRec := ClassRec;
  end;
end;

procedure TTitanClassList.InitStaticFields(StartRecNo: Integer);
var
  I: Integer;
begin
  for I:=StartRecNo to Count - 1 do
    Records[I].InitStaticFields;
end;

function TTitanClassList.AddClass(ClassID: Integer;
                                const AClassName, OwnerName, AncestorName: String;
                                ml: TTitanModifierList; ck: TTitanClassKind;
                                UpCase: Boolean): TTitanClassRec;
var
  ClassRec: TTitanClassRec;
begin
//  ClassRec := FindClass(TTitanBaseScripter(Scripter).SymbolTable.RootNamespaceID);

  if Count = 0 then
    ClassRec := nil
  else
    ClassRec := Records[0];

  if ClassRec <> nil then
  begin
    if ClassRec.ClassID = ClassID then
    begin
      result := ClassRec;
      Exit;
    end;

    if (modSTATIC in ml) and ClassRec.IsNestedClass(ClassID) then
    begin
      result := FindClass(ClassID);
      if result <> nil then
        Exit;
    end;
  end;

  result := TTitanClassRec.Create(Scripter, ck);
  result.Name := AClassName;
  result.OwnerName := OwnerName;
  result.AncestorName := AncestorName;
  result.ModuleID := TTitanBaseScripter(Scripter).Code.ModuleID;
  result.ClassID := ClassID;
  result.ml := ml;
  result.ck := ck;

  result.NameIdx := CreateNameIndex(result.Name, Scripter);
  if UpCase then
    result.UpCaseIndex := NameIndexToUpperCaseIndex(result.NameIdx, Scripter);

  AddObject(ClassID, result);
end;

function TTitanClassList.FindClass(ClassID: Integer): TTitanClassRec;
begin
  result := TTitanClassRec(GetObject(ClassID));
end;

function TTitanClassList.FindImportedClass(PClass: TClass): TTitanClassRec;
var
  I: Integer;
  R: TTitanClassRec;
begin
  for I:=0 to Count - 1 do
  begin
    R := Records[I];
    if R.IsImported then
      if R.fClassDef.PClass = PClass then
      begin
        result := R;
        Exit;
      end;
  end;
  result := nil;
end;

function TTitanClassList.ExistsClassByName(const Name: String): boolean;
var
  I: Integer;
begin
  for I:=Count - 1 downto 0 do
  if StrEql(Records[I].Name, Name) then
  begin
    result := true;
    Exit;
  end;
  result := false;
end;

function TTitanClassList.FindClassByName(const Name: String): TTitanClassRec;
var
  I: Integer;
  D: TTitanClassDefinition;
  C: TTitanClassRec;
  Lst: TStringList;
  M: TTitanMemberRec;
  NameIndex: Integer;
  S: String;
begin
  result := nil;
  if Name = '' then
    Exit;

  for I:=Count - 1 downto 0 do
  if StrEql(Records[I].Name, Name) then
  begin
    result := Records[I];
    Exit;
  end;

  if not SignLoadOnDemand then
    Exit;

  D := GlobalDefinitionList.FindClassDefByName(Name);
  if D <> nil then
  begin
    if D.Owner <> nil then
    begin
      if not ExistsClassByName(D.Owner.Name) then
        D.Owner.AddToScripter(scripter);
    end;

    C := nil;
    if D.PClass <> nil then
      if D.PClass.ClassParent <> nil then
          C := FindClassByName(D.PClass.ClassParent.ClassName);

    D.AddToScripter(scripter);

    result := Records[Count - 1];
    result.AncestorClassRec := C;

    if D.PClass <> nil then
    begin
      Lst := TStringList.Create;
      GetPublishedProperties(D.PClass, Lst);
      for I:=0 to Lst.Count - 1 do
      begin
        M := TTitanMemberRec.Create(0, result);
        M.IsPublished := true;
        M.Kind := KindPROP;
        NameIndex := CreateNameIndex(Lst[I], Scripter);
        M.UpCaseIndex := NameIndexToUpperCaseIndex(NameIndex, Scripter);
        result.MemberList.AddObject(NameIndex, M);
      end;
      Lst.Free;
    end;

    for I:=0 to GlobalDefinitionList.Count - 1 do
    begin
      if GlobalDefinitionList.Records[I].Owner = D then
      begin
        if GlobalDefinitionList.Records[I].DefKind <> dkClass then
           GlobalDefinitionList.Records[I].AddToScripter(scripter)
      end;
    end;
  end;

  if result = nil then
  begin
    S := FindTypeAlias(Name, true);
    if S = '' then
      Exit;
    result := FindClassByName(S);
  end;
end;

function TTitanClassList.FindClassByInstance(Instance: TObject): TTitanClassRec;
var
  C: TClass;
begin
  result := FindClassByName(Instance.ClassName);
  if result = nil then
  begin
    C := Instance.ClassType;

    repeat

      C := C.ClassParent;
      if C = TObject then
        Exit;
      result := FindClassByName(C.ClassName);
      if result <> nil then
        Exit;

    until false;
  end;
end;

function TTitanClassList.FindClassByGuid(const guid: TGUID): TTitanClassRec;
var
  I: Integer;
begin
  result := nil;
  for I:=Count - 1 downto 0 do
    if Records[I].fClassDef <> nil then
      if Records[I].fClassDef.ClassKind = ckInterface then
        if IsEqualGuid(Records[I].fClassDef.guid, guid) then
        begin
          result := Records[I];
          Exit;
        end;
end;


function TTitanClassList.FindNestedClass(OwnerList: TStringList; const Name: String): TTitanClassRec;
var
  I, ID, TempIndex, NameIndex: Integer;
  ClassRec: TTitanClassRec;
  S: String;
begin
  result := nil;

  if Name = RootNamespaceName then
    Exit;

  NameIndex := CreateNameIndex(Name, Scripter);

  ClassRec := Records[0];
  if OwnerList = nil then
  begin
    ID := ClassRec.FindNestedClassID(NameIndex);
    if ID = 0 then
      Exit;

    result := FindClass(ID);
    Exit;
  end;

  if OwnerList.Count = 0 then
  begin
    ID := ClassRec.FindNestedClassID(NameIndex);
    if ID = 0 then
      Exit;

    result := FindClass(ID);
    Exit;
  end;

  if OwnerList[0] = RootNamespaceName then
  begin
    ID := ClassRec.FindNestedClassID(NameIndex);
    if ID = 0 then
      Exit;

    result := FindClass(ID);
    Exit;
  end;

  for I:=0 to OwnerList.Count - 1 do
  begin
    S := OwnerList[I];
    TempIndex := CreateNameIndex(S, Scripter);
    ID := ClassRec.FindNestedClassID(TempIndex);
    if ID = 0 then
      Exit;
    ClassRec := FindClass(ID);
    if ClassRec = nil then
      Exit;
  end;

  ID := ClassRec.FindNestedClassID(NameIndex);
  if ID = 0 then
    Exit;

  result := FindClass(ID);
end;

function TTitanClassList.FindMember(ClassID, MemberNameIndex: Integer; ma: TTitanMemberAccess): TTitanMemberRec;
var
  ClassRec: TTitanClassRec;
begin
  ClassRec := FindClass(ClassID);
  if ClassRec = nil then
    result := nil
  else
    result := ClassRec.FindMember(MemberNameIndex, ma);
end;

function TTitanClassList.CreateScriptObject(ClassID: Integer): TTitanScriptObject;
var
  ClassRec: TTitanClassRec;
begin
  ClassRec := FindClass(ClassID);
  if ClassRec = nil then
    result := nil
  else
    result := ClassRec.CreateScriptObject;
end;

procedure TTitanClassList.Reset;
var
  I: Integer;
begin
  for I:=0 to Count - 1 do
    Records[I].Free;
  Objects.Clear;
  Clear;
end;

procedure _GetPublishedPropertiesEx(AClass: TClass;
                                   result: TStrings;
                                   ClassList: TTitanClassList);
var
  pti: PTypeInfo;
  ptd: PTypeData;
  Loop, nProps: Integer;
  pProps: PPropList;
  ppi: PPropInfo;
  S, T: String;
begin
  pti := AClass.ClassInfo;
  if pti = nil then Exit;
  ptd := GetTypeData(pti);
  nProps := ptd^.PropCount;
  if nProps > 0 then begin
    GetMem(pProps, SizeOf(PPropInfo) * nProps);
    GetPropInfos(pti, pProps);
  end
  else pProps := nil;
  for Loop:=0 to nProps - 1 do begin
    ppi := pProps[Loop];
    S := ppi^.Name;
    T := ppi^.PropType^.Name;

    if PaxTypes.IndexOf(UpperCase(T)) >= 0 then
      result.Add(UpperCase(S))
    else if ClassList.FindClassByName(T) <> nil then
      result.Add(UpperCase(S));
  end;
  if pProps <> nil then
    FreeMem(pProps, SizeOf(PPropInfo) * nProps);
end;

{
procedure TTitanClassList.AddDefinitionList(L: TTitanDefinitionList);
var
  I, J, NameIndex: Integer;
  D: TTitanDefinition;
  Lst: TStringList;
  R: TTitanClassRec;
  M: TTitanMemberRec;
begin
  for I:=0 to L.Count - 1 do
  begin
    D := L[I];
    if D <> nil then
      D.AddToScripter(Scripter);
  end;
  Lst := TStringList.Create;
  for I:=0 to Count - 1 do
  begin
    R := Records[I];
    if R.fClassDef <> nil then
      if R.fClassDef.PClass <> nil then
      begin
        GetPublishedProperties(R.fClassDef.PClass, Lst);
        for J:=0 to Lst.Count - 1 do
        begin
          M := TTitanMemberRec.Create(0, R);
          M.IsPublished := true;
          M.Kind := KindPROP;
          NameIndex := CreateNameIndex(Lst[J], Scripter);
          M.UpCaseIndex := NameIndexToUpperCaseIndex(NameIndex, Scripter);
          R.MemberList.AddObject(NameIndex, M);
        end;
        Lst.Clear;
      end;
  end;
  Lst.Free;
end;
}

procedure TTitanClassList.AddDefinitionList(L: TTitanDefinitionList; N: Integer);
var
  I, J, NameIndex: Integer;
  D: TTitanDefinition;
  Lst: TStringList;
  R: TTitanClassRec;
  M: TTitanMemberRec;
begin
  for I:=0 to N - 1 do
  begin
    D := L[I];
    if D <> nil then
      D.AddToScripter(Scripter);
  end;
  Lst := TStringList.Create;
  for I:=0 to Count - 1 do
  begin
    R := Records[I];
    if R.fClassDef <> nil then
      if R.fClassDef.PClass <> nil then
      begin
        GetPublishedProperties(R.fClassDef.PClass, Lst);
        for J:=0 to Lst.Count - 1 do
        begin
          M := TTitanMemberRec.Create(0, R);
          M.IsPublished := true;
          M.Kind := KindPROP;
          NameIndex := CreateNameIndex(Lst[J], Scripter);
          M.UpCaseIndex := NameIndexToUpperCaseIndex(NameIndex, Scripter);
          R.MemberList.AddObject(NameIndex, M);
        end;
        Lst.Clear;
      end;
  end;
  Lst.Free;
end;


function TTitanClassList.CreateBooleanObject(const AValue: Variant): TTitanScriptObject;
begin
  if (BooleanClass = nil) or (BooleanClassRec = nil) then
  begin
    BooleanClassRec := FindClassByName('Boolean');
    if (BooleanClass = nil) or (BooleanClassRec = nil) then
    raise TTitanScriptFailure.Create(Format(errClassIsNotFound, ['Boolean']));
  end;
  result := BooleanClass.Create(BooleanClassRec);

  result.SetDefaultValue(toBoolean(AValue));
  result.RefCount := 1;
end;

function TTitanClassList.CreateNumberObject(const AValue: Variant): TTitanScriptObject;
begin
  if (NumberClass = nil) or (NumberClassRec = nil) then
  begin
    NumberClassRec := FindClassByName('Number');
    if (NumberClass = nil) or (NumberClassRec = nil) then
    raise TTitanScriptFailure.Create(Format(errClassIsNotFound, ['Number']));
  end;
  result := NumberClass.Create(NumberClassRec);

  result.SetDefaultValue(toNumber(AValue));
  result.RefCount := 1;
end;

function TTitanClassList.CreateStringObject(const AValue: Variant): TTitanScriptObject;
begin
  if (StringClass = nil) or (StringClassRec = nil) then
  begin
    StringClassRec := FindClassByName('String');
    if (StringClass = nil) or (StringClassRec = nil) then
    raise TTitanScriptFailure.Create(Format(errClassIsNotFound, ['String']));
  end;
  result := StringClass.Create(StringClassRec);

  result.SetDefaultValue(String(AValue));
  result.RefCount := 1;
end;

function TTitanClassList.CreateDateObject(const AValue: Variant): TTitanScriptObject;
begin
  if (DateClass = nil) or (DateClassRec = nil) then
  begin
    DateClassRec := FindClassByName('Date');
    if (DateClass = nil) or (DateClassRec = nil) then
    raise TTitanScriptFailure.Create(Format(errClassIsNotFound, ['Date']));
  end;
  result := DateClass.Create(DateClassRec);

  result.SetDefaultValue(toDate(AValue));
  result.RefCount := 1;
end;

function TTitanClassList.CreateFunctionObject(const AValue: Variant): TTitanScriptObject;
begin
  if (FunctionClass = nil) or (FunctionClassRec = nil) then
  begin
    FunctionClassRec := FindClassByName('Function');
    if (FunctionClass = nil) or (FunctionClassRec = nil) then
    raise TTitanScriptFailure.Create(Format(errClassIsNotFound, ['Function']));
  end;
  result := FunctionClass.Create(FunctionClassRec);

  result.SetDefaultValue(AValue);
  result.RefCount := 1;
end;

procedure TTitanClassList.Dump(const FileName: String);
var
  I, J: Integer;
  T: TextFile;
  ClassRec, OwnerRec, AncestorRec: TTitanClassRec;

  MemberRec: TTitanMemberRec;
  S, S1, S2: String;
begin
  AssignFile(T, FileName);
  Rewrite(T);
  try
    for I:=0 to Count - 1 do
    begin
      ClassRec := Records[I];
      writeln(T, '');
      writeln(T, ClassRec.Name +
                 '[', ClassRec.ClassID:4, ']');
      writeln(T, '***********************************');
      AncestorRec := ClassRec.AncestorClassRec;
      if AncestorRec <> nil then
      begin
        writeln(T, 'Ancestor: ', AncestorRec.Name +
                   '[', AncestorRec.ClassID:4, ']');
      end;
      OwnerRec := ClassRec.OwnerClassRec;
      if OwnerRec <> nil then
      begin
        writeln(T, 'Owner: ', OwnerRec.Name +
                   '[', OwnerRec.ClassID:4, ']');
      end;

      writeln(T, 'Fields:');
      with ClassRec do
      begin
        for J:=0 to MemberList.Count - 1 do
        if (MemberList[J].Kind = KindVAR) or (MemberList[J].Kind = KindCONST) then
        begin
          MemberRec := MemberList[J];
          if MemberRec.IsStatic then
            S := ' static'
          else
            S := '';
          if MemberRec.IsSource then
            S2 := ' [SOURCE]'
          else
            S2 := '';
          with MemberRec do
            writeln(T, '[', ID:4, ']', TTitanBaseScripter(Scripter).GetName(MemberList.NameID[J]),
                       ' Init: ', MemberRec.InitN, S, S2);
        end;
      end;

      writeln(T, 'Methods:');
      with ClassRec do
      begin
        for J:=0 to MemberList.Count - 1 do
        if MemberList[J].Kind = KindSUB then
        begin
          MemberRec := MemberList[J];
          if MemberRec.IsStatic then
            S := ' static'
          else
            S := '';
          if MemberRec.IsSource then
            S2 := ' [SOURCE]'
          else
            S2 := '';
          with MemberRec do
            writeln(T, '[', ID:4, ']', TTitanBaseScripter(Scripter).GetName(MemberList.NameID[J]), S, S2);
        end;
      end;
      writeln(T, 'Properties:');
      with ClassRec, TTitanBaseScripter(Scripter).SymbolTable do
      begin
        for J:=0 to MemberList.Count - 1 do
        if MemberList[J].Kind = KindPROP then
        begin
          MemberRec := MemberList[J];
          if MemberRec.IsStatic then
            S := ' static'
          else
            S := '';

          if MemberRec.IsSource then
            S2 := ' [SOURCE]'
          else
            S2 := '';

          if MemberRec.IsImported then
          begin
            S := S + ' imported';
            S1 := Norm(MemberRec.Definition.Name, 10);
          end
          else
          begin
            S1 := Norm(Name[MemberRec.ID], 10);
          end;

          with MemberRec do
            writeln(T, '[', ID:4, ']', S1,
                       ' ReadID = [', ReadID:4, ']', Norm(Name[ReadID], 10),
                       ' WriteID = [', WriteID:4, ']', Norm(Name[WriteID], 10), S, S2);
        end;
      end;
      writeln(T, 'Nested classes:');
      with ClassRec do
      begin
        for J:=0 to MemberList.Count - 1 do
        if MemberList[J].Kind = KindTYPE then
        begin
          MemberRec := MemberList[J];
          if MemberRec.IsStatic then
            S := ' static'
          else
            S := '';
          with MemberRec do
            writeln(T, '[', ID:4, ']', TTitanBaseScripter(Scripter).GetName(MemberList.NameID[J]), S);
        end;
      end;
    end;
  finally
    Close(T);
  end;
end;

function ScriptObjectToVariant(const Value: TTitanScriptObject): Variant;
begin
  result := PtrInt(Value);
  TVarData(result).VType := varScriptObject;
end;

function VariantToScriptObject(const Value: Variant): TTitanScriptObject;
begin
  if not IsObject(Value) then
  begin
    if VarType(Value) <> varUndefined then
      raise TTitanScriptFailure.Create(errIncompatibleTypes)
    else
    begin
      raise TTitanScriptFailure.Create(errVariantToScriptObjectConversion);
    end;
  end;
  result := TTitanScriptObject(TVarData(Value).vint64);  // ct9999
end;

function ArrayToStr(const V: Variant): String; forward;

function VariantToString(const Value: Variant): String;
var
  T: Integer;
begin
  T := VarType(Value);
  if      T = varEmpty then
     result := 'null'
  else if T = varNull then
     result := 'null'
  else if T = varError then
     result := 'Error:' + IntToStr(TVarData(Value).VError)
  else if T = varDispatch then
     result := 'Dispatch:' + Format('%p', [TVarData(Value).VDispatch])
  else if IsObject(Value) then
    result := VariantToScriptObject(Value).ToString()
  else if T > varArray then
     result := ArrayToStr(Value)
  else if T = varBoolean then
  begin
    if Value then
      result := 'true'
    else
      result := 'false';
  end
  else
    result := VarToStr(Value);
end;

function ArrayToStr(const V: Variant): String;
var
  L: Integer;
  B1, B2: array[1..100] of Integer;

function F(P: Pointer; I: Integer; var SZ: Integer): String;
var
  K, TempSZ: Integer;
begin
  if I > L then
  begin
    result := VariantToString(Variant(P^));
    SZ := SizeOf(Variant);
    Exit;
  end;

  result := '[';

  SZ := 0;

  for K:=B1[I] to B2[I] do
  begin
    result := result + F(P, I + 1, TempSZ);
    if K < B2[I] then
      result := result + ',';
    Inc(PtrInt(P), TempSZ);
    Inc(SZ, TempSZ);
  end;

  result := result + ']';
end;

var
  I, SZ: Integer;
  P: Pointer;
begin
  L := VarArrayDimCount(V);
  for I:=1 to L do
  begin
    B1[I] := VarArrayLowBound(V, L - I + 1);
    B2[I] := VarArrayHighBound(V, L - I + 1);
  end;
  P := VarArrayLock(V);
  try
    result := F(P, 1, SZ);
  finally
    VarArrayUnlock(V);
  end;
end;

function ToStr(Scripter: Pointer; const V: Variant): String;
begin
  TTitanBaseScripter(Scripter).Visited.Clear;
  result := _ToStr(Scripter, V);
  TTitanBaseScripter(Scripter).Visited.Clear;
end;

function _ToStr(Scripter: Pointer; const V: Variant): String;
var
  SO: TTitanScriptObject;
begin
  if IsObject(V) then
  begin
    SO := VariantToScriptObject(V);
    if TTitanBaseScripter(Scripter).AssignedObject(SO) then
      result := SO.ToString()
    else
      result := 'nil'
  end
  else
    result := VariantToString(V);

  if result = '\r\n' then
    result := #13#10;
end;

function ToInteger(const V: Variant): Integer;
begin
  result := ToInt32(V);
end;

constructor TTitanScriptObjectList.Create(_Scripter: Pointer);
begin
  Self._Scripter := _Scripter;
  fPaxObjects := TList.Create;
end;

destructor TTitanScriptObjectList.Destroy;
begin
  Clear;

  fPaxObjects.Free;
  inherited;
end;

function TTitanScriptObjectList.GetItem(I: Integer): TTitanScriptObject;
begin
  result := TTitanScriptObject(fPaxObjects[I]);
end;

procedure TTitanScriptObjectList.Add(SO: TTitanScriptObject);
begin
  fPaxObjects.Add(SO);
end;

procedure TTitanScriptObjectList.Delete(I: Integer);
var
  SO: TTitanScriptObject;
begin
  SO := TTitanScriptObject(fPaxObjects[I]);
  Pointer(SO.Intf) := nil;

  if SO.Instance <> nil then
    if SO.ClassRec.Name = 'RegExp' then
      SO.Instance.Free;

  SO.Free;
  fPaxObjects.Delete(I);
end;

function TTitanScriptObjectList.GetCount: Integer;
begin
  result := fPaxObjects.Count;
end;

procedure TTitanScriptObjectList.Clear;
var
  I: Integer;
begin
  for I := Count - 1 downto 0 do
    Delete(I);
end;

function TTitanScriptObjectList.IndexOfDelphiObject(DelphiInstance: TObject): Integer;
var
  SO: TTitanScriptObject;
  I: Integer;
  S: String;
begin
  result := -1;
  If DelphiInstance = Nil then
    exit;

  if not IsDelphiObject(DelphiInstance) then
    Exit;

  S := DelphiInstance.ClassName;
  for I:=Count - 1 downto 0 do
  begin
    SO := TTitanScriptObject(fPaxObjects[I]);
    if SO.Instance = DelphiInstance then
    if SO.ClassRec.Name = S then
    begin
      result := I;
      Exit;
    end;
  end;
end;

function TTitanScriptObjectList.FindScriptObject(DelphiInstance: TObject): TTitanScriptObject;
var
  I: Integer;
begin
  I := IndexOfDelphiObject(DelphiInstance);
  if I = -1 then
    result := nil
  else
    result := TTitanScriptObject(fPaxObjects[I]);
end;

function TTitanScriptObjectList.FindScriptObjectByIntf(PIntf: PUnknown): TTitanScriptObject;
var
  I: Integer;
  SO: TTitanScriptObject;
begin
  for I:=0 to Count - 1 do
  begin
    SO := TTitanScriptObject(fPaxObjects[I]);
    if SO.PIntf <> nil then
      if SO.PIntf = PIntf then
      begin
        result := SO;
        Exit;
      end;
  end;
  result := nil;
end;

constructor TTitanDelegate.Create(Scripter: Pointer;
                                SubID, N: Integer; D: TTitanMethodDefinition);
begin
  if Scripter <> nil then
    inherited Create(TTitanBaseScripter(Scripter).CompileTimeHeap);

  Self.N := N;
  Self.SubID := SubID;
  Self.D := D;
  fName := '';
end;

destructor TTitanDelegate.Destroy;
begin
  if D <> nil then
    if D.DefList = nil then
      D.Free;
  inherited;
end;

procedure TTitanCompileTimeHeap.ResetCompileStage;
var
  D: TTitanDelegate;
  I: Integer;
begin
  for I:=Count - 1 downto 0 do
  begin
    D :=  TTitanDelegate(Items[I]);
    if D.SubID >= FirstSymbolCard then
      Delete(I);
  end;
end;

procedure TTitanScriptObjectList.ResetCompileStage(FreeHost: Boolean = false);
var
  SO: TTitanScriptObject;
  D: TTitanDelegate;
  I, J: Integer;
  S: TTitanBaseScripter;
  ClassList: TTitanClassList;
  ClassRec: TTitanClassRec;
begin
  S := _Scripter;
  ClassList := S.ClassList;

  for J := 1 to 2 do
  for I:=Count - 1 downto 0 do
  begin
    SO := TTitanScriptObject(fPaxObjects[I]);
    ClassRec := SO.ClassRec;

    if FreeHost then
      if ClassRec <> ClassList.FunctionClassRec then
      if ClassRec <> ClassList.DelegateClassRec then
      if ClassRec <> ClassList.ObjectClassRec then
      if ClassRec <> ClassList.BooleanClassRec then
      if ClassRec <> ClassList.StringClassRec then
      if ClassRec <> ClassList.NumberClassRec then
      if ClassRec <> ClassList.DateClassRec then
      if ClassRec <> ClassList.ArrayClassRec then
      if ClassRec <> ClassList.ActiveXClassRec then
      if (SO.Instance <> nil) and (SO.RefCount <> -1) then
      if not StrEql(ClassRec.Name, 'TTitanError') then
      begin

        If IsDelphiObject(SO.Instance) then
          SO.Instance.Free;

        SO.Instance := nil;
      end;

    if SO.PClass = TTitanDelegate then
    begin
      D := TTitanDelegate(SO.Instance);
      if D.SubID >= FirstSymbolCard then
        Delete(I);
    end
    else if SO.ClassRec.ClassID <> FirstSymbolCard then
    begin
{
      if SO.RefCount >= 1 then
      begin
        if SO.Intf <> nil then
        begin
          try
            if IntfRefCount(SO.Intf) > 1 then
              SO.Intf._Release;
          except
            PtrInt(SO.Intf) := 0;
          end;
        end;
        {
          if IntfRefCount(SO.Intf) > 1 then
            SO.Intf._Release;
        }
//        Delete(I);
//     end;

    end;
  end;
end;

procedure TTitanScriptObjectList.ResetRunStage(FreeHost: Boolean = false);
var
  SO: TTitanScriptObject;
  I, J: Integer;
  S: TTitanBaseScripter;
  ClassList: TTitanClassList;
  ClassRec: TTitanClassRec;
begin
  S := _Scripter;
  ClassList := S.ClassList;

  for J:=1 to 2 do begin
  for I:=Count - 1 downto 0 do
  begin
    SO := TTitanScriptObject(fPaxObjects[I]);
    ClassRec := SO.ClassRec;

    if FreeHost then
    if ClassRec <> ClassList.FunctionClassRec then
    if ClassRec <> ClassList.DelegateClassRec then
    if ClassRec <> ClassList.ObjectClassRec then
    if ClassRec <> ClassList.BooleanClassRec then
    if ClassRec <> ClassList.StringClassRec then
    if ClassRec <> ClassList.NumberClassRec then
    if ClassRec <> ClassList.DateClassRec then
    if ClassRec <> ClassList.ArrayClassRec then
    if ClassRec <> ClassList.ActiveXClassRec then
    if (SO.Instance <> nil) and (SO.RefCount <> -1) then
    if not StrEql(ClassRec.Name, 'TTitanError') then
    begin
      SO.Instance.Free;
      SO.Instance := nil;
    end;

    if SO.RefCount >= 1 then
    begin
//      if SO.ClassRec.ck = ckArray then
//        SO.Instance.Free;
      if (SO.ClassRec.ck = ckInterface) and (SO.Intf <> nil) then
      begin
{
        try
          if IntfRefCount(SO.Intf) = 1 then
          begin
            //SO.Intf := nil;
            Pointer (SO.Intf) := Nil //--jgv
          end
          else
            SO.Intf._Release;
        except
          Pointer(SO.Intf) := nil;
        end;
}
      end;

      SO.CallAutoDestructor;
      Delete(I);
    end;
  end;
  end;
end;

procedure TTitanScriptObjectList.RemoveObject(SO: TTitanScriptObject);
var
  I: Integer;
  tempSO: TTitanScriptObject;
begin
  for I:=Count - 1 downto 0 do
  begin
    tempSO := TTitanScriptObject(fPaxObjects[I]);
    if tempSO._ObjCount = SO._ObjCount then
    begin
      SO.CallAutoDestructor;
      Delete(I);
      Exit;
    end;
  end;
end;

function IsGlobalObject(SO: TTitanScriptObject): Boolean;
var
  SymbolTable: TTitanSymbolTable;
  I: Integer;
  tempSO: TTitanScriptObject;
begin
  SymbolTable := TTitanBaseScripter(SO.Scripter).SymbolTable;
  with SymbolTable do
    for I:= RootNamespaceID to Card do
      if not IsLocal(I) then
        if IsObject(PVariant(Address[I])^) then
        begin
          tempSO := VariantToScriptObject(PVariant(Address[I])^);
          if SO = tempSO then
          begin
            result := true;
            Exit;
          end;
       end;

  if (SO.Instance <> nil) and IsDelphiObject(SO.Instance) then
    if SO.Instance.InheritsFrom(TComponent) then
    begin
      if (SO.Instance as TComponent).Owner <> nil then
        result := true
      else
        result := false;

      Exit;
    end;

  result := false;
end;

procedure TTitanScriptObjectList.RemoveTail(K: Integer; ThreadID: Cardinal = 0);
var
  SO: TTitanScriptObject;
  I: Integer;
begin
  for I := Count - 1 downto K do
  begin
    SO := TTitanScriptObject(fPaxObjects[I]);
    if SO.RefCount = 1 then
    if SO.ThreadID = ThreadID then
    if not IsGlobalObject(SO) then
    begin
      SO.CallAutoDestructor;
      Delete(I);
    end;
  end;
end;

function TTitanScriptObjectList.HasObject(SO: TTitanScriptObject): Boolean;
var
  I: Integer;
  tempSO: TTitanScriptObject;
  ok: Boolean;
begin
  if SO = nil then
  begin
    result := false;
    Exit;
  end;

  ok := false;
  for I:=Count - 1 downto 0 do
  begin
    if TTitanScriptObject(fPaxObjects[I]) = SO then
    begin
      ok := true;
      break;
    end;
  end;

  if not ok then
  begin
    result := false;
    Exit;
  end;

  if not IsDelphiObject(SO) then
  begin
    result := false;
    Exit;
  end;

  for I:=Count - 1 downto 0 do
  begin
    tempSO := TTitanScriptObject(fPaxObjects[I]);
    if tempSO._ObjCount = SO._ObjCount then
    begin
      result := true;
      Exit;
    end;
  end;
  result := false;
end;

procedure TTitanScriptObjectList.RemoveStructure(const V: Variant);
var
  SO: TTitanScriptObject;
begin
  if IsObject(V) then
  begin
    SO := VariantToscriptObject(V);
    if SO.ClassRec.ck = ckStructure then
      RemoveObject(SO);
  end;
end;

function DuplicateObject(const V: Variant): Variant;
var
  SO: TTitanScriptObject;
begin
  if IsObject(V) then
  begin
    SO := VariantToScriptObject(V);
    if SO.ClassRec.ck = ckStructure then
      result := ScriptObjectToVariant(SO.Duplicate)
    else
      result := V;
  end
  else
    result := V;
end;

/////////////// CONVERSIONS ///////////////////////////////

function ToPrimitive(const V: Variant): Variant;
begin
  if IsObject(V) then
    Result := VariantToScriptObject(V).DefaultValue
  else
    result := V;
end;

function ToString(const Value: Variant): String;
var
  P: Integer;
  D: Double;
  SO: TTitanScriptObject;
begin
  if IsObject(Value) then
  begin
    SO := VariantToScriptObject(Value);
    TTitanBaseScripter(SO.Scripter).Visited.Clear;
    Result := SO.ToString;
  end
  else
  begin
    case VarType(Value) of
      varDouble:
      begin
        D := Value;
        if D = NaN then
        begin
          result := 'NaN';
          Exit;
        end
        else if D = Infinity then
        begin
          result := 'infinity';
          Exit;
        end
        else if D = - Infinity then
        begin
          result := '-infinity';
          Exit;
        end;
      end;
{$IFDEF VARIANTS}
     varInt64:
     begin
       result := VarToStr(Value);
       Exit;
     end;
{$ENDIF}
    end;

    result := VarToStr(Value);
    if VarType(Value) = varDouble then
    begin
      P := Pos(',', result);
      if P > 0 then
        result[P] := '.';
    end;
  end;
end;

function ToNumber(const V: Variant): Variant;
var
  S: String;
begin
  case VarType(V) of
    varEmpty: result := NaN;
    varNull: result := 0;
    varInteger, varDouble, varDate: result := V;
{$IFDEF VARIANTS}
     varInt64: result := V;
{$ENDIF}
    varBoolean:
      if V then
        result := 1
      else
        result := 0;
    varString, varOleStr:
       if IsNumericString(V) then
       begin
         S := StringReplace(V, '.', DecimalSeparator, []);
         result := SysUtils.StrToFloat(S);
       end
       else
         result := NaN;
    else
      if IsObject(V) then
        result := ToNumber(VariantToScriptObject(V).DefaultValue)
      else
        result := V;
  end;
end;

function ToDate(const V: Variant): Variant;
var
  D: Double;
begin
  D := DelphiDateTimeToEcmaTime(Now);
  case VarType(V) of
    varDouble: result := V;
    varDate:
    begin
      D := DelphiDateTimeToEcmaTime(V);
      result := D;
    end;
    else
      if IsObject(V) then
        result := ToDate(VariantToScriptObject(V).DefaultValue)
      else
        result := D;
  end;
end;

function ToInt32(const V: Variant): Variant;
var
  N: Variant;
  D: Double;
  I: Integer;
begin
  N := ToNumber(V);
  case VarType(N) of
    varDouble: begin
      D := N;
      if (D = NaN) or (D = NEGATIVE_INFINITY) or (D = POSITIVE_INFINITY) then
        result := 0
      else
      begin
{$IFDEF VARIANTS}

          result := Round(PtrInt(N));

{$ELSE}
          I := Round(N);
          result := I;
{$ENDIF}
      end;
    end;
    varInteger, varByte: Result := N;
{$IFDEF VARIANTS}
    varInt64:
    begin
      I := Round(PtrInt(N));
      result := I;
    end;
{$ENDIF}
  end;
end;

function ToInt64(const V: Variant): Variant;
var
  N: Variant;
  D: Double;
{$IFNDEF VARIANTS}
  I: Integer;
{$ENDIF}
begin
  N := ToNumber(V);
  case VarType(N) of
    varDouble: begin
      D := N;
      if (D = NaN) or (D = NEGATIVE_INFINITY) or (D = POSITIVE_INFINITY) then
        result := 0
      else
      begin
{$IFDEF VARIANTS}

        result := Round(PtrInt(N));

{$ELSE}
        I := Round(N);
        result := I;
{$ENDIF}
      end;
    end;
    varInteger: Result := N;
{$IFDEF VARIANTS}
    varInt64:
      result := N;
{$ENDIF}
  end;
end;

function ToBoolean(const V: Variant): Variant;
begin
  result := true;
  case VarType(V) of
    varEmpty, varNull: result := false;
    varBoolean: result := V;
{$IFDEF VARIANTS}
    varInt64,
{$ENDIF}
    varDouble,varInteger: Result := V<> 0;
    varString, VarOleStr: result := V <> '';
    else if IsObject(V) then
      result := ToBoolean(VariantToScriptObject(V).DefaultValue)
    else
      result := V <> 0;
  end;
end;

function IsNaN(const V: Variant): boolean;
var
  D: Double;
begin
  D := ToNumber(V);
  result := (D = NEGATIVE_INFINITY) or
            (D = POSITIVE_INFINITY) or
            (D = NaN);
end;

function RelationalComparison(const V1, V2: Variant): Variant;
                              //performs x < y comparison
var
  I, L: Integer;
  S1, S2: String;
  P1, P2, N1, N2: Variant;
begin
  P1 := ToPrimitive(V1);
  P2 := ToPrimitive(V2);
  if IsString(P1) and IsString(P2) then
  begin
    S1 := P1;
    S2 := P2;
    if Pos(S1, S2) > 0 then
      result := true
    else if Pos(S2, S1) > 0 then
      result := false
    else
    begin
      L := Length(S1);
      if Length(S2) < L then
        L := Length(S2);
      for I:=1 to L do
       if S1[I] <> S2[I] then
       begin
         if Ord(S1[I]) < Ord(S2[I]) then
           result := true
         else
           result := false;
         Exit;
       end;
      result := false;
    end;
  end
  else
  begin
    N1 := ToNumber(P1);
    N2 := ToNumber(P2);
    if N1 = NaN then
      Exit;
    if N2 = NaN then
      Exit;
    result := N1 < N2;
  end;
end;

function EqualityComparison(const V1, V2: Variant): TBoolean;
var
  T1, T2: Integer;
  SO1, SO2: TTitanScriptObject;
begin
  T1 := VarType(V1);
  T2 := VarType(V2);
  if T1 = T2 then begin
    if T1 = varUndefined then
    begin
      result := true;
      Exit;
    end;
    if T1 = varNull then
    begin
      result := true;
      Exit;
    end;
    if T1 = varScriptObject then
    begin
      SO1 := VariantToScriptObject(V1);
      SO2 := VariantToScriptObject(V2);
      result := SO1.ClassRec = SO2.ClassRec;
      Exit;
    end;
    if IsNumber(V1) then
    begin
      if IsNaN(V1) or IsNan(V2) then
      begin
        result := false;
        Exit;
      end;
      result := V1 = V2;
      Exit;
    end;

    result := V1 = V2;
  end
  else
  begin
    if (T1 = varNull) and (T2 = varUndefined) then
      result := true
    else if (T2 = varNull) and (T1 = varUndefined) then
     result := true
    else if IsNumber(V1) and IsString(V2) then
      result := EqualityComparison(V1, ToNumber(V2))
    else if IsNumber(V2) and IsString(V1) then
      result := EqualityComparison(V2, ToNumber(V1))
    else if IsNumber(V1) and IsBoolean(V2) then
      result := EqualityComparison(V1, ToNumber(V2))
    else if IsNumber(V2) and IsBoolean(V1) then
      result := EqualityComparison(V2, ToNumber(V1))
    else if IsObject(V1) and (IsNumber(V2) or IsBoolean(V2) or IsString(V2)) then
      result := EqualityComparison(ToPrimitive(V1), V2)
    else if IsObject(V2) and (IsNumber(V1) or IsBoolean(V1) or IsString(V1)) then
      result := EqualityComparison(ToPrimitive(V2), V1)
    else if IsNumber(V1) and IsNumber(V2) then
      result := V1 = V2
    else
      result := false;
  end;
end;

function ToObject(const Value: Variant; Scripter: Pointer): Variant;
var
  SO: TTitanScriptObject;
  X: ActiveXObject;
begin
  case VarType(Value) of
    varBoolean:
    begin
      SO := TTitanBaseScripter(Scripter).ClassList.CreateBooleanObject(Value);
      result := ScriptObjectToVariant(SO);
    end;
{$IFDEF VARIANTS}
    varInt64,
{$ENDIF}
    varDouble, varInteger:
    begin
      SO := TTitanBaseScripter(Scripter).ClassList.CreateNumberObject(Value);
      result := ScriptObjectToVariant(SO);
    end;
    varString, varOleStr:
    begin
      SO := TTitanBaseScripter(Scripter).ClassList.CreateStringObject(Value);
      result := ScriptObjectToVariant(SO);
    end;
    varDate:
    begin
      SO := TTitanBaseScripter(Scripter).ClassList.CreateDateObject(Value);
      result := ScriptObjectToVariant(SO);
    end;
    varEmpty: begin end;
    varNull: begin end;
    varDispatch:
    begin
      X := ActiveXObject.Create(Scripter);
      X.D := Value;
      result := ScriptObjectToVariant(DelphiInstanceToScriptObject(X, Scripter));
    end;
    varUnknown:
    begin
    end
    else
    begin
      if IsObject(Value) then
        result := Value
      else
        raise TTitanScriptFailure.Create(errIncompatibleTypes);
    end;
   end;
end;

function StrictEqualityComparison(const V1, V2: Variant): TBoolean;
begin
  if VarType(V1) <> VarType(V2) then
    result := false
  else
    result:= EqualityComparison(V1, V2);
end;

procedure SortVariants(var A: array of Variant);
var
  I, J, N: Integer;
  X: Variant;
begin
  N := Length(A) - 1;
  for I:=1 to N do
    for J:=N downto I do
      if RelationalComparison(A[J], A[J-1]) then
      begin
        X := A[J-1];
        A[J-1] := A[J];
        A[J] := X;
      end;
end;

function StringValueToVariant(const S: String; const Dest: Variant): Variant;
begin
  case VarType(Dest) of
    varString, varOleStr: result := S;
    varInteger: result := toInteger(S);
{$IFDEF VARIANTS}
    varInt64: result := toInt64(S);
{$ENDIF}
    varDouble: result := toNumber(S);
    varEmpty, varNull: result := S;
  else
    raise TTitanScriptFailure.Create(errIncompatibleTypes);
  end;
end;

function UntypedValueToVariant(P: Pointer; Count: Integer; const Dest: Variant): Variant;
var
  PC: PChar;
begin
  case VarType(Dest) of
    varString:
    begin
      PC := StrAlloc(Count + 1);
      Move(P^, PC^, Count);
      result := String(P);
    end;
    varInteger, varDouble:
    case Count of
      1: result := Byte(P^);
      2: result := Word(P^);
      4: result := PtrInt(P^);
      SizeOf(Double): result := Double(P^);
    end;
{$IFDEF VARIANTS}
    varInt64: result := Int64(P^);
{$ENDIF}
    else
      raise TTitanScriptFailure.Create(errIncompatibleTypes);
  end;
end;

procedure VariantToUntypedValue(const V: Variant; P: Pointer; Count: Integer);
var
  S: String;
begin
  case VarType(V) of
    varInteger:
      Move(TVarData(V).VInteger, P^, Count);
    varDouble:
      Move(TVarData(V).VDouble, P^, Count);
    varBoolean:
      Move(TVarData(V).VBoolean, P^, Count);
    varString:
    begin
      S := V;
      Move(Pointer(S)^, P^, Count);
    end;
{$IFDEF VARIANTS}
    varInt64: Move(TVarData(V).VInt64, P^, Count);
{$ENDIF}
    else
      raise TTitanScriptFailure.Create(errIncompatibleTypes);
  end;
end;

function TVarRecToVariant(const P: TVarRec; Scripter: Pointer): Variant;
var
  Self: TTitanBaseScripter;
begin
  Self := TTitanBaseScripter(Scripter);

  case P.vType of
    vtInteger: result := P.VInteger;
    vtBoolean: result := P.VBoolean;
    vtChar: result := P.VChar + '';
    vtExtended: result := P.VExtended^;
    vtString: result := P.VString^;
    vtPointer: result := PtrInt(P.VPointer);
    vtPChar: result := String(P.VPChar);
    vtClass: result := ScriptObjectToVariant(DelphiClassToScriptObject(P.VClass, Self));
    vtWideChar: result := P.VWideChar;
    vtPWideChar: result := WideString(P.VPWideChar);
    vtAnsiString: result := String(P.VAnsiString);
    vtCurrency: result := P.VCurrency^;
    vtVariant: result := P.VVariant^;
    vtObject:
      if P.VObject = nil then
        result := Undefined
      else if P.VObject.ClassType = TTitanScriptObject then
        result := ScriptObjectToVariant(TTitanScriptObject(P.VObject))
      else
        result := ScriptObjectToVariant(DelphiInstanceToScriptObject(P.VObject, Self));
    vtInterface: begin end;
    vtWideString: result := WideString(P.VWideString);
{$IFDEF VARIANTS}
    vtInt64: result := P.VInt64^;
{$ELSE}
    vtInt64: result := PtrInt(P.VInt64^);
{$ENDIF}
  end;
end;

constructor ActiveXObject.Create(Scripter: Pointer);
begin
  TTitanBaseScripter(Scripter).ActiveXObjectList.Add(Self);
end;

