{%mainunit ../BASE_Engine.pas}

{***************************************************
 Package pl_TitanScript.pkg
 This file is part of CodeTyphon Studio
****************************************************}

//============================= TTitanEventHandlerList ================================================
destructor TTitanEventHandlerList.Destroy;
begin
  ClearHandlers;
  inherited;
end;

procedure TTitanEventHandlerList.ClearHandlers;
var
  I: Integer;
  M: TMethod;
  H: TTitanEventHandler;
begin
  M.Code := nil;
  M.Data := nil;
  for I:=0 to Count - 1 do
  begin
    H := TTitanEventHandler(Items[I]);

    if IsDelphiObject(H.DelphiInstance) then
      SetMethodProp(H.DelphiInstance, H.PropInfo, M);

    H.Free;
  end;
  Clear;
end;

Procedure TTitanEventHandlerList.DeleteEvent(aEvent:TTitanEventHandler);
var
  I: Integer;
  M: TMethod;
  H: TTitanEventHandler;
 begin
   if aEvent=nil then exit;
   M.Code := nil;
   M.Data := nil;

   for I:=Count - 1 Downto 0 do
   begin
    H := TTitanEventHandler(Items[I]);
    if H=aEvent then
     begin
      if IsDelphiObject(H.DelphiInstance) then SetMethodProp(H.DelphiInstance, H.PropInfo, M);
      H.Free;
      self.Delete(i);
      exit;
     end;
  end;
   
 end;

//============================ TTitanEventHandler ====================================================
constructor TTitanEventHandler.Create(Scripter: Pointer;
                                    pti: PTypeInfo;
                                    SubID: Integer;
                                    const This: Variant);
type
  TParamData = record
    Flags: TParamFlags;
    ParamName, TypeName: ShortString;
  end;
  PParamData = ^TParamData;
var
  ptd: PTypeData;
  PParam: PParamData;
  PTypeString: ^ShortString;
  I: Integer;
begin
  inherited Create;

  OverrideHandlerMode := 0;

  with TTitanBaseScripter(Scripter) do
  begin
    if SymbolTable.Kind[SubID] = KindTYPE then
      while SymbolTable.Kind[SubID] <> KindSUB do
        Inc(SubID);
  end;

  Self.SubID := SubID;
  Self.Scripter := Scripter;
  Self.This := This;

  ptd := GetTypeData(pti);
  ParamCount := ptd^.ParamCount;
  SetLength(Parameters, ParamCount);
  ParamTypes := TStringList.Create;

  PParam := PParamData(@(ptd^.ParamList));
  for I:=0 to ParamCount - 1 do
  begin
    PTypeString := ShiftPointer(PParam, SizeOf(TParamFlags) + Length(PParam^.ParamName) + 1);
    ParamTypes.Add(PTypeString^);
    PParam := ShiftPointer(PTypeString, Length(PTypeString^) + 1);
  end;
end;

destructor TTitanEventHandler.Destroy;
begin
  ParamTypes.Free;
  inherited Destroy;
end;

var
  PThis: PVariant;

procedure TTitanEventHandler.Invoke;

type
  PWord = ^Word;
  PShortInt = ^ShortInt;
  PSmallInt = ^SmallInt;
  PByte = ^Byte;

procedure Adjust(var Val: Longint);
type
  T = array[1..4] of Byte;
begin
  T(Val)[2] := 0;
  T(Val)[3] := 0;
  T(Val)[4] := 0;
end;

var
  J:LongInt;
  I,  Index: integer;
  S: String;
  R, V: Variant;
  SO: TTitanScriptObject;
  ClassRec: TTitanClassRec;
  RTTITypeDefinition: TTitanRTTITypeDefinition;
  pti: PTypeInfo;
  ptd: PTypeData;
  SZ: Integer;
  
  IsDelphiObject: Boolean;
  TypeID, ParamID: Integer;
  ByRef: Boolean;

  Variants: array[0..30] of Variant;
  Ptrs: array[0..30] of Pointer;
  Integers: array[0..30] of Integer;

  Types, ExtraTypes, Sizes: array of Integer;
  Pointers: array of Pointer;
  ByRefs: array of Boolean;

  CreatedRecords: TList;

  InstanceClassRec: TTitanClassRec;
  fScripter:TTitanBaseScripter;

begin
  if not TTitanBaseScripter(Scripter).AllowedEvents then Exit;
  CreatedRecords := TList.Create;

  try

   //================ initialaze =======================================
  if Pointers=nil then
   begin

    fScripter:=TTitanBaseScripter(Scripter); //set Scripter from Pointer

    SetLength(Pointers, ParamCount + 1);
    SetLength(Types, ParamCount + 1);
    SetLength(ExtraTypes, ParamCount + 1);
    SetLength(Sizes, ParamCount + 1);
    SetLength(ByRefs, ParamCount + 1);

    for I:=0 to ParamCount do
    begin
      Types[I] := typeCLASS;
      ExtraTypes[I] := 0;
      Sizes[I] := 4;
      Pointers[I] := Self;
      ByRefs[I]:=false;
    end;

    Types[ParamCount] := 0;
    SZ := 4;
    RetSize := 0;
    _P := ShiftPointer(_P, SZ*(ParamCount - 2));
    SetLength(Parameters, ParamCount);

    if ParamCount > 2 then
    case ParamCount - 2 of
      1: RetSize := 4;
      2: RetSize := 8;
      3: RetSize := 12;
      4: RetSize := 16;
      5: RetSize := 20;
      6: RetSize := 24;
      7: RetSize := 28;

      8: RetSize := 32;
      9: RetSize := 36;
      10: RetSize := 40;
    else
      raise Exception.Create('Too many parameters in event call - scripter limitation');
    end;

  end;
  //======================================================================


  for I:=0 to ParamCount - 1 do   //--------- For ParamCount Loop start -------------
    begin

      with fScripter do
      begin
        ParamID := SymbolTable.GetParamID(SubID, I+1);
        ByRef  :=(SymbolTable.ByRef[ParamID]=1);
        ByRefs[I] := ByRef;
      end;


      if ByRef then
        case I of
          0:
          begin
            J := _EDX;
            Ptrs[I + 1] := Pointer(J);
            J := LongInt(Ptrs[I + 1]^);
          end;
          1:
          begin
            J := _ECX;
            Ptrs[I + 1] := Pointer(J);
            J := LongInt(Ptrs[I + 1]^);
          end;
        else
          begin
            // Set up the variable out of the stack
            J := LongInt(_P^);
            Ptrs[I + 1] := Pointer(J);
            J := LongInt(Ptrs[I + 1]^);
            // Shift the pointer so we get the right spot next time...
            _P := ShiftPointer(_P, -SZ);
            SZ := 4;
          end;
         end
     else
        case I of
          0: J := _EDX;
          1: J := _ECX;
        else
          begin
            J := LongInt(_P^);
            _P := ShiftPointer(_P, -SZ);
            SZ := 4;
          end;
        end;    


      S := ParamTypes[I];
      ClassRec := fScripter.ClassList.FindClassByName(S);

      IsDelphiObject := ClassRec <> nil;
      if StrEql(S, 'Boolean') or StrEql(S, 'String') then IsDelphiObject := false;
      if IsDelphiObject then IsDelphiObject := ClassRec.ck in [ckClass];

      //............. Structure .....................................
      if (Assigned(ClassRec)) and (ClassRec.ck=ckStructure)
      then begin
        SO := ClassRec.CreateScriptObject;
        Parameters[I].VType := vtObject;
        Parameters[I].VObject := SO;

        if (ByRef)
        then begin
          SO.ExtraPtr := Ptrs[I + 1];  // We're pointing *right* at the structure, not a reference to it...  J's not.
          SO.ExternalExtraPtr := TRUE;
        end
        else begin

          SO.ExtraPtr := AllocMem(SO.ExtraPtrSize);

          {.$IFDEF CPU64}
            //  Move(J, SO.ExtraPtr^, SO.ExtraPtrSize); //?????????????????
          {.$ELSE}
             Move(LongInt(J), SO.ExtraPtr^, SO.ExtraPtrSize);
          {.$ENDIF}

        end;

        CreatedRecords.Add(SO);

        Types[I] := typeRECORD;
        Pointers[I] := SO.ExtraPtr;
      end else
      //............. for Set .....................................
      if (ClassRec <> nil) and ClassRec.isSet then
      begin
        Integers[I] := LongInt(J);
        // pti := ClassRec.PtiSet;


        Adjust(LongInt(J));


        //Variants[I] := SetToVariantArray(J, pti);
        Variants[I] := ByteSetToPaxArray(TByteSet(J), Scripter);
        Parameters[I].VType := vtVariant;
        Parameters[I].VVariant := @Variants[I];
        Types[I] := typeSET;
        Pointers[I] := @Integers[I];
      end else
      //............. for Interface .....................................
      if (ClassRec <> nil) and (ClassRec.ck = ckInterface) then
      begin
        SO := InterfaceToScriptObject(IUnknown(J), Scripter, S);

        Parameters[I].VType := vtObject;
        Parameters[I].VObject := SO;

        Types[I] := typeCLASS;
        Pointers[I] := @SO.Instance;
      end else
      //............. for IsDelphiObject .....................................
      if (IsDelphiObject)
      then begin
        SO := nil;
        Index := fScripter.ScriptObjectList.IndexOfDelphiObject(TObject(J));
        if Index = -1 then
        begin
          if J = 0 then
          begin
            Integers[I] := 0;
            Parameters[I].VType := vtObject;
            Parameters[I].VObject := nil;
            Types[I] := typeCLASS;
            Pointers[I] := @Integers[I];
            continue;
          end
          else
          begin
            InstanceClassRec := fScripter.ClassList.FindClassByName(TObject(J).ClassName);
            if InstanceClassRec = nil then  InstanceClassRec := ClassRec;
            SO := InstanceClassRec.CreateScriptObject;
            SO.Instance := TObject(j);
           //SO := ClassRec.CreateScriptObject;
           //SO.Instance := TObject(J);
          end;
        end
        else begin
          SO := TTitanScriptObject(fScripter.ScriptObjectList.PaxObjects[Index]);
        end;

        Parameters[I].VType := vtObject;
        Parameters[I].VObject := SO;

        Types[I] := typeCLASS;
        Pointers[I] := @SO.Instance;
      end else
      //............. for RTTIType .....................................
      
      begin
        RTTITypeDefinition := GlobalDefinitionList.FindRTTITypeDefByName(S);

        if RTTITypeDefinition <> nil then
        begin

          pti := RTTITypeDefinition.pti;
          case pti^.Kind of
            tkEnumeration:
            begin
              Integers[I] := J;
              Adjust(J);
              ptd := GetTypeData(pti);
           {$ifndef fpc}
              if ptd^.BaseType^ = TypeInfo(Boolean) then
              begin
                Parameters[I].VType := vtBoolean;
                if J = 0 then
                  Parameters[I].VBoolean := false else
                  Parameters[I].VBoolean := true;

                Types[I] := typeBOOLEAN;
                Pointers[I] := @Parameters[I].VBoolean;
              end
              else
            {$endif}
              begin
               // Parameters[I].VType := vtAnsiString;
               // Parameters[I].VAnsiString := Pointer(AnsiString(GetEnumName(pti, J)));

                Parameters[I].VType := vtInteger;
                Parameters[I].VInteger := J;
                Types[I] := typeENUM;
                Pointers[I] := @Integers[I];
              end;
            end;
            tkSet:
            begin
              Integers[I] := J;
              Adjust(J);
              Variants[I] := SetToVariantArray(J, pti);
              Parameters[I].VType := vtVariant;
              Parameters[I].VVariant := @Variants[I];
              Types[I] := typeSET;
              Pointers[I] := @Integers[I];
            end;
            tkChar:
            begin
              Adjust(J);
              Parameters[I].VType := vtInteger;
              Parameters[I].VType := vtChar;
              Parameters[I].VInteger := J;
              Types[I] := typeCHAR;
              Pointers[I] := @ Parameters[I].VInteger;
            end;

            else
            begin
              Parameters[I].VType := vtInteger;
              Parameters[I].VInteger := J;
              Types[I] := typeINTEGER;
              Pointers[I] := @Parameters[I].VInteger;
            end;
          end;
        end else
        //............. for base types .....................................
        begin
          if StrEql(ParamTypes[I], 'String') then
          begin
            Parameters[I].VType := vtAnsiString;
            Parameters[I].VAnsiString := Pointer(J);
            Types[I] := typeSTRING;
            Pointers[I] := Parameters[I].VAnsiString;
          end
          else if StrEql(ParamTypes[I], 'WideString') then
          begin
            Parameters[I].VType := vtWideString;
            Parameters[I].VWideString := Pointer(J);
            Types[I] := typeWIDESTRING;
            Pointers[I] := Parameters[I].VWideString;
          end
          else if StrEql(ParamTypes[I], 'Boolean') then
          begin
            Parameters[I].VType := vtBoolean;
            Parameters[I].VBoolean := Boolean(Byte(J));
            Types[I] := typeBOOLEAN;
            Pointers[I] := @Parameters[I].VBoolean;
          end
          else if StrEql(ParamTypes[I], 'Word') then
          begin
            Parameters[I].VType := vtInteger;
            Parameters[I].VInteger := J;
            Types[I] := typeWORD;
            Pointers[I] := @Parameters[I].VInteger;
          end
          else if StrEql(ParamTypes[I], 'Char') then
          begin
            Adjust(J);
            Parameters[I].VType := vtInteger;
            Parameters[I].VInteger := J;
            Types[I] := typeCHAR;
            Pointers[I] := @Parameters[I].VInteger;
          end
          else if StrEql(ParamTypes[I], 'Variant') then
          begin
            Parameters[I].VType := vtVariant;
            if ByRef then
              PVariant(Parameters[I].VVariant) := PVariant(Ptrs[I + 1]) else
              PVariant(Parameters[I].VVariant) := PVariant(J);

            Types[I] := typeVARIANT;
            Pointers[I] := Parameters[I].VVariant;
          end
         { else if StrEql(ParamTypes[I],'Double') then //By Sternas stefanos
          begin
            Parameters[I].VType := vtExtended;
           // if ByRef then
            //  Parameters[I].VExtended := pExtended(Ptrs[I + 1]);// else
              Parameters[I].VExtended := pExtended(_P^);
            Types[I] := typeExtended;
            Pointers[I] := Parameters[I].VExtended;
          end    }
          //..........................................
          else  
          begin
            Parameters[I].VType := vtInteger;
            Parameters[I].VInteger := J;
            Types[I] := typeINTEGER;
            Pointers[I] := @Parameters[I].VInteger;
          end;
        end;
      end;
    end;    //--------- For ParamCount Loop end -------------


    //-------------------------------- Call Before Exist Event------------------------------------------------
    if OverrideHandlerMode = 1 then
      if HostHandler.Code <> nil then
        BASE_Engine.Call(HostHandler.Code,
                       nil,
                       HostHandler.Data,
                       false,
                       _ccRegister,
                       Pointers,
                       Types,
                       ExtraTypes,
                       ByRefs,
                       Sizes,
                       false);
    //-------------------------------------------------------------------------------------------------------
    PThis:=@This;

    R := fScripter.CallMethod(SubID, This, Parameters,TRUE); //Base Call to Scripter function


    //Get then variables after Call to Scripter function

      for I:=1 to fScripter.SymbolTable.Count[SubID] do
      begin
        ParamID := fScripter.SymbolTable.GetParamID(SubID, I);
        TypeID := fScripter.SymbolTable.PType[ParamID];

        if fScripter.SymbolTable.ByRef[ParamID] = 1 then
        begin
          V := fScripter.SymbolTable.VariantValue[ParamID];

          case TypeID of
            typeBOOLEAN  :PBoolean(Ptrs[I])^ := V;
            typeINTEGER  :PInteger(Ptrs[I])^ := V;
            typeWORD     :PWord(Ptrs[I])^ := V;
            typeSHORTINT :PShortInt(Ptrs[I])^ := V;
            typeSMALLINT :PSmallInt(Ptrs[I])^ := V;
            typeBYTE     :PByte(Ptrs[I])^ := V;
            typeCHAR     :begin
                          s := v;
                          PByte(Ptrs[i])^ := Byte(char(s[1]));
                         end;
            typeDOUBLE   :PDouble(Ptrs[I])^ := V;
            typeEXTENDED :PExtended(Ptrs[I])^ := V;
            typeSTRING   :PString(Ptrs[I])^ := V;
            typeWIDESTRING:PWideString(Ptrs[I])^ := V;
            typeVARIANT :PVariant(Ptrs[I])^ := V;

          else
          if (TypeID < 0) then
          begin
             if isObject(V) then
             begin
               PInteger(Ptrs[I])^ := PtrInt(VariantToScriptObject(V).Instance);
               VariantToScriptObject(V).RefCount := -1;
               continue;
             end;

            // It's a structure - we're already directly modifying it.  Don't write anything, but we do need to
            // clean up after ourselves..
            ClassRec := fScripter.ClassList.FindClass(TypeID);
            if ClassRec <> nil then
              if ClassRec.ck = ckEnum then
                      PByte(Ptrs[I])^ := V;
          end else
          begin

            if isObject(V) then
              PInteger(Ptrs[I])^ := PtrInt(VariantToScriptObject(V).Instance) else
              PByte(Ptrs[I])^ := V;

          end;
        end;
      end;

     end;


    //-------------------------------- Call After Exist Event------------------------------------------------
    if OverrideHandlerMode = 2 then
      if HostHandler.Code <> nil then
        BASE_Engine.Call(HostHandler.Code,
                       nil,
                       HostHandler.Data,
                       false,
                       _ccRegister,
                       Pointers,
                       Types,
                       ExtraTypes,
                       ByRefs,
                       Sizes,
                       false);
    //-------------------------------------------------------------------------------------------------------

  finally   

   for I:=0 to CreatedRecords.Count - 1 do
      begin
        SO := TTitanScriptObject(CreatedRecords[I]);
        TTitanBaseScripter(SO.Scripter).ScriptObjectList.RemoveObject(SO); //is not same with fScripter
      end;

    CreatedRecords.Free; // We need to clean up after ourselves!!!

  end;

  if Assigned(fScripter.OnHalt) then
   if fScripter.Code.SignHaltGlobal then
     fScripter.OnHalt(fScripter.Owner);

end;

{$IFNDEF SOLARIS}

{$IFDEF CPU64}  //========================= 64bit ===============================
procedure TTitanEventHandler.HandleEvent;
const
  LocalFrameSize = 40;
asm
  mov dword ptr Self._EAX, rax
  mov dword ptr Self._EDX, rdx
  mov dword ptr Self._ECX, rcx
  mov dword ptr Self._P, rsp

  push rbp
  mov rbp, rsp

  sub rsp, LocalFrameSize

  mov [rbp-12], rcx
  mov [rbp- 8], rdx
  mov [rbp- 4], rax

  push rax
  call Invoke
  pop rax

  mov rcx, self.RetSize

  mov rsp, rbp
  pop rbp

  cmp rcx, 0

  jnz @@Ret4
  ret

@@Ret4:
  cmp rcx, 4
  jnz @@Ret8
  ret 4

@@Ret8:
  cmp rcx, 8
  jnz @@Ret12
  ret 8

@@Ret12:
  cmp rcx, $0c
  jnz @@Ret16
  ret $0c

@@Ret16:
  cmp rcx, $10
  jnz @@Ret20
  ret $10

@@Ret20:
  cmp rcx, $14
  jnz @@Ret24
  ret $14

@@Ret24:
  cmp rcx, $18
  jnz @@Ret28
  ret $18

@@Ret28:
  ret $1C
end;

{$ENDIF}

{$IFDEF CPUi386}       //========================= 32bit ===============================

procedure TTitanEventHandler.HandleEvent;
const
  LocalFrameSize = 40;
asm
  mov dword ptr Self._EAX, eax
  mov dword ptr Self._EDX, edx
  mov dword ptr Self._ECX, ecx
  mov dword ptr Self._P, esp

  push ebp
  mov ebp, esp

  sub esp, LocalFrameSize

  mov [ebp-12], ecx
  mov [ebp- 8], edx
  mov [ebp- 4], eax

  push eax
  call Invoke
  pop eax

  mov ecx, self.RetSize

  mov esp, ebp
  pop ebp

  cmp ecx, 0

  jnz @@Ret4
  ret

@@Ret4:
  cmp ecx, 4
  jnz @@Ret8
  ret 4

@@Ret8:
  cmp ecx, 8
  jnz @@Ret12
  ret 8

@@Ret12:
  cmp ecx, $0c
  jnz @@Ret16
  ret $0c

@@Ret16:
  cmp ecx, $10
  jnz @@Ret20
  ret $10

@@Ret20:
  cmp ecx, $14
  jnz @@Ret24
  ret $14

@@Ret24:
  cmp ecx, $18
  jnz @@Ret28
  ret $18

@@Ret28:
  ret $1C
end;

{$ENDIF}

{$IFDEF CPUARM}       //========================= ARM ===============================

procedure TTitanEventHandler.HandleEvent;
const
  LocalFrameSize = 40;
asm

end;

{$ENDIF}   //========================= end  32bit ===============================
{$ELSE} // NOT SOLARIS

procedure TTitanEventHandler.HandleEvent;
const
  LocalFrameSize = 40;
asm

end;
{$ENDIF}

