{%mainunit ../BASE_Engine.pas}

{***************************************************
 Package pl_TitanScript.pkg
 This file is part of CodeTyphon Studio
****************************************************}

var VarHParser:TPascalParser;

function StringToByteSet(pti: PTypeInfo; const V: String): TByteSet;
var
  ptd: PTypeData;
  I: Byte;
  S: String;
begin
  result := [];
  if pti = nil then Exit;
  ptd := GetTypeData(pti);
  if ptd = nil then Exit;

  pti := ptd^.CompType;
  
  if pti = nil then Exit;

  ptd := GetTypeData(pti);
  if ptd = nil then Exit;

  for I:= ptd.MinValue to ptd.MaxValue do
  begin
    S :=GetEnumName(pti, I);
    if Pos(S, V) > 0 then
      result := result + [TByteInt(I)];
  end;
end;

function PaxArrayToByteSet(const V: Variant): TByteSet;
var
  SO: TTitanScriptObject;
  PaxArray: TTitanArray;
  I, L: Integer;
  Val: Byte;
begin
  SO := VariantToScriptObject(V);
  PaxArray := TTitanArray(SO.Instance);
  L := PaxArray.HighBound(1);
  result := [];
  for I:=0 to L - 1 do
  begin
    Val := PaxArray.Get([I]);
    Include(result, TByteInt(Val));
  end;
end;

function ByteSetToPaxArray(const S: TByteSet; Scripter: Pointer): Variant;
var
  I, K, L: Integer;
  PaxArray: TTitanArray;
  SO: TTitanScriptObject;
begin
  L := 0;
  for I:=0 to 255 do
    if TByteInt(I) in S then
      Inc(L);

  PaxArray := TTitanArray.Create([L-1]);
  PaxArray.Scripter := Scripter;

  K := -1;
  for I:=0 to 255 do
    if TByteInt(I) in S then
    begin
      Inc(K);
      PaxArray.Put([K], I);
    end;

  SO := TTitanBaseScripter(Scripter).ArrayClassRec.CreateScriptObject;
  SO.RefCount := 1;
  SO.Instance := PaxArray;
  result := ScriptObjectToVariant(SO);

end;

function _IndexOf(Scripter: Pointer; L: TTitanVarList; const Value: Variant): Integer;
var
  I: Integer;
  P: PVariant;
begin
  result := -1;
  for I:=1 to L.Count do
  begin
    P := L.GetAddress(I);
    if EqualVariants(Scripter, P^, Value) then
    begin
      result := I;
      Exit;
    end;
  end;
end;

function AddSets(const S1, S2: Variant): Variant;
var
  SO, SO1, SO2: TTitanScriptObject;
  A, A1, A2: TTitanArray;
  L: TTitanVarList;
  I, Index: Integer;
  P: PVariant;
  Scripter: Pointer;
begin
  SO1 := VariantToScriptObject(S1);
  SO2 := VariantToScriptObject(S2);
  A1 := TTitanArray(SO1.Instance);
  A2 := TTitanArray(SO2.Instance);

  Scripter := SO1.Scripter;

  L := TTitanVarList.Create;

  if A1.Length > 0 then
  for I:=0 to A1.Length - 1 do
  begin
    P := A1.GetPtr([I]);
    Index := _IndexOf(Scripter, L, P^);
    if Index = -1 then
      L.Add(P^);
  end;

  if A2.Length > 0 then
  for I:=0 to A2.Length - 1 do
  begin
    P := A2.GetPtr([I]);
    Index := _IndexOf(Scripter, L, P^);
    if Index = -1 then
      L.Add(P^);
  end;

  A := TTitanArray.Create([L.Count - 1]);
  A.Scripter := Scripter;
  for I:=1 to L.Count do
    A.Put([I-1], L[I]);

  L.Free;

  SO := TTitanBaseScripter(Scripter).ArrayClassRec.CreateScriptObject;
  SO.RefCount := 1;
  SO.Instance := A;
  result := ScriptObjectToVariant(SO);
end;

function SubSets(const S1, S2: Variant): Variant;
var
  SO, SO1, SO2: TTitanScriptObject;
  A, A1, A2: TTitanArray;
  L: TTitanVarList;
  I, Index: Integer;
  P: PVariant;
  Scripter: Pointer;
begin
  SO1 := VariantToScriptObject(S1);
  SO2 := VariantToScriptObject(S2);
  A1 := TTitanArray(SO1.Instance);
  A2 := TTitanArray(SO2.Instance);

  Scripter := SO1.Scripter;

  L := TTitanVarList.Create;

  if A1.Length > 0 then
  for I:=0 to A1.Length - 1 do
  begin
    P := A1.GetPtr([I]);
    Index := _IndexOf(Scripter, L, P^);
    if Index = -1 then
      L.Add(P^);
  end;

  if A2.Length > 0 then
  for I:=0 to A2.Length - 1 do
  begin
    P := A2.GetPtr([I]);
    Index := _IndexOf(Scripter, L, P^);
    if Index <> -1 then
      L.Delete(Index);
  end;

  A := TTitanArray.Create([L.Count - 1]);
  A.Scripter := Scripter;
  for I:=1 to L.Count do
    A.Put([I-1], L[I]);

  L.Free;

  SO := TTitanBaseScripter(Scripter).ArrayClassRec.CreateScriptObject;
  SO.RefCount := 1;
  SO.Instance := A;
  result := ScriptObjectToVariant(SO);
end;

function IntersectSets(const S1, S2: Variant): Variant;
var
  SO, SO1, SO2: TTitanScriptObject;
  A, A1, A2: TTitanArray;
  L: TTitanVarList;
  I, J, Index: Integer;
  P1, P2: PVariant;
  Scripter: Pointer;
begin
  SO1 := VariantToScriptObject(S1);
  SO2 := VariantToScriptObject(S2);
  A1 := TTitanArray(SO1.Instance);
  A2 := TTitanArray(SO2.Instance);

  Scripter := SO1.Scripter;

  L := TTitanVarList.Create;

  if A1.Length > 0 then
  for I:=0 to A1.Length - 1 do
  begin
    P1 := A1.GetPtr([I]);
    Index := _IndexOf(Scripter, L, P1^);
    if Index = -1 then
    begin
      for J:=0 to A2.Length - 1 do
      begin
        P2 := A2.GetPtr([J]);
        if EqualVariants(Scripter, P1^, P2^) then
        begin
          Index := J;
          Break;
        end;
      end;
      if Index <> -1 then
        L.Add(P1^);
    end;
  end;

  A := TTitanArray.Create([L.Count - 1]);
  A.Scripter := Scripter;
  for I:=1 to L.Count do
    A.Put([I-1], L[I]);

  L.Free;

  SO := TTitanBaseScripter(Scripter).ArrayClassRec.CreateScriptObject;
  SO.RefCount := 1;
  SO.Instance := A;
  result := ScriptObjectToVariant(SO);
end;

function IsSubSet(const S1, S2: Variant): Boolean;
var
  SO1, SO2: TTitanScriptObject;
  A1, A2: TTitanArray;
  I, J: Integer;
  P1, P2: PVariant;
  Scripter: Pointer;
  Found: Boolean;
begin
  SO1 := VariantToScriptObject(S1);
  SO2 := VariantToScriptObject(S2);
  A1 := TTitanArray(SO1.Instance);
  A2 := TTitanArray(SO2.Instance);

  Scripter := SO1.Scripter;

  result := true;

  if A1.Length > 0 then
  for I:=0 to A1.Length - 1 do
  begin
    P1 := A1.GetPtr([I]);
    Found := false;
    for J:=0 to A2.Length - 1 do
    begin
      P2 := A2.GetPtr([J]);
      Found := EqualVariants(Scripter, P1^, P2^);
      if Found then
        Break;
    end;
    if not Found then
    begin
      result := false;
      Exit;
    end;
  end;
end;

function IsStrictSubSet(const S1, S2: Variant): Boolean;
var
  SO1, SO2: TTitanScriptObject;
  A1, A2: TTitanArray;
begin
  SO1 := VariantToScriptObject(S1);
  SO2 := VariantToScriptObject(S2);
  A1 := TTitanArray(SO1.Instance);
  A2 := TTitanArray(SO2.Instance);
  result := IsSubSet(S1, S2) and (A1.Length < A2.Length);
end;

function AreEqualSets(const S1, S2: Variant): Boolean;
var
  SO1, SO2: TTitanScriptObject;
  A1, A2: TTitanArray;
  I, J: Integer;
  P1, P2: PVariant;
  Found: Boolean;
  Scripter: Pointer;
begin
  SO1 := VariantToScriptObject(S1);
  SO2 := VariantToScriptObject(S2);
  A1 := TTitanArray(SO1.Instance);
  A2 := TTitanArray(SO2.Instance);

  Scripter := SO1.Scripter;

  result := true;

  if A1.Length > 0 then
  for I:=0 to A1.Length - 1 do
  begin
    P1 := A1.GetPtr([I]);
    Found := false;
    for J:=0 to A2.Length - 1 do
    begin
      P2 := A2.GetPtr([J]);
      Found := EqualVariants(Scripter, P1^, P2^);
      if Found then
        Break;
    end;
    if not Found then
    begin
      result := false;
      Exit;
    end;
  end;

  if A2.Length > 0 then
  for I:=0 to A2.Length - 1 do
  begin
    P1 := A2.GetPtr([I]);
    Found := false;
    for J:=0 to A1.Length - 1 do
    begin
      P2 := A1.GetPtr([J]);
      Found := EqualVariants(Scripter, P1^, P2^);
      if Found then
        Break;
    end;
    if not Found then
    begin
      result := false;
      Exit;
    end;
  end;

end;

function InSet(const Value, ASet: Variant): Boolean;
var
  SO: TTitanScriptObject;
  A: TTitanArray;
  I: Integer;
  P: PVariant;
  Scripter: Pointer;
begin
  SO := VariantToScriptObject(ASet);
  A := TTitanArray(SO.Instance);
  Scripter := SO.Scripter;

  result := false;

  if IsPaxArray(Value) then
  begin
    result := true;
    SO := VariantToScriptObject(Value);
    A := TTitanArray(SO.Instance);
    for I:=0 to A.Length - 1 do
    begin
      P := A.GetPtr(I);
      result := result and InSet(P^, ASet);
      if not result then
        Break;
    end;
  end
  else
    for I:=0 to A.Length - 1 do
    begin
      P := A.GetPtr(I);
      if EqualVariants(Scripter, P^, Value) then
      begin
        result := true;
        Exit;
      end;
    end;
end;

function EqualVariants(Scripter: Pointer; const V1, V2: Variant;
                       T1: Integer = -1;
                       T2: Integer = -1): TBoolean;
var
  SO1, SO2: TTitanScriptObject;
  IsSet: Boolean;
begin
  IsSet := (T1 = typeSET) or (T2 = typeSET);

  T1 := VarType(V1);
  T2 := VarType(V2);
  if T1 = T2 then
  begin
    if T1 in [varEmpty, varNull] then
    begin
      result := true;
      Exit;
    end;

    if T1 = varScriptObject then
    begin
      SO1 := VariantToScriptObject(V1);
      SO2 := VariantToScriptObject(V2);
      if IsPaxArray(V1) and IsPaxArray(V2) then
      begin
        if IsSet then
          result := AreEqualSets(V1, V2)
        else
          result := ComparePaxArrays(TTitanArray(SO1.Instance),
                                     TTitanArray(SO2.Instance));
        Exit;
      end
      else
      begin
        result := SO1 = SO2;
        Exit;
      end;
    end;
    result := V1 = V2;
  end
  else
  begin
    if T1 in [varEmpty, varNull] then
      result := false
    else if T2 in [varEmpty, varNull] then
      result := false
    else
      result := V1 = V2;
  end;

{

  else if (T1 in [varByte, varInteger, varDouble, varVariant, varCurrency]) and (T2 in [varByte, varInteger, varDouble, varVariant, varCurrency]) then
    result := V1 = V2
  else if ((T1 = varString) or (T1 = varOleStr)) and ((T2 = varString) or (T2 = varOleStr)) then
    result := V1 = V2;
}
end;

function PaxArrayToDynamicArray(const V: Variant; ElTypeID: Integer): Pointer;
var
  L, I, ElSize: PtrInt;
  P: Pointer;
  SO: TTitanScriptObject;
  PaxArray: TTitanArray;
  Val: Variant;
begin
  if not IsObject(V) then
  begin
    result := AllocMem(2*SizeOf(Integer));
    PtrInt(result^) := 1;
    P := ShiftPointer(result, SizeOf(Integer));
    PtrInt(P^) := 0;
    P := ShiftPointer(P, SizeOf(Integer));
    result := P;
    Exit;
  end;

  SO := VariantToScriptObject(V);
  PaxArray := TTitanArray(SO.Instance);

  L := PaxArray.HighBound(1);
  ElSize := PaxTypes.GetSize(ElTypeID);
  result := AllocMem(2*SizeOf(Integer) +  L * ElSize);
  P := ShiftPointer(result, SizeOf(Integer));
  PtrInt(P^) := L;
  P := ShiftPointer(P, SizeOf(Integer));
  result := P;
  for I:=0 to L - 1 do
  begin
    Val := PaxArray.Get([I]);
    if IsPaxArray(Val) then
      Pointer(P^) := PaxArrayToDynamicArray(Val, ElTypeID)
    else
      PutVariantValue(SO.Scripter, P, Val, ElTypeID);
    P := ShiftPointer(P, ElSize);
  end;
end;

function ComparePaxArrays(A1, A2: TTitanArray): Boolean;
var
  I: Integer;
  P1, P2: PVariant;
  Scripter: Pointer;
begin
  result := false;

  if A1.N <> A2.N then
    Exit;

  if A1.L <> A2.L then
    Exit;

  for I:=0 to A1.L - 1 do
    if A1.fBounds[I] <> A2.fBounds[I] then
      Exit;

  Scripter := A1.Scripter;

  P1 := A1.P;
  P2 := A2.P;
  for I:=0 to A1.N - 1 do
  begin
    if not EqualVariants(Scripter, P1^, P2^) then
      Exit;
    P1 := ShiftPointer(P1, _SizeVariant);
    P2 := ShiftPointer(P2, _SizeVariant);
  end;
  result := true;
end;

function UnaryPlus(const V: Variant): Variant;
var
  SO: TTitanScriptObject;
begin
  if IsObject(V) then
  begin
    SO := VariantToScriptObject(V);
    if IsPaxArray(V) then
    begin
      result := TTitanArray(SO.Instance).Duplicate;
      Exit;
    end;
  end;
  result := V;
end;

//========================= TTitanArray =================================================

constructor TTitanArray.Create(Bounds: array of Integer; typeID: Integer = typeVARIANT);
var
  I: Integer;
begin
  _ElMin:=0;
  _ElMax:=0;

  if typeID = 0 then
    typeID := typeVARIANT;
  if TypeID > PaxTypes.Count then
    typeID := typeVARIANT;

  Self.TypeID := typeID;
  _ElSize := _SizeVariant;

  if TypeID < PaxTypes.Count then
    if TypeID <> typeVARIANT then
      _ElSize := PaxTypes.GetSize(typeID);

  L := System.Length(Bounds);
  SetLength(Self.fBounds, L);
  N := 1;
  for I:=0 to L - 1 do
  begin
    Self.fBounds[I] := Bounds[I] + 1;
    N := N * Self.fBounds[I];
  end;
  P := AllocMem(N * _ElSize);

  fIndexes := TTitanIds.Create(true);

  Scripter := nil;
end;

function TTitanArray.Typed: Boolean;
begin
  result := typeID <> typeVARIANT;
end;

function TTitanArray.Duplicate: Variant;
var
  I: Integer;
  Q, R: PVariant;
  PaxArray: TTitanArray;
  SO: TTitanScriptObject;
  B: array of Integer;
  Q1, R1: Pointer;
begin
  SetLength(B, L);
  for I:=0 to L - 1 do
    B[I] := fBounds[I] - 1;

  PaxArray := TTitanArray.Create(B);
  PaxArray.Scripter := Scripter;

  if _ElSize = _SizeVariant then
  begin
    Q := P;
    R := PaxArray.P;
    for I:=0 to N - 1 do
    begin
      if IsPaxArray(Q^) then
      begin
        SO := VariantToScriptObject(Q^);
        R^ := TTitanArray(SO.Instance).Duplicate;
      end
      else
        R^ := Q^;
      Q := ShiftPointer(Q, _ElSize);
      R := ShiftPointer(R, _ElSize);
    end;
  end
  else
  begin
    Q1 := P;
    R1 := PaxArray.P;
    for I:=0 to N - 1 do
    begin
      if typeID = typeSTRING then
        String(R1^) := String(Q1^)
      else
        Move(Q1^, R1^, _ElSize);
      Q1 := ShiftPointer(Q1, _ElSize);
      R1 := ShiftPointer(R1, _ElSize);
    end;
  end;

  result := ScriptObjectToVariant(DelphiInstanceToScriptObject(PaxArray, Scripter));
end;

procedure TTitanArray.ReDim(Bounds: array of Integer);
var
  I, OldN, MinN: Integer;
  Q: Pointer;
begin
  OldN := N;

  L := System.Length(Bounds);
  SetLength(Self.fBounds, L);
  N := 1;
  for I:=0 to L - 1 do
  begin
    Self.fBounds[I] := Bounds[I] + 1;
    N := N * Self.fBounds[I];
  end;
  Q := AllocMem(N * _ElSize);

  MinN := OldN;
  if N < OldN then
    MinN := N;
  Move(P^, Q^, MinN * _ElSize);

  FreeMem(P, OldN);
  P := Q;
end;

function TTitanArray.GetLength: Integer;
begin
  result := fBounds[0];
end;

procedure TTitanArray.PutLength(Value: Integer);
begin
  if GetLength <> Value then
  begin
    fBounds[0] := Value - 1;
    ReDim(fBounds);
  end;
end;

procedure TTitanArray._ReDim;
var
  I, OldN, MinN: Integer;
  Q: Pointer;
begin
  OldN := N;

  L := fIndexes.Count;
  SetLength(Self.fBounds, L);
  N := 1;
  for I:=0 to L - 1 do
  begin
    Self.fBounds[I] := fIndexes[I] + 1;
    N := N * Self.fBounds[I];
  end;
  Q := AllocMem(N * _ElSize);

  MinN := OldN;
  if N < OldN then
    MinN := N;
  Move(P^, Q^, MinN * _ElSize);

  FreeMem(P, OldN);
  P := Q;
end;

destructor TTitanArray.Destroy;
var
  I, VT: Integer;
  Ptr: Pointer;
  SO: TTitanScriptObject;
begin
  Ptr := P;

  if typeID = typeVARIANT then
  begin
    for I:=0 to N - 1 do
    begin
      VT := VarType(Variant(Ptr^));
      if VT = varString then
         Variant(Ptr^) := ''
      else if VT = varScriptObject then
      begin
        SO := VariantToScriptObject(Variant(Ptr^));
        if TTitanBaseScripter(scripter).ScriptObjectList.HasObject(SO) then
          SO.RefCount := 1;
        VarClear(Variant(Ptr^));
      end;

      Inc(PtrInt(Ptr), _ElSize);
    end;
  end
  else
  begin
    for I:=0 to N - 1 do
    begin
      if typeID = typeSTRING then
         String(Ptr^) := '';
      Inc(PtrInt(Ptr), _ElSize);
    end;
  end;

  FreeMem(P, N * _ElSize);
  fIndexes.Free;
  inherited;
end;

function TTitanArray.HighBound(Dim: Integer): Integer;
begin
  result := fBounds[Dim - 1];
end;

function TTitanArray.ToString: String;

var
  B: array[0..100] of Integer;

function F(Q: Pointer; I: Integer; var SZ: Integer): String;
var
  K, TempSZ: Integer;
begin
  if I >= L then
  begin
    if Scripter = nil then
      result := VarToStr(GetTerminal(Q)^)
    else
      result := _ToStr(Scripter, GetTerminal(Q)^);
    SZ := _ElSize;
    Exit;
  end;

  with TTitanBaseScripter(Scripter).Visited do
  if (IndexOf(Q) = -1) then
    Add(Q)
  else
  begin
    SZ := _ElSize;
    result := '...';
    Exit;
  end;

  result := '[';

  SZ := 0;

  for K:=1 to B[I] do
  begin
    result := result + F(Q, I + 1, TempSZ);
    if K < B[I] then
      result := result + ',';
    Inc(PtrInt(Q), TempSZ);
    Inc(SZ, TempSZ);
  end;

  result := result + ']';
end;

var
  I, SZ: Integer;
begin
  for I:=0 to L - 1 do
    B[I] := fBounds[L - 1 - I];

  if typeID = typeVARIANT then
    result := F(P, 0, SZ)
  else
    result := 'array';
end;

function TTitanArray.AddressOfElement: Pointer;
var
  I, J, R: Integer;
begin
  if L > fIndexes.Count then
    raise TTitanScriptFailure.Create(errNotEnoughParameters)
  else if L < fIndexes.Count then
    raise TTitanScriptFailure.Create(errTooManyParameters);

  J := fIndexes[0];
  R := 1;
  for I:=1 to L - 1 do
  begin
    R := R * fBounds[I-1];
    Inc(J, R * fIndexes[I]);
  end;
  result := Pointer(PtrInt(P) + J * _ElSize);
end;

function TTitanArray._Get: Variant;
var
  Q: Pointer;
begin
  Q := AddressOfElement;

  if TypeID = typeVARIANT then
    result := PVariant(Q)^
  else
    result := GetVariantValue(scripter, Q, typeID);
end;

function TTitanArray._GetPtr: PVariant;
begin
  result := AddressOfElement;
end;

function TTitanArray._GetEx: Variant;
begin
  if not _CheckIndexes then
    _ReDim;
  result := _Get;
end;

function TTitanArray._GetPtrEx: PVariant;
begin
  if not _CheckIndexes then
    _ReDim;

  result := AddressOfElement;
end;

procedure TTitanArray._Put(const Value: Variant);
var
  Q: Pointer;
begin
  Q := AddressOfElement;

  if TypeID = typeVARIANT then
    PVariant(Q)^ := Value
  else
    PutVariantValue(scripter, Q, Value, typeID);
end;

procedure TTitanArray._PutEx(const Value: Variant);
begin
  if not _CheckIndexes then
    _ReDim;

  _Put(Value);
end;

procedure TTitanArray.ClearIndexes;
begin
  fIndexes.Clear;
end;

function TTitanArray.AddIndex(I: Integer): Integer;
begin
  result := fIndexes.Add(I);
end;

procedure TTitanArray.InsertIndex(I: Integer);
begin
  fIndexes.Insert(0, Pointer(I));
end;

function TTitanArray.Get(const Indexes: array of Integer): Variant;
var
  I: Integer;
begin
  fIndexes.Clear;
  for I:=0 to System.Length(Indexes) - 1 do
    fIndexes.Add(Indexes[I]);
  result := _Get;
end;

function TTitanArray.GetPtr(const Indexes: array of Integer): PVariant;
var
  I: Integer;
begin
  fIndexes.Clear;
  for I:=0 to System.Length(Indexes) - 1 do
    fIndexes.Add(Indexes[I]);
  result := AddressOfElement;
end;

procedure TTitanArray.Put(const Indexes: array of Integer; const Value: Variant);
var
  I: Integer;
begin
  fIndexes.Clear;
  for I:=0 to System.Length(Indexes) - 1 do
    fIndexes.Add(Indexes[I]);
  _Put(Value);
end;

function TTitanArray._CheckIndexes: Boolean;
var
  I, Idx: Integer;
begin
  for I:=0 to fIndexes.Count - 1 do
  begin
    Idx := fIndexes[I] + 1;
    if Idx > fBounds[I] then
    begin
      result := False;
      Exit;
    end;
  end;
  result := true;
end;

function TTitanArray.CheckIndexes(const Indexes: array of Integer): Boolean;
var
  I, Idx: Integer;
begin
  result := true;
  for I:=0 to L - 1 do
  begin
    Idx := Indexes[I] + 1;
    if Idx > fBounds[I] then
    begin
      result := False;
      Exit;
    end;
  end;
end;

procedure TTitanArray.PutEx(const Indexes: array of Integer; const Value: Variant);
begin
  if not CheckIndexes(Indexes) then
    ReDim(Indexes);
  Put(Indexes, Value);
end;

function TTitanArray.GetEx(const Indexes: array of Integer): Variant;
begin
  if not CheckIndexes(Indexes) then
    ReDim(Indexes);
  result := Get(Indexes);
end;

Procedure TTitanArray.SetMinMax(const aMin,aMax:integer);
  begin
   _ElMin:=aMin;
   _ElMax:=aMax;
  end;
//==========================================================================


function IsScripters: Boolean;
begin
  if Assigned(ScripterList) then
    result := ScripterList.Count > 0
  else
    result := false;
end;

procedure _Dump(MethodBody: TTitanMethodBody);
begin
  with MethodBody do
    TTitanBaseScripter(Scripter).Dump;
end;

procedure _IOResult(MethodBody: TTitanMethodBody);
begin
  with MethodBody do
    result.AsInteger := IOResult;
end;

procedure _ScriptObjectListCount(MethodBody: TTitanMethodBody);
begin
  with MethodBody do
    result.AsInteger := TTitanBaseScripter(Scripter).ScriptObjectList.Count;
end;

procedure _Destroy(MethodBody: TTitanMethodBody);
var
  SO: TTitanScriptObject;
  P: PVariant;
begin
  with MethodBody do
  begin
    P := Params[0].PValue;
    if IsObject(P^) then
    begin
      SO := VariantToScriptObject(P^);
      TTitanBaseScripter(Scripter).ScriptObjectList.RemoveObject(SO);
    end;
    VarClear(P^);
  end;
end;

procedure _PaxArray(MethodBody: TTitanMethodBody);
var
  PaxArray: TTitanArray;
  Bounds: array of Integer;
  I: Integer;
  SO: TTitanScriptObject;
begin
  with MethodBody do
  begin
    SetLength(Bounds, ParamCount);
    for I:=0 to ParamCount - 1 do
      Bounds[I] := Params[I].AsInteger - 1;

    PaxArray := TTitanArray.Create(Bounds);
    PaxArray.Scripter := Scripter;

    SO := DelphiInstanceToScriptObject(PaxArray, Scripter);
    result.AsVariant := ScriptObjectToVariant(SO);
  end;
end;

procedure _GetTickCount(MethodBody: TTitanMethodBody);
begin
{$IFDEF WINDOWS}
  with MethodBody do
  begin
    result.AsInteger := GetTickCount;
  end;
{$ENDIF}
end;

procedure _GetTotalAllocated(MethodBody: TTitanMethodBody);
var
  D: Double;
begin
{$IFDEF WINDOWS}
  with MethodBody do
  begin
    D := GetHeapStatus.TotalAllocated;
    result.AsVariant := D;
  end;
{$ELSE}
  with MethodBody do
  begin
    result.AsVariant := 0;
  end;
{$ENDIF}

end;

procedure _Sleep(MethodBody: TTitanMethodBody);
begin
{$IFDEF WINDOWS}
  with MethodBody do
    Sleep(Params[0].AsInteger);
{$ENDIF}
end;

procedure _CreateArray(MethodBody: TTitanMethodBody);
var
  HBound: Integer;
begin
  with MethodBody do
  begin
    HBound := Params[0].AsInteger;
    result.AsVariant := varArrayCreate([0, HBound], varVariant);
  end;
end;

procedure _DoNotDestroy(MethodBody: TTitanMethodBody);
var
  V: Variant;
  SO: TTitanScriptObject;
begin
  with MethodBody do
  begin
    V := Params[0].AsVariant;
    SO := VariantToScriptObject(V);
    SO.RefCount := 0;
  end;
end;

procedure _Assigned(MethodBody: TTitanMethodBody);
var
  V: Variant;
  SO: TTitanScriptObject;
begin
  with MethodBody do
  begin
    V := Params[0].AsVariant;
    if IsObject(V) then
    begin
      SO := VariantToScriptObject(V);
      result.AsBoolean := TTitanBaseScripter(Scripter).ScriptObjectList.HasObject(SO);
    end
    else
      result.AsBoolean := false;
  end;
end;

constructor TTitanParameter.Create(AScripter: Pointer);
begin
  Scripter := AScripter;
  Clear;
end;

destructor TTitanParameter.Destroy;
begin
  inherited;
end;

procedure TTitanParameter.Clear;
begin
  PValue := nil;
  VType := varUndefined;
end;

function TTitanParameter.GetAsString: String;
var
  V: Variant;
begin
  V := Variant(PValue^);
  result := ToStr(Scripter, V);
end;

procedure TTitanParameter.SetAsString(const Value: String);
var
  S: String;
begin
  S := ToStr(Scripter, Value);
  Variant(PValue^) := S;
  VType := varString;
end;

function TTitanParameter.GetAsBoolean: Boolean;
var
  V: Variant;
begin
  V := Variant(PValue^);
  result := ToBoolean(V);
end;

procedure TTitanParameter.SetAsBoolean(Value: Boolean);
begin
  Variant(PValue^) := ToBoolean(Value);
  VType := varBoolean;
end;

function TTitanParameter.GetAsInteger: Integer;
begin
  result := ToInteger(PValue^);
end;

procedure TTitanParameter.SetAsInteger(Value: Integer);
begin
  Variant(PValue^) := Value;
end;

function ToCardinal(const V: Variant): Cardinal;
var
  D: Double;
begin
  case VarType(V) of
    varInteger: result := V;
    varDouble:
    begin
      D := V;
      result := Round(D);
    end;
    else
      raise TTitanScriptFailure.Create(errIncompatibleTypes);
  end;
end;

function TTitanParameter.GetAsCardinal: Cardinal;
begin
  result := toCardinal(PValue^);
end;

procedure TTitanParameter.SetAsCardinal(Value: Cardinal);
var
  Dbl: Double;
begin
  if Value > Cardinal(MaxInt) then
  begin
    Dbl := Value;
    Variant(PValue^) := Dbl;
  end
  else
    Variant(PValue^) := PtrInt(Value);
end;

function TTitanParameter.GetAsPointer: Pointer;
var
  SO: TTitanScriptObject;
begin
  if IsObject(PValue^) then
  begin
    SO := VariantToScriptObject(PValue^);
    if SO.ClassRec.fClassDef <> nil then
      result := SO.Instance
    else
      result := SO;  
  end
  else
    result := Pointer(PtrInt(PValue^));
end;

procedure TTitanParameter.SetAsPointer(Value: Pointer);
var
  Instance: TObject;
  AClass: TClass;
begin
  if IsDelphiObject(Value) then
  begin
    Instance := TObject(Value);
    if Instance.ClassType = TTitanScriptObject then
    begin
      Variant(PValue^) := PtrInt(Instance);
      TVarData(Variant(PValue^)).VType := varScriptObject;
    end
    else
      Variant(PValue^) := ScriptObjectToVariant(DelphiInstanceToScriptObject(Instance, Scripter));
  end
  else if IsDelphiClass(Value) then
  begin
    AClass := TClass(Value);
    Variant(PValue^) := ScriptObjectToVariant(DelphiClassToScriptObject(AClass, Scripter));
  end
  else
    Variant(PValue^) := PtrInt(Value);
end;

function TTitanParameter.GetAsDouble: Double;
begin
  result := PValue^;
end;

procedure TTitanParameter.SetAsDouble(Value: Double);
begin
  Variant(PValue^) := Value;
end;

function TTitanParameter.GetAsTObject: TObject;
var
  V: Variant;
begin
  V := Variant(PValue^);
  result := TObject(ToInteger(V));
  VType := varScriptObject;
end;

procedure TTitanParameter.SetAsTObject(Value: TObject);
begin
  Variant(PValue^) := PtrInt(Value);
  VType := varScriptObject;
end;

function TTitanParameter.GetAsVariant: Variant;
begin
  result := Variant(PValue^);
end;

procedure TTitanParameter.SetAsVariant(const Value: Variant);
begin
  Variant(PValue^) := Value;
  VType := VarType(Value);
end;

constructor TTitanParameterList.Create(AScripter: Pointer);
var
  I: Integer;
begin
  inherited Create;
  Scripter := AScripter;
  for I:=0 to MaxParams - 1 do
    Add(TTitanParameter.Create(AScripter));
end;

procedure TTitanParameterList.Clear;
var
  I: Integer;
begin
  for I:=0 to Count - 1 do
    TTitanParameter(Items[I]).Free;

  inherited;
end;

constructor TTitanMethodBody.Create(Scripter: Pointer);
begin
  Self.Scripter := Scripter;
  fResult := TTitanParameter.Create(Scripter);
  PSelf := nil;
  fParameterList := TTitanParameterList.Create(Scripter);
  fParamCount := 0;
end;

procedure TTitanMethodBody.AddParameters(K: Integer);
begin
  while fParameterList.Count < K do
    fParameterList.Add(TTitanParameter.Create(Scripter));
end;

function TTitanMethodBody.GetSelf: TObject;
begin
  if PSelf = nil then
    result := nil
  else
    result := TObject(PSelf^);
end;

procedure TTitanMethodBody.SetSelf(Value: TObject);
begin
  if PSelf <> nil then
    TObject(PSelf^) := Value;
end;

destructor TTitanMethodBody.Destroy;
begin
  fResult.Free;
  fParameterList.Free;
end;

function TTitanMethodBody.FindNestedClass(const NamespaceName, ClassName: String): Pointer;
var
  OwnerList: TStringList;
begin
  OwnerList := TStringList.Create;
  if NamespaceName <> '' then
    OwnerList.Add(NamespaceName);
  result := TTitanBaseScripter(Scripter).ClassList.FindNestedClass(OwnerList, ClassName);
  OwnerList.Free;
end;

function TTitanMethodBody.DefaultValue: Variant;
begin
  result := Undefined;
  if PSelf = nil then
    Exit;
  if GetSelf().InheritsFrom(TTitanScriptObject) then
   result := TTitanScriptObject(GetSelf()).DefaultValue;
end;

procedure TTitanMethodBody.Clear;
begin
  fResult.Clear;
  PSelf := nil;
end;

function TTitanMethodBody.GetParameter(I: Integer): TTitanParameter;
begin
  if I >= fParameterList.Count then
    AddParameters(I + 10);

  result := fParameterList[I];
end;

procedure TTitanMethodBody.SetParameter(I: Integer; Value: TTitanParameter);
begin
  if I >= fParameterList.Count then
    AddParameters(I + 10);

  fParameterList[I] := Value;
end;

function TTitanMethodBody.GetField(I: Integer): TTitanParameter;
begin
  if I >= fParameterList.Count then
    AddParameters(I + 10);

  result := fParameterList[I];
end;

procedure TTitanMethodBody.SetField(I: Integer; Value: TTitanParameter);
begin
  if I >= fParameterList.Count then
    AddParameters(I + 10);

  fParameterList[I] := Value;
end;

constructor TTitanDefinition.Create(DefList: TTitanDefinitionList;
                               const Name: String; Owner: TTitanDefinition;
                               DefKind: TTitanDefKind);
begin
  Self.DefList := DefList;
  Self.Name := Name;
  Self.Owner := Owner;
  Self.DefKind := DefKind;
  UserData := 0;
  Module := -1;
  ResultType := '';
end;

function TTitanDefinition.FullName: String;
begin
  if Owner <> nil then
    result := Owner.FullName + '.' + Name
  else
    result := Name;
end;

function TTitanDefinition.GetValue: Variant;
begin
  result := DefValue;
end;

function TTitanDefinition.PValue: PVariant;
begin
  result := @ DefValue;
end;

procedure TTitanDefinition.SetValue(const AValue: Variant);
begin
  DefValue := AValue;
end;

procedure TTitanDefinition.AddToScripter(Scripter: Pointer);
begin
end;

procedure TTitanDefinition.AddToScripterList;
var
  I: Integer;
begin
  if Assigned(ScripterList) then
  begin
    for I:=0 to ScripterList.Count - 1 do
      AddToScripter(ScripterList[I]);
  end;
end;

constructor TTitanFieldDefinition.Create(DefList: TTitanDefinitionList; PClass: TClass; const FieldName,
                                       FieldType: String; Offset: Integer);
var
  Owner: TTitanDefinition;
  S: String;
begin
  Owner := DefList.FindClassDef(PClass);
  inherited Create(DefList, FieldName, Owner, dkField);
  Self.PClass := PClass;
  Self.FieldType := FieldType;
  Self.Offset := Offset;
  Self.TypeID := PAXTypes.IndexOf(UpperCase(FieldType));
  if Self.TypeID = -1 then
  begin
    S := FindTypeAlias(FieldType, true);
    Self.TypeID := PAXTypes.IndexOf(UpperCase(S));
  end;
end;

function TTitanFieldDefinition.GetFieldAddress(SO: pointer): Pointer;
var
  Address: Pointer;
begin
  Address := ShiftPointer(TTitanScriptObject(SO).Instance, Offset);
  DefValue := GetVariantValue(TTitanScriptObject(SO).Scripter, Address, TypeID);
  result := @DefValue;
end;

function TTitanFieldDefinition.GetFieldValue(SO: pointer): Variant;
begin
  GetFieldAddress(SO);
  result := DefValue;
end;

procedure TTitanFieldDefinition.SetFieldValue(SO: pointer; const Value: Variant);
var
  Address: Pointer;
begin
  Address := ShiftPointer(TTitanScriptObject(SO).Instance, Offset);
  PutVariantValue(TTitanScriptObject(SO).Scripter, Address, Value, TypeID);
end;

procedure TTitanFieldDefinition.Dump(var T: TextFile);
begin
  writeln(T, 'Field =', Name,  ' Type =', FieldType, ' Class =', PClass.ClassName,
             ' Offset =', Offset,
             ' Index =', Index);
end;

procedure TTitanFieldDefinition.AddToScripter(Scripter: Pointer);
var
  ClassRec: TTitanClassRec;
  ClassList: TTitanClassList;
begin
  ClassList := TTitanBaseScripter(Scripter).ClassList;
  ClassRec := ClassList.FindImportedClass(PClass);

  if (SignLoadOnDemand and (ClassRec = nil)) then
    ClassRec := TTitanBaseScripter(Scripter).ClassList.FindClassByName(Owner.Name);
  
  if ClassRec <> nil then
    ClassRec.AddHostField(Self)
  else
    raise TTitanScriptFailure.Create(Format(errClassIsNotImported, [PClass.ClassName]));
end;

constructor TTitanRecordFieldDefinition.Create(DefList: TTitanDefinitionList;
                                             Owner: TTitanClassDefinition;
                                             const FieldName, FieldType: String;
                                             Offset: Integer);
var
  FieldTypeDef: TTitanClassDefinition;
begin
  FieldTypeDef := DefList.FindClassDefByName(FieldType);

  inherited Create(DefList, FieldName, Owner, dkRecordField);
  Self.FieldType := FieldType;
  Self.Offset := Offset;
  Self.TypeID := PAXTypes.IndexOf(UpperCase(FieldType));

  if FieldTypeDef <> nil then
  begin
    if FieldTypeDef.ClassKind = ckEnum then
      Self.TypeID := typeENUM
    else if FieldTypeDef.ClassKind = ckStructure then
      Self.TypeID := typeRECORD
    else if FieldTypeDef.ClassKind = ckDynamicArray then
      Self.TypeID := typeARRAY;
  end;
end;

function TTitanRecordFieldDefinition.GetFieldAddress(Scripter, Address: pointer): Pointer;
begin
  Address := ShiftPointer(Address, Offset);
  DefValue := GetVariantValue(Scripter, Address, TypeID, FieldType);
  result := @DefValue;
end;

function TTitanRecordFieldDefinition.GetFieldValue(Scripter, Address: pointer): Variant;
begin
  GetFieldAddress(Scripter, Address);
  result := DefValue;
end;

procedure TTitanRecordFieldDefinition.SetFieldValue(Scripter, Address: pointer; const Value: Variant);
begin
  Address := ShiftPointer(Address, Offset);
  PutVariantValue(Scripter, Address, Value, TypeID);
end;

procedure TTitanRecordFieldDefinition.Dump(var T: TextFile);
begin
  writeln(T, 'RecordField =', Name,  ' Type =', FieldType, ' Offset =', Offset,
             ' Index =', Index);
end;

procedure TTitanRecordFieldDefinition.AddToScripter(Scripter: Pointer);
var
  ClassRec: TTitanClassRec;
  ClassList: TTitanClassList;
begin
  ClassList := TTitanBaseScripter(Scripter).ClassList;
  ClassRec := ClassList.FindClassByName(Owner.Name);

  if (SignLoadOnDemand and (ClassRec = nil)) then
    ClassRec := TTitanBaseScripter(Scripter).ClassList.FindClassByName(Owner.Name);
  
  if ClassRec <> nil then
    ClassRec.AddHostRecordField(Self)
  else
    raise TTitanScriptFailure.Create(Format(errClassIsNotImported, [Owner.Name]));
end;

constructor TTitanClassDefinition.Create(DefList: TTitanDefinitionList;
                                       const Name: String; Owner, Ancestor: TTitanDefinition);
begin
  inherited Create(DefList, Name, Owner, dkClass);
  Self.Ancestor := Ancestor;
  Self.GetPropDef := nil;
  Self.PutPropDef := nil;
  Self.ClassKind := ckClass;
  RecordSize := -1;
  Self.pti := nil;
  FillChar(guid, SizeOf(guid), 0);
  IsSet := false;
  PtiSet:= nil;
  IsStaticArray := false;

  ElMin:=-1;
  ElMax:=-1;
  
  OwnerList := nil;
end;

destructor TTitanClassDefinition.Destroy;
begin
  if OwnerList <> nil then
    OwnerList.Free;
end;

procedure TTitanClassDefinition.Dump(var T: TextFile);
var
  StrAncestor: String;
begin
  if Ancestor = nil then
    StrAncestor := 'nil'
  else
    StrAncestor := Ancestor.FullName;
  writeln(T, 'Class =', FullName,  ' Ancestor =', StrAncestor,
          ' Index =', Index);
end;

procedure TTitanClassDefinition.CreateOwnerList;
var
  P: TTitanDefinition;
begin
  OwnerList := TStringList.Create;

  P := Owner;
  while P <> nil do
  begin
    OwnerList.Insert(0, P.Name);
    P := P.Owner;
  end;
end;

function TTitanClassDefinition.FindClassRec(Scripter: Pointer): Pointer;
begin
  if OwnerList = nil then
     CreateOwnerList;

  result := TTitanBaseScripter(Scripter).ClassList.FindNestedClass(OwnerList, Name);
end;

procedure TTitanClassDefinition.AddToScripter(Scripter: Pointer);
var
  ClassRec, OwnerRec: TTitanClassRec;
  ClassList: TTitanClassList;
  ClassID: Integer;
  OwnerName, AncestorName: String;
begin
  ClassList := TTitanBaseScripter(Scripter).ClassList;
  ClassRec := FindClassRec(Scripter);

  if ClassRec <> nil then
    Exit;

  ClassID := - Index;

  if Ancestor <> nil then
    AncestorName := Ancestor.Name
  else
    AncestorName := '';

  if Owner <> nil then
  begin
    OwnerName := Owner.Name;
    OwnerRec := TTitanClassDefinition(Owner).FindClassRec(Scripter);
  end
  else
  begin
    OwnerName := RootNamespaceName;
    OwnerRec := ClassList.Records[0];
  end;

  ClassRec := ClassList.AddClass(ClassID, Name, OwnerName, AncestorName, ml, ClassKind, true);
  ClassRec.fClassDef := Self;
  ClassRec.ck := ClassKind;
  ClassRec.IsStaticArray := IsStaticArray;

  if (GetPropDef <> nil) or (PutPropDef <> nil) then
    ClassRec.fHasRunTimeProperties := true;

  ClassRec.isSet := IsSet;
  ClassRec.PtiSet := PtiSet;

  if OwnerRec <> nil then
    OwnerRec.AddNestedClass(ClassID, ml);

  if StrEql('TTitanArray', Name) then
    TTitanBaseScripter(Scripter).ArrayClassRec := ClassRec;
end;

constructor TTitanRTTITypeDefinition.Create(DefList: TTitanDefinitionList; pti: PTypeInfo);
begin
  inherited Create(DefList, pti^.Name, nil, dkRTTIType);
  Self.pti := pti;
  Self.FinalType := GetFinalType;
end;

procedure TTitanRTTITypeDefinition.Dump(var T: TextFile);
begin
  writeln(T, 'RTTIType =', pti^.Name, ' Index =', Index);
end;

function TTitanRTTITypeDefinition.GetFinalType: Integer;
begin
  result := 0;
  case pti^.Kind of
    tkInteger: result := typeINTEGER;
    tkChar: result := typeCHAR;
    tkEnumeration: result := typeENUM;
    tkFloat: result := typeDOUBLE;
    tkString: result := typeSHORTSTRING;
    tkLString: result := typeSTRING;
    tkWString: result := typeWIDESTRING;
    tkClass: result := typeCLASS;
    tkWChar: result := typeWIDECHAR;
    tkMethod: result := typeMETHOD;
    tkVariant: result := typeVARIANT;
    tkSet: result := typeSET;
    tkInt64: result := typeINT64;
    tkInterface: result := typeINTERFACE;
  end;
  if StrEql(pti^.Name, 'TDateTime') then
    result := typeDOUBLE;
end;

constructor TTitanMethodDefinition.Create(DefList: TTitanDefinitionList;
                                     const Name: String; Proc: TTitanMethodImpl;
                                     NP: Integer; Owner: TTitanDefinition;
                                     IsStatic: Boolean);
begin
  inherited Create(DefList, Name, Owner, dkMethod);
  Self.Proc := Proc;
  Self.NP := NP;
  Self.PClass := nil;
  Self.DirectProc := nil;
  Self.CallConv := _ccRegister;
  Self.Fake := false;
  Self.NewFake := false;
  Self.IsStatic := IsStatic;
  DefParamList := TTitanVarList.Create;
  ReturnsDynamicArray := false;
  Self.IsIntf := false;
  MethodIndex := 0;
  intf_pti := nil;
  FillChar(guid, SizeOf(guid), 0);
end;

destructor TTitanMethodDefinition.Destroy;
begin
  DefParamList.Free;
  inherited;
end;

function TTitanMethodDefinition.Duplicate: TTitanMethodDefinition;
var
  I: Integer;
begin
  result := TTitanMethodDefinition.Create(DefList, Name, Proc, NP, Owner, true);

  result.PClass := PClass;
  result.DirectProc := DirectProc;
  result.CallConv := CallConv;
  result.Header := Header;
  SetLength(result.Types, Length(Types));
  SetLength(result.ExtraTypes, Length(Types));
  SetLength(result.StrTypes, Length(Types));
  SetLength(result.ParamNames, Length(Types));
  SetLength(result.Sizes, Length(Types));
  SetLength(result.ByRefs, Length(Types));
  SetLength(result.Consts, Length(Types));
  result.TypeSub := TypeSub;
  result.Fake := Fake;
  result.NewFake := NewFake;
  result.IsStatic := IsStatic;

  for I:=0 to Length(Types) - 1 do
    result.Types[I] := Types[I];
  for I:=0 to Length(Types) - 1 do
    result.ExtraTypes[I] := ExtraTypes[I];
  for I:=0 to Length(Types) - 1 do
    result.StrTypes[I] := StrTypes[I];
  for I:=0 to Length(Types) - 1 do
    result.ParamNames[I] := ParamNames[I];
  for I:=0 to Length(Types) - 1 do
    result.Sizes[I] := Sizes[I];
  for I:=0 to Length(ByRefs) - 1 do
    result.ByRefs[I] := ByRefs[I];
  for I:=0 to Length(Consts) - 1 do
    result.Consts[I] := Consts[I];
end;


procedure TTitanMethodDefinition.AddToScripter(Scripter: Pointer);
var
  ClassRec: TTitanClassRec;
  ClassList: TTitanClassList;
begin
  ClassList := TTitanBaseScripter(Scripter).ClassList;

  if Owner = nil then
    ClassRec := ClassList[0]
  else
  begin
    ClassRec := TTitanClassDefinition(Owner).FindClassRec(Scripter);
  end;

  if (SignLoadOnDemand and (ClassRec = nil)) then
    ClassRec := TTitanBaseScripter(Scripter).ClassList.FindClassByName(Owner.Name);

  if ClassRec <> nil then
    ClassRec.AddHostMethod(Self)
  else
    raise TTitanScriptFailure.Create(Format(errClassIsNotImported, [Owner.Name]));
end;

function TTitanMethodDefinition.LoadFromDll(Scripter: Pointer; SubID: Integer): Boolean;

var
  BaseScripter: TTitanBaseScripter;

function GetFinalType(ID: Integer): Integer;
var
  TypeIndex: Integer;
  RTTIDef: TTitanRTTITypeDefinition;
  S: String;
begin
  with BaseScripter do
  begin
    result := SymbolTable.PType[ID];
    TypeIndex := SymbolTable.TypeNameIndex[ID];
    if TypeIndex > 0 then
    begin
      result := typeCLASS;
      S := _GetName(TypeIndex, Scripter);
      RTTIDef := GlobalDefinitionList.FindRTTITypeDefByName(S);
      if RTTIDef <> nil then
        result := RTTIDef.FinalType;
    end;
  end;
end;

var
  DllName, DllProcName: String;
  DllID, DllProcID: Integer;
  DllRec: TTitanDllRec;
  I, ParamID, ResultID: Integer;
begin
  BaseScripter := TTitanBaseScripter(Scripter);
  with BaseScripter.SymbolTable do
  begin
    DllID := GetDllID(SubID);
    DllName := Name[DllID];
    DllProcID := GetDllProcID(SubID);
    DllProcName := Name[DllProcID];

    DirectProc := nil;
    if Assigned(TTitanBaseScripter(Scripter).OnLoadDll) then
        TTitanBaseScripter(Scripter).OnLoadDll(TTitanBaseScripter(Scripter).Owner, DllName,
                                             DllProcName, DirectProc);
    if DirectProc = nil then
    begin
      DllRec := DllList.LoadDll(DllName, scripter);
      if DllRec = nil then
        raise TTitanScriptFailure.Create(Format(errCannotLoadDll, [DllName]));


      DirectProc := dynlibs.GetProcedureAddress(DllRec.Handle, PChar(DllProcName));


    end;

    if DirectProc = nil then
      raise TTitanScriptFailure.Create(Format(errDllFunctionIsNotFound,
                                            [DllProcName, DllName]));

    NP := Count[SubID];
    SetLength(Types, NP + 1);
    SetLength(ExtraTypes, NP + 1);
    SetLength(StrTypes, NP + 1);
    SetLength(ParamNames, NP + 1);
    SetLength(Sizes, NP + 1);
    SetLength(ByRefs, NP + 1);
    SetLength(Consts, NP + 1);
    for I:=1 to NP do
    begin
      ParamID := GetParamID(SubID, I);
      Types[I - 1] := GetFinalType(ParamID);
      ExtraTypes[I - 1] := 0;
      StrTypes[I - 1] := '';
      ParamNames[I - 1] := '';
      ByRefs[I - 1] := ByRef[ParamID] <> 0;
      Consts[I - 1] := false;
      Sizes[I - 1] := -1;
    end;
    ResultID := GetResultID(SubID);
    Types[NP] := GetFinalType(ResultID);
    ExtraTypes[NP] := 0;
    StrTypes[NP] := '';
    ParamNames[NP] := '';
    ByRefs[NP] := true;
    Consts[NP] := false;
    Sizes[NP] := -1;

    Self.CallConv := CallConv[SubID];

    ml := [modSTATIC];
    result := true;
  end;
end;

function TTitanMethodDefinition.GetScriptType(ParamNumber: Integer): Integer;
begin
  case Types[ParamNumber] of
    typeINTEGER, typeBYTE, typeSMALLINT, typeSHORTINT: result := typeINTEGER;
    typeSTRING, typeCHAR, typePCHAR, typeWIDESTRING: result := typeSTRING;
    typeBOOLEAN, typeWORDBOOL, typeLONGBOOL: result := typeBOOLEAN;
    typeDOUBLE, typeEXTENDED, typeSINGLE, typeCURRENCY: result := typeDOUBLE;
  else
    result := typeVARIANT;
  end;
end;

procedure TTitanMethodDefinition.Dump(var T: TextFile);
begin
  writeln(T, 'Method =', FullName, ' Index =', Index);
end;

constructor TTitanPropertyDefinition.Create(DefList: TTitanDefinitionList;
                                       const Name: String; Owner: TTitanDefinition;
                                       PClass: TClass; ReadDef, WriteDef: TTitanDefinition);
begin
  inherited Create(DefList, Name, Owner, dkMethod);
  Self.PClass := PClass;
  Self.ReadDef := TTitanMethodDefinition(ReadDef);
  Self.WriteDef := TTitanMethodDefinition(WriteDef);
  Self.DefKind := dkProperty;
  Self.PropDef := '';
  Self.PropTYpe := '';
  NP := 0;
end;

procedure TTitanPropertyDefinition.Dump(var T: TextFile);
begin
  writeln(T, 'Property =', FullName, ' Index =', Index);
end;

procedure TTitanPropertyDefinition.AddToScripter(Scripter: Pointer);
var
  ClassRec: TTitanClassRec;
  ClassList: TTitanClassList;
begin
  ClassList := TTitanBaseScripter(Scripter).ClassList;

  if Owner = nil then
    ClassRec := ClassList[0]
  else
    ClassRec := TTitanClassDefinition(Owner).FindClassRec(Scripter);

  if (SignLoadOnDemand and (ClassRec = nil)) then
    ClassRec := TTitanBaseScripter(Scripter).ClassList.FindClassByName(Owner.Name);

  if ClassRec <> nil then
    ClassRec.AddHostProperty(Self)
  else
    raise TTitanScriptFailure.Create(Format(errClassIsNotImported, [Owner.Name]));
end;

constructor TTitanVariableDefinition.Create(DefList: TTitanDefinitionList;
                                       const Name, TypeName: String;
                                       Address: Pointer;
                                       Owner: TTitanDefinition);
begin
  inherited Create(DefList, Name, Owner, dkMethod);
  Self.DefKind := dkVariable;
  Self.Address := Address;
  Self.TypeID := PAXTypes.IndexOf(UpperCase(TypeName));
  Self.TypeName := TypeName;
end;

function TTitanVariableDefinition.GetAddress: Pointer;
begin
  DefValue := GetVariantValue(nil, Address, TypeID);
  result := @DefValue;
end;

procedure TTitanVariableDefinition.Dump(var T: TextFile);
begin
  writeln(T, 'Variable =', FullName, ' Index =', Index);
end;

procedure TTitanVariableDefinition.AddToScripter(Scripter: Pointer);
var
  ClassRec: TTitanClassRec;
  ClassList: TTitanClassList;
begin
  ClassList := TTitanBaseScripter(Scripter).ClassList;

  if Owner = nil then
    ClassRec := ClassList[0]
  else
//  ClassRec := ClassList.FindClassByName(Owner.Name);
    ClassRec := TTitanClassDefinition(Owner).FindClassRec(Scripter);

  if (SignLoadOnDemand and (ClassRec = nil)) then
    ClassRec := TTitanBaseScripter(Scripter).ClassList.FindClassByName(Owner.Name);

  if ClassRec <> nil then
    ClassRec.AddHostVariable(Self)
  else
    raise TTitanScriptFailure.Create(Format(errClassIsNotImported, [Owner.Name]));
end;

function TTitanVariableDefinition.GetValue(Scripter: Pointer): Variant;
begin
  result := GetVariantValue(Scripter, Address, TypeID);
end;

procedure TTitanVariableDefinition.SetValue(Scripter: Pointer; const AValue: Variant);
begin
  if DefList.Global then
  begin
    _BeginWrite;
    try
      PutVariantValue(Scripter, Address, AValue, TypeID);
    finally
      _EndWrite;
    end;
  end
  else
    PutVariantValue(Scripter, Address, AValue, TypeID);
end;

constructor TTitanObjectDefinition.Create(DefList: TTitanDefinitionList;
                                        const Name: String;
                                        Instance: TObject;
                                        Owner: TTitanDefinition);
begin
  inherited Create(DefList, Name, Owner, dkMethod);
  Self.DefKind := dkObject;
  Self.Instance := Instance;
end;

procedure TTitanObjectDefinition.Dump(var T: TextFile);
begin
  writeln(T, 'Object =', FullName, ' Index =', Index);
end;

procedure TTitanObjectDefinition.AddToScripter(Scripter: Pointer);
var
  ClassRec: TTitanClassRec;
  ClassList: TTitanClassList;
begin
  ClassList := TTitanBaseScripter(Scripter).ClassList;

  if Owner = nil then
    ClassRec := ClassList[0]
  else
    ClassRec := TTitanClassDefinition(Owner).FindClassRec(Scripter);

  if (SignLoadOnDemand and (ClassRec = nil)) then
    ClassRec := TTitanBaseScripter(Scripter).ClassList.FindClassByName(Owner.Name);

  if ClassRec <> nil then
    ClassRec.AddHostObject(Self)
  else
    raise TTitanScriptFailure.Create(Format(errClassIsNotImported, [Owner.Name]));
end;

constructor TTitanVirtualObjectDefinition.Create(DefList: TTitanDefinitionList;
                                               const Name: String;
                                               Owner: TTitanDefinition);
begin
  inherited Create(DefList, Name, Owner, dkMethod);
  Self.DefKind := dkVirtualObject;
end;

procedure TTitanVirtualObjectDefinition.Dump(var T: TextFile);
begin
  writeln(T, 'Virtual Object =', FullName, ' Index =', Index);
end;

procedure TTitanVirtualObjectDefinition.AddToScripter(Scripter: Pointer);
var
  ClassRec: TTitanClassRec;
  ClassList: TTitanClassList;
begin
  ClassList := TTitanBaseScripter(Scripter).ClassList;

  if Owner = nil then
    ClassRec := ClassList[0]
  else
    ClassRec := TTitanClassDefinition(Owner).FindClassRec(Scripter);

  if (SignLoadOnDemand and (ClassRec = nil)) then
    ClassRec := TTitanBaseScripter(Scripter).ClassList.FindClassByName(Owner.Name);

  if ClassRec <> nil then
    ClassRec.AddVirtualObject(Self)
  else
    raise TTitanScriptFailure.Create(Format(errClassIsNotImported, [Owner.Name]));
end;

constructor TTitanInterfaceVarDefinition.Create(DefList: TTitanDefinitionList;
                                           const Name: String;
                                           PIntf: PUnknown;
                                           const guid: TGuid;
                                           Owner: TTitanDefinition);
begin
  inherited Create(DefList, Name, Owner, dkVariable);
  Self.DefKind := dkInterface;
  Self.PIntf := PIntf;
  Self.guid := guid;
end;

procedure TTitanInterfaceVarDefinition.Dump(var T: TextFile);
begin
  writeln(T, 'InterfaceVar =', FullName, ' Index =', Index);
end;

procedure TTitanInterfaceVarDefinition.AddToScripter(Scripter: Pointer);
var
  ClassRec: TTitanClassRec;
  ClassList: TTitanClassList;
begin
  ClassList := TTitanBaseScripter(Scripter).ClassList;

  if Owner = nil then
    ClassRec := ClassList[0]
  else
    ClassRec := TTitanClassDefinition(Owner).FindClassRec(Scripter);

  if (SignLoadOnDemand and (ClassRec = nil)) then
    ClassRec := TTitanBaseScripter(Scripter).ClassList.FindClassByName(Owner.Name);

  if ClassRec <> nil then
    ClassRec.AddHostInterfaceVar(Self)
  else
    raise TTitanScriptFailure.Create(Format(errClassIsNotImported, [Owner.Name]));
end;

constructor TTitanConstantDefinition.Create(DefList: TTitanDefinitionList;
                                          const Name: String;
                                          const Value: Variant;
                                          Owner: TTitanDefinition);
begin
  inherited Create(DefList, Name, Owner, dkMethod);
  Self.DefKind := dkConstant;
  Self.DefValue := Value;
  case VarType(Value) of
    varInteger: Self.TypeID := typeINTEGER;
    varByte: Self.TypeID := typeBYTE;
    varDouble: Self.TypeID := typeDOUBLE;
    varSingle: Self.TypeID := typeDOUBLE;
    varCurrency: Self.TypeID := typeDOUBLE;
    varBoolean: Self.TypeID := typeBOOLEAN;
    varString: Self.TypeID := typeSTRING;
  else
    Self.TypeID := 0;
  end;
  if TypeID > 0 then
    resultType := PaxTypes[TypeID];
end;

procedure TTitanConstantDefinition.Dump(var T: TextFile);
var
  S: String;
begin
  if Owner <> nil then
    S := Owner.Name
  else
    S := '';

  if VarType(DefValue) in [varEmpty, varNull] then
    writeln(T, 'Constant =', FullName, ' Index =', Index, ' Value = Undefined')
  else
    writeln(T, 'Constant =', FullName, ' Index =', Index, ' Value =', DefValue, ' Owner =', S);
end;

function TTitanConstantDefinition.GetAddress: Pointer;
begin
  result := @ DefValue;
end;

procedure TTitanConstantDefinition.AddToScripter(Scripter: Pointer);
var
  ClassRec: TTitanClassRec;
  ClassList: TTitanClassList;
begin
  ClassList := TTitanBaseScripter(Scripter).ClassList;

  if Owner = nil then
    ClassRec := ClassList[0]
  else
    ClassRec := TTitanClassDefinition(Owner).FindClassRec(Scripter);
//    ClassRec := ClassList.FindClassByName(Owner.Name);

  if (SignLoadOnDemand and (ClassRec = nil)) then
    ClassRec := TTitanBaseScripter(Scripter).ClassList.FindClassByName(Owner.Name);

  if ClassRec <> nil then
    ClassRec.AddHostConstant(Self)
  else
    raise TTitanScriptFailure.Create(Format(errClassIsNotImported, [Owner.Name]));
end;

constructor TTitanDefinitionList.Create(Global: Boolean);

  //-- jgv
  function doCreateList: TStringList;
  begin
    Result := TStringList.Create;
    With Result do begin
      Sorted := True;
{$IFDEF VARIANTS}
      CaseSensitive := False;
{$ENDIF}
    end;
  end;

begin
  inherited Create;
  _LastIndex := -1;
  ClassIdx := TTitanIds.Create(false);
  ConstIdx := TTitanIds.Create(false);
  self.Global := Global;

  //-- jgv
  MethodIdx := DoCreateList;
  FieldIdx := DoCreateList;
  RTTITypeDefIdx := DoCreateList;
  ConstantIdx := DoCreateList;
  GUIDIdx := DoCreateList;

  VarHParser:=TPascalParser.Create;
  fListOfClasses:=THClassesList.Create;
  fListOfRTTIs:=THGenList.Create;
end;

destructor TTitanDefinitionList.Destroy;
begin
  //VarHParser.free;

  fListOfClasses.Clear;
  fListOfRTTIs.Clear;

  fListOfClasses.Free;
  fListOfRTTIs.Free;


  MethodIdx.Free; // jgv
  FieldIdx.Free; // jgv
  RTTITypeDefIdx.Free; // jgv
  ConstantIdx.Free;
  GUIDIdx.Free;

  ClassIdx.Free;
  ConstIdx.Free;
  inherited;
end;

function TTitanDefinitionList.GetRecord(Index: Integer): TTitanDefinition;
begin
  result := TTitanDefinition(Items[Index]);
end;

function TTitanDefinitionList.FindMethod(const ClassName, MethodName: String): TTitanMethodDefinition;
var
  I: Integer;
  D: TTitanDefinition;
begin
  I := MethodIdx.IndexOf (ClassName + '.' + MethodName);
  If I >= 0 then
    Result := MethodIdx.Objects [i] as TTitanMethodDefinition
  else
    Result := Nil;

  Exit;

  for I:=Count - 1 downto 0 do
  begin
    D := Records[I];
    if D <> nil then
    begin
      if D.DefKind = dkMethod then
      if StrEql(D.Name, MethodName) then
      begin
        if D.Owner <> nil then
          if StrEql(D.Owner.Name, ClassName) then
          begin
            result := TTitanMethodDefinition(D);
            Exit;
          end;
      end;
    end;
  end;
  result := nil;
end;

function TTitanDefinitionList.FindProperty(const ClassName,PropertyName:String):TTitanPropertyDefinition;
 var
  I:Integer;
  D:TTitanDefinition;
begin
  for I:=Count - 1 downto 0 do
  begin
    D := Records[I];
    if D <> nil then
    begin
      if D.DefKind =dkProperty then
      if StrEql(D.Name,PropertyName) then
      begin
        if D.Owner <> nil then
          if StrEql(D.Owner.Name,ClassName) then
          begin
            result := TTitanPropertyDefinition(D);
            Exit;
          end;
      end;
    end;
  end;
  result := nil;
end;

function TTitanDefinitionList.FindField(const ClassName, FieldName: String): TTitanFieldDefinition;
var
  I: Integer;
begin
  i := FieldIdx.IndexOf (ClassName + '.' + FieldName);
  If i >= 0 then
    Result := FieldIdx.Objects [i] as TTitanFieldDefinition
  else
    Result := Nil;
  Exit;

  for I:=0 to Count - 1 do
  begin
    result := TTitanFieldDefinition(Records[I]);
    if Result <> nil then
    begin
      if result.DefKind = dkField then
        if StrEql(result.Name, FieldName) then
          Exit;
    end;
  end;
  result := nil;
end;


Function  TTitanDefinitionList.FindRecordField(ClassDef:TTitanClassDefinition;Const FieldName:string):TTitanRecordFieldDefinition;
var I:Integer;
    pd:TTitanRecordFieldDefinition;
begin
  result:=nil;

  if ClassDef=nil then exit;
  if FieldName='' then exit;

  for I:=ClassDef.Index+1 to Count - 1 do
   begin
    pd := TTitanRecordFieldDefinition(Records[I]);

    if pd <> nil then
      if pd.DefKind=dkRecordField then
        if pd.Owner=ClassDef then
         if SameText(pd.Name,FieldName) then
           begin
            result:=pd;
            exit;
           end;
   end;
end;

Procedure TTitanDefinitionList.FindRecordFields(aOutlist:TStringList;ClassDef:TTitanClassDefinition);
var I:Integer;
    pd:TTitanRecordFieldDefinition;
begin
  if aOutlist=nil then exit;
  aOutlist.Clear;
  for I:=0 to Count - 1 do
   begin
    pd := TTitanRecordFieldDefinition(Records[I]);
    if pd <> nil then
      if pd.DefKind=dkRecordField then
        if pd.Owner=ClassDef then aOutlist.AddObject(pd.Name,TTitanRecordFieldDefinition(pd));

   end;

end;


function TTitanDefinitionList.FindClassDef(PClass: TClass): TTitanClassDefinition;
var
  I: Integer;
  D: TTitanDefinition;
begin
  result := nil;

  for I:=ClassIdx.Count - 1 downto 0 do
  begin
    D := Records[ClassIdx[I]];
    if D.DefKind = dkClass then
      if TTitanClassDefinition(D).PClass = PClass then
      begin
        result := TTitanClassDefinition(D);
        Break;
      end;
  end;
end;

function TTitanDefinitionList.FindInterfaceTypeDef(const Guid: TGuid): TTitanClassDefinition;
var
  I: Integer;
  D: TTitanDefinition;
  DC: TTitanClassDefinition;
begin
{$IFDEF VARIANTS}
  i := GUIDIdx.IndexOf (GUIDToString (GUID));
  If i >= 0 then
    Result := GUIDIdx.Objects [i] as TTitanClassDefinition
  else
    Result := Nil;
  Exit;

{$ENDIF}

  result := nil;

  for I:=ClassIdx.Count - 1 downto 0 do
  begin
    D := Records[ClassIdx[I]];
    if D.DefKind = dkClass then
    begin
      DC := TTitanClassDefinition(D);
      if DC.ClassKind = ckInterface then
        if IsEqualGuid(DC.Guid, Guid) then
        begin
          result := DC;
          Break;
        end;
    end;
  end;
end;

function TTitanDefinitionList.FindObjectDef(Instance: TObject; Owner: TTitanDefinition): TTitanObjectDefinition;
var
  I: Integer;
  D: TTitanDefinition;
begin
  for I:=0 to Count - 1 do
  begin
    D := Records[I];
    if D.DefKind = dkObject then
    if D.Owner = Owner then
    begin
      if (D as TTitanObjectDefinition).Instance = Instance then
        begin
          result := D as TTitanObjectDefinition;
          Exit;
        end;
    end;
  end;
  result := nil;
end;

function TTitanDefinitionList.FindClassDefByPTI(pti: PTypeInfo): TTitanClassDefinition;
var
  I: Integer;
  D: TTitanDefinition;
begin
  result := nil;

  for I:=ClassIdx.Count - 1 downto 0 do
  begin
    D := Records[ClassIdx[I]];
    if D.DefKind = dkClass then
      if TTitanClassDefinition(D).pti = pti then
      begin
        result := TTitanClassDefinition(D);
        Break;
      end;
  end;
end;

function TTitanDefinitionList.FindClassDefByName(const Name: String): TTitanClassDefinition;
var
  I: Integer;
  D: TTitanDefinition;
begin
  result := nil;

  for I:=ClassIdx.Count - 1 downto 0 do
  begin
    D := Records[ClassIdx[I]];
    if D.DefKind = dkClass then
      if StrEql(D.Name, Name) then
      begin
        result := TTitanClassDefinition(D);
        Break;
      end;
  end;
end;

function TTitanDefinitionList.FindRTTITypeDef(pti: PTypeInfo): TTitanRTTITypeDefinition;
var
  I: Integer;
  D: TTitanDefinition;
begin
  i := RTTITypeDefIdx.IndexOf (pti^.Name);
  If i >= 0 then
    Result := RTTITypeDefIdx.Objects [i] as TTitanRTTITypeDefinition
  else
    Result := Nil;
  Exit;


  result := nil;

  for I:=ClassIdx.Count - 1 downto 0 do
  begin
    D := Records[ClassIdx[I]];
    if D.DefKind = dkRTTIType then
      if TTitanRTTITypeDefinition(D).pti = pti then
      begin
        result := TTitanRTTITypeDefinition(D);
        Break;
      end;
  end;
end;

function TTitanDefinitionList.FindRTTITypeDefByName(const Name: String): TTitanRTTITypeDefinition;
var
  I: Integer;
  D: TTitanDefinition;
begin
  i := RTTITypeDefIdx.IndexOf(Name);
  If i >= 0 then
    Result := RTTITypeDefIdx.Objects [i] as TTitanRTTITypeDefinition
  else
    Result := Nil;
  Exit;

  result := nil;

  for I:=ClassIdx.Count - 1 downto 0 do
  begin
    D := Records[ClassIdx[I]];
    if D.DefKind = dkRTTIType then
      if StrEql(TTitanRTTITypeDefinition(D).pti^.Name, Name) then
      begin
        result := TTitanRTTITypeDefinition(D);
        Break;
      end;
  end;
end;

function TTitanDefinitionList.AddNamespace(const Name: String;
                                         Owner: TTitanClassDefinition): TTitanClassDefinition;
begin
  result := TTitanClassDefinition(Lookup(Name, dkClass, Owner));
  if result <> nil then
    Exit;

  result := TTitanClassDefinition.Create(Self, Name, Owner, nil);
  TTitanClassDefinition(result).PClass := nil;
  Add(result);
  result.ml := result.ml + [modSTATIC];
  result.Index := NewIndex;

  ClassIdx.Add(Count - 1);
end;

function TTitanDefinitionList.AddClass1(const Name: String;
                                  Owner, Ancestor: TTitanClassDefinition;
                                  ReadProp, WriteProp: TTitanMethodImpl): TTitanClassDefinition;
begin
  result := FindClassDefByName(Name);

  if result = nil then
  begin
    result := TTitanClassDefinition.Create(Self, Name, Owner, Ancestor);
    Add(result);
    result.Index := NewIndex;

    ClassIdx.Add(Count - 1);
  end;

  TTitanClassDefinition(result).PClass := nil;

  if Assigned(ReadProp) then
    TTitanClassDefinition(result).GetPropDef :=
      TTitanMethodDefinition(AddMethod3('', ReadProp, -1, result, true));

  if Assigned(WriteProp) then
    TTitanClassDefinition(result).PutPropDef :=
      TTitanMethodDefinition(AddMethod3('', WriteProp, -1, result, true));
end;

function TTitanDefinitionList.AddClass2(PClass: TClass;
                                      Owner: TTitanClassDefinition;
                                      ReadProp, WriteProp: TTitanMethodImpl;
                                      ClassKind: TTitanClassKind = ckClass): TTitanClassDefinition;
var
  D: TTitanDefinition;
  ParentClass: TClass;
begin
  result := FindClassDef(PClass);
  if result <> nil then
  begin
    if Assigned(ReadProp) then
      TTitanClassDefinition(result).GetPropDef :=
        TTitanMethodDefinition(AddMethod3('', ReadProp, -1, result, true));

    if Assigned(WriteProp) then
      TTitanClassDefinition(result).PutPropDef :=
        TTitanMethodDefinition(AddMethod3('', WriteProp, -1, result, true));
    Exit;
  end;

  ParentClass := PClass.ClassParent;
  if ParentClass = nil then
    D := nil
  else
  begin
    D := FindClassDef(ParentClass);
    if D = nil then
      D := AddClass2(ParentClass, nil, nil, nil);
  end;

  result := AddClass1(PClass.ClassName, Owner, TTitanClassDefinition(D), ReadProp, WriteProp);
  result.PClass := PClass;
  result.ClassKind := ClassKind;
end;

function TTitanDefinitionList.AddMethod1(const Header: String; Address: Pointer;
                                       Owner: TTitanDefinition;
                                       Static: Boolean = false): TTitanMethodDefinition;
var
  P: TPascalParser;
  D: TTitanMethodDefinition;
begin
  result := TTitanMethodDefinition.Create(Self, '', nil, 0, Owner, Static);

  if Pos('__', Header) > 0 then
    result.NewFake := true;

  result.Header := Header;
  P := TPascalParser.Create;
  P.D := result;
  P.Parse_Heading;
  P.Free;

  if CheckDup then begin
    if Pos('OVERLOAD;', UpperCase(Header)) = 0 then begin
      //jgv
      If Owner <> Nil then
        D := FindMethod (Owner.Name, Result.Name)
      else
        D := TTitanMethodDefinition(Lookup(result.Name, dkMethod, Owner));
      if D <> nil then begin
        result.Free;
        result := D;
        result.DirectProc := Address;
        Exit;
      end;
    end;
  end;

  result.DirectProc := Address;
  if Static then
    result.ml := result.ml + [modSTATIC];
  Add(result);
  result.Index := NewIndex;
  result.Owner := Owner;

  If Owner <> Nil then
    MethodIdx.AddObject (Owner.Name + '.' + Result.Name, Result);
end;

function TTitanDefinitionList.AddMethod2(PClass: TClass; const Header: String; Address: Pointer;
              Static: Boolean = false): TTitanMethodDefinition;
begin
  result := AddMethod1(Header, Address, FindClassDef(PClass), false);
  result.PClass := PClass;
end;

function TTitanDefinitionList.AddMethod3(const Name: String;
                                   Proc: TTitanMethodImpl;
                                   NP: Integer;
                                   Owner: TTitanClassDefinition;
                                   Static: Boolean = false): TTitanMethodDefinition;
begin
  if CheckDup then
  begin
    //-- jgv
    If Owner <> Nil then
      Result := FindMethod (Owner.Name, Name)
    else
      result := TTitanMethodDefinition(Lookup(Name, dkMethod, Owner));
    if result <> nil then
    begin
      result.Proc := Proc;
      Exit;
    end;
  end;

  result := TTitanMethodDefinition.Create(Self, Name, Proc, NP, Owner, Static);
  if Static then
    result.ml := result.ml + [modSTATIC];
  Add(result);
  result.Index := NewIndex;

  If Owner <> Nil then
    MethodIdx.AddObject (Owner.Name + '.' + Name, Result);
end;

function TTitanDefinitionList.AddMethod4(PClass: TClass;
                                       const Name: String;
                                       Proc: TTitanMethodImpl;
                                       NP: Integer;
                                       Static: Boolean = false): TTitanMethodDefinition;
begin
  result := AddMethod3(Name, Proc, NP, FindClassDef(PClass), Static);
  result.PClass := PClass;
  if Static then
    result.ml := result.ml + [modSTATIC];
end;

function TTitanDefinitionList.AddMethod5(pti: PTypeInfo; const Header: String): TTitanMethodDefinition;
function CreateMethodIndex: Integer;
var
  D: TTitanMethodDefinition;
  I: Integer;
  curr_pti: PTypeInfo;
  TypeData: PTypeData;
begin
  result := 0;
  curr_pti := pti;
  TypeData := nil;

  repeat

    for I:= 0 to Count - 1 do
      if Records[I].DefKind = dkMethod then
      begin
        D := TTitanMethodDefinition(Records[I]);
        if D.IsIntf then
          if D.intf_pti = curr_pti then
            Inc(result);
      end;

    if (TypeData.IntfParent <> nil) then
    begin
  {$ifndef fpc}
     curr_pti := TypeData.IntfParent^;
  {$else}
     curr_pti := TypeData.IntfParent;
  {$endif}
      if curr_pti = nil then
        break;
    end
    else
      Break;

  until false;
end;

var
  RttiDef: TTitanRTTITypeDefinition;
begin
  RttiDef := FindRTTITypeDef(pti);
  result := AddMethod1(Header, nil, RttiDef, false);
  result.PClass := nil;
  result.IsIntf := true;
  result.intf_pti := pti;
  result.MethodIndex := CreateMethodIndex;

{$ifndef fp}
  if pti <> nil then
    result.Guid := GetTypeData(pti).Guid;
 {$endif}
end;

function GUIDToString(const GUID: TGUID): string;
begin
  SetLength(Result, 38);
  StrLFmt(PChar(Result),
38,'{%.8x-%.4x-%.4x-%.2x%.2x-%.2x%.2x%.2x%.2x%.2x%.2x}',
    [GUID.D1, GUID.D2, GUID.D3, GUID.D4[0], GUID.D4[1], GUID.D4[2],
GUID.D4[3],
    GUID.D4[4], GUID.D4[5], GUID.D4[6], GUID.D4[7]]);
end;

function TTitanDefinitionList.AddMethod6(const Guid: TGuid; const Header: String;
                                       InitMethodIndex: Integer = -1): TTitanMethodDefinition;

var
  ClassDef: TTitanClassDefinition;

function CreateMethodIndex: Integer;
var
  D: TTitanMethodDefinition;
  I: Integer;
  curr_class_def: TTitanClassDefinition;
begin
  result := 0;
  curr_class_def := ClassDef;
  if curr_class_def = nil then
    raise TTitanScriptFailure.Create(Format(errInterfaceNotImported, [GuidToString(Guid)]));

  repeat

    for I:= curr_class_def.Index + 1 to Count - 1 do // from the beginning of interface class definition
      if Records[I].DefKind = dkMethod then
      begin
        D := TTitanMethodDefinition(Records[I]);
        if D.IsIntf then
          if IsEqualGuid(D.Guid, curr_class_def.guid) then
            Inc(result);
        if D.DefKind = dkClass then // we have found another class definition
          break;
      end;

    if curr_class_def.Ancestor <> nil then
      curr_class_def := TTitanClassDefinition(curr_class_def.Ancestor)
    else
      Break;

  until false;
end;

begin
  ClassDef := FindInterfaceTypeDef(Guid);
  result := AddMethod1(Header, nil, ClassDef, false);
  result.PClass := nil;
  result.IsIntf := true;
  result.intf_pti := nil;
  result.guid := Guid;

  if InitMethodIndex = -1 then
    result.MethodIndex := CreateMethodIndex
  else
    result.MethodIndex := InitMethodIndex;
end;

function TTitanDefinitionList.AddFakeMethod(PClass: TClass; const Header: String; Address: Pointer;
              Static: Boolean = false): TTitanMethodDefinition;
var
  C_Duplicate, Basic_Duplicate: TTitanMethodDefinition;
begin
  result := AddMethod1(Header, Address, FindClassDef(PClass), false);
  result.PClass := PClass;
  result.Fake := true;

  if result.TypeSub = tsConstructor then
  begin
    C_Duplicate := result.Duplicate;
    C_Duplicate.Name := result.PClass.ClassName;
    Add(C_Duplicate);
    C_Duplicate.Index := NewIndex;

    Basic_Duplicate := result.Duplicate;
    Basic_Duplicate.Name := 'New';
    Add(Basic_Duplicate);
    Basic_Duplicate.Index := NewIndex;
  end;
end;

function TTitanDefinitionList.AddProperty1(PClass: TClass;
                                        const Name: String;
                                        ProcRead: TTitanMethodImpl;
                                        NPRead: Integer;
                                        ProcWrite: TTitanMethodImpl;
                                        NPWrite: Integer;
                                        Default: Boolean = false): TTitanPropertyDefinition;
var
  Owner: TTitanClassDefinition;
  ReadDef, WriteDef: TTitanMethodDefinition;
begin
  Owner := FindClassDef(PClass);

  if CheckDup then
  begin
    result := TTitanPropertyDefinition(Lookup(Name, dkProperty, Owner));
    if result <> nil then
      Exit;
  end;

  if Assigned(ProcRead) then
    ReadDef := AddMethod3('', ProcRead, NPRead, Owner, false)
  else
    ReadDef := nil;

  if Assigned(ProcWrite) then
    WriteDef := AddMethod3('', ProcWrite, NPWrite, Owner, false)
  else
    WriteDef := nil;

  result := TTitanPropertyDefinition.Create(Self, Name, Owner, PClass, ReadDef, WriteDef);
  if Default then
    result.ml := result.ml + [modDEFAULT];
  Add(result);
  result.Index := NewIndex;
end;

function TTitanDefinitionList.AddProperty2(PClass: TClass; const Name: String;
                                     ReadDef, WriteDef: TTitanMethodDefinition;
                                     Default: Boolean = false): TTitanPropertyDefinition;
var
  Owner: TTitanClassDefinition;
begin
  Owner := FindClassDef(PClass);

  if CheckDup then
  begin
    result := TTitanPropertyDefinition(Lookup(Name, dkProperty, Owner));
    if result <> nil then
      Exit;
  end;

  result := TTitanPropertyDefinition.Create(Self, Name, Owner, PClass, ReadDef, WriteDef);
  if Default then
    result.ml := result.ml + [modDEFAULT];
  Add(result);
  result.Index := NewIndex;
end;

function TTitanDefinitionList.AddProperty3(PClass: TClass; const PropDef: String): TTitanPropertyDefinition;
var
  P: TPascalParser;
  ReadName, WriteName: String;
  Def: Boolean;
  Owner: TTitanClassDefinition;
begin
  Owner := FindClassDef(PClass);

  result := TTitanPropertyDefinition.Create(Self, '', Owner, PClass, nil, nil);
  result.PropDef := PropDef;

  P := TPascalParser.Create;
  P.Scanner.SourceCode := PropDef;
  P.Call_SCANNER;
  result.Name := P.Parse_Property(ReadName, WriteName, Def, Self);
  result.ReadDef := FindMethod(PClass.ClassName, ReadName);
  if result.ReadDef = nil then
    result.ReadDef := FindField(PClass.ClassName, ReadName);
  result.WriteDef := FindMethod(PClass.ClassName, WriteName);
  if result.WriteDef = nil then
    result.WriteDef := FindField(PClass.ClassName, WriteName);
  result.NP := P.NP;
  result.PropType := P.ResultType;
  P.Free;

  if result.ReadDef <> nil then
    if result.ReadDef.ResultType <> '' then
{
      if result.ReadDef.ResultType[1] in ['i','I'] then
      begin
        if not StrEql(result.ReadDef.ResultType, 'Integer') then
        begin
          result.Free;
          Exit;
        end;
      end
      else } if StrEql(result.ReadDef.ResultType, 'TWndMethod') then
      begin
        result.Free;
        Exit;
      end;

  if Def then
    result.ml := result.ml + [modDEFAULT];

  Add(result);
  result.Index := NewIndex;
end;



//========================================================================

function TTitanDefinitionList.AddClassSX(PClass:TClass):TTitanClassDefinition;
var
  D:TTitanDefinition;
  ParentClass:TClass;
  inx:Integer;
begin

  ParentClass := PClass.ClassParent;
  if ParentClass=nil then
    D := nil
  else
  begin
    D := FindClassDef(ParentClass);
    if D=nil then  D:=AddClassSX(ParentClass);
  end;

  result := AddClass1(PClass.ClassName,nil,TTitanClassDefinition(D),nil,nil);
  result.PClass := PClass;
  result.ClassKind := ckClass;

   //...........................................................
  fListOfClasses.AddClass(result.pClass,result.Index);
end;

function TTitanDefinitionList.AddClass2SX(const Name:String; Owner,Ancestor:TTitanClassDefinition; RProp,WProp:TTitanMethodImpl):TTitanClassDefinition;
begin
   result := TTitanClassDefinition.Create(Self,Name,Owner,Ancestor);
   Add(result);
   result.Index := NewIndex;
   ClassIdx.Add(Count - 1);
   TTitanClassDefinition(result).PClass := nil;

  if Assigned(RProp) then
    TTitanClassDefinition(result).GetPropDef :=
      TTitanMethodDefinition(AddMethod3('',RProp,-1,result,true));

  if Assigned(WProp) then
    TTitanClassDefinition(result).PutPropDef :=
      TTitanMethodDefinition(AddMethod3('',WProp,-1,result,true));

end;


function TTitanDefinitionList.AddClassSXFromType(const Name:String; Owner,Ancestor:TTitanClassDefinition):TTitanClassDefinition;
begin
 result := TTitanClassDefinition.Create(Self,Name,Owner,Ancestor);
 Add(result);
 result.Index := NewIndex;
 ClassIdx.Add(Count - 1);

 TTitanClassDefinition(result).PClass := nil;
end;


function TTitanDefinitionList.AddRoutineSX(const Header:String; Address:Pointer):TTitanMethodDefinition;
  var D:TTitanMethodDefinition;
begin
  result := TTitanMethodDefinition.Create(Self,'',nil,0,nil,true);

  result.Header := Header;

  VarHParser.D := result;
  VarHParser.Parse_Heading_ForRoutine;

  if CheckDup then
   begin
    if Pos('OVERLOAD',UpperCase(Header))=0 then
    begin
      D := TTitanMethodDefinition(Lookup(result.Name,dkMethod,nil));
      if D <> nil then
      begin
        result.Free;
        result := D;
        result.DirectProc := Address;
        Exit;
      end;
    end;
  end;

  result.DirectProc := Address;
  result.ml := result.ml + [modSTATIC];
  Add(result);
  result.Index := NewIndex;
  result.Owner :=nil;
end;

function TTitanDefinitionList.AddMethodSX(ClassID:integer; const Header:String; Address:Pointer):TTitanMethodDefinition;
 var D:TTitanMethodDefinition;
begin

  result := TTitanMethodDefinition.Create(Self,'',nil,0,Records[ClassID],false);
  result.Header := Header;

  VarHParser.D := result;
  VarHParser.Parse_Heading;

  if CheckDup then
   begin
    if Pos('OVERLOAD',UpperCase(Header))=0 then
    begin
      D := TTitanMethodDefinition(Lookup(result.Name,dkMethod,Records[ClassID]));
      if D <> nil then
      begin
        result.Free;
        result := D;
        result.DirectProc := Address;
        Exit;
      end;
    end;
  end;

  result.DirectProc := Address;
  Add(result);
  result.Index := NewIndex;
  result.Owner := Records[ClassID];
  result.PClass := TTitanClassDefinition(Records[ClassID]).PClass;
end;


function TTitanDefinitionList.AddFakeMethodSX(ClassID:integer; const Header:String; Address:Pointer):TTitanMethodDefinition;
  var  C_Duplicate,Basic_Duplicate:TTitanMethodDefinition;
begin
  result := TTitanMethodDefinition.Create(Self,'',nil,0,Records[ClassID],false);
  result.Header := Header;

  VarHParser.D := result;
  VarHParser.Parse_Heading_ForFakeMethod;

  result.DirectProc := Address;
  Add(result);
  result.Index := NewIndex;
  result.Owner := Records[ClassID];
  result.PClass := TTitanClassDefinition(Records[ClassID]).PClass; 
  result.Fake := true;

 if result.TypeSub=tsConstructor then
  begin
    C_Duplicate := result.Duplicate;
    C_Duplicate.Name := result.PClass.ClassName;
    Add(C_Duplicate);
    C_Duplicate.Index := NewIndex;

    Basic_Duplicate := result.Duplicate;
    Basic_Duplicate.Name := 'New';
    Add(Basic_Duplicate);
    Basic_Duplicate.Index := NewIndex;
  end;
end;

function TTitanDefinitionList.AddProperty3SXi(ClassID:Integer; GR,GW:TTitanMethodDefinition; const PropDef:String):TTitanPropertyDefinition;
var
  ReadName,WriteName:String;
  Def:Boolean;
  Owner:TTitanClassDefinition;
begin
  Owner := TTitanClassDefinition(Records[ClassID]);

  result := TTitanPropertyDefinition.Create(Self,'',Owner,Owner.PClass,nil,nil);
  result.PropDef := PropDef;

  VarHParser.Scanner.SourceCode := PropDef;
  VarHParser.Call_SCANNER;
  result.Name := VarHParser.Parse_PropertySXi(Def,Self);
  result.ReadDef :=GR;
  result.WriteDef := GW;
  result.NP := VarHParser.NP;
  result.PropType := VarHParser.ResultType;


  if result.ReadDef <> nil then
    if result.ReadDef.ResultType <> '' then
     if StrEql(result.ReadDef.ResultType,'TWndMethod') then
      begin
        result.Free;
        Exit;
      end;

  if Def then  result.ml := result.ml + [modDEFAULT];

  Add(result);
  result.Index := NewIndex;
end;

function TTitanDefinitionList.AddProperty3SX(ClassID:Integer; GR,GW:TTitanMethodDefinition; const aName,aResType:String):TTitanPropertyDefinition;
var Owner:TTitanClassDefinition;
begin
  Owner := TTitanClassDefinition(Records[ClassID]);
  result := TTitanPropertyDefinition.Create(Self,'',Owner,Owner.PClass,nil,nil);
  result.PropDef := 'Property '+aName+':'+aResType+';';
  result.Name := aName;
  result.ReadDef :=GR;
  result.WriteDef := GW;
  result.NP :=0;
  result.PropType :=aResType;

  if result.ReadDef <> nil then
    if result.ReadDef.ResultType <> '' then
     if StrEql(result.ReadDef.ResultType,'TWndMethod') then
      begin
        result.Free;
        Exit;
      end;

  Add(result);
  result.Index := NewIndex;
end;


function TTitanDefinitionList.AddVariableSX(const Name,TypeName:String; Address:Pointer):TTitanVariableDefinition;
var TypeDef:TTitanRTTITypeDefinition;
begin

    result := TTitanVariableDefinition.Create(Self,Name,TypeName,Address,nil);

    TypeDef := GlobalDefinitionList.FindRTTITypeDefByName(TypeName);
    if TypeDef <> nil then
      if TypeDef.pti^.Kind=tkSet then
        result.TypeID := typeSET;

    Add(result);
    result.Index := NewIndex;
    result.ml := result.ml + [modSTATIC];
end;


//=====================================================================




function TTitanDefinitionList.AddInterfaceProperty(const guid: TGUID; const PropDef: String): TTitanPropertyDefinition;
var
  P: TPascalParser;
  ReadName, WriteName: String;
  Def: Boolean;
  Owner: TTitanClassDefinition;
begin
  Owner := FindInterfaceTypeDef(guid);

  result := TTitanPropertyDefinition.Create(Self, '', Owner, nil, nil, nil);
  result.PropDef := PropDef;

  P := TPascalParser.Create;
  P.Scanner.SourceCode := PropDef;
  P.Call_SCANNER;
  result.Name := P.Parse_Property(ReadName, WriteName, Def, Self);
  result.ReadDef := FindMethod(Owner.Name, ReadName);
  if result.ReadDef = nil then
    result.ReadDef := FindField(Owner.Name, ReadName);
  result.WriteDef := FindMethod(Owner.Name, WriteName);
  if result.WriteDef = nil then
    result.WriteDef := FindField(Owner.Name, WriteName);
  result.NP := P.NP;
  result.PropType := P.ResultType;
  P.Free;

  if result.ReadDef <> nil then
    if result.ReadDef.ResultType <> '' then
{
      if result.ReadDef.ResultType[1] in ['i','I'] then
      begin
        if not StrEql(result.ReadDef.ResultType, 'Integer') then
        begin
          result.Free;
          Exit;
        end;
      end
      else } if StrEql(result.ReadDef.ResultType, 'TWndMethod') then
      begin
        result.Free;
        Exit;
      end;

  if Def then
    result.ml := result.ml + [modDEFAULT];

  Add(result);
  result.Index := NewIndex;
end;

function TTitanDefinitionList.AddVariable(const Name, TypeName: String;
                                        Address: Pointer;
                                        Owner: TTitanClassDefinition;
                                        Scripter: Pointer = nil): TTitanVariableDefinition;
var
  I: Integer;
  TypeDef: TTitanRTTITypeDefinition;
begin
  result := TTitanVariableDefinition(Lookup(Name, dkVariable, Owner, Scripter));

  if result = nil then
  begin
    result := TTitanVariableDefinition.Create(Self, Name, TypeName, Address, Owner);

    TypeDef := GlobalDefinitionList.FindRTTITypeDefByName(TypeName);
    if TypeDef <> nil then
      if TypeDef.pti^.Kind = tkSet then
        result.TypeID := typeSET;

    I := IndexOf(nil);
    if I = -1 then
      Add(result)
    else
      Items[I] := result;

    result.Index := NewIndex;
    result.ml := result.ml + [modSTATIC];
  end
  else
    result.Address := Address;

  if Scripter <> nil then
    if TTitanBaseScripter(Scripter).fState in
      [_ssPaused, _ssRunning, _ssTerminated] then
        result.AddToScripter(Scripter);
end;

function TTitanDefinitionList.AddObject(const Name: String;
                                     Instance: TObject;
                                     Owner: TTitanClassDefinition;
                                     Scripter: Pointer = nil): TTitanObjectDefinition;
var
  I: Integer;
begin
  result := TTitanObjectDefinition(Lookup(Name, dkObject, Owner, Scripter));

  if result = nil then
  begin
    result := TTitanObjectDefinition.Create(Self, Name, Instance, Owner);

    I := IndexOf(nil);
    if I = -1 then
      Add(result)
    else
      Items[I] := result;

    result.Index := NewIndex;
    result.ml := result.ml + [modSTATIC];
  end
  else
    result.Instance := Instance;

  if Scripter <> nil then
    if TTitanBaseScripter(Scripter).fState in [_ssPaused, _ssRunning, _ssTerminated] then
        result.AddToScripter(Scripter);
end;

function TTitanDefinitionList.AddVirtualObject(const Name: String;
                                             Owner: TTitanClassDefinition;
                                             Scripter: Pointer = nil): TTitanVirtualObjectDefinition;
var
  I: Integer;
begin
  result := TTitanVirtualObjectDefinition(Lookup(Name, dkObject, Owner, Scripter));

  if result = nil then
  begin
    result := TTitanVirtualObjectDefinition.Create(Self, Name, Owner);

    I := IndexOf(nil);
    if I = -1 then
      Add(result)
    else
      Items[I] := result;

    result.Index := NewIndex;
    result.ml := result.ml + [modSTATIC];
  end;

  if Scripter <> nil then
    if TTitanBaseScripter(Scripter).fState in [_ssPaused, _ssRunning, _ssTerminated] then
        result.AddToScripter(Scripter);
end;

function TTitanDefinitionList.AddInterfaceVar(const Name: String;
                                            PIntf: PUnknown;
                                            const guid: TGUID;
                                            Owner: TTitanClassDefinition;
                                            Scripter: Pointer = nil): TTitanInterfaceVarDefinition;
var
  I: Integer;
begin
  result := TTitanInterfaceVarDefinition(Lookup(Name, dkInterface, Owner, Scripter));

  if result = nil then
  begin
    result := TTitanInterfaceVarDefinition.Create(Self, Name, PIntf, guid, Owner);

    I := IndexOf(nil);
    if I = -1 then
      Add(result)
    else
      Items[I] := result;

    result.Index := NewIndex;
    result.ml := result.ml + [modSTATIC];
  end
  else
  begin
    if result.PIntf <> nil then
      result.PIntf^ := nil;
    result.PIntf := PIntf;
  end;

  if Scripter <> nil then
    if TTitanBaseScripter(Scripter).fState in [_ssPaused, _ssRunning, _ssTerminated] then
        result.AddToScripter(Scripter);
end;

function TTitanDefinitionList.AddField(PClass: TClass; const FieldName, FieldType: String;
                                     Offset: Integer): TTitanFieldDefinition;
var
  I: Integer;
begin
  result := TTitanFieldDefinition.Create(Self, PClass, FieldName, FieldType, Offset);
  I := IndexOf(nil);
  if I = -1 then
    Add(result)
  else
    Items[I] := result;

  result.Index := NewIndex;
  result.ml := result.ml;

  FieldIdx.AddObject (PClass.ClassName + '.' + FieldName, Result);
end;

function TTitanDefinitionList.AddRecordField(Owner: TTitanClassDefinition; const FieldName,
                                           FieldType: String;
                                           Offset: Integer): TTitanRecordFieldDefinition;
var
  I: Integer;
begin
  result := TTitanRecordFieldDefinition.Create(Self, Owner, FieldName, FieldType, Offset);
  I := IndexOf(nil);
  if I = -1 then
    Add(result)
  else
    Items[I] := result;
  result.Index := NewIndex;
  result.ml := result.ml;
end;

function TTitanDefinitionList.AddConstant(const Name: String;
                                     const Value: Variant;
                                     Owner: TTitanClassDefinition;
                                     Static: Boolean = true;
                                     ck: TTitanClassKind = ckNone;
                                     Scripter: Pointer = nil): TTitanConstantDefinition;
var
  I: Integer;
begin
  result := TTitanConstantDefinition(Lookup(Name, dkConstant, Owner, Scripter));

  if result = nil then
  begin
    result := TTitanConstantDefinition.Create(Self, Name, Value, Owner);

    I := IndexOf(nil);
    if I = -1 then
      Add(result)
    else
      Items[I] := result;

    result.Index := NewIndex;
    if Static then
      result.ml := result.ml + [modSTATIC];
    result.ck := ck;

    If Owner <> Nil then
      ConstantIdx.AddObject (Owner.Name + '.' + Name, Result)
    else
      ConstantIdx.AddObject (Name, Result)

  end
  else
    result.SetValue(Value);

  ConstIdx.Add(Count - 1);

  if Scripter <> nil then
    if TTitanBaseScripter(Scripter).fState in
      [_ssPaused, _ssRunning, _ssTerminated] then
        result.AddToScripter(Scripter);
end;


function TTitanDefinitionList.AddConstantSX(const Name:String; const Value:Variant):TTitanConstantDefinition;
begin
    result := TTitanConstantDefinition.Create(Self,Name,Value,nil);
    Add(result);
    result.Index := NewIndex;
    result.ml := result.ml + [modSTATIC];
    result.ck := ckNone;
    ConstIdx.Add(Count - 1);
end;


procedure DynaGet(M: TTitanMethodBody); forward;
procedure DynaPut(M: TTitanMethodBody); forward;

function TTitanDefinitionList.AddDynamicArrayType(const TypeName, ElementTypeName: String;
                                                Owner: TTitanDefinition)
                                            :TTitanClassDefinition;
var
  ElTypeDef: TTitanClassDefinition;
begin
  ElTypeDef := FindClassDefByName(ElementTypeName);

  result := AddClass1(TypeName, TTitanClassDefinition(Owner), nil, DynaGet, DynaPut);
  result.ClassKind := ckDynamicArray;
  result.ElType := PaxTypes.GetTypeID(ElementTypeName);
  if result.ElType = -1 then
    result.ElType := typeCLASS;
  result.ElSize := PaxTypes.GetSize(result.ElType);
  result.ElTypeName := ElementTypeName;

  if ElTypeDef <> nil then
  begin
    if ElTypeDef.ClassKind = ckStructure then
    begin
      result.ElType := typeRECORD;
      result.ElSize := ElTypeDef.RecordSize;
    end;
  end;
end;


function TTitanDefinitionList.AddDynArrayTypeSX(const TypeName,ElementTypeName:String):TTitanClassDefinition;
var
  ElTypeDef:TTitanClassDefinition;
begin
  ElTypeDef := FindClassDefByName(ElementTypeName);

  result := AddClass2SX(TypeName,nil,nil,DynaGet,DynaPut);
  result.ClassKind := ckDynamicArray;
  result.ElType := PaxTypes.GetTypeID(ElementTypeName);
  if result.ElType=-1 then result.ElType := typeCLASS;
  result.ElSize := PaxTypes.GetSize(result.ElType);
  result.ElTypeName := ElementTypeName;

  if ElTypeDef <> nil then
  begin
    if ElTypeDef.ClassKind=ckStructure then
    begin
      result.ElType := typeRECORD;
      result.ElSize := ElTypeDef.RecordSize;
    end;
  end;
end;

function TTitanDefinitionList.AddArrayTypeSX(const TypeName,ElementTypeName:String; Const aMin,aMax:Integer):TTitanClassDefinition;
var
  ElTypeDef:TTitanClassDefinition;
begin
  ElTypeDef := FindClassDefByName(ElementTypeName);

   result := TTitanClassDefinition.Create(Self,TypeName,nil,nil);
   Add(result);
   result.Index := NewIndex;
   ClassIdx.Add(Count - 1);
   TTitanClassDefinition(result).PClass:=nil;

  result.ClassKind :=ckArray;
  result.ElType := PaxTypes.GetTypeID(ElementTypeName);
  if result.ElType=-1 then result.ElType := typeCLASS;
  result.ElSize := PaxTypes.GetSize(result.ElType);
  result.ElTypeName := ElementTypeName;

  if ElTypeDef <> nil then
  begin
    if ElTypeDef.ClassKind=ckStructure then
    begin
      result.ElType := typeRECORD;
      result.ElSize := ElTypeDef.RecordSize;
    end;
  end;

  result.ElMin:=aMin;
  result.ElMax:=aMax;
  result.IsStaticArray := true;

end;

function TTitanDefinitionList.AddRecordType(const TypeName: String; Size: Integer;
                                          Owner: TTitanDefinition):TTitanClassDefinition;
begin
  result := AddClass1(TypeName, TTitanClassDefinition(Owner), nil, nil, nil);
  result.ClassKind := ckStructure;
  result.RecordSize := Size;
end;

function TTitanDefinitionList.AddInterfaceType(const Name: String; Guid: TGuid;
                                             const ParentName: String; ParentGuid: TGUID;
                                             Owner: TTitanClassDefinition): TTitanClassDefinition;
var
  ParentDef: TTitanClassDefinition;
  DummyGuid: TGuid;
begin
//  result := FindInterfaceTypeDef(Guid);
//  result := FindClassDefByName(Name);
//  if result <> nil then
//    Exit;

  if ParentName = '' then
    ParentDef := nil
  else
  begin
    ParentDef := FindInterfaceTypeDef(ParentGuid);
    if ParentDef = nil then
      ParentDef := AddInterfaceType(ParentName, ParentGuid, '', DummyGuid, Owner);
  end;

  result := AddClass1(Name, Owner, ParentDef, nil, nil);
  result.ClassKind := CkInterface;
  result.Guid := Guid;

  GUIDIdx.AddObject (GUIDToString (Guid), Result);
end;

function TTitanDefinitionList.AddRTTIType(pti: PTypeInfo): TTitanRTTITypeDefinition;
var
  C: TTitanClassDefinition;
  TypeData: PTypeData;
  I: Integer;
  parent_pti: PTypeInfo;
  ParentDef: TTitanClassDefinition;
begin
  result := FindRTTITypeDef(pti);
  if result <> nil then Exit;

  result := TTitanRTTITypeDefinition.Create(Self, pti);
  Add(result);
  result.Index := NewIndex;

  ClassIdx.Add(Count - 1);

  if result.FinalType = typeENUM then
  begin
    TypeData := GetTypeData(pti);
    C := AddClass1(result.Name, nil, nil, nil, nil);
    C.ClassKind := ckEnum;

    for I:= TypeData.MinValue to TypeData.MaxValue do
      AddConstant(GetEnumName(pti, I), I, nil, true, ckEnum);
    for I:= TypeData.MinValue to TypeData.MaxValue do
      AddConstant(GetEnumName(pti, I), I, C, true, ckEnum);
  end
  else if result.FinalType = typeINTERFACE then
  begin
    TypeData := GetTypeData(pti);
    ParentDef := nil;
    if (TypeData.IntfParent <> nil) then
    begin

  {$ifndef fpc}
     parent_pti := TypeData.IntfParent^;
  {$else}
     parent_pti := TypeData.IntfParent;
  {$endif}

      if parent_pti <> nil then
      begin
        ParentDef := FindClassDefByName(parent_pti^.Name);
        if (ParentDef = nil) and StrEql('IInterface', parent_pti^.Name) then
          ParentDef := FindClassDefByName('IUnknown');
      end;
    end;

    C := AddClass1(result.Name, nil, ParentDef, nil, nil);
    C.pti := pti;
    C.ClassKind := ckInterface;
{$ifndef fpc}
    C.Guid := TypeData^.Guid;
{$endif}
  end
  else if result.FinalType = typeSET then
  begin
    TypeData := GetTypeData(pti);
    if Assigned(TypeData^.CompType) then
{$ifdef fpc}
      AddRTTIType(TypeData^.CompType);
{$else}
      AddRTTIType(TypeData^.CompType^);
{$endif}
    C := AddClass1(result.Name, nil, nil, nil, nil);
    C.IsSet := true;
    C.PtiSet := pti;
  end   
  else if result.FinalType=typeINTEGER then
    GlobalTypeAliases.AddObject(pti^.Name,'INTEGER',TObject(typeINTEGER))
  else if result.FinalType=typeCARDINAL then
    GlobalTypeAliases.AddObject(pti^.Name,'CARDINAL',TObject(typeCARDINAL))
  else if result.FinalType=typeWORD then
    GlobalTypeAliases.AddObject(pti^.Name,'WORD',TObject(typeWORD))
  else if result.FinalType=typeWORDBOOL then
    GlobalTypeAliases.AddObject(pti^.Name,'WORDBOOL',TObject(typeWORDBOOL))
  else if result.FinalType=typeLONGBOOL then
    GlobalTypeAliases.AddObject(pti^.Name,'LONGBOOL',TObject(typeLONGBOOL));
  //...........
  fListOfRTTIs.AddItem(Result.Name,result.Index);
end;

function TTitanDefinitionList.AddEnumTypeByDef(const TypeDef: WideString): TTitanClassDefinition;
var
  I, P, PE, Value, Code1: Integer;
  S, TypeName, StrValue, ValName: String;
  L: TStringList;
begin
  S := TypeDef;
  P := Pos('=', S);
  TypeName := Trim(Copy(S, 1, P - 1));
  System.Delete(S, 1, P);

  result := AddClass1(TypeName, nil, nil, nil, nil);
  result.ClassKind := ckEnum;

  System.Delete(S, 1, Pos('(', S));

  L := TStringList.Create;

  P := Pos(',', S);
  if P = 0 then
    P := Pos(')', S);

  I := 0;
  while P > 0 do
  begin
    PE := Pos('=', S);

    if (PE > 0) and (PE < P) then
    begin
      ValName := Trim(Copy(S, 1, PE - 1));
      StrValue := Trim(Copy(S, PE + 1, P - PE - 1));
      Val(StrValue, Value, Code1);
      if Code1 <> 0 then
        Value := 0;
      System.Delete(S, 1, P);

      AddConstant(ValName, Value, nil, true, ckEnum);
      AddConstant(ValName, Value, result, true, ckEnum);
    end
    else
    begin
      ValName := Trim(Copy(S, 1, P - 1));
      System.Delete(S, 1, P);
      AddConstant(ValName, I, nil, true, ckEnum);
      AddConstant(ValName, I, result, true, ckEnum);
      Inc(I);
    end;

    P := Pos(',', S);

    if P = 0 then
      P := Pos(')', S);
  end;

  L.Free;
end;

function TTitanDefinitionList.NewIndex: Integer;
begin
  Inc(_LastIndex);
  result := _LastIndex;

  if _LastIndex <> Count - 1 then
    _LastIndex := result;
end;

procedure TTitanDefinitionList.Clear;
var
  I: Integer;
begin
  for I:=0 to Count - 1 do
    if Items[I] <> nil then
       TTitanDefinition(Items[I]).Free;

  inherited;
end;

function TTitanDefinitionList.LookupFullName(const FullName: String): TTitanDefinition;
var
  I: Integer;
begin
  result := nil;

  for I:=0 to Count - 1 do
    if Records[I] <> nil then
    if StrEql(Records[I].FullName, FullName) then
    begin
      result := Records[I];
      Exit;
    end;
end;

function TTitanDefinitionList.Lookup(const Name: String;
                                   DefKind: TTitanDefKind;
                                   Owner: TTitanDefinition;
                                   Scripter: Pointer = nil): TTitanDefinition;
var
  I: Integer;
  D: TTitanDefinition;
begin
  result := nil;

    if DefKind = dkConstant then
    begin
      If Owner <> Nil then
        i := ConstantIdx.IndexOf (Owner.Name + '.' + Name)
      else
        i := ConstantIdx.IndexOf (Name);

      If i >= 0 then
        Result := ConstantIdx.Objects [i] as TTitanDefinition
      else
        Result := Nil;
      Exit;

      for I:=0 to ConstIdx.Count - 1 do
      begin
        D := Records[ConstIdx[I]];
        if D.Owner = Owner then
            if StrEql(D.Name, Name) then
            begin
              result := D;
              Exit;
            end;
      end;
      Exit;
    end;

    if DefKind = dkClass then
    begin
      for I:=0 to ClassIdx.Count - 1 do
      begin
        D := Records[ClassIdx[I]];
        if D.Owner = Owner then
            if StrEql(D.Name, Name) then
            begin
              result := D;
              Exit;
            end;
      end;
      Exit;
    end;

    for I:=0 to Count - 1 do
    if Records[I] <> nil then
    begin
      D := Records[I];
      if (D.DefKind = DefKind) or (DefKind = dkAny) then
        if D.Owner = Owner then
          if StrEql(D.Name, Name) then
          begin
            result := D;
            Exit;
          end;
    end;
end;

function TTitanDefinitionList.LookupAll(const Name: String;
                                      DefKind: TTitanDefKind;
                                      Owner: TTitanDefinition;
                                      Scripter: Pointer = nil): TList;
var
  I: Integer;
  D: TTitanDefinition;
begin
  result := TList.Create;

    if DefKind = dkConstant then
    begin
      for I:=0 to ConstIdx.Count - 1 do
      begin
        D := Records[ConstIdx[I]];
        if D.Owner = Owner then
            if StrEql(D.Name, Name) then
            begin
              result.Add(D);
              Exit;
            end;
      end;
      Exit;
    end;

    if DefKind = dkClass then
    begin
      for I:=0 to ClassIdx.Count - 1 do
      begin
        D := Records[ClassIdx[I]];
        if D.Owner = Owner then
            if StrEql(D.Name, Name) then
            begin
              result.Add(D);
              Exit;
            end;
      end;
      Exit;
    end;

    for I:=0 to Count - 1 do
    if Records[I] <> nil then
    begin
      D := Records[I];
      if (D.DefKind = DefKind) or (DefKind = dkAny) then
        if D.Owner = Owner then
          if StrEql(D.Name, Name) then
          begin
            result.Add(D);
          end;
    end;
end;

procedure TTitanDefinitionList.UnregisterObject(Instance: TObject; Owner: TTitanDefinition);
var
  I: Integer;
  D: TTitanDefinition;
begin
  for I:=0 to Count - 1 do
  if Records[I] <> nil then
  begin
    D := Records[I];
    if D.DefKind = dkObject then
      if D.Owner = Owner then
        if (D as TTitanObjectDefinition).Instance = Instance then
        begin
          D.Free;
          Items[I] := nil;
          Exit;
        end;
  end;
end;

procedure TTitanDefinitionList.Unregister(const Name: String;
                                        DefKind: TTitanDefKind;
                                        Owner: TTitanDefinition;
                                        Scripter: Pointer = nil);
var
  I, J: Integer;
  D: TTitanDefinition;
begin

  try
    if DefKind = dkConstant then
    begin
      for I:=0 to ConstIdx.Count - 1 do
      begin
        J := ConstIdx[I];
        D := Records[J];
        if D.Owner = Owner then
            if D.Name = Name then
            begin
              D.Free;
              Items[J] := nil;

              If Owner <> Nil then
                J := ConstantIdx.IndexOf (Owner.Name + '.' + Name)
              else
                J := ConstantIdx.IndexOf (Name);
              if J >= 0 then
                ConstantIdx.Delete(J);

              Exit;
            end;
      end;
      Exit;
    end;

    if DefKind = dkClass then
    begin
      for I:=0 to ClassIdx.Count - 1 do
      begin
        J := ClassIdx[I];
        D := Records[J];
        if D.Owner = Owner then
            if D.Name = Name then
            begin
              D.Free;
              Items[J] := nil;
              Exit;
            end;
      end;
      Exit;
    end;

    for I:=0 to Count - 1 do
    if Records[I] <> nil then
    begin
      D := Records[I];
      if D.DefKind = DefKind then
        if D.Owner = Owner then
          if D.Name = Name then
          begin
            D.Free;
            Items[I] := nil;
            Exit;
          end;
    end;

  finally

    ConstIdx.Clear;
    ClassIdx.Clear;

    for I:=0 to Count - 1 do
      if Records[I] <> nil then
      case Records[I].DefKind of
        dkClass, dkRTTIType: ClassIdx.Add(I);
        dkConstant: ConstIdx.Add(I);
      end;
    end;
end;

procedure TTitanDefinitionList.UnregisterAll(DefKind: TTitanDefKind;
                                           Scripter: Pointer);

procedure RemoveFromClassList(O: TTitanDefinition; const Name: String);
var
  ClassRec: TTitanClassRec;
begin
  if O = nil then
     ClassRec := TTitanBaseScripter(Scripter).ClassList[0]
   else
     ClassRec := TTitanBaseScripter(Scripter).ClassList.FindImportedClass(TTitanClassDefinition(O).PClass);
   ClassRec.MemberList.DeleteMember(Name, true);
end;

var
  I, J: Integer;
  D: TTitanDefinition;
begin
  if Scripter = nil then
    Exit;

  if DefKind = dkConstant then
  begin
    for I:=0 to ConstIdx.Count - 1 do
    begin
      J := ConstIdx[I];
      D := Records[J];
      RemoveFromClassList(D.Owner, D.Name);
      D.Free;
      Items[J] := nil;
    end;
    Exit;
  end;

  if DefKind = dkClass then
  begin
    for I:=0 to ClassIdx.Count - 1 do
    begin
      J := ClassIdx[I];
      D := Records[J];
      RemoveFromClassList(D.Owner, D.Name);
      D.Free;
      Items[J] := nil;
    end;
    Exit;
  end;

  for I:=0 to Count - 1 do
  if Records[I] <> nil then
  begin
    D := Records[I];
    if D.DefKind = DefKind then
      if D.Name <> 'ErrorObject' then
      begin
        RemoveFromClassList(D.Owner, D.Name);
        D.Free;
        Items[I] := nil;
      end;
  end;

  ConstIdx.Clear;
  ClassIdx.Clear;

  for I:=0 to Count - 1 do
    if Records[I] <> nil then
    case Records[I].DefKind of
      dkClass, dkRTTIType: ClassIdx.Add(I);
      dkConstant: ConstIdx.Add(I);
    end;
end;

procedure TTitanDefinitionList.Dump(const FileName: String);
var
  I: Integer;
  T: TextFile;
begin

  AssignFile(T, FileName);
  Rewrite(T);
  for I:=0 to Count - 1 do
  if Records[I] <> nil then
  begin
    writeln(T, I);
    Records[I].Dump(T);
  end;
  CloseFile(T);
end;

function TTitanDefinitionList.GetRecordByIndex(Index: Integer): TTitanDefinition;
var
  I: Integer;
begin
  if (Index >= 0) and (Index < Count) then
    if Records[Index] <> nil then
    if Records[Index].Index = Index then
    begin
      result := Records[Index];
      Exit;
    end;
  for I:=0 to Count - 1 do
    if Records[I] <> nil then
    if Records[I].Index = Index then
    begin
      result := Records[I];
      Exit;
    end;

  CurrScripter.Dump();
  raise Exception.Create(errGetRecordByIndex);
end;

function TTitanDefinitionList.GetKind(Index: Integer): Integer;
begin
  result := KindVAR;
  case GetRecordByIndex(Index).DefKind of
    dkMethod: result := KindSUB;
    dkClass, dkRTTIType: result := KindTYPE;
    dkProperty: result := KindPROP;
    dkVariable, dkObject: result := KindVAR;
    dkConstant: result := KindCONST;
  end;
end;

function TTitanDefinitionList.GetCount(Index: Integer): Integer;
var
  D: TTitanDefinition;
begin
  D := GetRecordByIndex(Index);

  case D.DefKind of
    dkMethod:
    begin
      result := TTitanMethodDefinition(D).NP;
    end;
  else
    result := -1;
  end;
end;

function TTitanDefinitionList.GetParamTypeName(Index, ParamIndex: Integer): String;
var
  D: TTitanDefinition;
  MD: TTitanMethodDefinition;
begin
  D := GetRecordByIndex(Index);

  case D.DefKind of
    dkMethod:
    begin
      MD := TTitanMethodDefinition(D);
      Dec(ParamIndex);
      if (ParamIndex < Length(MD.StrTypes)) and (ParamIndex >= 0) then
        result := MD.StrTypes[ParamIndex]
      else
        result := 'Variant';
    end;
  else
    result := '';
  end;
end;

function TTitanDefinitionList.GetParamName(Index, ParamIndex: Integer): String;
var
  D: TTitanDefinition;
  MD: TTitanMethodDefinition;
begin
  D := GetRecordByIndex(Index);

  case D.DefKind of
    dkMethod:
    begin
      MD := TTitanMethodDefinition(D);
      Dec(ParamIndex);
      if (ParamIndex < Length(MD.ParamNames)) and (ParamIndex >= 0) then
        result := MD.ParamNames[ParamIndex]
      else
        result := 'X' + IntToStr(ParamIndex);
    end;
  else
    result := '';
  end;
end;

function TTitanDefinitionList.GetTypeName(Index: Integer): String;
var
  D: TTitanDefinition;
  MD: TTitanMethodDefinition;
  PD: TTitanPropertyDefinition;
  CD: TTitanConstantDefinition;
  FD: TTitanFieldDefinition;
  RD: TTitanRecordFieldDefinition; // krus (1C)
  OD: TTitanObjectDefinition;
  VD: TTitanVariableDefinition;    // krus (1C)
begin
  D := GetRecordByIndex(Index);

  case D.DefKind of
    dkMethod:
    begin
      MD := TTitanMethodDefinition(D);
      result := MD.ResultType;
    end;
    dkProperty:
    begin
      PD := TTitanPropertyDefinition(D);
      result := PD.PropType;
    end;
    dkConstant:
    begin
      CD := TTitanConstantDefinition(D);
      result := CD.ResultType;
    end;
    dkField:
    begin
      FD := TTitanFieldDefinition(D);
      result := FD.FieldType;
    end;
    dkRecordField:                           // krus (1C)
    begin                                    // krus (1C)
      RD := TTitanRecordFieldDefinition(D);    // krus (1C)
      result := RD.FieldType;                // krus (1C)
    end;                                     // krus (1C)
    dkVariable:                              // krus (1C)
    begin                                    // krus (1C)
      VD := TTitanVariableDefinition(D);       // krus (1C)
      result := VD.TypeName;                 // krus (1C)
    end;                                     // krus (1C)
    dkObject:
    begin
      OD := TTitanObjectDefinition(D);
      result := OD.Instance.ClassName;
    end;
  else
    result := '';
  end;
end;

function TTitanDefinitionList.GetNameIndex(Index: Integer; Scripter: Pointer): Integer;
begin
  result := CreateNameIndex(GetRecordByIndex(Index).Name, Scripter);
end;

function TTitanDefinitionList.GetName(Index: Integer): String;
begin
  result := GetRecordByIndex(Index).Name;
end;

function TTitanDefinitionList.GetFullName(Index: Integer): String;
begin
  result := GetRecordByIndex(Index).FullName;
end;

function TTitanDefinitionList.GetAddress(Scripter: Pointer; Index: Integer): Pointer;
var
  D: TTitanDefinition;
begin
  D := GetRecordByIndex(Index);

  if D.DefKind in [dkClass, dkRTTIType] then
    result := TTitanBaseScripter(Scripter).NegVarList.GetAddress(Index)
  else
    result := @ (D.DefValue);
end;

function TTitanDefinitionList.GetUserData(Scripter: Pointer; Index: Integer): Integer;
var
  D: TTitanDefinition;
begin
  D := GetRecordByIndex(Index);
  result := D.UserData;
end;

function TTitanDefinitionList.GetVariant(Scripter: Pointer; Index: Integer): Variant;
var
  D: TTitanDefinition;
  SO: TTitanScriptObject;
begin
  D := GetRecordByIndex(Index);

  if D.DefKind in [dkClass, dkRTTIType] then
    result := TTitanBaseScripter(Scripter).NegVarList.GetAddress(Index)^
  else if D.DefKind = dkVariable then
    result := D.Value
  else if D.DefKind = dkMethod then
  begin
    SO := TTitanScriptObject.Create(TTitanBaseScripter(Scripter).ClassList.DelegateClassRec);
    SO.Instance := TTitanDelegate.Create(Scripter, - Index, TTitanBaseScripter(Scripter).Code.N, TTitanMethodDefinition(D));
    SO.PClass := TTitanDelegate;
    SO.PutProperty(CreateNameIndex('Name', Scripter), D.Name, 0);
    result := ScriptObjectToVariant(SO);
  end
  else
    result := D.Value;
end;

function TTitanDefinitionList.GetPVariant(Scripter: Pointer; Index: Integer): PVariant;
var
  D: TTitanDefinition;
begin
  D := GetRecordByIndex(Index);

  if D.DefKind in [dkClass, dkRTTIType] then
    result := TTitanBaseScripter(Scripter).NegVarList.GetAddress(Index)
  else if D.DefKind = dkVariable then
    result := D.PValue
  else
    result := D.PValue;
end;

procedure TTitanDefinitionList.PutVariant(Scripter: Pointer; Index: Integer; const Value: Variant);
var
  D: TTitanDefinition;
begin
  D := GetRecordByIndex(Index);

  if D.DefKind in [dkClass, dkRTTIType] then
    TTitanBaseScripter(Scripter).NegVarList.GetAddress(Index)^ := Value
  else if D.DefKind = dkVariable then
    D.Value := Value
  else
    D.DefValue := Value;
end;

function TTitanDefinitionList.UnregisterNamespace(const Namespacename: String): Boolean;
var
  D, D1: TTitanDefinition;
  DC: TTitanClassDefinition;
  I: Integer;
  L: TList;
begin
  result := false;
  D := LookupFullName(Namespacename);
  if D = nil then
    Exit;

  L := TList.Create;

  for I:=Count - 1 downto 0 do
  begin
    D1 := Records[i];
    if (D1.Owner = D) or (D1 = D) then
      L.Add(D1);
    if (D1.Owner <> nil) then
      if D1.Owner.Owner = D then
        L.Add(D1);
  end;

  for I:=0 to Count - 1 do
  begin
    D1 := Records[i];
    if L.IndexOf(D1.Owner) >= 0 then
       D1.Owner := nil;
    if D1.DefKind = dkClass then
    begin
      DC := D1 as TTitanClassDefinition;
      if DC.Ancestor <> nil then
        if L.IndexOf(DC.Ancestor) >= 0 then
          DC.Ancestor := nil;
    end;
  end;

  for I:=Count - 1 downto 0 do
  begin
    D1 := Records[i];
    if L.IndexOf(D1) >= 0 then
    begin
      Delete(I);
      D1.Free;
    end;
  end;

  L.Free;

  for I:=0 to Count - 1do
    Records[i].Index := I;

 _LastIndex := Count - 1;

  ConstIdx.Clear;
  ClassIdx.Clear;

  for I:=0 to Count - 1 do
    if Records[I] <> nil then
    case Records[I].DefKind of
      dkClass, dkRTTIType: ClassIdx.Add(I);
      dkConstant: ConstIdx.Add(I);
    end;

  result := true;
end;

constructor TTitanFieldList.Create;
begin
  fList := TList.Create;
end;

procedure TTitanFieldList.Clear;
var
  I: Integer;
begin
  for I:=0 to Count - 1 do
    Records[I].Free;
  fList.Clear;
end;

destructor TTitanFieldList.Destroy;
begin
  Clear;
  fList.Free;
end;

function TTitanFieldList.Count: Integer;
begin
  result := fList.Count;
end;

function TTitanFieldList.GetRecord(Index: Integer): TTitanFieldRec;
begin
  result := TTitanFieldRec(fList[Index]);
end;

procedure TTitanFieldList.AddRec(const ObjectName: String;
                               ObjectType: String;
                               FieldName: String;
                               FieldType: String;
                               Address: Pointer);
var
  R: TTitanFieldRec;
begin
  R := TTitanFieldRec.Create;
  R.ObjectName := ObjectName;
  R.ObjectType := ObjectType;
  R.FieldName := FieldName;
  R.FieldType := FieldType;
  R.Address := Address;
  fList.Add(R);
end;

function TTitanFieldList.FindRecord(const ObjectName: String;
                                  ObjectType: String;
                                  FieldName: String): TTitanFieldRec;
var
  I: Integer;
  R: TTitanFieldRec;
begin
  result := nil;
  for I:=0 to Count - 1 do
  begin
    R := Records[I];
    if StrEql(R.ObjectName, ObjectName) and
       StrEql(R.ObjectType, ObjectType) and
       StrEql(R.FieldName, FieldName) then
       begin
         result := R;
         Exit;
       end;
  end;
end;

procedure _HasProperty(MethodBody: TTitanMethodBody);
var
  S: String;
  ID: Integer;
  SO: TTitanScriptObject;
begin
  with MethodBody do
  begin
    S := Params[0].AsString;
    with TTitanBaseScripter(Scripter) do
    begin
      ID := ClassList[0].ClassID;
      SO := VariantToScriptObject(SymbolTable.GetVariant(ID));
      result.AsVariant := SO.HasProperty(S);
    end;
  end;
end;

procedure _PropExists(MethodBody: TTitanMethodBody);
var
  SO: TTitanScriptObject;
  S: String;
begin
  with MethodBody do
  begin
    SO := VariantToScriptObject(Params[0].AsVariant);
    S := Params[1].AsString;
    result.AsVariant := SO.HasProperty(S);
  end;
end;

procedure _ToInteger(MethodBody: TTitanMethodBody);
begin
  with MethodBody do
    result.AsVariant := toInteger(Params[0].AsVariant);
end;

procedure _ToNumber(MethodBody: TTitanMethodBody);
begin
  with MethodBody do
    result.AsVariant := toNumber(Params[0].AsVariant);
end;

procedure _ToBoolean(MethodBody: TTitanMethodBody);
begin
  with MethodBody do
    result.AsVariant := toBoolean(Params[0].AsVariant);
end;

procedure _ToString(MethodBody: TTitanMethodBody);
begin
  with MethodBody do
    //result.AsVariant := toString(Params[0].AsVariant);
    result.AsVariant := String(Params[0].AsVariant); //by 9999
end;

procedure _IsUndefined(MethodBody: TTitanMethodBody);
begin
  with MethodBody do
    result.AsVariant := IsUndefined(Params[0].PValue^);
end;

procedure _IsString(MethodBody: TTitanMethodBody);
begin
  with MethodBody do
    result.AsVariant := IsString(Params[0].PValue^) or
                        IsStringObject(Params[0].PValue^);
end;

procedure _IsBoolean(MethodBody: TTitanMethodBody);
begin
  with MethodBody do
    result.AsVariant := IsBoolean(Params[0].PValue^) or
                        IsBooleanObject(Params[0].PValue^);
end;

procedure _IsNumber(MethodBody: TTitanMethodBody);
begin
  with MethodBody do
    result.AsVariant := IsNumber(Params[0].PValue^) or
                        IsNumberObject(Params[0].PValue^);
end;

procedure _IsObject(MethodBody: TTitanMethodBody);
begin
  with MethodBody do
    result.AsVariant := IsObject(Params[0].PValue^);
end;

procedure _IsPaxArray(MethodBody: TTitanMethodBody);
begin
  with MethodBody do
    result.AsVariant := IsPaxArray(Params[0].PValue^);
end;

procedure _IsVBArray(MethodBody: TTitanMethodBody);
begin
  with MethodBody do
    result.AsVariant := IsVBArray(Params[0].PValue^);
end;

procedure _IsFunction(MethodBody: TTitanMethodBody);
var
  SO: TTitanScriptObject;
begin
  with MethodBody do
    if IsObject(Params[0].PValue^) then
    begin
      SO := VariantToScriptObject(Params[0].PValue^);
      result.AsVariant := (SO.ClassRec.AncestorName = 'Function') or
                          (SO.ClassRec.Name = 'TTitanDelegate');
    end
    else
      result.AsVariant := false;
end;

procedure TTitanArray_GetOwner(MethodBody: TTitanMethodBody);
var
  PaxArray: TTitanArray;
begin
  with MethodBody do
  begin
    PaxArray := TTitanArray(Self);
    result.AsVariant := PaxArray.Owner;
  end;
end;

procedure TTitanArray_PutOwner(MethodBody: TTitanMethodBody);
var
  PaxArray: TTitanArray;
begin
  with MethodBody do
  begin
    PaxArray := TTitanArray(Self);
    PaxArray.Owner := Params[0].PValue^;
  end;
end;

procedure _Prev(MethodBody: TTitanMethodBody);
begin
  with MethodBody do
  begin
    result.AsVariant := Params[0].AsVariant - 1;
  end;
end;

procedure _Succ(MethodBody: TTitanMethodBody);
begin
  with MethodBody do
  begin
    result.AsVariant := Params[0].AsVariant + 1;
  end;
end;

procedure _Low(MethodBody: TTitanMethodBody);
var
  V: Variant;
  SO: TTitanScriptObject;
begin
  with MethodBody do
  begin
    V := Params[0].AsVariant;
    if IsPaxArray(V) then
      result.AsVariant := 0
    else if IsDynamicArray(V) then
      result.AsVariant := 0
    else if IsObject(V) then
    begin
      SO := VariantToScriptObject(V);

      if SO.HasProperty('L1') then
        result.AsVariant := SO.GetProperty(CreateNameIndex('L1', Scripter), 0);
    end;
  end;
end;

procedure _High(MethodBody: TTitanMethodBody);
var
  V: Variant;
  SO: TTitanScriptObject;
  P: Pointer;
begin
  with MethodBody do
  begin
    V := Params[0].AsVariant;
    if IsPaxArray(V) then
    begin
      SO := VariantToScriptObject(V);
      result.AsVariant := TTitanArray(SO.Instance).fBounds[0];
    end
    else if IsDynamicArray(V) then
    begin
      SO := VariantToScriptObject(V);
      if SO.ExtraPtr = nil then
        raise Exception.Create(errDynamicArrayNotInitialized);
      P := ShiftPointer(SO.ExtraPtr, - 4);
      result.AsVariant := PtrInt(P^);
    end
    else if IsObject(V) then
    begin
      SO := VariantToScriptObject(V);

      if SO.HasProperty('H1') then
        result.AsVariant := SO.GetProperty(CreateNameIndex('H1', Scripter), 0);
    end;
  end;
end;

procedure _Length(MethodBody: TTitanMethodBody);
var
  V: Variant;
  SO: TTitanScriptObject;
  D: TTitanClassDefinition;
  P: Pointer;
begin
  with MethodBody do
  begin
    V := Params[0].AsVariant;
    if IsObject(V) then
    begin
      SO := VariantToScriptObject(V);
      D := SO.ClassRec.GetClassDef;
      if D = nil then
        raise Exception.Create(errIncompatibleTypes);
      if D.ClassKind <> ckDynamicArray then
        raise Exception.Create(errIncompatibleTypes);
      P := SO.ExtraPtr;
      if P = nil then
        result.AsInteger := 0
      else
      begin
        P := ShiftPointer(P, - SizeOf(Integer));
        result.AsInteger := PtrInt(P^);
      end;
    end
    else if IsUndefined(V) then
    begin
      result.AsInteger := 0;
    end
    else
      result.AsInteger := Length(Params[0].AsString);
  end;
end;

procedure _SetLength(MethodBody: TTitanMethodBody);
var
  I, L, OldL: Integer;
  V: Variant;
  S: String;
  SO: TTitanScriptObject;
  D: TTitanClassDefinition;
  P: Pointer;
  temp: boolean;
  ClassRec: TTitanClassRec;
  buff: Pointer;
  lbuff: Integer;
begin
  buff := nil;
  with MethodBody do
  begin
    V := Params[0].AsVariant;
    L := Params[1].AsInteger;
    OldL := 0;
    lbuff := 0;
    if IsObject(V) then
    begin
      SO := VariantToScriptObject(V);
      SO.Instance := SO;

      D := SO.ClassRec.GetClassDef;
      if D = nil then
        raise Exception.Create(errIncompatibleTypes);
      if D.ClassKind <> ckDynamicArray then
        raise Exception.Create(errIncompatibleTypes);

      if SO.ExtraPtr <> nil then
      begin
        OldL := PtrInt(ShiftPointer(SO.ExtraPtr, - SizeOf(Pointer))^);
        if (OldL > 0) and (OldL < L) then
        begin
          lbuff := SO.ExtraPtrSize - 8;
          GetMem(buff, lbuff);
          Move(SO.ExtraPtr^, buff^, lbuff);
        end;
      end;

      temp := SO.ExternalExtraPtr;
      SO.ExternalExtraPtr := false;
      SO.FreeExtraPtr();
      SO.ExternalExtraPtr := temp;

      ClassRec := TTitanBaseScripter(Scripter).ClassList.FindClassByName(D.ElTypeName);
      if ClassRec <> nil then
      if ClassRec.fClassDef = nil then
      begin
        D.ElSize := _SizeVariant;
        D.ElType := typeVariant;
      end;

      P := AllocMem(2*SizeOf(Integer) +  L * D.ElSize);
      PtrInt(P^) := 1;
      P := ShiftPointer(P, SizeOf(Integer));
      PtrInt(P^) := L;
      P := ShiftPointer(P, SizeOf(Integer));
      SO.ExtraPtr := P;

      if lbuff > 0 then
        Move(buff^, P^, lbuff);

      SO.ExtraPtrSize := 2*SizeOf(Integer) +  L * D.ElSize;

      if ClassRec <> nil then
      if ClassRec.fClassDef = nil then
      begin
        for I:= 0 to L - 1 do
        begin
          if I > OldL - 1 then
          begin
            SO := ClassRec.CreateScriptObject;
            V := ScriptObjectToVariant(SO);
            Variant(P^) := V;
          end;
          P := ShiftPointer(P, _SizeVariant);
        end;
      end;

      if lbuff > 0 then
        FreeMem(buff, lbuff);
    end
    else
    begin
      //S := ToString(V);
      S := String(V); //by 9999
      SetLength(S, L);
      Params[0].AsVariant := S;
    end;
  end;
end;

procedure DynaGet(M: TTitanMethodBody);
var
  SO: TTitanScriptObject;
  D: TTitanClassDefinition;
  I: Integer;
  P: Pointer;
begin
  with M do
  begin
    SO := TTitanScriptObject(Self);
    if SO.ExtraPtr = nil then
      raise Exception.Create(errDynamicArrayNotInitialized);

    D := SO.ClassRec.GetClassDef;
    I := StrToInt(Name);
    P := ShiftPointer(SO.ExtraPtr, I * D.ElSize);
    result.AsVariant := GetVariantValue(SO.Scripter, P, D.ElType, D.ElTypeName);
    PSelf := nil;
  end;
end;

procedure DynaPut(M: TTitanMethodBody);
var
  SO: TTitanScriptObject;
  D: TTitanClassDefinition;
  I: Integer;
  P: Pointer;
  V: Variant;
begin
  with M do
  begin
    SO := TTitanScriptObject(Self);
    if SO.ExtraPtr = nil then
      raise Exception.Create(errDynamicArrayNotInitialized);

    D := SO.ClassRec.GetClassDef;
    I := StrToInt(Name);
    V := Params[0].AsVariant;
    P := ShiftPointer(SO.ExtraPtr, I * D.ElSize);
    PutVariantValue(SO.Scripter, P, V, D.ElType);
    PSelf := nil;
  end;
end;
   {
procedure AddTypeAlias(const T1, T2: String);
var
  S, Dest: String;
begin
  Dest := T2;
  repeat
    S := FindTypeAlias(Dest, true);
    if S <> '' then
      Dest := S;
  until S = '';
  GlobalTypeAliases.Add(T1 + '=' + Dest);
end;

function FindTypeAlias(const TypeName: String; UpCase: Boolean): String;
var
  I: Integer;
  S: String;
  b: Boolean;
begin

  If UpCase then begin
    Result := '';
    // locate name
    S := TypeName + TypeAliases.NameValueSeparator;
    if Not TypeAliases.Find (S, I) and
       (I < TypeAliases.Count) and
       (StrEql (Copy(TypeAliases[I], 1, Length(S)), S)) then begin
      //S := TypeAliases [i];
      Result := TypeAliases.ValueFromIndex [i];
    end;

    Exit;
  end;


end;
   }

//==============================================================

procedure AddTypeAlias(const T1,T2:String);
begin
  if GlobalTypeAliases.IndexOf(T1)>-1 then exit;
  GlobalTypeAliases.AddItem(T1,T2);
end;

function FindTypeAlias(const TypeName: String; UpCase: Boolean): String;
begin
 result := '';
 result:=GlobalTypeAliases.IndexOfByName(TypeName);
end;

//====================================================================================
//====================================================================================
//====================================================================================

procedure _Abort;
begin
  Abort;
end;

procedure Initialization_BASE_EXTERN;
var
  C: TTitanClassDefinition;
  Inx:integer;
begin
  GlobalDefinitionList := TTitanDefinitionList.Create(true);
  ArrayParamMethods := Th2StringsList.Create;
  GlobalTypeAliases:=Th2StringsList.Create;
  UnresolvedTypes:=Th2StringsList.Create;




 with GlobalDefinitionList do
 begin
   AddDynamicArrayType('DynArrayVariant', 'Variant', nil);

   AddRTTIType(TypeInfo(HRESULT));

   AddTypeAlias('Longint', 'Integer');
   AddTypeAlias('LongWord', 'Cardinal');
   AddTypeAlias('DWord', 'LongWord');
   AddTypeAlias('UInt', 'DWord');
   AddTypeAlias('ULong', 'Cardinal');
   AddTypeAlias('THandle', 'Cardinal');
   AddTypeAlias('TDateTime', 'Double');
   AddTypeAlias('Real', 'Double');
   AddTypeAlias('TCLSID', 'TGUID');

   C := AddRecordType('TGUID', SizeOf(TGUID), nil);
   AddRecordField(C, 'D1', 'Cardinal', 0);
   AddRecordField(C, 'D2', 'Word', 4);
   AddRecordField(C, 'D3', 'Word', 6);
   AddRecordField(C, 'D40', 'byte', 8);
   AddRecordField(C, 'D41', 'byte', 9);
   AddRecordField(C, 'D42', 'byte', 10);
   AddRecordField(C, 'D43', 'byte', 11);
   AddRecordField(C, 'D44', 'byte', 12);
   AddRecordField(C, 'D45', 'byte', 13);
   AddRecordField(C, 'D46', 'byte', 14);
   AddRecordField(C, 'D47', 'byte', 15);



  {
   AddClass2(TObject, nil, nil, nil, ckClass);
   AddMethod2(TObject, 'constructor Create;', @TObject.Create);
   AddMethod2(TObject, 'class function ClassInfo: Pointer;', @TObject.ClassInfo);
   AddMethod2(TObject, 'class function ClassName: ShortString;', @TObject.ClassName);
   AddMethod2(TObject, 'class function ClassParent: TClass;', @TObject.ClassParent);
   AddMethod2(TObject, 'function ClassType: TClass;', @TObject.ClassType);
   AddMethod2(TObject, 'destructor Destroy; virtual;', @TObject.Destroy);
   AddMethod2(TObject, 'destructor Free;', @TObject.Free);
   AddMethod2(TObject, 'class function InheritsFrom(AClass: TClass): Boolean;', @TObject.InheritsFrom);
   AddMethod2(TObject, 'class function InstanceSize: Longint;', @TObject.InstanceSize);
   AddMethod2(TObject, 'function GetInterface(const IID: TGUID; out Obj): Boolean;', @TObject.GetInterface);
      }
    inx:=AddClassSX(TObject).Index;
    AddMethodSX(inx,'constructor Create;',@TObject.Create);
    AddMethodSX(inx,'procedure Free;',@TObject.Free);
    AddMethodSX(inx,'class function InitInstance(Instance: Pointer): TObject;',@TObject.InitInstance);
    AddMethodSX(inx,'procedure CleanupInstance;',@TObject.CleanupInstance);
    AddMethodSX(inx,'function ClassType: TClass;',@TObject.ClassType);
    AddMethodSX(inx,'class function ClassName: ShortString;',@TObject.ClassName);
    AddMethodSX(inx,'class function ClassNameIs(const Name: string): Boolean;',@TObject.ClassNameIs);
    AddMethodSX(inx,'class function ClassParent: TClass;',@TObject.ClassParent);
    AddMethodSX(inx,'class function ClassInfo: Pointer;',@TObject.ClassInfo);
    AddMethodSX(inx,'class function InstanceSize: Longint;',@TObject.InstanceSize);
    AddMethodSX(inx,'class function InheritsFrom(AClass: TClass): Boolean;',@TObject.InheritsFrom);
    AddMethodSX(inx,'class function MethodAddress(const Name: ShortString): Pointer;',@TObject.MethodAddress);
    AddMethodSX(inx,'class function MethodName(Address: Pointer): ShortString;',@TObject.MethodName);
    AddMethodSX(inx,'function FieldAddress(const Name: ShortString): Pointer;',@TObject.FieldAddress);
    AddMethodSX(inx,'function GetInterface(const IID: TGUID; out Obj): Boolean;',@TObject.GetInterface);
   // AddMethodSX(inx,'class function GetInterfaceEntry(const IID: TGUID): PInterfaceEntry;',@TObject.GetInterfaceEntry);
   // AddMethodSX(inx,'class function GetInterfaceTable: PInterfaceTable;',@TObject.GetInterfaceTable);
    AddMethodSX(inx,'function SafeCallException(ExceptObject: TObject; ExceptAddr: Pointer): HResult; virtual;',@TObject.SafeCallException);
    AddMethodSX(inx,'procedure AfterConstruction; virtual;',@TObject.AfterConstruction);
    AddMethodSX(inx,'procedure BeforeDestruction; virtual;',@TObject.BeforeDestruction);
    AddMethodSX(inx,'procedure Dispatch(var Message); virtual;',@TObject.Dispatch);
    AddMethodSX(inx,'procedure DefaultHandler(var Message); virtual;',@TObject.DefaultHandler);
    AddMethodSX(inx,'class function NewInstance: TObject; virtual;',@TObject.NewInstance);
    AddMethodSX(inx,'procedure FreeInstance; virtual;',@TObject.FreeInstance);
    AddMethodSX(inx,'destructor Destroy; virtual;',@TObject.Destroy);


   AddInterfaceType('IUnknown', IUnknown, '', IUnknown, nil);
   AddMethod6(IUnknown, 'function QueryInterface(const IID: TGUID; out Obj): HResult; stdcall;');
   AddMethod6(IUnknown, 'function _AddRef: Integer; stdcall;');
   AddMethod6(IUnknown, 'function _Release: Integer; stdcall;');

   AddMethod3('PaxArray', _PaxArray, -1, nil, true);
   AddMethod3('Dump', _Dump, 0, nil, true);
   AddMethod3('IOResult', _IOResult, 0, nil, true);
   AddMethod3('ScriptObjectListCount', _ScriptObjectListCount, 0, nil, true);
   AddMethod3('Destroy', _Destroy, 0, nil, true);
   AddMethod3('GetTickCount', _GetTickCount, 0, nil, true);
   AddMethod3('GetTotalAllocated', _GetTotalAllocated, 0, nil, true);
   AddMethod3('Sleep', _Sleep, 1, nil, true);
   AddMethod3('DoNotDestroy', _DoNotDestroy, 1, nil, true);
   AddMethod3('Assigned', _Assigned, 1, nil, true);
   AddMethod3('Keep', _DoNotDestroy, 1, nil, true);
   AddMethod3('CreateArray', _CreateArray, 1, nil, true);
   AddMethod3('Length', _Length, 1, nil, true);
   AddMethod3('SetLength', _SetLength, 2, nil, true);
   AddMethod3('HasProperty', _HasProperty, 1, nil, true);
   AddMethod3('PropExists', _PropExists, 2, nil, true);
   AddMethod3('toInteger', _toInteger, 1, nil, true);
   AddMethod3('toString', _toString, 1, nil, true);
   AddMethod3('toNumber', _toNumber, 1, nil, true);
   AddMethod3('toBoolean', _toBoolean, 1, nil, true);

   AddMethod3('isUndefined', _IsUndefined, 1, nil, true);
   AddMethod3('isString', _IsString, 1, nil, true);
   AddMethod3('isNumber', _IsNumber, 1, nil, true);
   AddMethod3('isBoolean', _IsBoolean, 1, nil, true);
   AddMethod3('isObject', _IsObject, 1, nil, true);
   AddMethod3('isPaxArray', _IsPaxArray, 1, nil, true);
   AddMethod3('isVBArray', _IsVBArray, 1, nil, true);

   AddMethod3('Prev', _Prev, 1, nil, true);
   AddMethod3('Pred', _Prev, 1, nil, true);
   AddMethod3('Succ', _Succ, 1, nil, true);

   AddMethod3('Low', _Low, 1, nil, true);
   AddMethod3('High', _High, 1, nil, true);

   AddClass2(TTitanDelegate, nil, nil, nil);
   AddClass2(TTitanError, nil, nil, nil);

   AddClass2(TTitanArray, nil, nil, nil, ckArray);
   AddMethod2(TTitanArray, 'constructor Create(Bounds: array of Integer);',@TTitanArray.Create);
   AddMethod2(TTitanArray, 'function Get(const Indexes: array of Integer): Variant;',@TTitanArray.Get);
   AddMethod2(TTitanArray, 'procedure Put(const Indexes: array of Integer; const Value: Variant);',@TTitanArray.Put);
   AddMethod2(TTitanArray, 'procedure ClearIndexes;', @TTitanArray.ClearIndexes);
   AddMethod2(TTitanArray, 'function AddIndex(I: Integer): Integer;',@TTitanArray.AddIndex);
   AddMethod2(TTitanArray, 'function _Get: Variant;',@TTitanArray._Get);
   AddMethod2(TTitanArray, 'procedure _Put(const Value: Variant);',@TTitanArray._Put);

   AddConstant('null', Undefined, nil);
   AddConstant('MaxInt', MaxInt, nil);
   AddConstant('DEFAULT_CHARSET', 1, nil);


   AddConstant('S_OK', 0, nil);
   AddConstant('S_FALSE', 1, nil);
   AddConstant('E_NOINTERFACE', PtrInt($80004002), nil);
   AddConstant('E_UNEXPECTED', PtrInt($8000FFFF), nil);
   AddConstant('E_NOTIMPL', PtrInt($80004001), nil);

   AddConstant('varEmpty',    $0000, nil);
   AddConstant('varNull',     $0001, nil);
   AddConstant('varSmallInt', $0002, nil);
   AddConstant('varInteger',  $0003, nil);
   AddConstant('varSingle',   $0004, nil);
   AddConstant('varDouble',   $0005, nil);
   AddConstant('varCurrency', $0006, nil);
   AddConstant('varDate',     $0007, nil);
   AddConstant('varOleStr',   $0008, nil);
   AddConstant('varDispatch', $0009, nil);
   AddConstant('varError',    $000A, nil);
   AddConstant('varBoolean',  $000B, nil);
   AddConstant('varVariant',  $000C, nil);
   AddConstant('varUnknown',  $000D, nil);
   AddConstant('varShortInt', $0010, nil);
   AddConstant('varByte',     $0011, nil);
   AddConstant('varWord',     $0012, nil);
   AddConstant('varLongWord', $0013, nil);
   AddConstant('varInt64',    $0014, nil);



   AddMethod1('procedure Abort;', @_Abort, nil, true);
   //............ Strings Functions .............................
   AddMethod1('function Format(const Format:string; const Args:array of const):string;',@Format,nil,true);
   AddMethod1('function IntfRefCount(const I:IUnknown):Integer;',@IntfRefCount,nil,true);
   AddMethod1('function IntToStr(Value:Integer):string;',@IntToStr,nil,true);
   AddMethod1('function StrToInt(const S:string):Integer;',@StrToInt,nil,true);
   AddMethod1('function IntToHex(Value:Integer; Digits:Integer):string;',@IntToHex,nil,true);
   AddMethod1('function UpperCase(const S:string):string;',@UpperCase,nil,true);
   AddMethod1('function LowerCase(const S:string):string;',@LowerCase,nil,true);
   AddMethod1('function CompareStr(const S1,S2:string):Integer;',@CompareStr,nil,true);
   AddMethod1('function CompareText(const S1,S2:string):Integer;',@CompareText,nil,true);
   AddMethod1('function SameText(const S1,S2:string):Boolean;',@SameText,nil,true);
   AddMethod1('function Trim(const S:string):string;',@Trim,nil,true);
      //............ DateTime Functions .............................
   AddMethod1('function TimeToStr(const DateTime: TDateTime): string;', @TimeToStr,nil,true);
   AddMethod1('function DateToStr(const DateTime: TDateTime): string;', @DateToStr,nil,true);
   AddMethod1('function DateTimeToStr(const DateTime: TDateTime): string;', @DateTimeToStr,nil,true);
   AddMethod1('function FloatToDateTime(const Value: Extended): TDateTime;', @FloatToDateTime,nil,true);
   AddMethod1('function StrToDateTime(const S: string): TDateTime;', @StrToDateTime,nil,true);
   AddMethod1('function StrToDate(const S: string): TDateTime;', @StrToDate,nil,true);
   AddMethod1('function StrToTime(const S: string): TDateTime;', @StrToTime,nil,true);
   AddMethod1('function Date: TDateTime;', @Date,nil,true);
   AddMethod1('function Time: TDateTime;', @Time,nil,true);
   AddMethod1('function Now: TDateTime;', @Now,nil,true);
   AddMethod1('function DayOfWeek(const DateTime: TDateTime): Word;', @DayOfWeek,nil,true);
   AddMethod1('function CurrentYear: Word;', @CurrentYear,nil,true);
   AddMethod1('function FormatDateTime(const Format:string; DateTime:TDateTime):string;', @FormatDateTime,nil,true);
   //................ Files Functions ................................
   AddMethod1('function ExtractFilePath(const FileName: string): string;', @ExtractFilePath,nil,true);
   AddMethod1('function ExtractFileDir(const FileName: string): string;', @ExtractFileDir,nil,true);
   AddMethod1('function ExtractFileName(const FileName: string): string;', @ExtractFileName,nil,true);
   AddMethod1('function ExtractFileExt(const FileName: string): string;', @ExtractFileExt,nil,true);
   AddMethod1('function ChangeFileExt(const FileName, Extension: string):string;', @ChangeFileExt,nil,true);
   AddMethod1('function FileExists(const FileName: string): Boolean;', @FileExists,nil,true);
   AddMethod1('function DirectoryExists(const Directory: string): Boolean;', @DirectoryExists,nil,true);


 end;
 GlobalDefListInitialCount := GlobalDefinitionList.Count;

end;

procedure Finalization_BASE_EXTERN;
begin
  GlobalDefinitionList.Free;
  ArrayParamMethods.Free;
  GlobalTypeAliases.Free;
  UnresolvedTypes.Free;
end;

 
