{%mainunit ../BASE_Engine.pas}

{***************************************************
 Package pl_TitanScript.pkg
 This file is part of CodeTyphon Studio
****************************************************}

constructor TTitanModule.Create(Scripter: TTitanBaseScripter);
begin
  inherited Create;
  Self.Scripter := Scripter;
  Namespaces := TTitanIds.Create(False);
  IsSource := True;
  BuffStream := TMemoryStream.Create;
end;

destructor TTitanModule.Destroy;
begin
  Namespaces.Free;
  BuffStream.Free;
  inherited;
end;

procedure TTitanModule.AddNamespace(ID: integer);
begin
  if Namespaces.IndexOf(ID) = -1 then
    Namespaces.Add(ID);
end;

function TTitanModule.GetTextPos(LineNumber, Position: integer): integer;
var
  S: string;
  P, LineCount: integer;
begin
  Result := 0;
  P := 1;
  LineCount := 1;

  S := Text + #255;

  while S[P] <> #255 do
  begin
    if S[P] in [#10, #13] then
    begin
      Inc(LineCount);
      if S[P] = #13 then
        Inc(P);
    end;

    if LineCount = LineNumber then
    begin
      while S[P] in [#10, #13] do
        Inc(P);

      Result := P + Position;
      Exit;
    end;

    Inc(P);
  end;
end;

procedure TTitanModule._SaveToStream(S: TStream);
begin
  SaveString(Name, S);
  SaveString(FileName, S);
  SaveString(LanguageName, S);
  SaveInteger(Count, S);
  Namespaces.SaveToStream(S);

  SaveInteger(S1, S);
  SaveInteger(S2, S);

  SaveInteger(P1, S);
  SaveInteger(P2, S);

  SaveInteger(C1, S);
  SaveInteger(C2, S);
end;

procedure TTitanModule._LoadFromStream(S: TStream);
var
  I, K: integer;
  St: string;
begin
  Name := LoadString(S);
  FileName := LoadString(S);
  LanguageName := LoadString(S);
  K := LoadInteger(S);
  Namespaces.LoadFromStream(S);

  S1 := LoadInteger(S);
  S2 := LoadInteger(S);

  P1 := LoadInteger(S);
  P2 := LoadInteger(S);

  C1 := LoadInteger(S);
  C2 := LoadInteger(S);

  if Assigned(Scripter.fOnLoadSourceCode) then
  begin
    Scripter.fOnLoadSourceCode(Scripter.Owner, Name, FileName, St);
    Text := St;
  end
  else
  begin
    if FileExists(FileName) then
      LoadFromFile(FileName)
    else
      for I := 0 to K - 1 do
        Add('##' + IntToStr(I) + '##');
  end;

  IsSource := False;
end;

constructor TTitanModules.Create(Scripter: TTitanBaseScripter);
begin
  inherited Create;
  Self.Scripter := Scripter;
end;

destructor TTitanModules.Destroy;
var
  I: integer;
begin
  for I := 0 to Count - 1 do
    Items[I].Free;
  inherited;
end;

procedure TTitanModules._SaveToStream(S: TStream);
var
  I: integer;
begin
  SaveInteger(Count, S);
  for I := 0 to Count - 1 do
    Items[I]._SaveToStream(S);
end;

procedure TTitanModules._LoadFromStream(S: TStream);
var
  I, K: integer;
  M: TTitanModule;
begin
  Clear;
  K := LoadInteger(S);
  for I := 1 to K do
  begin
    M := TTitanModule.Create(Scripter);
    M._LoadFromStream(S);
    AddObject(M.Name, M);
  end;
end;

function TTitanModules.GetModule(Index: integer): TTitanModule;
begin
  if (Index < 0) or (Index >= Count) then
    Result := nil
  else
    Result := TTitanModule(Objects[Index]);
end;

procedure TTitanModules.Delete(Index: integer);
begin
  if (Index < 0) or (Index >= Count) then
    Exit;
  Items[Index].Free;
  inherited;
end;

procedure TTitanModules.Clear;
var
  I: integer;
begin
  for I := Count - 1 downto 0 do
    Delete(I);
  inherited;
end;

function TTitanModules.GetSourceCode: string;
var
  I: integer;
begin
  Result := '';
  for I := 0 to Count - 1 do
    Result := Result + GetModule(I).Text;
end;

function TTitanModules.IndexOf(const Name: string): integer;
var
  Module: TTitanModule;
  I: integer;
begin
  Result := -1;
  for I := 0 to Count - 1 do
  begin
    Module := Items[I];
    if StrEql(Name, Module.Name) then
    begin
      Result := I;
      Exit;
    end;
  end;
end;

function TTitanModules.Add(const Name, LanguageName: string): integer;
var
  Module: TTitanModule;
begin
  Result := IndexOf(Name);
  if Result >= 0 then
    Exit;
  Module := TTitanModule.Create(Scripter);
  Module.Name := Name;
  Module.FileName := '';
  Module.LanguageName := LanguageName;
  Result := AddObject(Name, Module);
end;

procedure TTitanModules.Dump(const FileName: string);
var
  T: TextFile;
  I: integer;
  M: TTitanModule;
begin
  if _IsDump = False then
    Exit;

  AssignFile(T, FileName);
  Rewrite(T);
  try
    for I := 0 to Count - 1 do
    begin
      M := GetModule(I);
      writeln(T, M.Name);
      writeln(T, 'S1 =', M.S1);
      writeln(T, 'S2 =', M.S2);
      writeln(T, 'P1 =', M.P1);
      writeln(T, 'P2 =', M.P2);
      writeln(T, 'C1 =', M.C1);
      writeln(T, 'C2 =', M.C2);
      writeln(T, '**************************************');
      writeln(T, M.Text);
      writeln(T, '**************************************');
    end;
  finally
    CloseFile(T);
  end;
end;              
