{%mainunit ../BASE_Engine.pas}

{***************************************************
 Package pl_TitanScript.pkg
 This file is part of CodeTyphon Studio
****************************************************}

constructor TTitanSymbolTable.Create(AScripter: Pointer);
var
  I: Integer;
begin
  StateStack := TTitanStack.Create;

  Scripter := AScripter;

  MemSize := FirstMemSize;

  Mem := AllocMem(MemSize);
  MemBoundVar  := 0;

  SetLength(A, FirstSymbolCard);
  for I:=0 to FirstSymbolCard - 1 do
    FillChar(A[I], SizeOf(TTitanSymbolRec), 0);

  fCard := -1;


  for I:=0 to PAXTypes.Count - 1 do
  begin
    AppVariant(Undefined, true);
    SetName(I, PAXTypes[I]);
    SetKind(I, kindTYPE);
  end;

  SetName(AppVariantConst(ord(maAny)), '_maAny_');
  SetName(AppVariantConst(ord(maMyBase)), '_maMyBase_');
  SetName(AppVariantConst(ord(maMyClass)), '_maMyClass_');

  IDtrue := AppVariant(true);
  SetName(IDtrue, 'true');

  IDfalse := AppVariant(false);
  SetName(IDfalse, 'false');

  IDundefined := AppVariant(undefined);

  RootNameSpaceID := AppVariant(Undefined);
  Name[RootNameSpaceID] := RootNamespaceName;
  Kind[RootNameSpaceID] := KindTYPE;

  CreateCard := Card;
  CreateMemBoundVar := MemBoundVar;

  MaxLocalVars := 0;
end;

procedure TTitanSymbolTable.SetCard(Value: Integer);
begin
  fCard := Value;
  CheckLength;
end;

procedure TTitanSymbolTable.IncCard;
begin
  Inc(fCard);
  CheckLength;
end;

procedure TTitanSymbolTable.DecCard;
begin
  Dec(fCard);
end;

procedure TTitanSymbolTable.CheckLength;
var
  L, I: Integer;
begin
  L := Length(A);
  if L < Card + 10 then
  begin
    SetLength(A, Card + DeltaSymbolCard);
    for I:=fCard + 1 to Length(A) - 1 do
      FillChar(A[I], SizeOf(TTitanSymbolRec), 0);
  end;
end;

procedure TTitanSymbolTable.CheckLength(Value: Integer);
var
  L, I: Integer;
begin
  L := Length(A);
  if value >= L then
  begin
    SetLength(A, value + DeltaSymbolCard);
    for I:=fCard + 1 to Length(A) - 1 do
      FillChar(A[I], SizeOf(TTitanSymbolRec), 0);
  end;
end;

procedure TTitanSymbolTable.SaveToStream(S: TStream; I1, I2: Integer);
var
  I, J, K, L: Integer;
  ClassRec: TTitanClassRec;
  ClassList: TTitanClassList;
  MemberRec: TTitanMemberRec;
  SO: TTitanScriptObject;
  FInstance: TTitanDelegate;
  ClassRecList: TTitanIds;
begin
  ClassList := TTitanBaseScripter(Scripter).ClassList;

  ClassRecList := TTitanIds.Create(false);

  SaveInteger(I1, S);
  SaveInteger(I2, S);
  SaveInteger(MaxLocalVars, S);

  for I:=I1 to I2 do
    if Kind[I] = KindTYPE then
    begin
      ClassRec := ClassList.FindClass(I);
      if ClassRec <> nil then
        ClassRecList.Add(I);
    end;

  for I:=I1 to I2 do
  begin
    S.WriteBuffer(A[I], SizeOF(A[I]));
    SaveString(Name[I], S);
  end;

  SaveInteger(ClassRecList.Count, S);
  for J:=0 to ClassRecList.Count - 1 do
  begin
    I := ClassRecList[J];
    L := Level[I];

    ClassRec := ClassList.FindClass(I);
    ClassRec.SaveToStream(S);
    if Kind[L] = KindTYPE then
    begin
      ClassRec := ClassList.FindClass(L);
      MemberRec := ClassRec.GetMember(NameIndex[I]);
      MemberRec.SaveToStream(S);
    end;
  end;

  ClassRecList.Free;

  for I:=I1 to I2 do
  begin
    K := Kind[I];
    L := Level[I];

    if K = KindVAR then
    begin
      if L = RootNamespaceID then
      begin
        ClassRec := ClassList[0];
        MemberRec := ClassRec.GetMember(NameIndex[I]);
        if MemberRec = nil then
          SaveInteger(0, S)
        else
        begin
          if MemberRec.IsSource then
          begin
            SaveInteger(1, S);
            MemberRec.SaveToStream(S);
          end
          else
            SaveInteger(0, S);
        end;
      end;
    end
    else if K = KindCONST then
    begin
      SaveVariant(GetVariant(I), S);
    end
    else if K = KindSUB then
    begin
      SO := VariantToScriptObject(GetVariant(I));
      FInstance := TTitanDelegate(SO.Instance);
      SaveInteger(FInstance.N, S);

      if L = RootNamespaceID then
      begin
        ClassRec := ClassList[0];
        MemberRec := ClassRec.GetMember(NameIndex[I]);
        MemberRec.SaveToStream(S);
      end;
    end;
  end;
end;

procedure TTitanSymbolTable.LoadFromStream(S: TStream;
                                         DS: Integer = 0; DP: Integer = 0);
var
  I1, I2, I, J, K, L, Temp: Integer;
  St: String;
  ClassRec: TTitanClassRec;
  ClassList: TTitanClassList;
  MemberRec: TTitanMemberRec;

  Shift: Boolean;
begin
  Shift := (DS <> 0) or (DP <> 0);

  ClassList := TTitanBaseScripter(Scripter).ClassList;

  I1 := LoadInteger(S);
  I2 := LoadInteger(S);
  MaxLocalVars := LoadInteger(S);

  ReallocateMem((I2 + MaxLocalVars * 2) * _SizeVariant + DeltaMemSize);

  for I:=I1 to I2 do
  begin
    IncCard;
    S.ReadBuffer(A[Card], SizeOf(A[Card]));
    St := LoadString(S);
    Name[Card] := St;

    AllocateVar(Card);

    if Shift then
    begin
      if A[Card].Level > RootNamespaceID then
        Inc(A[Card].Level, DS);
      if A[Card].Next > RootNamespaceID then
        Inc(A[Card].Next, DS);
      if A[Card].PType > RootNamespaceID then
        Inc(A[Card].PType, DS);
    end;
  end;

  K := LoadInteger(S);
  for J:=1 to K do
  begin
    ClassRec := TTitanClassRec.Create(Scripter, ckClass);
    ClassRec.LoadFromStream(S, DS, DP);
    ClassList.AddObject(ClassRec.ClassID, ClassRec);

    I := ClassRec.ClassID;
    L := Level[I];

    if Kind[L] = KindTYPE then
    begin
      ClassRec := ClassList.FindClass(L);
      MemberRec := TTitanMemberRec.Create(I, ClassRec);
      MemberRec.LoadFromStream(S, DS, DP);

      ClassRec.MemberList.AddObject(NameIndex[I], MemberRec);
    end;
  end;

  for I:=I1 + DS to I2 + DS do
  begin
    K := Kind[I];
    L := Level[I];

    if K = KindVAR then
    begin
      if L = RootNamespaceID then
      begin
        Temp := LoadInteger(S);
        if Temp <> 0 then
        begin
          ClassRec := ClassList[0];
          MemberRec := TTitanMemberRec.Create(I, ClassRec);
          MemberRec.LoadFromStream(S, DS, DP);
          ClassRec.MemberList.AddObject(NameIndex[I], MemberRec);
        end;
      end;
    end
    else if K = KindCONST then
    begin
      PutVariant(I, LoadVariant(S));
    end
    else if K = KindSUB then
    begin
      Temp := LoadInteger(S);
      if Temp > 0 then
        Inc(Temp, DP);

      PutVariant(I, Temp);

      if L = RootNamespaceID then
      begin
        ClassRec := ClassList[0];
        MemberRec := TTitanMemberRec.Create(I, ClassRec);
        MemberRec.LoadFromStream(S, DS, DP);
        ClassRec.MemberList.AddObject(NameIndex[I], MemberRec);
      end;
    end;
  end;
end;

procedure TTitanSymbolTable.Reset;
begin
  FillChar(Mem^, MemSize, 0);
  MemBoundVar := CreateMemBoundVar;

  EraseTail(CreateCard);
  StateStack.Clear;
end;

procedure TTitanSymbolTable.EraseTail(K: Integer);
var
  I: Integer;
  P: Pointer;
begin
  for I:=Card downto K + 1 do
  if Level[I] >= 0 then
  begin
    P := A[I].Address;

    if P <> nil then
    begin
      try
        if IsInsideMemAddress(P) then
           ClearVariant(I);
      finally
        FillChar(A[I], SizeOf(TTitanSymbolRec), 0);
      end;
    end;
  end;
  Card := K;
end;

function TTitanSymbolTable.IsFormalParamID(ID: Integer): Boolean;
var
  I, SubID: Integer;
begin
  result := false;
  if ID <= 0 then
    Exit;
  SubID := GetLevel(ID);
  if GetKind(SubID) <> KindSUB then
    Exit;
  for I:=1 to GetCount(SubID) do
    if GetParamID(SubID, I) = ID then
      result := true;
end;

function TTitanSymbolTable.CodeStringConst(const Val: String): Integer;
var
  I: Integer;
begin
  for I:=PAXTypes.Count + 1 to Card do
    if GetType(I) > 0 then
    if GetKind(I) = kindCONST then
      if VarType(Variant(GetAddr(I)^)) = varString then
        if Val = Variant(GetAddr(I)^) then
        begin
          result := I;
          Exit;
        end;

//  result := AppVariantConst(Val);

  IncCard;

  with A[Card] do
  begin
    PName := 0;
    PType := GeTTitantype(Val);
    Kind := kindCONST;

    Address := Pointer(PtrInt(Mem) + MemBoundVar);
    Inc(MemBoundVar, SizeOf(Variant));
    FillChar(Address^, SizeOf(Variant), 0);
    Variant(Address^) := Val;// StrPas(PChar(Val));

    Module := -1;
    Position := -1;
  end;
  result := Card;
end;

destructor TTitanSymbolTable.Destroy;
begin
  EraseTail(0);

  FreeMem(Mem, MemSize);
  StateStack.Free;

  inherited;
end;

procedure TTitanSymbolTable.SetName(I: Integer; const Name: String);
begin
  A[I].PName := TTitanBaseScripter(Scripter).NameList.Add(Name);
end;


function TTitanSymbolTable.GetName(I: Integer): String;
var
  Index: Integer;
begin
  if I < 0 then
  begin
    result := GlobalDefinitionList.GetName(-I);
    Exit;
  end;

  if I > Card then
  begin
    result := '';
    Exit;
  end;

  Index := A[I].PName;
  if Index > 0 then
  begin
    if Index < NameList.Count then
      result := NameList[Index]
    else
      result := '';
  end
  else
    result := '';
end;

function TTitanSymbolTable.GetFullName(I: Integer): String;
var
  L: Integer;
begin
  if I < 0 then
  begin
    result := GlobalDefinitionList.GetFullName(-I);
    Exit;
  end;
  result := GetName(I);
  L := GetLevel(I);
  if (L > 0) and (L < Card) and (L <> RootNamespaceID) then
    result := GetName(L) + '.' + result;
end;


procedure TTitanSymbolTable.SetLocal(ID: Integer);
begin
  A[ID].Local := 1;
end;

function TTitanSymbolTable.IsLocal(ID: Integer): Boolean;
begin
  result := A[ID].Local > 0;
end;

procedure TTitanSymbolTable.SetNameIndex(I, Value: Integer);
begin
  A[I].PName := Value;
end;

function TTitanSymbolTable.GetNameIndex(I: Integer): Integer;
begin
  if I > 0 then
    result := A[I].PName
  else
  begin
    result := GlobalDefinitionList.GetNameIndex(-I, Scripter);
  end;
end;

procedure TTitanSymbolTable.SetKind(I, AKind: Integer);
begin
  if I > 0 then
    A[I].Kind := AKind;
end;

function TTitanSymbolTable.GetKind(I: Integer): Integer;
begin
  if I > 0 then
    result := A[I].Kind
  else
    result := GlobalDefinitionList.GetKind(-I);
end;

procedure TTitanSymbolTable.SetImported(I: Integer; Value: Boolean);
begin
  if I > 0 then
    A[I].Imported := Value;
end;

function TTitanSymbolTable.GetImported(I: Integer): Boolean;
begin
  if I > 0 then
    result := A[I].Imported
  else
    result := false;
end;

procedure TTitanSymbolTable.SetGlobal(I: Integer; Value: Boolean);
begin
  if I > 0 then
    A[I].Global := Value;
end;

function TTitanSymbolTable.GetGlobal(I: Integer): Boolean;
begin
  if I > 0 then
    result := A[I].Global
  else
    result := false;
end;


function TTitanSymbolTable.GetStrKind(I: Integer): String;
var
  J: Integer;
begin
  J := GetKind(I);
  if (J >= 1) and (J <= PAXKinds.Count - 1) then
    result := PAXKinds[J]
  else
    result := ' ';
end;

procedure TTitanSymbolTable.SetCount(I, ACount: Integer);
begin
  A[I].Count := ACount;
end;

function TTitanSymbolTable.GetCount(I: Integer): Integer;
begin
  if I > 0 then
    result := A[I].Count
  else
    result := GlobalDefinitionList.GetCount(-I);
end;

function TTitanSymbolTable.GetParamTypeName(SubID, ParamIndex: Integer): String;
var
  ParamID, TypeID: Integer;
begin
  if SubID > 0 then
  begin
    ParamID := GetParamID(SubID, ParamIndex);
    TypeID := GetType(ParamID);
    result := GetName(TypeID);
  end
  else
  begin
    result := GlobalDefinitionList.GetParamTypeName(-SubID, ParamIndex);
  end;
end;

function TTitanSymbolTable.GetParamName(SubID, ParamIndex: Integer): String;
var
  ParamID: Integer;
begin
  if SubID > 0 then
  begin
    ParamID := GetParamID(SubID, ParamIndex);
    result := GetName(ParamID);
  end
  else
  begin
    result := GlobalDefinitionList.GetParamName(-SubID, ParamIndex);
  end;
end;

function TTitanSymbolTable.GetTypeName(ID: Integer): String;
var
  TypeID: Integer;
begin
  if ID > 0 then
  begin
    TypeID := GetType(ID);
    result := GetName(TypeID);
  end
  else
  begin
    result := GlobalDefinitionList.GetTypeName(-ID);
  end;
end;

procedure TTitanSymbolTable.SetRank(I, ARank: Integer);
begin
  if I > 0 then
    A[I].Rank := ARank;
end;

function TTitanSymbolTable.GetRank(I: Integer): Integer;
begin
  if I > 0 then
    result := A[I].Rank
  else
    result := 0;
end;

procedure TTitanSymbolTable.SetNext(I, Value: Integer);
begin
  A[I].Next := Value;
end;

function TTitanSymbolTable.GetNext(I: Integer): Integer;
begin
  result := A[I].Next;
end;

procedure TTitanSymbolTable.SetModule(I, Value: Integer);
begin
  A[I].Module := Value;
end;

function TTitanSymbolTable.GetModule(I: Integer): Integer;
begin
  result := A[I].Module;
end;

procedure TTitanSymbolTable.SetPosition(I, Value: Integer);
begin
  A[I].Position := Value;
end;

function TTitanSymbolTable.GetPosition(I: Integer): Integer;
begin
  result := A[I].Position;
end;

procedure TTitanSymbolTable.SetStartPosition(I, Value: Integer);
begin
  A[I].StartPosition := Value;
end;

function TTitanSymbolTable.GetStartPosition(I: Integer): Integer;
begin
  result := A[I].StartPosition;
end;

procedure TTitanSymbolTable.SetLevel(I: Integer; Value: Integer);
begin
  if I > 0 then
    A[I].Level := Value;
end;

function TTitanSymbolTable.GetLevel(I: Integer): Integer;
begin
  if I > 0 then
    result := A[I].Level
  else if I < 0 then
    result := -1
  else
    result := 0;
end;

procedure TTitanSymbolTable.SetCallConv(I, Value: Integer);
begin
  A[I].CallConv := Value;
end;

function TTitanSymbolTable.GetCallConv(I: Integer): Integer;
begin
  result := A[I].CallConv;
end;

procedure TTitanSymbolTable.SetTypeNameIndex(I, Value: Integer);
begin
  A[I].TypeNameIndex := Value;
end;

function TTitanSymbolTable.GetTypeNameIndex(I: Integer): Integer;
begin
  result := A[I].TypeNameIndex;
end;

procedure TTitanSymbolTable.SetType(I, AType: Integer);
begin
  if I > 0 then
  A[I].PType := AType;
end;

function TTitanSymbolTable.GetType(I: Integer): Integer;
begin
  if I > 0 then
    result := A[I].PType
  else
    result := typeVARIANT;
end;

procedure TTitanSymbolTable.SetAddr(I: Integer; Address: Pointer);
begin
  A[I].Address := Address;
end;

function TTitanSymbolTable.GetAddr(I: Integer): Pointer;
begin
  if I > 0 then
    result := A[I].Address
  else if I = 0 then
    result := @Undefined
  else
    result := GlobalDefinitionList.GetAddress(Scripter, -I);
end;

procedure TTitanSymbolTable.SetByRef(ParamID: Integer; Value: Integer);
begin
  A[ParamID].Misc := Value;
end;

function TTitanSymbolTable.GetByRef(ParamID: Integer): Integer;
begin
  result := A[ParamID].Misc;
end;

procedure TTitanSymbolTable.SetTypeSub(SubID: Integer; Value: TTitanTypeSub);
begin
  A[SubID].Misc := Ord(Value);
end;

function TTitanSymbolTable.GetTypeSub(SubID: Integer): TTitanTypeSub;
var
  D: TTitanDefinition;
begin
  result := tsNone;
  if SubID > 0 then
    result := TTitanTypeSub(A[SubID].Misc)
  else
  begin
    D := GlobalDefinitionList[-SubID];
    if D.DefKind = dkMethod then
      result := TTitanMethodDefinition(D).TypeSub;
  end;
end;

function TTitanSymbolTable.GetStrType(I: Integer): String;
var
  J: Integer;
begin
  J := GetType(I);
  if (J >= 1) and (J <= Card) then
    result := GetName(J)
  else if J < 0 then
    result := '-' + GlobalDefinitionList.GetName(-J)
  else
    result := ' ';
end;

function TTitanSymbolTable.GetStrVal(I: Integer): String;
begin
  if not IsInsideMemAddress(GetAddr(I)) then
    result := '***'
  else if GetType(I) > 0 then
    result := ToStr(Scripter, GetVariant(I))
  else
    result := '';
end;

function TTitanSymbolTable.GetSizeOf(I: Integer): Integer;
begin
  result := _SizeVariant;
end;

function TTitanSymbolTable.AppLabel: Integer;
begin
  result := AppVariant(Undefined);
  A[result].Kind := KindLABEL;
end;

function TTitanSymbolTable.GetAlias(ID: Integer): Variant;
begin
  result := CreateAlias(GetAddr(ID));
end;

function TTitanSymbolTable.GetVariant(ID: Integer): Variant;
begin
  if ID > 0 then
    result := GetTerminal(GetAddr(ID))^
  else if ID = 0 then
  begin
  end
  else
    result := GlobalDefinitionList.GetVariant(Scripter, -ID);
end;

procedure TTitanSymbolTable.ClearVariant(ID: Integer);
begin
  VarClear(Variant(GetAddr(ID)^));
end;

procedure TTitanSymbolTable.ClearVariantValue(ID: Integer);
var
  Base: Variant;
  SO: TTitanScriptObject;
begin
  if GetKind(ID) <> kindREF then
  begin
    VarClear(Variant(GetAddr(ID)^));
    Exit;
  end;

  Base := GetVariant(ID);
  if VarIsNull(Base) or VarIsEmpty(Base) then
    raise TTitanScriptFailure.Create(errIncompatibleTypes);

  SO := VariantToScriptObject(Base);
  SO.ClearProperty(NameIndex[ID]);
end;

procedure TTitanSymbolTable.PutVariant(ID: Integer; const Val: Variant);
begin
  if ID > 0 then
    GetTerminal(GetAddr(ID))^ := Val
  else if ID = 0 then
  begin
  end
  else
    GlobalDefinitionList.PutVariant(Scripter, -ID, Val);
end;

function TTitanSymbolTable.AppVariant(const Val: Variant; HasAddress: Boolean = true): Integer;
begin
  CheckMem;

  IncCard;

  with A[Card] do
  begin
    PName := 0;
    PType := GeTTitantype(Val);
    Kind := kindVAR;

    if HasAddress then
    begin
      Address := Pointer(PtrInt(Mem) + MemBoundVar);    // 64bit fix
      Inc(MemBoundVar, SizeOf(Variant));
      FillChar(Address^, SizeOf(Variant), 0);
      Variant(Address^) := Val;
    end
    else
      Address := @Undefined;

    Module := -1;
    Position := -1;
  end;
  result := Card;
end;

function TTitanSymbolTable.AppVariantConst(const Val: Variant; Dup: Boolean = false): Integer;
var
  R: Double;
  I: Integer;
begin
  if Dup = false then
  begin
    result := LookupConstID(Val);
    if result > 0 then
      Exit;
  end;

  CheckMem;

  if VarType(Val) = varByte then
  begin
    I := Val;
    result := AppVariant(I);
  end
  else
    result := AppVariant(Val);

  SetKind(result, KindCONST);
  SetName(result, VarToStr(Val));

  if GetType(result) = typeDOUBLE then
  begin
    R := Frac(Val);
    if R = 0 then
      SetType(result, typeInt64);
  end;
end;

function TTitanSymbolTable.AllocateVar(I: Integer): Pointer;
var
  MemCnt: Integer;
begin
  CheckMem;

  result := ShiftPointer(Mem, MemBoundVar);
  A[I].Address := result;
  MemCnt := GetSizeOf(I);
  Inc(MemBoundVar, MemCnt);

  FillChar(result^, MemCnt, 0);
end;

procedure TTitanSymbolTable.CheckMem;
begin
  if MemBoundVar > MemSize - 256 then
    ReallocateMem(MemSize + DeltaMemSize);
end;

procedure TTitanSymbolTable.ReallocateMem(NewSize: Integer);
var
  I: Integer;
  P, Q, Adr: Pointer;
  V: Variant;
  K: Integer;
begin
  if NewSize = MemSize then
    Exit
  else if NewSize < MemSize then
  begin
//    ReallocMem(Mem, NewSize);
//    MemSize := NewSize;
    Exit;
  end;

  P := AllocMem(NewSize);
  Q := P;

  K := 0;
  for I:=1 to Card do
  begin
    Adr := A[I].Address;
    if (Adr <> nil) and (Adr <> @Undefined)
        and (not TTitanBaseScripter(Scripter).ParamList.HasAddress(Adr)) then
    begin
      V := Variant(Adr^);
      VarClear(Variant(Adr^));
      A[I].Address := Q;
      if not IsUndefined(V) then
        Variant(A[I].Address^) := V;

      Inc(PtrInt(Q), _SizeVariant);

      Inc(K, _SizeVariant);
    end;
  end;

  if K > MemBoundVar then
    MemBoundVar := K;

  FreeMem(Mem, MemSize);
  Mem := P;
  MemSize := NewSize;
end;

function TTitanSymbolTable.CodeNumberConst(Val: Variant): Integer;
var
  I, VT: Integer;
  TempVal: Variant;
begin
  VT := VarType(Val);
  for I:=PAXTypes.Count + 1 to Card do
    if GetType(I) > 0 then
    if GetKind(I) = kindCONST then
    begin
      TempVal := GetVariant(I);
      if VT = VarType(TempVal) then
        if Val = TempVal then
        begin
          result := I;
          Exit;
        end;
    end;

  result := AppVariantConst(Val);
  SetName(result, GetStrVal(result));
end;

function TTitanSymbolTable.LookUpID(const Name: String; aLevel: Integer; UpCase: Boolean = true): Integer;
var
  I, K: Integer;
  S: String;
  B: Boolean;
  SymbolRec: TTitanSymbolRec;
begin
  result := 0;
  for I:=Card downto 1 do
  begin
    SymbolRec := A[I];
    if SymbolRec.PName <> 0 then
    begin
      K := SymbolRec.Kind;
      if (K = KindVAR) or (K = KindSUB) or (K = KindTYPE) or
         (K = KindLABEL) or
         (K = KindHOSTOBJECT) or (K = KindHOSTVAR) or (K = KindHOSTCONST) or (K = KindHOSTINTERFACEVAR)
         then
        if SymbolRec.Level = aLevel then
        begin
          S := GetName(I);

          if UpCase then
            B := StrEql(Name, S)
          else
            B := Name = S;

          if B then
          begin
            result := I;
            Exit;
          end;
        end;
    end;
  end;
end;

function TTitanSymbolTable.LookUpSubID(const Name: String; aLevel: Integer; UpCase: Boolean = true): Integer;
var
  I, K: Integer;
  S: String;
  B: Boolean;
  SymbolRec: TTitanSymbolRec;
begin
  result := 0;
  for I:=Card downto 1 do
  begin
    SymbolRec := A[I];
    if SymbolRec.PName <> 0 then
    begin
      K := SymbolRec.Kind;
      if K = KindSUB then
        if SymbolRec.Level = aLevel then
        begin
          S := GetName(I);

          if UpCase then
            B := StrEql(Name, S)
          else
            B := Name = S;

          if B then
          begin
            result := I;
            Exit;
          end;
        end;
    end;
  end;
end;

function TTitanSymbolTable.LookupConstID(const Value: Variant): Integer;
var
  I, VT: Integer;
begin
  VT := VarType(Value);
  result := 0;
  for I:=1 to Card do
    if GetKind(I) = KindCONST then
      if VarType(Variant(GetAddr(I)^)) = VT then
        if Variant(GetAddr(I)^) = Value then
        begin
          result := I;
          Exit;
        end;
end;

function TTitanSymbolTable.IsOutsideMemAddress(A: Pointer): boolean;
var
  N, B1, B2: PtrInt;
begin
  N  := PtrInt(A);
  B1 := PtrInt(Mem) + MemBoundVar;
  B2 := PtrInt(Mem) + MemSize;
  result := (N >= B1) and (N <= B2);
end;

function TTitanSymbolTable.IsInsideMemAddress(A: Pointer): boolean;
var
  N, B1, B2: PtrInt;
begin
  N  := PtrInt(A);
  B1 := PtrInt(Mem);
  B2 := B1 + MemBoundVar;
  result := (N >= B1) and (N < B2);
end;

function TTitanSymbolTable.IsExternalAddress(A: Pointer): boolean;
var
  N, B1, B2: PtrInt;
begin
  N  := PtrInt(A);
  B1 := PtrInt(Mem);
  B2 := PtrInt(Mem) + MemSize;
  result := (N < B1) or (N > B2);
end;

function TTitanSymbolTable.IsInternalAddress(A: Pointer): boolean;
var
  N, B1, B2: PtrInt;
begin
  N  := PtrInt(A);
  B1 := PtrInt(Mem);
  B2 := PtrInt(Mem) + MemSize - _SizeVariant;
  result := (N >= B1) and (N <= B2);
end;

procedure TTitanSymbolTable.LinkVariables(SubID: Integer; HasResult: Boolean);
var
  I, J, PrevID, FirstCode, ParamID, K, _SubID: Integer;
  Code: TTitanCode;
  Found: Boolean;
  D: Integer;
  Def: TTitanDefinition;
  DefMethod: TTitanMethodDefinition;
begin
  Found := false;

  Code := TTitanBaseScripter(Scripter).Code;
  FirstCode := GetVariant(SubID);

  for I:=1 to Count[SubID] do
  begin
    ParamID := GetParamID(SubID, I);
    if ByRef[ParamID] = 0 then
    begin
      for J:=FirstCode to Code.Card do
      if Code.Prog[J].Res = ParamID then
      begin
        Found := true;
        Break;
      end;
      if not Found then
        SetByRef(ParamID, 2);
    end
    else if ByRef[ParamID] = 2 then
    begin
      for J:=FirstCode to Code.Card do
        if Code.Prog[J].Res = ParamID then
        begin
          Code.Card := J;
          raise TTitanScriptFailure.Create(errLeftSideCannotBeAssignedTo);
        end
        else if (Code.Prog[J].Arg1 = ParamID) and (Code.Prog[J].OP = OP_PUSH) then
        begin
          _SubID := Code.Prog[J].Res;
          K := Code.Prog[J].Arg2;
          if _SubID > 0 then
          begin
            if Kind[_SubID] = KindSUB then
            begin
              ParamID := GetParamID(_SubID, K);
              if GetByRef(ParamID) = 1 then
              begin
                Code.Card := J;
                raise TTitanScriptFailure.Create(errLeftSideCannotBeAssignedTo);
              end;
            end;
          end
          else if _SubID < 0 then
          begin
            Def := GlobalDefinitionList[- _SubID];
            if Def.DefKind = dkMethod then
            begin
              DefMethod := TTitanMethodDefinition(Def);
              if Length(DefMethod.ByRefs) >= K then
                if DefMethod.ByRefs[K-1] = true then
                begin
                  Code.Card := J;
                  raise TTitanScriptFailure.Create(errLeftSideCannotBeAssignedTo);
                end;
            end;
          end;
        end;
    end;
  end;

  if HasResult then
    D := 1
  else
    D := 3;

  K := 0;
  PrevID := SubID;
  for I:=SubID + D to Card do
    if Level[I] = SubID then
    begin
      if Kind[I] = KindVAR then
        if ByRef[I] = 0 then
        begin
          Inc(K);
          Next[PrevID] := I;
          Next[I] := -1;
          PrevID := I;
        end;
   end
//
   else if Level[I] = -1 then
   begin
      if Kind[I] = KindREF then
        if ByRef[I] = 0 then
        begin
          Inc(K);
          Next[PrevID] := I;
          Next[I] := -1;
          PrevID := I;
        end;
   end;
//
  if K = 0 then
    Next[SubID] := 0;

  if K > MaxLocalVars then
    MaxLocalVars := K;
end;

function TTitanSymbolTable.GetResultID(SubID: Integer): Integer;
begin
  result := SubID + 1;
end;

function TTitanSymbolTable.GetThisID(SubID: Integer): Integer;
begin
  result := SubID + 2;
end;

function TTitanSymbolTable.GetParamID(SubID: Integer; N: Integer): Integer;
begin
  result := SubID + 2 + N;
end;

function TTitanSymbolTable.GetDllID(SubID: Integer): Integer;
begin
  result := SubID + 3 + Count[SubID];
end;

function TTitanSymbolTable.GetDllProcID(SubID: Integer): Integer;
begin
  result := GetDllID(SubID) + 1;
end;

procedure TTitanSymbolTable.AllocateSubroutines;
var
  I, LastBound: Integer;
label Start;
begin
  LastBound := MemBoundVar;
  Start:
  for I:=1 to Card do
    if Kind[I] = KindSUB then
    begin
      if MemBoundVar > MemSize - 1024 then
      begin
         ReallocateMem(MemSize + DeltaMemSize);
         MemBoundVar := LastBound;
         goto Start;
      end;
      AllocateSub(I);
    end;
end;

function TTitanSymbolTable.GetSubCount: Integer;
var
  I: Integer;
begin
  result := 0;
  for I:=1 to Card do
    if Kind[I] = KindSUB then
      Inc(Result);
end;

procedure TTitanSymbolTable.AllocateSub(SubID: Integer);
var
  P: Pointer;
  I: Integer;
begin
  I := A[SubID].Next;

  if I = 0 then
    Exit;

   P := Pointer(PtrInt(Mem) + MemBoundVar); // ct9999

{
  Pointer(P^) := A[I].Address;
  Inc(PtrInt(P), SizeOf(Pointer));
}

  FillChar(P^, _SizeVariant, 0);
  Variant(P^) := PtrInt(A[I].Address);
  Inc(PtrInt(P), _SizeVariant);

  repeat
    PtrInt(P^) := 0;
    PtrInt(Pointer(PtrInt(P)+4)^) := 0;
    PtrInt(Pointer(PtrInt(P)+8)^) := 0;
    PtrInt(Pointer(PtrInt(P)+12)^) := 0;

    with A[I] do
    begin
      Address := P;
      I := Next;
    end;
    Inc(PtrInt(P), _SizeVariant);
  until (I = -1);

  MemBoundVar := PtrInt(P) - PtrInt(Mem);
end;

procedure TTitanSymbolTable.DeallocateSub(SubID: Integer);
var
  P: Pointer;
  I: Integer;
begin
  I := A[SubID].Next;

  if I = 0 then
    Exit;

{
  P := Pointer(PtrInt(A[I].Address) - SizeOf(Pointer));
  P := Pointer(P^);
}

  P := Pointer(PtrInt(A[I].Address) - _SizeVariant);
  P := Pointer(PtrInt(Variant(P^)));


  repeat
    with A[I] do
    begin
      with TVarData(Address^) do
        if VType = varString then
        begin
          if (Level = SubID) then
             String(VString) := ''

          else if (Kind = KindVar) and (Level = -1) then
          begin
             String(VString) := '';
          end;

        end;

      Address := P;
      I := Next;
    end;

    Dec(MemBoundVar, _SizeVariant);
    if (I=-1) then
      Break;

    Inc(PtrInt(P), _SizeVariant);
  until false;

{
  Dec(MemBoundVar, SizeOf(Pointer));
}
  Dec(MemBoundVar, _SizeVariant);
end;

function TTitanSymbolTable.IsLocalVar(SubID, ID: Integer): Boolean;
var
  I: Integer;
begin
  result := false;
  I := Next[SubID];
  repeat
    if I = ID then
    if IsLocal(I) then
    begin
      result := true;
      Exit;
    end;
    I := Next[I];
  until I = -1;
end;

procedure TTitanSymbolTable.Erase(I: Integer);
var
  K: Integer;
begin
  K := GetKind(I);
  if (K = KindVar) or (K = KindConst) or (K = 0) then
      ClearVariant(I);
end;

procedure TTitanSymbolTable.SaveState;
begin
  StateStack.Push(MemBoundVar);
  StateStack.Push(Card);
end;

procedure TTitanSymbolTable.RestoreState;
var
  K, I: Integer;
  V: Variant;
begin
  K := Card;
  Card := StateStack.Pop;
  for I := K downto Card + 1 do
  begin
    if TTitanBaseScripter(Scripter).EvalCount = 0 then
    if Kind[I] <> KindREF then
    begin
      V := GetVariant(I);
      if IsObject(V) then
        TTitanBaseScripter(Scripter).ScriptObjectList.RemoveObject(VariantToScriptObject(V));
    end;
    ClearVariant(I);
    FillChar(A[I], SizeOf(A[I]), 0);
  end;

  MemBoundVar := StateStack.Pop;
end;

procedure TTitanSymbolTable.InitRunStage;
begin
  SaveState;
end;

procedure TTitanSymbolTable.ResetRunStage;
var
  I, K, SubID: Integer;
  P, PType, PSub: Pointer;
begin
  RestoreState;
  StateStack.Clear;
  SubID := 0;
  PSub := nil;

  PType := nil;

  for I:=CreateCard to Card do
  begin
    K := GetKind(I);

    if K = KindType then
      PType := GetAddr(I)
    else if K = KindSub then
    begin
      PSub := GetAddr(I);
      SubID := I;
    end;

    if K = KindREF then
      Kind[I] := KindVAR;

    if Kind[I] = KindVAR then
      if GetLevel(I) >= -1 then
      begin
        P := GetAddr(I);
        if (P <> PType) and (P <> PSub) and (I <> GetThisID(SubID)) and
           (not Imported[I]) then
             if (not Global[I]) then
             begin
{
               if IsObject(PVariant(A[I].Address)^) then
               begin
                 SO := VariantToScriptObject(PVariant(A[I].Address)^);
                 SO.RefCount := 1;
               end;
 }
               ClearVariant(I);
             end;
      end;
  end;
end;

procedure TTitanSymbolTable.ResetCompileStage;
var
  K: Integer;
  SO: TTitanScriptObject;
  V: Variant;
  P: Pointer;
begin
  V := GetVariant(RootNamespaceID);
  if IsObject(V) then
    SO := VariantToScriptObject(V)
  else
    SO := nil;

  while Card > FirstSymbolCard do
  begin
    P := Address[Card];
    if IsInternalAddress(P) then
    begin
      K := GetKind(Card);
      if K = KindREF then
      begin
        Kind[Card] := KindVAR;
      end;
      if GetLevel(Card) >= -1 then
      begin
        ClearVariant(Card);
      end;
    end;

    FillChar(A[Card], SizeOf(A[Card]), 0);
    DecCard;
  end;
  if SO <> nil then
    TTitanBaseScripter(Scripter).ScriptObjectList.RemoveObject(SO);
  ClearVariant(RootNamespaceID);

  MemBoundVar := CreateMemBoundVar;
end;

function TTitanSymbolTable.GetMemberAccess(RefID: Integer): TTitanMemberAccess;
begin
  result := TTitanMemberAccess(A[RefID].Misc);
end;

procedure TTitanSymbolTable.SetMemberAccess(RefID: Integer; Value: TTitanMemberAccess);
begin
  A[RefID].Misc := Ord(Value);
end;

function TTitanSymbolTable.GetMemberRec(RefID: Integer; ma: TTitanMemberAccess): TTitanMemberRec;
var
  SO: TTitanScriptObject;
begin
  SO := VariantToScriptObject(GetVariant(RefID));
  result := SO.ClassRec.FindMember(NameIndex[RefID], ma);

  if result <> nil then
    result.CheckAccess;
end;

function TTitanSymbolTable.GetMemberRecEx(RefID: Integer;
                                        ma: TTitanMemberAccess;
                                        var FoundInBaseClass: Boolean;
                                        var SO: TTitanScriptObject): TTitanMemberRec;
var
  Idx: Integer;
begin
  Idx := NameIndex[RefID];
  if RefID > 0 then
    SO := VariantToScriptObject(PVariant(A[RefID].Address)^)
  else
    SO := VariantToScriptObject(GetVariant(RefID));
  result := SO.ClassRec.FindMemberEx(Idx, ma, FoundInBaseClass);

  if result = nil then
    if (Idx = SO.ClassRec.NameIdx) or (Idx = NameIndex_New) then
       result := SO.ClassRec.FindMemberEx(NameIndex_Create, ma, FoundInBaseClass);

  if result <> nil then
    result.CheckAccess;
end;

function TTitanSymbolTable.GetOverloadedSubID(RefID: Integer; ma: TTitanMemberAccess;
                        ParamCount: Integer): Integer;
var
  SO: TTitanScriptObject;
begin
  SO := VariantToScriptObject(GetVariant(RefID));
  result := SO.ClassRec.FindOverloadedSubID(NameIndex[RefID], ma, ParamCount);
end;

function TTitanSymbolTable.AppReference(const Base: Variant; ANameIndex: Integer; ma: TTitanMemberAccess): Integer;
begin
  result := AppVariant(Base);
  NameIndex[result] := ANameIndex;
  Kind[result] := kindREF;
  MemberAccess[result] := ma;
end;

procedure TTitanSymbolTable.SetReference(RefID: Integer; const Base: Variant; ma: TTitanMemberAccess);
//var
//  P: PVariant;
begin
  PutVariant(RefID, Base);
{
   P := A[RefID].Address;
   TVarData(P^).VType := varScriptObject;
   TVarData(P^).VInteger := TVarData(Base).VInteger;
}

  A[RefID].Kind := kindREF;

//  MemberAccess[RefID] := ma;
  A[RefID].Misc := Ord(ma);
end;

function TTitanSymbolTable.GetUserData(ID: Integer): Integer;
var
  MemberRec: TTitanMemberRec;
  D: TTitanDefinition;
  DV: TTitanVariableDefinition;
  DC: TTitanConstantDefinition;
  DOB: TTitanObjectDefinition;
  DI: TTitanInterfaceVarDefinition;
  I: Integer;
  DefList: TTitanDefinitionList;
begin
  result := 0;
  case Kind[ID] of
    KindRef:
    begin
      MemberRec := GetMemberRec(ID, maAny);
      if MemberRec = nil then
        Exit;
      D := MemberRec.Definition;
      if D <> nil then
        result := D.UserData;
    end;
    KindHostVar:
    begin
      I := GetVariant(ID);
      if I > 0 then
      begin
        DefList := GlobalDefinitionList;
        DV := TTitanVariableDefinition(DefList[I]);
      end
      else
      begin
        DefList := TTitanBaseScripter(Scripter).LocalDefinitions;
        DV := TTitanVariableDefinition(DefList.GetRecordByIndex(- I));
      end;
      result := DV.UserData;
    end;
    KindHostConst:
    begin
      I := GetVariant(ID);
      if I > 0 then
      begin
        DefList := GlobalDefinitionList;
        DC := TTitanConstantDefinition(DefList[I]);
      end
      else
      begin
        DefList := TTitanBaseScripter(Scripter).LocalDefinitions;
        DC := TTitanConstantDefinition(DefList.GetRecordByIndex(- I));
      end;
      result := DC.UserData;
    end;
    KindHostObject:
    begin
      I := GetVariant(ID);
      if I > 0 then
      begin
        DefList := GlobalDefinitionList;
        DOB := TTitanObjectDefinition(DefList[I]);
      end
      else
      begin
        DefList := TTitanBaseScripter(Scripter).LocalDefinitions;
        DOB := TTitanObjectDefinition(DefList.GetRecordByIndex(- I));
      end;
      result := DOB.UserData;
    end;
    KindHostInterfaceVar:
    begin
      I := GetVariant(ID);
      if I > 0 then
      begin
        DefList := GlobalDefinitionList;
        DI := TTitanInterfaceVarDefinition(DefList[I]);
      end
      else
      begin
        DefList := TTitanBaseScripter(Scripter).LocalDefinitions;
        DI := TTitanInterfaceVarDefinition(DefList.GetRecordByIndex(- I));
      end;
      result := DI.UserData;
    end
    else
      if ID < 0 then
      begin
        D := GlobalDefinitionList[-ID];
        result := D.UserData;
      end;
  end;
end;

function TTitanSymbolTable.GetAddressEx(ID: Integer): Pointer;
var
  DV: TTitanVariableDefinition;
  DC: TTitanConstantDefinition;
  DOB: TTitanObjectDefinition;
  DI: TTitanInterfaceVarDefinition;
  I: Integer;
  DefList: TTitanDefinitionList;
begin
  case Kind[ID] of
    KindHostVar:
    begin
      I := GetVariant(ID);

      if I > 0 then
      begin
        DefList := GlobalDefinitionList;
        DV := TTitanVariableDefinition(DefList[I]);
      end
      else
      begin
        DefList := TTitanBaseScripter(Scripter).LocalDefinitions;
        DV := TTitanVariableDefinition(DefList.GetRecordByIndex(- I));
      end;

      result := DV.GetAddress;
    end;
    KindHostConst:
    begin
      I := GetVariant(ID);

      if I > 0 then
      begin
        DefList := GlobalDefinitionList;
        DC := TTitanConstantDefinition(DefList[I]);
      end
      else
      begin
        DefList := TTitanBaseScripter(Scripter).LocalDefinitions;
        DC := TTitanConstantDefinition(DefList.GetRecordByIndex(- I));
      end;

      result := DC.GetAddress;
    end;
    KindHostObject:
    begin
      I := GetVariant(ID);

      if I > 0 then
      begin
        DefList := GlobalDefinitionList;
        DOB := TTitanObjectDefinition(DefList[I]);
      end
      else
      begin
        DefList := TTitanBaseScripter(Scripter).LocalDefinitions;
        DOB := TTitanObjectDefinition(DefList.GetRecordByIndex(- I));
      end;

      result := @ DOB.DefValue;
    end;
    KindHostInterfaceVar:
    begin
      I := GetVariant(ID);

      if I > 0 then
      begin
        DefList := GlobalDefinitionList;
        DI := TTitanInterfaceVarDefinition(DefList[I]);
      end
      else
      begin
        DefList := TTitanBaseScripter(Scripter).LocalDefinitions;
        DI := TTitanInterfaceVarDefinition(DefList.GetRecordByIndex(- I));
      end;

      result := @ DI.DefValue;
    end
    else
      result := GetAddr(ID);
  end;
end;

function TTitanSymbolTable.GetValue(ID: Integer): Variant;
var
  Base: Variant;
  SO: TTitanScriptObject;
  DV: TTitanVariableDefinition;
  DC: TTitanConstantDefinition;
  DOB: TTitanObjectDefinition;
  DI: TTitanInterfaceVarDefinition;
  I: Integer;
  DefList: TTitanDefinitionList;
begin
  case Kind[ID] of
    KindRef:
    begin
      Base := GetVariant(ID);
      if VarIsNull(Base) or VarIsEmpty(Base) then
      begin
        TTitanBaseScripter(Scripter).Dump();
        raise TTitanScriptFailure.Create(errIncompatibleTypes);
      end;

      SO := VariantToScriptObject(Base);
      result := SO.GetProperty(NameIndex[ID], 0);
    end;
    KindHostVar:
    begin
      I := GetVariant(ID);

      if I > 0 then
      begin
        DefList := GlobalDefinitionList;
        DV := TTitanVariableDefinition(DefList[I]);
      end
      else
      begin
        DefList := TTitanBaseScripter(Scripter).LocalDefinitions;
        DV := TTitanVariableDefinition(DefList.GetRecordByIndex(- I));
      end;

      result := DV.GetValue(Scripter);
    end;
    KindHostConst:
    begin
      I := GetVariant(ID);

      if I > 0 then
      begin
        DefList := GlobalDefinitionList;
        DC := TTitanConstantDefinition(DefList[I]);
      end
      else
      begin
        DefList := TTitanBaseScripter(Scripter).LocalDefinitions;
        DC := TTitanConstantDefinition(DefList.GetRecordByIndex(- I));
      end;

      result := DC.Value;
    end;
    KindHostObject:
    begin
      I := GetVariant(ID);

      if I > 0 then
      begin
        DefList := GlobalDefinitionList;
        DOB := TTitanObjectDefinition(DefList[I]);
      end
      else
      begin
        DefList := TTitanBaseScripter(Scripter).LocalDefinitions;
        DOB := TTitanObjectDefinition(DefList.GetRecordByIndex(- I));
      end;

      result := DOB.Value;
    end;
    KindHostInterfaceVar:
    begin
      I := GetVariant(ID);

      if I > 0 then
      begin
        DefList := GlobalDefinitionList;
        DI := TTitanInterfaceVarDefinition(DefList[I]);
      end
      else
      begin
        DefList := TTitanBaseScripter(Scripter).LocalDefinitions;
        DI := TTitanInterfaceVarDefinition(DefList.GetRecordByIndex(- I));
      end;

      result := DI.Value;
    end
    else
      result := GetVariant(ID);
    end;
end;

procedure TTitanSymbolTable.PutValue(ID: Integer; const Val: Variant);
var
  Base: Variant;
  SO: TTitanScriptObject;
  DV: TTitanVariableDefinition;
  DOB: TTitanObjectDefinition;
  I: Integer;
  DefList: TTitanDefinitionList;
begin
  case GetKind(ID) of
    KindRef:
    begin
      Base := GetVariant(ID);
      if VarIsNull(Base) or VarIsEmpty(Base) then
        raise TTitanScriptFailure.Create(errIncompatibleTypes);

      SO := VariantToScriptObject(Base);
      SO.PutProperty(NameIndex[ID], Val, 0);
    end;
    KindHostVar:
    begin
      I := GetVariant(ID);

      if I > 0 then
      begin
        DefList := GlobalDefinitionList;
        DV := TTitanVariableDefinition(DefList[I]);
      end
      else
      begin
        DefList := TTitanBaseScripter(Scripter).LocalDefinitions;
        DV := TTitanVariableDefinition(DefList.GetRecordByIndex(- I));
      end;

      DV.SetValue(Scripter, Val);
    end;
    KindHostConst:
    begin
      raise TTitanScriptFailure.Create(Format(errCannotWriteReadOnlyProperty, [GetName(ID)]));
    end;
    KindHostObject:
    begin
      I := GetVariant(ID);

      if I > 0 then
      begin
        DefList := GlobalDefinitionList;
        DOB := TTitanObjectDefinition(DefList[I]);
      end
      else
      begin
        DefList := TTitanBaseScripter(Scripter).LocalDefinitions;
        DOB := TTitanObjectDefinition(DefList.GetRecordByIndex(- I));
      end;

      DOB.Value := Val;
    end;
    KindHostInterfaceVar:
    begin
    end
    else
      PutVariant(ID, Val);
    end;
end;

procedure TTitanSymbolTable.Dump(const FileName: String);
const
  V = '|';
var
  T: Text;
procedure WriteLine(I: Integer);

function StrAddr: String;
begin
  if A[I].Address = nil then
    result := 'nil'
  else
    result := IntToStr(PtrInt(A[I].Address)-PtrInt(Mem));
end;

function StrLocal: String;
begin
  if A[I].Local > 0 then
    result := 'L'
  else
    result := ' ';
end;

begin
  writeln( T, '' );

  if I > Card then
  begin
    write(T, I:6);
    Exit;
  end;

  try
  with A[I] do
  begin
    write(T, I:5, V,
            Norm(GetName(I), 12), V,
            Norm(GetStrKind(I), 3), V,
            StrLocal, V,
            Norm(GetStrType(I), 9), ':', PType:4, V,
            Misc:4, V,
            Count:4, V,
            Rank:2, V,
            Next:5, V,
            Level:5, V,
            Module:2, V,
            Position:5, V,
            Norm(StrAddr, 8), V,
            Owner:5, V,
            GetStrVal(I));
  end;
  except
    I := 0;
  end;
end;

procedure Header;
begin
  writeln(T,'____N', V,
            '        Name', V,
            'Knd', V,
            'L', V,
            '          Type', V,
            'Misc', V,
            ' Cnt', V,
            'Rn', V,
            ' Next', V,
            'Level', V,
            'Md', V,
            '  Pos', V,
            '   Addr', V,
            'Owner', V,
            'Value' );
end;

var
  I: Integer;
begin
  AssignFile(T, FileName);
  Rewrite(T);

  try
    Header;

    for I:=0 to Card do
    begin
      WriteLine(I);
    end;

    writeln(T, '');
    writeln(T, '');
    Header;

    writeln(T, '');
    writeln(T, '');
    write(T, 'CreateCard ', CreateCard);
    writeln(T, '');
    write(T, 'MemBoundVar ', MemBoundVar);
    writeln(T, '');

    writeln(T, '');
    writeln(T, '');
    for I:=0 to NameList.Count - 1 do
    begin
      writeln(T, I:5, ' ', NameList[I]);
    end;

  finally
    CloseFile(T);
  end;
end;

function _GetName(NameIndex: Integer; Scripter: Pointer): String;
begin
  result := TTitanBaseScripter(Scripter).NameList[NameIndex];
end;

procedure TTitanSymbolTable.SetupSubs(StartRecNo: Integer);
var
  ClassRec: TTitanClassRec;
  ClassList: TTitanClassList;
  I: Integer;
  V: Variant;
  D: TTitanMethodDefinition;
  J, N: Integer;
  SO: TTitanScriptObject;
begin
  ClassList := TTitanBaseScripter(Scripter).ClassList;

  if ClassList.DelegateClassRec = nil then
    raise TTitanScriptFailure.Create(Format(errClassIsNotImported, ['Delegate']));

  for I:=StartRecNo to Card do
    if Kind[I] = KindSUB then
    begin
      V := GetVariant(I);

      if not IsObject(V) then
      begin
        D := nil;
        N := V;
        if N = 0 then
        begin
          ClassRec := ClassList.FindClass(Level[I]);
          if ClassRec <> nil then
            for J:=0 to ClassRec.MemberList.Count - 1 do
              if ClassRec.MemberList[J].Kind = KindSUB then
              if StrEql(Name[I], _GetName(ClassRec.MemberList.NameID[J], Scripter)) then
              begin

                D := TTitanMethodDefinition.Create(nil, Name[GetDllProcID(I)], nil,
                                                 Count[I], nil, true);
                if D.LoadFromDll(Scripter, I) then
                  ClassRec.MemberList[J].Definition := D;
              end;
        end;
        SO := TTitanScriptObject.Create(ClassList.DelegateClassRec);
        SO.Instance := TTitanDelegate.Create(Scripter, I, N, D);
        SO.PClass := TTitanDelegate;
        SO.PutProperty(CreateNameIndex('Name', Scripter), Name[I], 0);
        PutVariant(I, ScriptObjectToVariant(SO));
      end;

    end;
end;

function TTitanSymbolTable.TypeCast(TypeID: Integer; const Value: Variant): Variant;
var
  C: TTitanClassRec;
  SO: TTitanScriptObject;
  S: String;
  I: IUnknown;
begin
  if (TypeID > 0) and (TypeID <= PaxTypes.Count) then
  case TypeID of
    typeINTEGER, typeWORD, typeSHORTINT, typeSMALLINT: result := ToInteger(Value);
    typeBOOLEAN: result := ToBoolean(Value);
    //typeSTRING: result := ToString(Value);
    typeSTRING: result := String(Value); //by 9999
    typeDOUBLE, typeCARDINAL, typeINT64: result := ToNumber(Value);
  end
  else
  begin
    result := Value;
    if IsObject(result) then
    begin
      S := GetName(TypeID);
      C := TTitanBaseScripter(Scripter).ClassList.FindClassByName(S);
      if C <> nil then
      begin
        SO := VariantToScriptObject(result);
        SO.ClassRec := C;

        if SO.Intf <> nil then if C.fClassDef <> nil then
        begin
          SO.Intf.QueryInterface(C.fClassDef.guid, I);
          SO.Intf := I;
        end;
      end;
    end;
  end;
end;

function TTitanSymbolTable.IsVirtual(SubId: Integer): Boolean;
var
  L: Integer;
  C: TTitanClassRec;
  M: TTitanMemberRec;
  D: TTitanDefinition;
begin
  result := false;
  if SubId > 0 then
  begin
    if A[SubId].IsVirtual then
    begin
      result := true;
      Exit;
    end;

    L := Level[SubId];
    C := TTitanBaseScripter(Scripter).ClassList.FindClass(L);
    if C <> nil then
    begin
      M := C.MemberList.GetMemberRecByID(SubId);
      if M <> nil then
        result := modVIRTUAL in M.ml;
    end;
  end
  else
  begin
    D := GlobalDefinitionList[-SubId];
    result := modVIRTUAL in D.ml;
  end;
end;

function TTitanSymbolTable.NameList: TTitanNameList;
begin
  result := TTitanBaseScripter(Scripter).NameList;
end;

procedure TTitanSymbolTable.InitGlobalVars;
var
  I, T: Integer;
  V: Variant;
begin
  for I:=1 to Card do
    if Kind[I] = KindVAR then
//      if not IsLocal(I) then
        if Pos('$', Name[I]) = 0 then
        begin
          V := GetVariant(I);
          if VarType(V) = varEmpty then
          begin
            T := PType[I];
            if T = typeINTEGER then
              PutVariant(I, 0)
            else if T = typeDOUBLE then
              PutVariant(I, 0.0)
            else if T = typeSTRING then
              PutVariant(I, '')
            else if T = typeBOOLEAN then
              PutVariant(I, false);
          end;
        end;
end;


function TTitanSymbolTable.GetJSIndex(I: Integer): Boolean;
begin
  if (I >= 1) and (I <= Card) then
    result := A[I].JSIndex
  else
    result := false;
end;

procedure TTitanSymbolTable.SetJSIndex(I: Integer; Value: Boolean);
begin
  if (I >= 1) and (I <= Card) then
    A[I].JSIndex := Value;
end;

