{%mainunit ../BASE_Engine.pas}

{***************************************************
 Package pl_TitanScript.pkg
 This file is part of CodeTyphon Studio
****************************************************}

constructor TTitanObjectList.Create;
begin
  fItems := TList.Create;
end;


procedure TTitanObjectList.Add(X: TObject);
begin
  fItems.Add(X);
end;

procedure TTitanObjectList.Clear;
var
  I: Integer;
begin
  for I:=0 to Count - 1 do
    TObject(fItems[I]).Free;
  fItems.Clear;
end;

destructor TTitanObjectList.Destroy;
begin
  Clear;
  fItems.Free;
  inherited;
end;

procedure TTitanObjectList.Delete(I: Integer);
begin
  TObject(fItems[I]).Free;
  fItems.Delete(I);
end;

function TTitanObjectList.Count: Integer;
begin
  result := fItems.Count;
end;

type
  IntegerArray  = array[0..$effffff] of Integer;
  PIntegerArray = ^IntegerArray;

function IsEqualGUID(const guid1, guid2: TGUID): Boolean;
var
  a, b: PIntegerArray;
begin
  a := PIntegerArray(@guid1);
  b := PIntegerArray(@guid2);
  Result := (a^[0] = b^[0]) and (a^[1] = b^[1]) and (a^[2] = b^[2]) and (a^[3] = b^[3]);
end;


function IntfRefCount(const I: IUnknown): Integer;
begin
  result := I._AddRef - 1;
  I._Release;
end;

function IsEqualBytes(P1, P2: Pointer; L: Integer): Boolean;
var
  I: Integer;
begin
  result := true;
  for I := 0 to L - 1 do
    if Byte(ShiftPointer(P1, I)^) <> Byte(ShiftPointer(P2, I)^) then
    begin
      result := false;
      Exit;
    end;
end;

function IsZeroBytes(P: Pointer; L: Integer): Boolean;
var
  I: Integer;
begin
  result := true;
  for I := 0 to L - 1 do
    if Byte(ShiftPointer(P, I)^) <> 0 then
    begin
      result := false;
      Exit;
    end;
end;

procedure ClearVar(V: PVariant);
begin
  if TVarData(V^).VType = varString then
    VarClear(V^)
  else if (TVarData(V^).VType and varArray) <> 0 then
    VarClear(V^)
  else
    FillChar(V^, SizeOf(TVarData), 0);
  TVarData(V^).VType := varEmpty;
end;

procedure SaveStringToTextFile(const S, FileName: String);
var
  T: TextFile;
begin
  AssignFile(T, FileName);
  Rewrite(T);
  writeln(T, S);
  CloseFile(T);
end;

constructor TOperatorList.Create;
begin
  inherited;
  fItems := TStringList.Create;
end;

destructor TOperatorList.Destroy;
begin
  fItems.Free;
  inherited;
end;

procedure TOperatorList.Add(const Name: String; Op: Integer);
begin
  fItems.AddObject(Name, TObject(Op));
end;

function TOperatorList.GetName(Op: Integer): String;
var
  I: Integer;
begin
  for I:=0 to fItems.Count - 1 do
    if PtrInt(fItems.Objects[I]) = Op then
    begin
      result := fItems[I];
      Exit;
    end;
  result := '';
end;

function TOperatorList.IndexOf(const Name: String): Integer;
begin
  result := fItems.IndexOf(Name);
end;

constructor TTitanAssociativeList.Create(DupYes: Boolean);
begin
  L1 := TTitanIds.Create(DupYes);
  L2 := TTitanIds.Create(DupYes);
end;

destructor TTitanAssociativeList.Destroy;
begin
  L1.Free;
  L2.Free;
  inherited;
end;

procedure TTitanAssociativeList.Clear;
begin
  L1.Clear;
  L2.Clear;
end;

function TTitanAssociativeList.Convert(ID: Integer): Integer;
var
  I: Integer;
begin
  result := ID;
  for I:=0 to Count - 1 do
    if L1[I] = ID then
      result := L2[I];
end;

function TTitanAssociativeList.Count: Integer;
begin
  result := L1.Count;
end;

function TTitanAssociativeList.Add(I1, I2: Integer): Integer;
begin
  result := L1.Add(I1);
  L2.Add(I2);
end;

function TTitanAssociativeList.GetFirst(I: Integer): Integer;
begin
  result := L1[I];
end;

function TTitanAssociativeList.GetSecond(I: Integer): Integer;
begin
  result := L2[I];
end;

constructor TTitanCallRec.Create;
begin
  CallN := 0;
  CallP := 0;
  ArgsN := TTitanIds.Create(false);
  ArgsP := TTitanIds.Create(false);
end;

destructor TTitanCallRec.Destroy;
begin
  ArgsN.Free;
  ArgsP.Free;
  inherited;
end;

function TTitanCallRecList.GetRecord(I: Integer): TTitanCallRec;
begin
  result := TTitanCallRec(Objects[I]);
end;

procedure TTitanCallRecList.AddObject(N: Integer; X: TTitanCallRec);
var
  I: Integer;
begin
  I := IndexOf(N);
  if I = -1 then
    inherited AddObject(N, X)
  else
    X.Free;
end;

function TTitanCallRecList.Top: TTitanCallRec;
begin
  result := TTitanCallRec(Objects[Count - 1]);
end;

constructor TTitanIDRec.Create(ID, N, Pos: Integer);
begin
  Self.ID := ID;
  Self.N := N;
  Self.Pos := Pos;
end;

function TTitanIDRecList.GetRecord(I: Integer): TTitanIDRec;
begin
  result := TTitanIDRec(Objects[I]);
end;

procedure TDefaultParameterRec.SaveToStream(S: TStream);
begin
  SaveInteger(SubID, S);
  SaveInteger(ID, S);
  SaveVariant(Value, S);
end;

procedure TDefaultParameterRec.LoadFromStream(S: TStream);
begin
  SubID := LoadInteger(S);
  ID := LoadInteger(S);
  Value := LoadVariant(S);
end;

destructor TDefaultParameterList.Destroy;
var
  I: Integer;
begin
  for I:=0 to Count - 1 do
    Records[I].Free;
  inherited;
end;

function TDefaultParameterList.GetRecord(I: Integer): TDefaultParameterRec;
begin
  result := TDefaultParameterRec(Items[I]);
end;

procedure TDefaultParameterList.AddParameter(SubID, ID: Integer; const Value: Variant);
var
  R: TDefaultParameterRec;
begin
  R := TDefaultParameterRec.Create;
  R.SubID := SubID;
  R.ID := ID;
  R.Value := Value;
  Add(R);
end;

function TDefaultParameterList.FindFirst(SubID: Integer): Integer;
var
  I: Integer;
begin
  for I:=0 to Count - 1 do
    if Records[I].SubID = SubID then
    begin
      result := I;
      Exit;
    end;
  result := -1;
end;

function TDefaultParameterList.FindNext(I, SubID: Integer): Integer;
begin
  result := -1;
  if I < Count - 1 then
  begin
    Inc(I);
    if Records[I].SubID = SubID then
      result := I;
  end;
end;

procedure TDefaultParameterList.SaveToStream(S: TStream);
var
  R: TDefaultParameterRec;
  I: Integer;
begin
  SaveInteger(Count, S);
  for I:=0 to Count - 1 do
  begin
    R := GetRecord(I);
    R.SaveToStream(S);
  end;
end;

procedure TDefaultParameterList.LoadFromStream(S: TStream);
var
  I, K: Integer;
  R: TDefaultParameterRec;
begin
  for I:=0 to Count - 1 do
    Records[I].Free;
  Clear;
  K := LoadInteger(S);
  for I:=0 to K - 1 do
  begin
    R := TDefaultParameterRec.Create;
    R.LoadFromStream(S);
    Add(R);
  end;
end;

function IsDigits(const S: String): Boolean;
var
  I: Integer;
begin
  result := true;
  for I:=1 to Length(S) do
    if not (S[I] in ['0'..'9']) then
    begin
      result := False;
      Exit;
    end;
end;

function CreateAlias(P: PVariant): Variant;
begin
  TVarData(result).VType := varAlias;
  TVarData(result).VInteger := PtrInt(P);
end;

function IsAlias(P: PVariant): Boolean;
begin
  result := VarType(P^) = varAlias;
end;

function GetTerminal(P: PVariant): PVariant;
begin
  result := P;
  while VarType(result^) = varAlias do
    result := PVariant(TVarData(result^).VInteger);
end;

constructor TTitanVariantStack.Create;
begin
  SetLength(A, FirstStackSize);
  L := Length(A) - 1;

  Card := 0;
end;

destructor TTitanVariantStack.Destroy;
begin
  Clear;
  inherited;
end;

procedure TTitanVariantStack.Clear;
begin
  while Card > 0 do
    Pop;
end;

function TTitanVariantStack.Push(const V: Variant): PVariant;
begin
  if Card = L then
  begin
    SetLength(A, Card + DeltaStackSize);
    L := Length(A) - 1;
  end;

  Inc(Card);
  A[Card] := V;
  result := @A[Card];
end;

function TTitanVariantStack.Top: Variant;
begin
  result := A[Card];
end;

procedure TTitanVariantStack.Pop;
begin
  VarClear(A[Card]);
  Dec(Card);
end;

function GeTTitanType(const V: Variant): Integer;
begin
  case VarType(V) of
    varInteger: result := typeINTEGER;
    varDouble: result := typeDOUBLE;
    varString: result := typeSTRING;
    varBoolean: result := typeBOOLEAN;
  else
    result := typeVARIANT;
  end;
end;

function CompareIntegers(P1, P2: Pointer): Integer;
begin
  if PtrInt(P1) > PtrInt(P2) then
    result := 1
  else if PtrInt(P1) = PtrInt(P2) then
    result := 0
  else
    result := -1;
end;

procedure SortList(L: TList; CompareItems: TCompareItems; I1, I2: Integer);
var
  P: Pointer;
  I, J: Integer;
  Done: Boolean;
begin
  for I:=I1 to I2 - 1 do
  begin
    Done := true;
    for J:=I2 downto I + 1 do
      if CompareItems(L[J], L[J-1]) < 0 then
      begin
        P := L[J];
        L[J] := L[J-1];
        L[J-1] := P;
        Done := false;
      end;
    if Done then
      Exit;
  end;
end;

procedure SaveVariant(const Value: Variant; S: TStream);
var
  VType: Integer;
begin
  VType := VarType(Value);
  SaveInteger(VType, S);
  case VType of
    varString:
      SaveString(Value, S);
    else
      S.WriteBuffer(Value, SizeOf(Variant));
  end;
end;

function LoadVariant(S: TStream): Variant;
var
  VType: Integer;
begin
  VType := LoadInteger(S);
  case VType of
    varString:
      result := LoadString(S);
    else
      S.ReadBuffer(result, SizeOf(Variant));
  end;
end;

function IsValidName(const S: String): Boolean;
begin
  result := S <> '';
  if result then
    result := PosCh('$', S) = 0;
end;

procedure SaveInteger(Value: Integer; S: TStream);
begin
  S.WriteBuffer(Value, SizeOf(Integer));
end;

function LoadInteger(S: TStream): Integer;
begin
  S.ReadBuffer(result, SizeOf(Integer));
end;

procedure SaveString(const Value: String; S: TStream);
var
  W: TWriter;
begin
  W := TWriter.Create(S, 1024);
  W.WriteString(Value);
  W.Free;
end;

function LoadString(S: TStream): String;
var
  R: TReader;
begin
  R := TReader.Create(S, 1024);
  result := R.ReadString;
  R.Free;
end;

constructor TTitanVarList.Create;
begin
  fItems := TList.Create;
end;

destructor TTitanVarList.Destroy;
var
  I: Integer;
begin
  for I:=0 to fItems.Count - 1 do
    FreeMem(fItems[I], _SizeVariant);
  fItems.Free;

  inherited;
end;

procedure TTitanVarList.Delete(Index: Integer);
var
  P: PVariant;
begin
  P := GetAddress(Index);
  FreeMem(P, _SizeVariant);
  fItems.Delete(Index - 1);
end;

function TTitanVarList.Add(const Value: Variant): Integer;
var
  P: Pointer;
begin
  P := AllocMem(_SizeVariant);
  result := fItems.Add(P);
  Variant(P^) := Value;
end;

function TTitanVarList.Get(Index: Integer): Variant;
begin
  result := GetAddress(Index)^;
end;

function TTitanVarList.IndexOf(const Value: Variant): Integer;
var
  I: Integer;
  P: PVariant;
begin
  for I:=1 to fItems.Count do
  begin
    P := GetAddress(I);
    if P^ = Value then
    begin
      result := I;
      Exit;
    end;
  end;
  result := -1;
end;

function TTitanVarList.GetAddress(Index: Integer): PVariant;
var
  P: Pointer;
begin
  while fItems.Count < Index do
  begin
    P := AllocMem(_SizeVariant);
    fItems.Add(P);
  end;
  result := fItems[Index - 1];
end;

function TTitanVarList.GetCount: Integer;
begin
  result := fItems.Count;
end;

function TTitanParamList.HasAddress(P: Pointer): Boolean;
var
  I: Integer;
begin
  for I:=0 to L.Count - 1 do
    if L.Objects[I] = P then
    begin
      result := true;
      Exit;
    end;
  result := false;
end;

constructor TTitanParamList.Create;
begin
  inherited;
  L := TStringList.Create;
end;

destructor TTitanParamList.Destroy;
var
  I: Integer;
  P: Pointer;
begin
  for I:=0 to L.Count - 1 do
  begin
    P := L.Objects[I];
    FreeMem(P, SizeOf(Variant));
  end;
  L.Free;
  inherited;
end;

function TTitanParamList.GetAddress(const ParamName: String): Pointer;
var
  I: Integer;
begin
  I := L.IndexOf(ParamName);
  if I = -1 then
    result := nil
  else
    result := L.Objects[I];
end;

function TTitanParamList.GetParam(const ParamName: String): Variant;
var
  P: Pointer;
begin
  P := GetAddress(ParamName);
  if P = nil then
    result := Undefined
  else
    result := Variant(P^);
end;

procedure TTitanParamList.SetParam(const ParamName: String; const Value: Variant);
var
  P: Pointer;
begin
  P := GetAddress(ParamName);
  if P = nil then
  begin
    P := AllocMem(SizeOf(Variant));
    L.AddObject(ParamName, P);
  end;
  Variant(P^) := Value;
end;

function HashNumber(const S: String): Integer;
var
  I, J: Integer;
  UpS: String;
begin
  if Length(S) = 0 then
  begin
    result := -1;
    Exit;
  end;

  UpS := UpperCase(S);

  I := 0;
  for J:=1 to Length(UpS) do
  begin
    I := I shl 1;
    I := I xor ord(UpS[J]);
  end;
  if I < 0 then I := - I;
  result := I mod MaxHash;
end;

constructor TTitanHashArray.Create;
var
  I: Integer;
begin
  for I:=0 to MaxHash do
    A[I] := TList.Create;
end;

procedure TTitanHashArray.Clear;
var
  I: Integer;
begin
  for I:=0 to MaxHash do
    A[I].Clear;
end;

destructor TTitanHashArray.Destroy;
var
  I: Integer;
begin
  for I:=0 to MaxHash do
    A[I].Free;
end;

procedure TTitanHashArray.AddName(const Name: String; NameIndex, _HashNumber: Integer);
begin
  with A[_HashNumber] do
    if IndexOf(Pointer(NameIndex)) = -1 then
       Add(Pointer(NameIndex));
end;

constructor TTitanHashTable.Create;
var
  I: Integer;
begin
  for I:=0 to MaxHash do
  begin
    Keys[I] := TList.Create;
    Values[I] := TList.Create;
  end;
end;

procedure TTitanHashTable.Clear;
var
  I: Integer;
begin
  for I:=0 to MaxHash do
  begin
    Keys[I].Clear;
    Values[I].Clear;
  end;
end;

destructor TTitanHashTable.Destroy;
var
  I: Integer;
begin
  for I:=0 to MaxHash do
  begin
    Keys[I].Free;
    Values[I].Free;
  end;
end;

procedure TTitanHashTable.Add(Key: Integer; Value: Integer);
var
  H: Integer;
begin
  H := Abs(Key) mod MaxHash;
  Keys[H].Add(Pointer(Key));
  Values[H].Add(Pointer(Value));
end;

function TTitanHashTable.FindValue(Key: Integer; var found: Boolean): Integer;
var
  H: Integer;
begin
  H := Abs(Key) mod MaxHash;
  result := Keys[H].IndexOf(Pointer(Key));
  if result >= 0 then
  begin
    found := true;
    result := PtrInt(Values[H][result]);
  end
  else
    found := false;
end;

procedure TTitanHashTable.DeleteValue(Value: Integer);
var
  I, Idx: Integer;
begin
  for I:=0 to MaxHash do
  begin
    Idx := Values[I].IndexOf(Pointer(Value));
    if Idx >= 0 then
    begin
      Keys[I].Delete(Idx);
      Values[I].Delete(Idx);
    end;
  end;
end;


function SetToVariantArray(Val: Integer; pti: PTypeInfo): Variant;
var
  S: TIntegerSet;
  TypeInfo: PTypeInfo;
  I: Integer;
  L: TStringList;
begin
  L := TStringList.Create;
  try
    Integer(S) := Val;
{$ifdef fpc}
    TypeInfo := GetTypeData(pti)^.CompType;
{$else}
    TypeInfo := GetTypeData(pti)^.CompType^;
{$endif}
    for I := 0 to SizeOf(Integer) * 8 - 1 do
      if I in S then
        L.Add(GetEnumName(TypeInfo, I));
    result := VarArrayCreate([0, L.Count - 1], varVariant);
    for I:=0 to L.Count - 1 do
      result[I] := L[I];
  finally
    L.Free;
  end;
end;

constructor TTitanIniFile.Create(const FileName: String);
begin
  Self.FileName := FileName;
  L := TStringList.Create;
  if FileExists(FileName) then
    L.LoadFromFile(FileName)
  else
    L.SaveToFile(FileName);
end;

destructor TTitanIniFile.Destroy;
var
  A: Integer;
begin
{$IFDEF UNIX}
  L.SaveToFile(FileName);
{$ELSE}
  A := FileGetAttr(FileName);
  if A and faReadOnly = 0 then
    L.SaveToFile(FileName);
{$ENDIF}
  L.Free;
end;

function TTitanIniFile.IndexOf(const Key: String): Integer;
var
  I, P: Integer;
  S: String;
begin
  result := -1;
  for I:=0 to L.Count - 1 do
  begin
    P := Pos('=', L[I]);
    if P >= 0 then
    begin
      S := Copy(L[I], 1, P - 1);
      if S = Key then
      begin
        result := I;
        Exit;
      end;
    end;
  end;
end;

function TTitanIniFile.GetValue(const Key: String): String;
var
  I: Integer;
begin
  result := '';
  I := IndexOf(Key);
  if I >= 0 then
    result := Trim(Copy(L[I], Pos('=', L[I]) + 1, Length(L[I])));
end;

procedure TTitanIniFile.SetValue(const Key, Value: String);
var
  I: Integer;
begin
  I := IndexOf(Key);
  if I >= 0 then
    L[I] := Key + '=' + Value
  else
    L.Add(Key + '=' + Value);
end;

procedure SetBit(var Value: Integer; const Bit: Integer);
begin
  Value := Value or (1 shl Bit);
end;

function TestBit(const Value: Integer; const Bit: Integer): Boolean;
begin
  Result := (Value and (1 shl Bit)) <> 0;
end;

procedure ClearBit(var Value: Integer; const Bit: Integer);
begin
  Value := Value and not (1 shl Bit);
end;

const
  BitsPerInteger  = SizeOf(Integer) * 8;

function BinaryToString(const Value: Integer): string;
var
  I, J: Integer;
  P: PChar;
begin
  SetLength(Result, BitsPerInteger);
  P := PChar(Result) + ((BitsPerInteger - 1) * SizeOf(Char));
  J := Value;
  for I := 0 to BitsPerInteger - 1 do
  begin
    P^ := Chr(48 + (J and $00000001));
    Dec(P);
    J := J shr 1;
  end;
end;

function StringToBinary(const S: String): Integer;
var
  K, I, P: Integer;
  c: Char;
begin
  K := Length(S);
  if K > BitsPerInteger then
    raise Exception.Create(err_Incorrect_string_in_StringToBinary);
  I := K;
  result := 0;
  P := 1;
  while I > 0 do
  begin
    c := S[I];
    case c of
      '1': result := result + P;
      '0': begin end;
     else
       raise Exception.Create(err_Incorrect_string_in_StringToBinary);
     end;
    Dec(I);
    P := P * 2;
  end;
end;

function StringToOctal(const S: String): Integer;
var
  K, I, P: Integer;
  c: Char;
begin
  K := Length(S);
  I := K;
  result := 0;
  P := 1;
  while I > 0 do
  begin
    c := S[I];
    case c of
      '0'..'7': result := result + (ord(S[I]) - ord('0')) * P;
     else
       raise Exception.Create(err_Incorrect_string_in_StringToOctal);
     end;
    Dec(I);
    P := P * 8;
  end;
end;

function ChPos(ch: Char; const S: String): Integer;
var
  I: Integer;
begin
  for I:=1 to Length(S) do
    if S[I] = ch then
    begin
      result := I;
      Exit;
    end;
  result := 0;
end;

function _Shr(X, Y: Integer): Variant;
var
  I, J: Integer;
begin
  J := X shr Y;
  if X >= 0 then
  begin
    result := J;
    Exit;
  end;

  if Y < 0 then
  begin
    result := NaN;
    Exit;
  end;

  for I:=31 downto 32 - Y do
    SetBit(J, I);

  result := J;
end;

procedure AdjustEnum(var Val: Integer);
type
  T = array[1..4] of Byte;
begin
  T(Val)[2] := 0;
  T(Val)[3] := 0;
  T(Val)[4] := 0;
end;

function IsCorrectAddress(Address: Pointer): Boolean;
var
  R: record
       W1, W2: Word;
     end;
begin
  Move(Address, R, SizeOf(Pointer));
  result := (R.W2 <> 0);
end;


{$ifdef fpc}
function _IsDelphiClass(Address: Pointer): Boolean;
begin
  result := true;
end;
{$else}
function _IsDelphiClass(Address: Pointer): Boolean; assembler;
asm
        CMP     Address, Address.vmtSelfPtr
        JNZ     @False
        MOV     Result, True
        JMP     @Exit
@False:
        MOV     Result, False
@Exit:
end;
{$endif}

function IsDelphiClass(Address: Pointer): Boolean;
begin
  if IsCorrectAddress(Address) then
    result := _IsDelphiClass(Address)
  else
    result := false;
end;


{$ifdef fpc}
function _IsDelphiObject(Address: Pointer): Boolean;
begin
  result := true;
end;
{$else}
function _IsDelphiObject(Address: Pointer): Boolean; assembler;
asm
// or IsDelphiClass(Pointer(Address^));
        MOV     EAX, [Address]
        CMP     EAX, EAX.vmtSelfPtr
        JNZ     @False
        MOV     Result, True
        JMP     @Exit
@False:
        MOV     Result, False
@Exit:
end;

{$endif}
  {
function IsDelphiObject(Address: Pointer): Boolean;
var
  VMT, SelfP : pointer;
begin
  Result := FALSE;
  if IsCorrectAddress(Address) then
    try
      VMT := PPointer(Address)^;
      if IsCorrectAddress(VMT) then
      begin
        SelfP := Pointer(DWORD(VMT) + vmtSelfPtr);
        if IsCorrectAddress(SelfP) then
        begin
          SelfP := PPointer(SelfP)^;
          if IsCorrectAddress(SelfP) then
            Result := VMT = SelfP;
        end;
      end;
    except
    end;
end;

}
function IsDelphiObject(Address: Pointer): Boolean;
begin
  if IsCorrectAddress(Address) then
  begin
    try
      result := _IsDelphiObject(Address);
    except

  {$IFNDEF SOLARIS}
    {$IFDEF CPUARM}
       // ct9999 MUST CODE THIS
    asm
    end;
    {$ELSE}

      asm
        MOV   EAX, [Address]
      end;

    {$ENDIF}
  {$ELSE} // NOT SOLARIS
         // ct9999 MUST CODE THIS
    asm
    end;
  {$ENDIF}

      result := false;
    end;
  end
  else
    result := false;
end;


procedure ErrMessageBox(const S:  String);
begin
{$IFDEF USECONSOLE}
  writeln(S);
  Exit;
{$ENDIF}

{$IFDEF MSWINDOWS} //ONLY for WIN32 WIN64
  MessageBox(GetActiveWindow(), PChar(S), PChar('TitanScript'), MB_ICONEXCLAMATION or MB_OK);
{$ENDIF}
{$IFDEF WINCE}
  MessageBox(GetActiveWindow(), PWideChar(S), PWideChar('TitanScript'), MB_ICONEXCLAMATION or MB_OK);
{$ENDIF}

{$IFDEF UNIX}
{$IFDEF USECONSOLE}
  writeln(S);
{$ELSE}
  Application.MessageBox(PChar(S), 'TitanScript', 1); { TODO : Must find 9999 }
{$ENDIF}
{$ENDIF}

end;

function IsVBArray(const V: Variant): boolean;
begin
  result := VarType(V) > varArray;
end;

function IsUndefined(const V: Variant): boolean;
var
  VT: Integer;
begin
  VT := VarType(V);
  result := (VT = varUndefined) or (VT = varNull);
end;

function IsString(const V: Variant): boolean;
begin
  result := VarType(V) = varString;
end;

function IsNumber(const V: Variant): boolean;
var
  VT: Integer;
begin
  VT := VarType(V);
  result := (VT = varInteger) or (VT = varDouble);
end;

function IsBoolean(const V: Variant): boolean;
begin
  result := VarType(V) = varBoolean;
end;

function IsObject(const V: Variant): boolean;
begin
  result := VarType(V) = varScriptObject;
end;

function GetOperName(OP: Integer): String;
begin
  if OP = 0 then
    result := 'UNKNOWN'
  else
    result := PAXOperators[BOUND_OPER - OP];
end;

function ShiftPointer(P: Pointer; D: Integer): Pointer;
begin
  result := Pointer(PtrInt(P) + D);
end;

function StrEql(Const S1, S2: String): Boolean;
begin
  Result := CompareText(S1, S2) = 0;
end;

function Norm(const S: String; L: Integer): String;
begin
  result := Copy(S, 1, L);
  while Length(result) < L do
    result := ' ' + result;
end;

procedure TTitanEntryStack.Push(ABreakLabel, AContinueLabel: Integer;
                           var AStringLabel: String);
var
  EntryRec: TTitanEntryRec;
begin
  EntryRec := TTitanEntryRec.Create;
  with EntryRec do
  begin
    BreakLabel := ABreakLabel;
    ContinueLabel := AContinueLabel;
    StringLabel := AStringLabel;
  end;
  Add(EntryRec);

  AStringLabel := '';
end;

procedure TTitanEntryStack.Pop;
begin
  TTitanEntryRec(Items[Count - 1]).Free;
  Delete(Count - 1);
end;

function TTitanEntryStack.TopBreakLabel(const AStringLabel: String = ''): Integer;
var
  I: Integer;
  R: TTitanEntryRec;
begin
  if AStringLabel <> '' then
  begin
    for I:=Count - 1 downto 0 do
    begin
      R := TTitanEntryRec(Items[I]);
      with R do
        if StringLabel = AStringLabel then
        begin
          result := BreakLabel;
          Exit;
        end;
    end;
    raise TTitanScriptFailure.Create(errLabelIsNotFound);
  end
  else
    with TTitanEntryRec(Items[Count - 1]) do
      result := BreakLabel;
end;

function TTitanEntryStack.TopContinueLabel(const AStringLabel: String = ''): Integer;
var
  I: Integer;
begin
  if AStringLabel <> '' then
  begin
    for I:=Count - 1 downto 0 do
    with TTitanEntryRec(Items[I]) do
      if StringLabel = AStringLabel then
      begin
        result := ContinueLabel;
        Exit;
      end;
    raise TTitanScriptFailure.Create(errLabelIsNotFound);
  end
  else
    with TTitanEntryRec(Items[Count - 1]) do
      result := ContinueLabel;
end;

constructor TTitanIndexedList.Create;
begin
  inherited;
  fItems := TList.Create;
  Objects := TList.Create;
end;

function TTitanIndexedList.Count: Integer;
begin
  result := fItems.Count;
end;

procedure TTitanIndexedList.Clear;
var
  I: Integer;
begin
  for I:=0 to Count - 1 do
    if Objects[I] <> nil then
    begin
      TObject(Objects[I]).Free;
      Objects[I] := nil;
    end;

  fItems.Clear;
  Objects.Clear;
end;

procedure TTitanIndexedList.Delete(I: Integer);
begin
  if Objects[I] <> nil then
    TObject(Objects[I]).Free;
  fItems.Delete(I);
  Objects.Delete(I);
end;

destructor TTitanIndexedList.Destroy;
begin
  Clear;
  fItems.Free;
  Objects.Free;
  inherited;
end;

function TTitanIndexedList.IndexOf(I: Integer): Integer;
var
  J: Integer;
begin
  result := -1;
  for J:=Count - 1 downto 0 do
    if PtrInt(fItems[J]) = I then
    begin
      result := J;
      Exit;
    end;
end;

function TTitanIndexedList.GetNameID(I: Integer): Integer;
begin
  result := PtrInt(fItems[I]);
end;

procedure TTitanIndexedList.SetNameID(I: Integer; Value: Integer);
begin
  fItems[I] := Pointer(Value);
end;

function TTitanIndexedList.AddObject(ID: Integer; AnObject: TObject): Integer;
begin
  result := fItems.Add(Pointer(ID));
  Objects.Add(AnObject);
end;

procedure TTitanIndexedList.DeleteObject(Index: Integer);
begin
  TObject(Objects[Index]).Free;

  Objects.Delete(Index);
  fItems.Delete(Index);
end;

function TTitanIndexedList.GetObject(ID: Integer): TObject;
var
  I: Integer;
begin
  I := IndexOf(ID);
  if I = - 1 then
    result := nil
  else
    result := Objects[I];
end;

constructor TTitanHashedIndexedList.Create;
begin
  inherited;
  HashTable := TTitanHashTable.Create;
end;

destructor TTitanHashedIndexedList.Destroy;
begin
  HashTable.Free;
  inherited;
end;

function TTitanHashedIndexedList.IndexOf(ID: Integer): Integer;
var
  found: boolean;
begin
  result := HashTable.FindValue(ID, found);
end;

function TTitanHashedIndexedList.AddObject(ID: Integer; AnObject: TObject): Integer;
begin
  result := fItems.Add(Pointer(ID));
  Objects.Add(AnObject);

  HashTable.Add(ID, result);
end;

function TTitanHashedIndexedList.GetObject(ID: Integer): TObject;
var
  I: Integer;
begin
  I := IndexOf(ID);
  if I = - 1 then
    result := nil
  else
    result := Objects[I];
end;

procedure TTitanHashedIndexedList.DeleteObject(Index: Integer);
var
  I: Integer;
begin
  TObject(Objects[Index]).Free;
  Objects.Delete(Index);
  fItems.Delete(Index);

  HashTable.Clear;
  for I := 0 to Count - 1 do
    HashTable.Add(PtrInt(fItems[I]), I);
end;

procedure TTitanHashedIndexedList.Clear;
begin
  inherited;
  HashTable.Clear;
end;

function TTitanTypes.AddType(const TypeName: String; TypeSize: Integer): Integer;
begin
  result := AddObject(TypeName, TObject(TypeSize));
end;

function TTitanTypes.GetSize(TypeID: Integer): Integer;
begin
  result := PtrInt(Objects[TypeID]);
end;

function TTitanTypes.GetTypeID(const TypeName: String): Integer;
var
  I: Integer;
begin
  for I:=0 to Count - 1 do
    if StrEql(Strings[I], TypeName) then
    begin
      result := I;
      Exit;
    end;
  result := -1;


end;

constructor TTitanStack.Create;
begin
  SetLength(fItems, FirstStackSize);
  L := Length(fItems) - 1;
  Card := 0;
end;

procedure TTitanStack.Clear;
begin
  Card := 0;
end;

function TTitanStack.GetItem(I: Integer): Integer;
begin
  result := fItems[I];
end;

function TTitanStack.Push(I: Integer): Integer;
begin
  if Card = L then
  begin
    SetLength(fItems, Card + DeltaStackSize);
    L := Length(fItems) - 1;
  end;

  Inc(Card);
  fItems[Card] := I;
  result := I;
end;

function TTitanStack.IndexOf(I: Integer): Integer;
var
  J: Integer;
begin
  for J:=1 to Card do
    if fItems[J] = I then
    begin
      result := J;
      Exit;
    end;
  result := -1;
end;

function TTitanStack.PushUnique(I: Integer): Integer;
begin
  result := I;
  if IndexOf(I) = -1 then
    Push(I);
end;

procedure TTitanStack.Delete(I: Integer);
var
  Index, J: Integer;
begin
  Index := IndexOf(I);
  if Index = -1 then
    Exit;
  for J := Index to Card - 1 do
    fItems[J] := fItems[J+1];
  Dec(Card);
end;

procedure TTitanStack.Push(P: Pointer);
begin
  if Card = L then
  begin
    SetLength(fItems, Card + DeltaStackSize);
    L := Length(fItems) - 1;
  end;

  Inc(Card);
  fItems[Card] := PtrInt(P);
end;

function TTitanStack.Pop: Integer;
begin
  result := fItems[Card];
  Dec(Card);
end;

function TTitanStack.PopPointer: Pointer;
begin
  result := Pointer(fItems[Card]);
  Dec(Card);
end;

function TTitanStack.Top: Integer;
begin
  result := fItems[Card];
end;

procedure TTitanStack.SaveToStream(f: TStream);
begin
  f.Write(Card, SizeOf(Card));
  f.Write(fItems[1], Card*SizeOf(Integer));
end;

procedure TTitanStack.LoadFromStream(f: TStream);
begin
  f.Read(Card, SizeOf(Card));

  SetLength(fItems, Card + DeltaStackSize);
  L := Length(fItems) - 1;

  f.Read(fItems[1], Card*SizeOf(Integer));
end;

function TTitanStack.StackPtr: Pointer;
begin
  if Card > 0 then
    result := @fItems[Card]
  else
    result := nil;
end;

procedure TTitanStack.CopyFrom(S: TTitanStack);
var
  I: Integer;
begin
  SetLength(fItems, S.Card + DeltaStackSize);
  L := Length(fItems) - 1;

  for I:=1 to S.Card do
    fItems[I] := S.fItems[I];
  Card := S.Card;
end;

constructor TTitanFastStack.Create;
begin
  Card := 0;
end;

procedure TTitanFastStack.Clear;
begin
  Card := 0;
end;

function TTitanFastStack.GetItem(I: Integer): Integer;
begin
  result := fItems[I];
end;

function TTitanFastStack.Push(I: Integer): Integer;
begin
  Inc(Card);
  fItems[Card] := I;
  result := I;
end;

procedure TTitanFastStack.Push(P: Pointer);
begin
  Inc(Card);
  fItems[Card] := PtrInt(P);
end;

function TTitanFastStack.Pop: Integer;
begin
  result := fItems[Card];
  Dec(Card);
end;

function TTitanFastStack.PopPointer: Pointer;
begin
  result := Pointer(fItems[Card]);
  Dec(Card);
end;

function TTitanFastStack.Top: Integer;
begin
  result := fItems[Card];
end;

function TTitanFastStack.IndexOf(I: Integer): Integer;
var
  J: Integer;
begin
  for J:=1 to Card do
    if fItems[J] = I then
    begin
      result := J;
      Exit;
    end;
  result := -1;
end;

function TTitanFastStack.PushUnique(I: Integer): Integer;
begin
  result := I;
  if IndexOf(I) = -1 then
    Push(I);
end;

procedure TTitanFastStack.Delete(I: Integer);
var
  Index, J: Integer;
begin
  Index := IndexOf(I);
  if Index = -1 then
    Exit;
  for J := Index to Card - 1 do
    fItems[J] := fItems[J+1];
  Dec(Card);
end;

constructor TTitanNameList.Create;
begin
  inherited Create;
  HashArray := TTitanHashArray.Create;
  inherited Add('');
  NameIndex_Create := Add('Create');
  NameIndex_New := Add('New');
end;

destructor TTitanNameList.Destroy;
begin
  HashArray.Free;
  inherited;
end;

function TTitanNameList.GetSize: Integer;
var
  I: Integer;
begin
  result := 0;
  for I:=0 to Count - 1 do
    result := result + Length(Strings[I]) + 8;
  for I:=0 to MaxHash do
    result := result + HashArray.A[I].Count * 4;
end;

function TTitanNameList.Add(const S: string): Integer;
var
  _HashNumber: Integer;

function IndexOfName: Integer;
var
  I, NameIndex: Integer;
begin
  result := -1;
  for I:=0 to HashArray.A[_HashNumber].Count - 1 do
  begin
    NameIndex := PtrInt(HashArray.A[_HashNumber].Items[I]);
    if NameIndex >= 0 then
    if NameIndex < Count then
    if S = Strings[NameIndex] then
    begin
      result := NameIndex;
      Exit;
    end;
  end;
end;

begin
  if S = '' then
  begin
    result := 0;
    Exit;
  end;

  _HashNumber := HashNumber(S);

  result := IndexOfName;
  if result = -1 then
  begin
    result := inherited Add(S);
    HashArray.AddName(S, result, _HashNumber);
  end;
end;

function ArrayCreate(const P: array of Integer): Variant;
var
  Bounds: array of PtrInt;
  I, J, L: Integer;
begin
  L := Length(P);
  SetLength(Bounds, L + L);
  J := 0;
  for I:=0 to L - 1 do
  begin
    Bounds[J] := 0;
    Bounds[J+1] := P[I];
    Inc(J, 2);
  end;
  result := VarArrayCreate(Bounds, varVariant);
end;

procedure ArrayPut(V: PVariant; const Indexes: array of Integer; const Val: Variant);
var
  I, J, L, K1, K2: Integer;
  P: PVariant;
  R: array[1..100] of Integer;
begin
  L := VarArrayDimCount(V^);

{ if L = 1 then
  begin
    I := Indexes[0];
    V^[I] := Val;
    Exit;
  end; }

  R[1] := 1;
  for I:=1 to L do
  begin
    K1 := VarArrayLowBound(V^, I);
    K2 := VarArrayHighBound(V^, I);

    R[I + 1] := R[I] * (K2 - K1 + 1);
  end;

  J := 0;
  for I:=1 to L do
    J := J + R[I] * Indexes[I-1];

  P := VarArrayLock(V^);
  try
    P := ShiftPointer(P, J * _SizeVariant);
    P^ := Val;
  finally
    VarArrayUnlock(V^);
  end;
end;

function ArrayGet(V: PVariant; const Indexes: array of Integer): PVariant;
var
  I, J, L, K1, K2: Integer;
  P: PVariant;
  R: array[1..100] of Integer;
begin
  L := VarArrayDimCount(V^);

  R[1] := 1;
  for I:=1 to L do
  begin
    K1 := VarArrayLowBound(V^, I);
    K2 := VarArrayHighBound(V^, I);

    R[I + 1] := R[I] * (K2 - K1 + 1);
  end;

  J := 0;
  for I:=1 to L do
    J := J + R[I] * Indexes[I-1];

  P := VarArrayLock(V^);
  try
    P := ShiftPointer(P, J * _SizeVariant);
    result := P;
  finally
    VarArrayUnlock(V^);
  end;
end;

procedure BackUpTextFile(const FileName: String);
var
  S: String;
  T1, T2: TextFile;
  P: Integer;
begin
  P := Pos('.', FileName);
  if P > 0 then
    S := Copy(FileName, 1, P) + '~' + Copy(FileName, P + 1, 3)
  else
    S := FileName + '.~';

  AssignFile(T1, FileName);
  AssignFile(T2, S);
  Reset(T1);
  Rewrite(T2);

  try
    while not EOF(T1) do
    begin
      Readln(T1, S);
      Writeln(T2, S);
    end;
  finally
    Close(T1);
    Close(T2);
  end;
end;

function HasPublishedProperty(C: TClass; const PropName: String; Scripter: Pointer; NeedCheckForForbidden: Boolean = true): boolean;
var
  pti: PTypeInfo;
  ptd: PTypeData;
  pProps: PPropList;
  I, nProps: Integer;
  ppi: PPropInfo;
begin
  result := false;

  if NeedCheckForForbidden then
  if Scripter <> nil then
  if TTitanBaseScripter(Scripter).ForbiddenPublishedProperties.IndexOf(C) >= 0 then
    Exit;

  pti := C.ClassInfo;

  if not Assigned(pti) then
    Exit;

  ptd := GetTypeData(pti);
  nProps := ptd^.PropCount;

  if nProps > 0 then
  begin
    GetMem(pProps, SizeOf(PPropInfo) * nProps);
    GetPropInfos(pti, pProps);
  end
  else
    pProps := nil;

  for I:=0 to nProps - 1 do
  begin
    ppi := pProps[I];
    if StrEql(ppi^.Name, PropName) then
    begin
      result := true;
      Break;
    end;
  end;

  if nProps > 0 then
   FreeMem(pProps, SizeOf(PPropInfo) * nProps);
end;

function HasPublishedPropertyEx(C: TClass; const PropName: String; var ClassType: String; Scripter: Pointer): boolean;
var
  pti: PTypeInfo;
  ptd: PTypeData;
  pProps: PPropList;
  I, nProps: Integer;
  ppi: PPropInfo;
begin
  result := false;

  if Scripter <> nil then
  if TTitanBaseScripter(Scripter).ForbiddenPublishedProperties.IndexOf(C) >= 0 then
    Exit;

  ClassType := '';

  pti := C.ClassInfo;

  if not Assigned(pti) then
    Exit;

  ptd := GetTypeData(pti);
  nProps := ptd^.PropCount;

  if nProps > 0 then
  begin
    GetMem(pProps, SizeOf(PPropInfo) * nProps);
    GetPropInfos(pti, pProps);
  end
  else
    pProps := nil;

  for I:=0 to nProps - 1 do
  begin
    ppi := pProps[I];
    if StrEql(ppi^.Name, PropName) then
    begin
{$ifdef fpc}
      if ppi^.PropType^.Kind = tkClass then
        ClassType := ppi^.PropType^.Name;
{$else}
      if ppi^.PropType^^.Kind = tkClass then
        ClassType := ppi^.PropType^^.Name;
{$endif}

      result := true;
      Break;
    end;
  end;

  if nProps > 0 then
   FreeMem(pProps, SizeOf(PPropInfo) * nProps);
end;

procedure GetPublishedPropertiesEx(AClass: TClass; result: TStrings);
var
  pti: PTypeInfo;
  ptd: PTypeData;
  Loop, nProps: Integer;
  pProps: PPropList;
  ppi: PPropInfo;
  S: String;
begin
  pti := AClass.ClassInfo;
  if pti = nil then Exit;
  ptd := GetTypeData(pti);
  nProps := ptd^.PropCount;
  if nProps > 0 then begin
    GetMem(pProps, SizeOf(PPropInfo) * nProps);
    GetPropInfos(pti, pProps);
  end
  else pProps := nil;
  for Loop:=0 to nProps - 1 do begin
    ppi := pProps[Loop];
    S := ppi^.Name;
    result.Add(S + ':' + ppi^.PropType^.Name);
  end;
  if pProps <> nil then
    FreeMem(pProps, SizeOf(PPropInfo) * nProps);
end;

procedure GetPublishedProperties(AClass: TClass; result: TStrings);
var
  pti: PTypeInfo;
  ptd: PTypeData;
  Loop, nProps: Integer;
  pProps: PPropList;
  ppi: PPropInfo;
  S: String;
begin
  pti := AClass.ClassInfo;
  if pti = nil then Exit;
  ptd := GetTypeData(pti);
  nProps := ptd^.PropCount;
  if nProps > 0 then begin
    GetMem(pProps, SizeOf(PPropInfo) * nProps);
    GetPropInfos(pti, pProps);
  end
  else pProps := nil;
  for Loop:=0 to nProps - 1 do begin
    ppi := pProps[Loop];
    S := ppi^.Name;
    result.Add(UpperCase(S));
  end;
  if pProps <> nil then
    FreeMem(pProps, SizeOf(PPropInfo) * nProps);
end;

function IsPublishedEventName(const S: String): boolean;
begin
  if Length(S) > 2 then
     result := (S[1] = 'O') and (S[2] = 'n') and (S[3] in ['A'..'Z', '0'..'9'])
  else
    result := false;
end;

procedure GetPublishedPropertiesEx2(AClass: TClass; result: TStrings);
var
  pti: PTypeInfo;
  ptd: PTypeData;
  Loop, nProps: Integer;
  pProps: PPropList;
  ppi: PPropInfo;
  S: String;
begin
  pti := AClass.ClassInfo;
  if pti = nil then Exit;
  ptd := GetTypeData(pti);
  nProps := ptd^.PropCount;
  if nProps > 0 then begin
    GetMem(pProps, SizeOf(PPropInfo) * nProps);
    GetPropInfos(pti, pProps);
  end
  else pProps := nil;
  for Loop:=0 to nProps - 1 do begin
    ppi := pProps[Loop];
    S := ppi^.Name;
    if not IsPublishedEventName(S) then
      result.Add(S);
  end;
  if pProps <> nil then
    FreeMem(pProps, SizeOf(PPropInfo) * nProps);
end;

procedure GetPublishedEvents(AClass: TClass; result: TStrings);
var
  pti: PTypeInfo;
  ptd: PTypeData;
  Loop, nProps: Integer;
  pProps: PPropList;
  ppi: PPropInfo;
  S: String;
begin
  pti := AClass.ClassInfo;
  if pti = nil then Exit;
  ptd := GetTypeData(pti);
  nProps := ptd^.PropCount;
  if nProps > 0 then begin
    GetMem(pProps, SizeOf(PPropInfo) * nProps);
    GetPropInfos(pti, pProps);
  end
  else pProps := nil;
  for Loop:=0 to nProps - 1 do begin
    ppi := pProps[Loop];
    S := ppi^.Name;
    if IsPublishedEventName(S) then
      result.Add(S);
  end;
  if pProps <> nil then
    FreeMem(pProps, SizeOf(PPropInfo) * nProps);
end;

procedure GetMethodList(FromClass: TClass; MethodList: TStrings);
type
  PPointer = ^Pointer;
  PMethodRec = ^TMethodRec;
  TMethodRec = packed record
    wSize: Word;
    pCode: Pointer;
    sName: ShortString;
  end;
var
  MethodTable: PChar;
  AClass: TClass;
  MethodRec: PMethodRec;
  wCount: Word;
  nMethod: integer;
begin
  MethodList.Clear;
  AClass := FromClass;
  while (AClass <> nil) do begin
    // Get a pointer to the class's published method table
    MethodTable := PChar( Pointer( PChar( AClass ) + vmtMethodTable )^ );
    if (MethodTable <> nil) then begin
      // Get the count of the methods in the table
      Move(MethodTable^, wCount, 2);
      // Position the MethodRec pointer at the first method in the table
      // (skip over the 2-byte method count)
      MethodRec := PMethodRec(MethodTable + 2);
      // Iterate through all the published methods of this class
      for nMethod := 0 to wCount - 1 do  begin
        // Add the method name and address to the MethodList TStrings
        MethodList.AddObject(MethodRec.sName, MethodRec.pCode);
        // Skip to the next method
        MethodRec := PMethodRec(PChar(MethodRec) + MethodRec.wSize);
      end;
    end;
    // Get the ancestor (parent) class
    AClass := AClass.ClassParent;
  end;
end;

function TTitanCodeList.IndexOfKey(const Key: String): Integer;
var
  I: Integer;
begin
  result := -1;
  for I:=0 to Count - 1 do
    if Records[I].Key = Key then
    begin
      result := I;
      Exit;
    end;
end;

function TTitanCodeList.AddCode(const Key, LanguageName, Code: String): TTitanLanCodeRec;
var
  Index: Integer;
  R: TTitanLanCodeRec;
begin
  Index := IndexOfKey(Key);
  if Index = -1 then
  begin
    R := TTitanLanCodeRec.Create;
    R.Key := Key;
    R.LanguageName := LanguageName;
    R.Code := Code;
    Index := Add(R);
  end;
  result := Records[Index];
end;

function TTitanCodeList.GetRecord(Index: Integer): TTitanLanCodeRec;
begin
  result := TTitanLanCodeRec(Items[Index]);
end;

procedure TTitanCodeList.Clear;
var
  I: Integer;
begin
  for I:=0 to Count - 1 do
    Records[I].Free;

  inherited;
end;

function TTitanCodeList.GetCode(LanguageName: String): String;
var
  I: Integer;
begin
  result := '';
  for I:=0 to Count - 1 do
    if Records[I].LanguageName = LanguageName then
      result := result + Records[I].Code;
end;

constructor TTitanIds.Create(DupYes: Boolean);
begin
  inherited Create;
  Self.DupYes := DupYes;
end;

procedure TTitanIds.SaveToStream(S: TStream);
var
  I: Integer;
begin
  SaveInteger(Count, S);
  for I:=0 to Count - 1 do
    SaveInteger(PtrInt(Items[I]), S);
end;

procedure TTitanIds.LoadFromStream(S: TStream);
var
  I, K: Integer;
begin
  Clear;
  K := LoadInteger(S);
  for I:=0 to K - 1 do
    Add(LoadInteger(S));
end;

function TTitanIds.GetID(Index: Integer): Integer;
begin
  result := PtrInt(Items[Index]);
end;

procedure TTitanIds.PutID(Index: Integer; Value: Integer);
begin
  Items[Index] := Pointer(Value);
end;

function TTitanIds.Add(ID: Integer): Integer;
begin
  if DupYes then
  begin
    result := inherited Add(Pointer(ID));
    Exit;
  end;

  result := IndexOf(ID);
  if result = -1 then
    result := inherited Add(Pointer(ID));
end;

function TTitanIds.IndexOf(ID: Integer): Integer;
begin
  result := inherited IndexOf(Pointer(ID));
end;

procedure AddOPResultType(T1, T2, R: Integer);
begin
  OpResultTypes[T1, T2] := R;
  OPResultTypes[T2, T1] := R;
end;

procedure AddAssResultType(T1, T2: Integer);
begin
  AssResultTypes[T1, T2] := true;
end;

function IsNumericString(const S: String): Boolean;
var
  I: Integer;
begin
  if S = '' then
  begin
    result := false;
    Exit;
  end;

  result := true;
  for I:=1 to Length(S) do
    if not (S[I] in ['0'..'9','.','+','-','e','E']) then
    begin
      result := false;
      Exit;
    end;
end;

function _Floor(X: Extended): Int64;
begin
  result := Trunc(X);
  if Frac(X) < 0 then
    Dec(result);
end;

function DelphiDateTimeToEcmaTime(const AValue: TDateTime): Int64;
var
  T: TTimeStamp;
  D1970: TDateTime;
begin
  D1970 := EncodeDate(1970,1,1);

  T := DateTimeToTimeStamp(AValue);
  Result := (_Floor(AValue) - _Floor(D1970)) * MSecsPerDay + T.Time;
end;

function EcmaTimeToDelphiDateTime(const AValue: Variant): TDateTime;
var
  TimeStamp: TTimeStamp;
  D1970: TDateTime;
begin
  D1970 := EncodeDate(1970,1,1);

  TimeStamp := DateTimeToTimeStamp(D1970);

  TimeStamp.Time := _Floor(AValue) mod MSecsPerDay;
  TimeStamp.Date := TimeStamp.Date + _Floor(AValue) div MSecsPerDay;

  result := TimeStampToDateTime(TimeStamp);
end;


function GetRefCountPtr(const S: String): Pointer;
begin
  if Pointer(S) <> nil then
    result := Pointer(PtrInt(Pointer(S)) - 8)
  else
    result := nil;
end;

function GetRefCount(const S: String): Integer;
begin
  result := PtrInt(GetRefCountPtr(S)^);
end;

function GetSizePtr(const S: String): Pointer;
begin
  if Pointer(S) <> nil then
    result := Pointer(PtrInt(Pointer(S)) - 4)
  else
    result := nil;
end;

constructor TTitanAssocList.Create;
begin
  fItems := TList.Create;
end;

destructor TTitanAssocList.Destroy;
begin
  Clear;
  fItems.Free;
end;

function TTitanAssocList.GetItem(I: Integer): TTitanAssocRec;
begin
  result := TTitanAssocRec(fItems[I]);
end;

function CompareTVarRecs(const R1, R2: TVarRec): Boolean;
begin
  result := false;
  if R1.VType <> R2.VType then
    Exit;
  case R1.VType of
    vtInteger: result := R1.VInteger = R2.VInteger;
    vtBoolean: result := R1.VBoolean = R2.VBoolean;
    vtChar: result := R1.VChar = R2.VChar;
    vtExtended: result := R1.VExtended = R2.VExtended;
    vtString: result := R1.VString = R2.VString;
    vtPointer: result := R1.VPointer = R2.VPointer;
    vtPChar: result := R1.VPChar = R2.VPChar;
    vtObject: result := R1.VObject = R2.VObject;
    vtClass: result := R1.VClass = R2.VClass;
    vtWideChar: result := R1.VWideChar = R2.VWideChar;
    vtPWideChar: result := R1.VPWideChar = R2.VPWideChar;
    vtAnsiString: result := R1.VAnsiString = R2.VAnsiString;
    vtCurrency: result := R1.VCurrency = R2.VCurrency;
    vtVariant: result := R1.VVariant = R2.VVariant;
    vtInterface: result := R1.VInterface = R2.VInterface;
    vtWideString: result := R1.VWideString = R2.VWideString;
    vtInt64: result := R1.VInt64 = R2.VInt64;
  end;
end;

function TTitanAssocList.FindObject(const Data: TVarRec): TObject;
var
  I: Integer;
  R: TTitanAssocRec;
begin
  for I:=0 to fItems.Count - 1 do
  begin
    R := Items[I];
    if CompareTVarRecs(R.Data, Data) then
    begin
      result := R.Obj;
      Exit;
    end;
  end;
  result := nil;
end;

function TTitanAssocList.Add(const Data: TVarRec; Obj: TObject): Integer;
var
  R: TTitanAssocRec;
begin
  R := TTitanAssocRec.Create;
  R.Data := Data;
  R.Obj := Obj;
  result := fItems.Add(R);
end;

procedure TTitanAssocList.Clear;
var
  I: Integer;
  R: TTitanAssocRec;
begin
  for I:=0 to fItems.Count - 1 do
  begin
    R := Items[I];
    R.Obj.Free;
    R.Free;
  end;
  fItems.Clear;
end;

function GetImplementorOfInterface(const I: IUnknown): TObject;
{ TODO -cDOC : Original code by Hallvard Vassbotn }
{ TODO -cTesting : Check the implemetation for any further version of compiler }
const
  AddByte = $04244483; // opcode for ADD DWORD PTR [ESP+4], Shortint
  AddLong = $04244481; // opcode for ADD DWORD PTR [ESP+4], Longint
type
  PAdjustSelfThunk = ^TAdjustSelfThunk;
  TAdjustSelfThunk = packed record
    case AddInstruction: Longint of
      AddByte: (AdjustmentByte: ShortInt);
      AddLong: (AdjustmentLong: Longint);
  end;
  PInterfaceMT = ^TInterfaceMT;
  TInterfaceMT = packed record
    QueryInterfaceThunk: PAdjustSelfThunk;
  end;
  TInterfaceRef = ^PInterfaceMT;
var
  QueryInterfaceThunk: PAdjustSelfThunk;
begin
  try
    Result := Pointer(I);
    if Assigned(Result) then
    begin
      QueryInterfaceThunk := TInterfaceRef(I)^.QueryInterfaceThunk;
      case QueryInterfaceThunk.AddInstruction of
        AddByte:
          Inc(PChar(Result), QueryInterfaceThunk.AdjustmentByte);
        AddLong:
          Inc(PChar(Result), QueryInterfaceThunk.AdjustmentLong);
      else
        Result := nil;
      end;
    end;
  except
    Result := nil;
  end;
end;

function PosCh(ch: Char; const S: String): Integer;
var
  I: integer;
begin
  for I:=1 to Length(S) do
    if ch = S[I] then
    begin
      result := I;
      Exit;
    end;
  result := 0;
end;


procedure Initialization_BASE_SYS;
begin
  Initialization_BASE_SYNC;

  PAXTypes := TTitanTypes.Create;
  with PAXTypes do
  begin
    AddType('VOID', 0);
    AddType('VARIANT', SizeOf(Variant));
    AddType('OLEVARIANT', SizeOf(TOleVariant));
    AddType('BYTE', SizeOf(TByte));
    AddType('CHAR', SizeOf(TChar));
    AddType('BOOLEAN', SizeOf(TBoolean));
    AddType('WORDBOOL', SizeOf(TWordBool));
    AddType('LONGBOOL', SizeOf(TLongBool));
    AddType('INTEGER', SizeOf(TInteger));
    AddType('CARDINAL', SizeOf(TCardinal));
    AddType('POINTER', SizeOf(TPointer));
    AddType('DOUBLE',    SizeOf(TDouble));
    AddType('STRING',  SizeOf(TString));
    AddType('#ENUMERATION', SizeOf(TInteger));
    AddType('RECORD',  -1);
    AddType('ARRAY', -1);
    AddType('SHORTSTRING',-1);
    AddType('TEXTFILE',SizeOf(TTextFile));
    AddType('#SUBRANGE', SizeOf(TInteger));
    AddType('SET', -1);
    AddType('CLASS',  SizeOf(TPointer));
    AddType('#CLASSREF',  SizeOf(TPointer));
    AddType('#DYNAMICARRAY', SizeOf(TPointer));
    AddType('PCHAR', SizeOf(Pointer));
    AddType('WORD',  SizeOf(TWord));
    AddType('SHORTINT', SizeOf(TShortInt));
    AddType('SMALLINT', SizeOf(TSmallInt));
    AddType('INT64',  SizeOf(TInt64));
    AddType('SINGLE', SizeOf(TSingle));
    AddType('CURRENCY', SizeOf(TCurrency));
    AddType('COMP', SizeOf(TComp));
    AddType('REAL48', SizeOf(TReal48));
    AddType('EXTENDED', SizeOf(TExtended));
    AddType('INTERFACE', SizeOf(Integer));
    AddType('#METHOD', SizeOf(TMethod));
    AddType('FILE', SizeOf(TFile));
    AddType('WIDECHAR',  SizeOf(TWideChar));
    AddType('WIDESTRING',  SizeOf(TWideString));
    AddType('PWIDECHAR', SizeOf(PWideChar));
    AddType('TVarRec', SizeOf(TVarRec));
  end;

  PAXKinds := TStringList.Create;
  with PAXKinds do
  begin
    Add('NONE');
    Add('VAR');
    Add('TYPE');
    Add('CONST');
    Add('SUB');
    Add('REF');
    Add('PROP');
    Add('LABEL');
    Add('HVAR');
    Add('HCONST');
    Add('HOBJECT');
    Add('HINTF');
    Add('VOBJECT');
  end;

  PAXOperators := TTitanOperators.Create;
  with PAXOperators do
  begin
    OP_SEPARATOR := BOUND_OPER - Add('SEPARATOR');
    OP_NOP := BOUND_OPER - Add('NOP');
    OP_SKIP := BOUND_OPER - Add('SKIP');
    OP_HALT := BOUND_OPER - Add('HALT');
    OP_HALT_GLOBAL := BOUND_OPER - Add('HALT GLOBAL');
    OP_HALT_OR_NOP := BOUND_OPER - Add('HALT OR NOP');
    OP_PRINT := BOUND_OPER - Add('PRINT');
    OP_PRINT_HTML := BOUND_OPER - Add('PRINT HTML');

    OP_GO := BOUND_OPER - Add('GO');
    OP_GO_FALSE := BOUND_OPER - Add('GO_FALSE');
    OP_GO_FALSE_EX := BOUND_OPER - Add('GO_FALSE(EX)');
    OP_GO_TRUE := BOUND_OPER - Add('GO_TRUE');
    OP_GO_TRUE_EX := BOUND_OPER - Add('GO_TRUE(EX)');

    OP_SET_LABEL := BOUND_OPER - Add('SET LABEL');

    OP_BEGIN_METHOD := BOUND_OPER - Add('BEGIN METHOD');
    OP_CREATE_ARRAY := BOUND_OPER - Add('CREATE ARRAY');
    OP_CREATE_DYNAMIC_ARRAY_TYPE := BOUND_OPER - Add('CREATE DYNAMIC ARRAY TYPE');
    OP_GET_FIELD := BOUND_OPER - Add('GET FIELD');

    OP_CREATE_SHORT_STRING := BOUND_OPER - Add('CREATE SHORT STRING');

    OP_DO_NOT_DESTROY := BOUND_OPER - Add('DO NOT DESTROY');

    OP_CREATE_OBJECT := BOUND_OPER - Add('CREATE OBJECT');
    OP_CREATE_RESULT := BOUND_OPER - Add('CREATE RESULT');
    OP_CHECK_CLASS := BOUND_OPER - Add('CHECK_CLASS');
    OP_SET_TYPE := BOUND_OPER - Add('SET_TYPE');
    OP_DESTROY_TEMP_OBJECTS := BOUND_OPER - Add('DESTROY TEMP OBJECTS');
    OP_DESTROY_HOST := BOUND_OPER - Add('DESTROY HOST');
    OP_DESTROY_OBJECT := BOUND_OPER - Add('DESTROY OBJECT');
    OP_DESTROY_LOCAL_VAR := BOUND_OPER - Add('DESTROY LOCAL VAR');
    OP_DESTROY_INTF := BOUND_OPER - Add('DESTROY INTF');
    OP_RELEASE := BOUND_OPER - Add('RELEASE');
    OP_CREATE_REF := BOUND_OPER - Add('CREATE REF');
    OP_USE_NAMESPACE := BOUND_OPER - Add('USE NAMESPACE');
    OP_END_OF_NAMESPACE := BOUND_OPER - Add('END OF NAMESPACE');
    OP_BEGIN_WITH := BOUND_OPER - Add('BEGIN WITH');
    OP_END_WITH := BOUND_OPER - Add('END WITH');
    OP_EVAL_WITH := BOUND_OPER - Add('EVAL WITH');
    OP_IN := BOUND_OPER - Add('IN');
    OP_IN_SET := BOUND_OPER - Add('IN SET');
    OP_INSTANCEOF := BOUND_OPER - Add('INSTANCEOF');
    OP_TYPEOF := BOUND_OPER - Add('TYPEOF');
    OP_GET_NEXT_PROP := BOUND_OPER - Add('GET NEXT PROP');
    OP_SAVE_RESULT := BOUND_OPER - Add('SAVE RESULT');
    OP_GET_ANCESTOR_NAME := BOUND_OPER - Add('GET ANCESTOR NAME');

    OP_PUSH := BOUND_OPER - Add('PUSH');
    OP_PUT_PROPERTY := BOUND_OPER - Add('PUT PROPERTY');
    OP_CALL := BOUND_OPER - Add('CALL');
    OP_CALL_CONSTRUCTOR := BOUND_OPER - Add('CALL CONSTRUCTOR');
    OP_TYPE_CAST := BOUND_OPER - Add('TYPE CAST');
    OP_RET := BOUND_OPER - Add('RET');
    OP_EXIT0 := BOUND_OPER - Add('EXIT0');
    OP_EXIT := BOUND_OPER - Add('EXIT');
    OP_RETURN := BOUND_OPER - Add('RETURN');
    OP_GET_PARAM_COUNT := BOUND_OPER - Add('GET PARAM COUNT');
    OP_GET_PARAM := BOUND_OPER - Add('GET PARAM');
    OP_GET_PUBLISHED_PROPERTY := BOUND_OPER - Add('GET PUBLISHED PROPERTY');
    OP_PUT_PUBLISHED_PROPERTY := BOUND_OPER - Add('PUT PUBLISHED PROPERTY');
    OP_RET_OPERATOR := BOUND_OPER - Add('RET OPERATOR');

    OP_GET_ITEM := BOUND_OPER - Add('GET ITEM');
    OP_PUT_ITEM := BOUND_OPER - Add('PUT ITEM');
    OP_GET_ITEM_EX := BOUND_OPER - Add('GET ITEM(ex)');
    OP_PUT_ITEM_EX := BOUND_OPER - Add('PUT ITEM(ex)');

    OP_GET_STRING_ELEMENT := BOUND_OPER - Add('GET STRING ELEMENT');
    OP_PUT_STRING_ELEMENT := BOUND_OPER - Add('PUT STRING ELEMENT');

    OP_FINALLY := BOUND_OPER - Add('FINALLY');
    OP_CATCH := BOUND_OPER - Add('CATCH');
    OP_TRY_ON := BOUND_OPER - Add('TRY_ON');
    OP_TRY_OFF := BOUND_OPER - Add('TRY_OFF');
    OP_THROW := BOUND_OPER - Add('THROW');
    OP_DISCARD_ERROR := BOUND_OPER - Add('DISCARD ERROR');
    OP_EXIT_ON_ERROR := BOUND_OPER - Add('EXIT ON ERROR');

    OP_ASSIGN := BOUND_OPER - Add(':=');
    OP_ASSIGN_ADDRESS := BOUND_OPER - Add('ASSIGN ADDRESS');
    OP_GET_TERMINAL := BOUND_OPER - Add('GET TERMINAL');
    OP_ASSIGN_RESULT := BOUND_OPER - Add(':=(RES)');

    OP_AND := BOUND_OPER - Add('AND');
    OP_OR := BOUND_OPER - Add('OR');
    OP_XOR := BOUND_OPER - Add('XOR');
    OP_NOT := BOUND_OPER - Add('NOT');

    OP_LEFT_SHIFT := BOUND_OPER - Add('LEFT SHIFT');
    OP_LEFT_SHIFT_EX := BOUND_OPER - Add('LEFT SHIFT(ex)');

    OP_RIGHT_SHIFT := BOUND_OPER - Add('RIGHT SHIFT');
    OP_RIGHT_SHIFT_EX := BOUND_OPER - Add('RIGHT SHIFT(ex)');

    OP_UNSIGNED_RIGHT_SHIFT := BOUND_OPER - Add('UNSIGNED RIGHT SHIFT');
    OP_UNSIGNED_RIGHT_SHIFT_EX := BOUND_OPER - Add('UNSIGNED RIGHT SHIFT(ex)');

    OP_PLUS := BOUND_OPER - Add('+');
    OP_PLUS_EX := BOUND_OPER - Add('+(ex)');

    OP_MINUS := BOUND_OPER - Add('-');
    OP_MINUS_EX := BOUND_OPER - Add('-(ex)');

    OP_UNARY_PLUS := BOUND_OPER - Add('+(unary)');

    OP_UNARY_MINUS := BOUND_OPER - Add('-(unary)');
    OP_UNARY_MINUS_EX := BOUND_OPER - Add('-(unary ex)');

    OP_MULT := BOUND_OPER - Add('*');
    OP_MULT_EX := BOUND_OPER - Add('*(ex)');

    OP_DIV := BOUND_OPER - Add('/');
    OP_DIV_EX := BOUND_OPER - Add('/(ex)');

    OP_INT_DIV := BOUND_OPER - Add('DIV');

    OP_MOD := BOUND_OPER - Add('MOD');
    OP_MOD_EX := BOUND_OPER - Add('MOD(ex)');

    OP_POWER := BOUND_OPER - Add('POWER');

    OP_LT := BOUND_OPER - Add('<');
    OP_LT_EX := BOUND_OPER - Add('<(ex)');

    OP_GT := BOUND_OPER - Add('>');
    OP_GT_EX := BOUND_OPER - Add('>(ex)');

    OP_EQ := BOUND_OPER - Add('=');
    OP_EQ_EX := BOUND_OPER - Add('=(ex)');

    OP_NE := BOUND_OPER - Add('<>');
    OP_NE_EX := BOUND_OPER - Add('<>(ex)');

    OP_ID := BOUND_OPER - Add('===');
    OP_ID_EX := BOUND_OPER - Add('===(ex)');

    OP_NI := BOUND_OPER - Add('!==');
    OP_NI_EX := BOUND_OPER - Add('!==(ex)');

    OP_GE := BOUND_OPER - Add('>=');
    OP_GE_EX := BOUND_OPER - Add('>=(ex)');

    OP_LE := BOUND_OPER - Add('<=');
    OP_LE_EX := BOUND_OPER - Add('<=(ex)');

    OP_IS := BOUND_OPER - Add('IS');
    OP_AS := BOUND_OPER - Add('AS');


    OP_TO_INTEGER := BOUND_OPER - Add('TO INTEGER');
    OP_TO_STRING := BOUND_OPER - Add('TO STRING');
    OP_TO_BOOLEAN := BOUND_OPER - Add('TO BOOLEAN');

    OP_DEFINE := BOUND_OPER - Add('DEFINE');

    OP_DECLARE_ON := BOUND_OPER - Add('DECLARE ON');
    OP_DECLARE_OFF := BOUND_OPER - Add('DECLARE OFF');

    OP_UPCASE_ON := BOUND_OPER - Add('UPCASE ON');
    OP_UPCASE_OFF := BOUND_OPER - Add('UPCASE OFF');

    OP_OPTIMIZATION_ON := BOUND_OPER - Add('OPTIM ON');
    OP_OPTIMIZATION_OFF := BOUND_OPER - Add('OPTIM OFF');

    OP_JS_OPERS_ON := BOUND_OPER - Add('JS OPERS ON');
    OP_JS_OPERS_OFF := BOUND_OPER - Add('JS OPERS OFF');

    OP_ZERO_BASED_STRINGS_ON := BOUND_OPER - Add('ZERO ON');
    OP_ZERO_BASED_STRINGS_OFF := BOUND_OPER - Add('ZERO OFF');

    OP_VBARRAYS_ON := BOUND_OPER - Add('VBARRAYS ON');
    OP_VBARRAYS_OFF := BOUND_OPER - Add('VBARRAYS OFF');

    OP_USE_LANGUAGE_NAMESPACE := BOUND_OPER - Add('USE LANG');

//-------------------------------------------------------
    FOP_INC1 := BOUND_OPER - Add('INC(FOP 1)');
    FOP_INC2 := BOUND_OPER - Add('INC(FOP 2)');

    FOP_PLUS_INTEGER1 := BOUND_OPER - Add('+(FOP INTEGER 1)');
    FOP_PLUS_INTEGER2 := BOUND_OPER - Add('+(FOP INTEGER 2)');
    FOP_PLUS_DOUBLE1 := BOUND_OPER - Add('+(FOP DOUBLE 1)');
    FOP_PLUS_DOUBLE2 := BOUND_OPER - Add('+(FOP DOUBLE 2)');
    FOP_PLUS_STRING1 := BOUND_OPER - Add('+(FOP STRING 1)');
    FOP_PLUS_STRING2 := BOUND_OPER - Add('+(FOP STRING 2)');

    FOP_MINUS_INTEGER1 := BOUND_OPER - Add('-(FOP INTEGER 1)');
    FOP_MINUS_INTEGER2 := BOUND_OPER - Add('-(FOP INTEGER 2)');
    FOP_MINUS_DOUBLE1 := BOUND_OPER - Add('-(FOP DOUBLE 1)');
    FOP_MINUS_DOUBLE2 := BOUND_OPER - Add('-(FOP DOUBLE 2)');

    FOP_MULT_INTEGER1 := BOUND_OPER - Add('*(FOP INTEGER 1)');
    FOP_MULT_INTEGER2 := BOUND_OPER - Add('*(FOP INTEGER 2)');
    FOP_MULT_DOUBLE1 := BOUND_OPER - Add('*(FOP DOUBLE 1)');
    FOP_MULT_DOUBLE2 := BOUND_OPER - Add('*(FOP DOUBLE 2)');

    FOP_DIV_INTEGER1 := BOUND_OPER - Add('/(FOP INTEGER 1)');
    FOP_DIV_INTEGER2 := BOUND_OPER - Add('/(FOP INTEGER 2)');
    FOP_DIV_DOUBLE1 := BOUND_OPER - Add('/(FOP DOUBLE 1)');
    FOP_DIV_DOUBLE2 := BOUND_OPER - Add('/(FOP DOUBLE 2)');

    FOP_MOD1 := BOUND_OPER - Add('MOD(FOP 1)');
    FOP_MOD2 := BOUND_OPER - Add('MOD(FOP 2)');

    FOP_LT_INTEGER1 := BOUND_OPER - Add('<(FOP INTEGER 1)');
    FOP_LT_INTEGER2 := BOUND_OPER - Add('<(FOP INTEGER 2)');
    FOP_LT_DOUBLE1 := BOUND_OPER - Add('<(FOP DOUBLE 1)');
    FOP_LT_DOUBLE2 := BOUND_OPER - Add('<(FOP DOUBLE 2)');

    FOP_LE_INTEGER1 := BOUND_OPER - Add('<=(FOP INTEGER 1)');
    FOP_LE_INTEGER2 := BOUND_OPER - Add('<=(FOP INTEGER 2)');
    FOP_LE_DOUBLE1 := BOUND_OPER - Add('<=(FOP DOUBLE 1)');
    FOP_LE_DOUBLE2 := BOUND_OPER - Add('<=(FOP DOUBLE 2)');

    FOP_GT_INTEGER1 := BOUND_OPER - Add('>(FOP INTEGER 1)');
    FOP_GT_INTEGER2 := BOUND_OPER - Add('>(FOP INTEGER 2)');
    FOP_GT_DOUBLE1 := BOUND_OPER - Add('>(FOP DOUBLE 1)');
    FOP_GT_DOUBLE2 := BOUND_OPER - Add('>(FOP DOUBLE 2)');

    FOP_GE_INTEGER1 := BOUND_OPER - Add('>=(FOP INTEGER 1)');
    FOP_GE_INTEGER2 := BOUND_OPER - Add('>=(FOP INTEGER 2)');
    FOP_GE_DOUBLE1 := BOUND_OPER - Add('>=(FOP DOUBLE 1)');
    FOP_GE_DOUBLE2 := BOUND_OPER - Add('>=(FOP DOUBLE 2)');

    FOP_EQ_INTEGER1 := BOUND_OPER - Add('=(FOP INTEGER 1)');
    FOP_EQ_INTEGER2 := BOUND_OPER - Add('=(FOP INTEGER 2)');
    FOP_EQ_DOUBLE1 := BOUND_OPER - Add('=(FOP DOUBLE 1)');
    FOP_EQ_DOUBLE2 := BOUND_OPER - Add('=(FOP DOUBLE 2)');

    FOP_NE_INTEGER1 := BOUND_OPER - Add('<>(FOP INTEGER 1)');
    FOP_NE_INTEGER2 := BOUND_OPER - Add('<>(FOP INTEGER 2)');
    FOP_NE_DOUBLE1 := BOUND_OPER - Add('<>(FOP DOUBLE 1)');
    FOP_NE_DOUBLE2 := BOUND_OPER - Add('<>(FOP DOUBLE 2)');

    FOP_BITWISE_AND1 := BOUND_OPER - Add('AND_BITWISE(FOP 1)');
    FOP_BITWISE_AND2 := BOUND_OPER - Add('AND_BITWISE(FOP 2)');

    FOP_BITWISE_OR1 := BOUND_OPER - Add('OR_BITWISE(FOP 1)');
    FOP_BITWISE_OR2 := BOUND_OPER - Add('OR_BITWISE(FOP 2)');

    FOP_BITWISE_XOR1 := BOUND_OPER - Add('XOR_BITWISE(FOP 1)');
    FOP_BITWISE_XOR2 := BOUND_OPER - Add('XOR_BITWISE(FOP 2)');

    FOP_LOGICAL_AND1 := BOUND_OPER - Add('AND_LOGICAL(FOP 1)');
    FOP_LOGICAL_AND2 := BOUND_OPER - Add('AND_LOGICAL(FOP 2)');

    FOP_LOGICAL_OR1 := BOUND_OPER - Add('OR_LOGICAL(FOP 1)');
    FOP_LOGICAL_OR2 := BOUND_OPER - Add('OR_LOGICAL(FOP 2)');

    FOP_LOGICAL_XOR1 := BOUND_OPER - Add('XOR_LOGICAL(FOP 1)');
    FOP_LOGICAL_XOR2 := BOUND_OPER - Add('XOR_LOGICAL(FOP 2)');

    FOP_BITWISE_NOT1 := BOUND_OPER - Add('NOT_BITWISE(FOP 1)');
    FOP_BITWISE_NOT2 := BOUND_OPER - Add('NOT_BITWISE(FOP 2)');

    FOP_LOGICAL_NOT1 := BOUND_OPER - Add('NOT_LOGICAL(FOP 1)');
    FOP_LOGICAL_NOT2 := BOUND_OPER - Add('NOT_LOGICAL(FOP 2)');

    FOP_UNARY_MINUS_INTEGER1 := BOUND_OPER - Add('-(unary FOP INTEGER 1)');
    FOP_UNARY_MINUS_INTEGER2 := BOUND_OPER - Add('-(unary FOP INTEGER 2)');

    FOP_UNARY_MINUS_DOUBLE1 := BOUND_OPER - Add('-(unary FOP DOUBLE 1)');
    FOP_UNARY_MINUS_DOUBLE2 := BOUND_OPER - Add('-(unary FOP DOUBLE 2)');

    FOP_SHL1 := BOUND_OPER - Add('SHL(FOP 1)');
    FOP_SHL2 := BOUND_OPER - Add('SHL(FOP 2)');

    FOP_SHR1 := BOUND_OPER - Add('SHR(FOP 1)');
    FOP_SHR2 := BOUND_OPER - Add('SHR(FOP 2)');

    FOP_USHR1 := BOUND_OPER - Add('USHR(FOP 1)');
    FOP_USHR2 := BOUND_OPER - Add('USHR(FOP 2)');

    FOP_ASSIGN := BOUND_OPER - Add(':=(FOP)');
    FOP_GO_FALSE1 := BOUND_OPER - Add('GO_FALSE(FOP 1)');
    FOP_GO_FALSE2 := BOUND_OPER - Add('GO_FALSE(FOP 2)');
    FOP_GO_TRUE1 := BOUND_OPER - Add('GO_TRUE(FOP 1)');
    FOP_GO_TRUE2 := BOUND_OPER - Add('GO_TRUE(FOP 2)');

//=========================================
    FOP_PUSH := BOUND_OPER - Add('PUSH(FOP)');
    FOP_CALL := BOUND_OPER - Add('CALL(FOP)');

    FOP_PLUS1 := BOUND_OPER - Add('+(FOP 1)');
    FOP_PLUS2 := BOUND_OPER - Add('+(FOP 2)');

    FOP_MINUS1 := BOUND_OPER - Add('-(FOP 1)');
    FOP_MINUS2 := BOUND_OPER - Add('-(FOP 2)');

    FOP_MULT1 := BOUND_OPER - Add('*(FOP 1)');
    FOP_MULT2 := BOUND_OPER - Add('*(FOP 2)');

    FOP_DIV1 := BOUND_OPER - Add('/(FOP 1)');
    FOP_DIV2 := BOUND_OPER - Add('/(FOP 2)');

    FOP_LT1 := BOUND_OPER - Add('<(FOP 1)');
    FOP_LT2 := BOUND_OPER - Add('<(FOP 2)');

    FOP_LE1 := BOUND_OPER - Add('<=(FOP 1)');
    FOP_LE2 := BOUND_OPER - Add('<=(FOP 2)');

    FOP_GT1 := BOUND_OPER - Add('>(FOP 1)');
    FOP_GT2 := BOUND_OPER - Add('>(FOP 2)');

    FOP_GE1 := BOUND_OPER - Add('>=(FOP 1)');
    FOP_GE2 := BOUND_OPER - Add('>=(FOP 2)');

    FOP_EQ1 := BOUND_OPER - Add('=(FOP 1)');
    FOP_EQ2 := BOUND_OPER - Add('=(FOP 2)');

    FOP_NE1 := BOUND_OPER - Add('<>(FOP 1)');
    FOP_NE2 := BOUND_OPER - Add('<>(FOP 2)');

    OP_ON_USES := BOUND_OPER - Add('ON USES');
    OP_DECLARE := BOUND_OPER - Add('DECLARE');

    OP_ASSIGN_SIMPLE := BOUND_OPER - Add('ASSIGN_SIMPLE');

    FillChar(OpResultTypes, SizeOf(OpResultTypes), 0);
    AddOpResultType(typeINTEGER, typeCARDINAL, typeINTEGER);
    AddOpResultType(typeINTEGER, typeINT64, typeINT64);
    AddOpResultType(typeINTEGER, typeDOUBLE, typeDOUBLE);
    AddOpResultType(typeINTEGER, typeCURRENCY, typeCURRENCY);
    AddOpResultType(typeINTEGER, typeEXTENDED, typeEXTENDED);
    AddOpResultType(typeINTEGER, typeSINGLE, typeSINGLE);
    AddOpResultType(typeINTEGER, typeBYTE, typeINTEGER);
    AddOpResultType(typeINTEGER, typeWORD, typeINTEGER);
    AddOpResultType(typeINTEGER, typeSHORTINT, typeINTEGER);
    AddOpResultType(typeINTEGER, typeSMALLINT, typeINTEGER);

    AddOpResultType(typeCARDINAL, typeINT64, typeINT64);
    AddOpResultType(typeCARDINAL, typeDOUBLE, typeDOUBLE);
    AddOpResultType(typeCARDINAL, typeCURRENCY, typeCURRENCY);
    AddOpResultType(typeCARDINAL, typeSINGLE, typeSINGLE);
    AddOpResultType(typeCARDINAL, typeBYTE, typeINTEGER);
    AddOpResultType(typeCARDINAL, typeWORD, typeINTEGER);
    AddOpResultType(typeCARDINAL, typeSHORTINT, typeINTEGER);
    AddOpResultType(typeCARDINAL, typeSMALLINT, typeINTEGER);

    AddOpResultType(typeINT64, typeDOUBLE, typeDOUBLE);
    AddOpResultType(typeINT64, typeCURRENCY, typeCURRENCY);
    AddOpResultType(typeINT64, typeSINGLE, typeSINGLE);
    AddOpResultType(typeINT64, typeBYTE, typeINT64);
    AddOpResultType(typeINT64, typeWORD, typeINT64);
    AddOpResultType(typeINT64, typeSHORTINT, typeINT64);
    AddOpResultType(typeINT64, typeSMALLINT, typeINT64);

    AddOpResultType(typeBYTE, typeDOUBLE, typeDOUBLE);
    AddOpResultType(typeBYTE, typeCURRENCY, typeCURRENCY);
    AddOpResultType(typeBYTE, typeSINGLE, typeSINGLE);
    AddOpResultType(typeBYTE, typeWORD, typeINTEGER);
    AddOpResultType(typeBYTE, typeSHORTINT, typeINTEGER);
    AddOpResultType(typeBYTE, typeSMALLINT, typeINTEGER);

    AddOpResultType(typeWORD, typeDOUBLE, typeDOUBLE);
    AddOpResultType(typeWORD, typeCURRENCY, typeCURRENCY);
    AddOpResultType(typeWORD, typeSINGLE, typeSINGLE);
    AddOpResultType(typeWORD, typeSHORTINT, typeINTEGER);
    AddOpResultType(typeWORD, typeSMALLINT, typeINTEGER);

    AddOpResultType(typeSHORTINT, typeDOUBLE, typeDOUBLE);
    AddOpResultType(typeSHORTINT, typeCURRENCY, typeCURRENCY);
    AddOpResultType(typeSHORTINT, typeSINGLE, typeSINGLE);
    AddOpResultType(typeSHORTINT, typeSMALLINT, typeINTEGER);

    AddOpResultType(typeSMALLINT, typeDOUBLE, typeDOUBLE);
    AddOpResultType(typeSMALLINT, typeCURRENCY, typeCURRENCY);
    AddOpResultType(typeSMALLINT, typeSINGLE, typeSINGLE);

    AddOpResultType(typeDOUBLE, typeSINGLE, typeDOUBLE);
    AddOpResultType(typeDOUBLE, typeCURRENCY, typeCURRENCY);
    AddOpResultType(typeDOUBLE, typeEXTENDED, typeEXTENDED);

    AddOPResultType(typeWIDESTRING, typeSTRING, typeSTRING);

    AddOpResultType(typeSINGLE, typeCURRENCY, typeCURRENCY);

    AddOpResultType(typeCHAR, typeSTRING, typeSTRING);

    FillChar(AssResultTypes, SizeOf(AssResultTypes), 0);

    AddAssResultType(typeINTEGER, typeCARDINAL);
    AddAssResultType(typeINTEGER, typeINT64);
    AddAssResultType(typeINTEGER, typeBYTE);
    AddAssResultType(typeINTEGER, typeWORD);
    AddAssResultType(typeINTEGER, typeSHORTINT);
    AddAssResultType(typeINTEGER, typeSMALLINT);

    AddAssResultType(typeCARDINAL, typeINTEGER);
    AddAssResultType(typeCARDINAL, typeINT64);
    AddAssResultType(typeCARDINAL, typeBYTE);
    AddAssResultType(typeCARDINAL, typeWORD);
    AddAssResultType(typeCARDINAL, typeSHORTINT);
    AddAssResultType(typeCARDINAL, typeSMALLINT);

    AddAssResultType(typeINT64, typeINTEGER);
    AddAssResultType(typeINT64, typeCARDINAL);
    AddAssResultType(typeINT64, typeBYTE);
    AddAssResultType(typeINT64, typeWORD);
    AddAssResultType(typeINT64, typeSHORTINT);
    AddAssResultType(typeINT64, typeSMALLINT);

    AddAssResultType(typeBYTE, typeINTEGER);
    AddAssResultType(typeBYTE, typeINT64);
    AddAssResultType(typeBYTE, typeCARDINAL);
    AddAssResultType(typeBYTE, typeWORD);
    AddAssResultType(typeBYTE, typeSHORTINT);
    AddAssResultType(typeBYTE, typeSMALLINT);

    AddAssResultType(typeWORD, typeINTEGER);
    AddAssResultType(typeWORD, typeINT64);
    AddAssResultType(typeWORD, typeCARDINAL);
    AddAssResultType(typeWORD, typeBYTE);
    AddAssResultType(typeWORD, typeSHORTINT);
    AddAssResultType(typeWORD, typeSMALLINT);

    AddAssResultType(typeSHORTINT, typeINTEGER);
    AddAssResultType(typeSHORTINT, typeINT64);
    AddAssResultType(typeSHORTINT, typeCARDINAL);
    AddAssResultType(typeSHORTINT, typeBYTE);
    AddAssResultType(typeSHORTINT, typeWORD);
    AddAssResultType(typeSHORTINT, typeSMALLINT);

    AddAssResultType(typeSMALLINT, typeINTEGER);
    AddAssResultType(typeSMALLINT, typeINT64);
    AddAssResultType(typeSMALLINT, typeCARDINAL);
    AddAssResultType(typeSMALLINT, typeBYTE);
    AddAssResultType(typeSMALLINT, typeWORD);
    AddAssResultType(typeSMALLINT, typeSHORTINT);

    AddAssResultType(typeDOUBLE, typeINTEGER);
    AddAssResultType(typeDOUBLE, typeCARDINAL);
    AddAssResultType(typeDOUBLE, typeINT64);
    AddAssResultType(typeDOUBLE, typeBYTE);
    AddAssResultType(typeDOUBLE, typeWORD);
    AddAssResultType(typeDOUBLE, typeSHORTINT);
    AddAssResultType(typeDOUBLE, typeEXTENDED);
    AddAssResultType(typeDOUBLE, typeSINGLE);
    AddAssResultType(typeDOUBLE, typeCURRENCY);

    AddAssResultType(typeEXTENDED, typeINTEGER);
    AddAssResultType(typeEXTENDED, typeCARDINAL);
    AddAssResultType(typeEXTENDED, typeINT64);
    AddAssResultType(typeEXTENDED, typeBYTE);
    AddAssResultType(typeEXTENDED, typeWORD);
    AddAssResultType(typeEXTENDED, typeSHORTINT);
    AddAssResultType(typeEXTENDED, typeDOUBLE);
    AddAssResultType(typeEXTENDED, typeSINGLE);
    AddAssResultType(typeEXTENDED, typeCURRENCY);

    AddAssResultType(typeSINGLE, typeINTEGER);
    AddAssResultType(typeSINGLE, typeCARDINAL);
    AddAssResultType(typeSINGLE, typeINT64);
    AddAssResultType(typeSINGLE, typeBYTE);
    AddAssResultType(typeSINGLE, typeWORD);
    AddAssResultType(typeSINGLE, typeSHORTINT);
    AddAssResultType(typeSINGLE, typeDOUBLE);
    AddAssResultType(typeSINGLE, typeEXTENDED);
    AddAssResultType(typeSINGLE, typeCURRENCY);

    AddAssResultType(typeCURRENCY, typeINTEGER);
    AddAssResultType(typeCURRENCY, typeCARDINAL);
    AddAssResultType(typeCURRENCY, typeINT64);
    AddAssResultType(typeCURRENCY, typeBYTE);
    AddAssResultType(typeCURRENCY, typeWORD);
    AddAssResultType(typeCURRENCY, typeSHORTINT);
    AddAssResultType(typeCURRENCY, typeDOUBLE);
    AddAssResultType(typeCURRENCY, typeEXTENDED);
    AddAssResultType(typeCURRENCY, typeSINGLE);

    AddAssResultType(typePOINTER, typeINTEGER);
    AddAssResultType(typePOINTER, typeCLASS);
    AddAssResultType(typePOINTER, typePCHAR);
    AddAssResultType(typePOINTER, typeVARIANT);

    AddAssResultType(typePCHAR, typeSTRING);
    AddAssResultType(typeSHORTSTRING, typeSTRING);
    AddAssResultType(typeWIDESTRING, typeSTRING);
    AddAssResultType(typePWIDECHAR, typeSTRING);

    AddAssResultType(typeCHAR, typeSTRING);
    AddAssResultType(typeSTRING, typeCHAR);

    SupportedPaxTypes := IntegerPaxTypes +
                         BooleanPaxTypes +
                         StringPaxTypes +
                         RealPaxTypes +
                         VarianTTitanTypes +
                         [typeSET] +
                         [typeSHORTSTRING];

    OverloadableOperators := TOperatorList.Create;
    with OverloadableOperators do
    begin
      Add('+', OP_PLUS);
      Add('+', OP_UNARY_PLUS);
      Add('-', OP_MINUS);
      Add('-', OP_UNARY_MINUS);
      Add('*', OP_MULT);
      Add('/', OP_DIV);
      Add('%', OP_MOD);
      Add('>>', OP_RIGHT_SHIFT);
      Add('<<', OP_LEFT_SHIFT);
      Add('==', OP_EQ);
      Add('!=', OP_NE);
      Add('<', OP_LT);
      Add('>', OP_GT);
      Add('<=', OP_LE);
      Add('>=', OP_GE);
    end;

    RelationalOperators := TOperatorList.Create;
    with RelationalOperators do
    begin
      Add('==', OP_EQ);
      Add('!=', OP_NE);
      Add('<', OP_LT);
      Add('>', OP_GT);
      Add('<=', OP_LE);
      Add('>=', OP_GE);
      Add('===', OP_ID);
      Add('!==', OP_NI);
    end;
  end;
end;

procedure Finalization_BASE_SYS;
begin
  PAXOperators.Free;
  PAXKinds.Free;
  PAXTypes.Free;
  OverloadableOperators.Free;
  RelationalOperators.Free;

  Finalization_BASE_SYNC;
end;


